Package: rixpress
Title: Build Reproducible Analytical Pipelines with 'Nix'
Version: 0.12.3
Authors@R: c(
  person(given = "Bruno", family = "Rodrigues", email = "bruno@brodrigues.co", role = c("aut", "cre"),
         comment = c(ORCID = "0000-0002-3211-3689")),
  person(given = c("William", "Michael"), family = "Landau", email = "will.landau.oss@gmail.com", role = "rev",
         comment = c("William reviewed the package (v. 0.2.0) for rOpenSci,
                      see <https://github.com/ropensci/software-review/issues/706>")),
  person(given = "Anthony", family = "Martinez", email = "ajmartinez@usgs.gov", role = "rev",
         comment = c(ORCID = "0000-0002-4295-0261", "Anthony reviewed the
                     package (v. 0.2.0) for rOpenSci,
                     see <https://github.com/ropensci/software-review/issues/625>"))
  )
Description: Streamlines the creation of reproducible analytical pipelines using
  'default.nix' expressions generated via the 'rix' package for reproducibility. Define
  derivations in 'R', 'Python' or 'Julia', chain them into a composition of pure
  functions and build the resulting pipeline using 'Nix' as the underlying
  end-to-end build tool. Functions to plot the pipeline as a directed acyclic
  graph are included, as well as functions to load and inspect intermediary results
  for interactive analysis. User experience heavily inspired by the 'targets' package.
License: GPL (>= 3)
Encoding: UTF-8
URL: https://github.com/ropensci/rixpress/,
        https://docs.ropensci.org/rixpress/
BugReports: https://github.com/ropensci/rixpress/issues/
Depends: R (>= 4.1.0)
Imports: cli, igraph, jsonlite, processx
RoxygenNote: 7.3.3
SystemRequirements: Nix
Language: en-GB
Suggests: chronicler, dplyr, ggdag, ggplot2, knitr, maybe, mockery,
        reticulate, rix, rmarkdown, testthat (>= 3.0.0), usethis,
        visNetwork
Config/testthat/edition: 3
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-05-19 06:51:21 UTC; brodrigues
Author: Bruno Rodrigues [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-3211-3689>),
  William Michael Landau [rev] (William reviewed the package (v. 0.2.0)
    for rOpenSci, see
    <https://github.com/ropensci/software-review/issues/706>),
  Anthony Martinez [rev] (ORCID: <https://orcid.org/0000-0002-4295-0261>,
    Anthony reviewed the package (v. 0.2.0) for rOpenSci, see
    <https://github.com/ropensci/software-review/issues/625>)
Maintainer: Bruno Rodrigues <bruno@brodrigues.co>
Repository: CRAN
Date/Publication: 2026-05-19 07:30:07 UTC
Built: R 4.6.0; ; 2026-05-19 10:27:41 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,rxp_derivation)
S3method(print,rxp_pipeline)
export(add_import)
export(adjust_import)
export(rxp_check_chronicles)
export(rxp_copy)
export(rxp_dag_for_ci)
export(rxp_export_artifacts)
export(rxp_ga)
export(rxp_gc)
export(rxp_ggdag)
export(rxp_import_artifacts)
export(rxp_init)
export(rxp_inspect)
export(rxp_jl)
export(rxp_jl_file)
export(rxp_list_logs)
export(rxp_load)
export(rxp_make)
export(rxp_pipeline)
export(rxp_populate)
export(rxp_py)
export(rxp_py2r)
export(rxp_py_file)
export(rxp_qmd)
export(rxp_r)
export(rxp_r2py)
export(rxp_r_file)
export(rxp_read)
export(rxp_rmd)
export(rxp_trace)
export(rxp_visnetwork)
export(rxp_write_dag)
importFrom(igraph,V)
importFrom(igraph,delete_vertex_attr)
importFrom(igraph,graph_from_data_frame)
importFrom(igraph,set_vertex_attr)
importFrom(igraph,write_graph)
importFrom(jsonlite,fromJSON)
importFrom(jsonlite,write_json)
importFrom(processx,run)
importFrom(stats,setNames)
importFrom(utils,capture.output)
importFrom(utils,globalVariables)
importFrom(utils,head)

# rixpress: Build Reproducible Analytical Pipelines with ‘Nix’ <a href="https://docs.ropensci.org/rixpress/"><img src="man/figures/logo.png" align="right" height="138" /></a>

<!-- Badges -->

[![R-hub
v2](https://github.com/ropensci/rixpress/actions/workflows/rhub.yaml/badge.svg)](https://github.com/ropensci/rixpress/actions/workflows/rhub.yaml/)
[![CRAN](https://www.r-pkg.org/badges/version/rixpress)](https://CRAN.R-project.org/package=rixpress)
[![runiverse-package
rixpress](https://ropensci.r-universe.dev/badges/rixpress?scale=1&color=pink&style=round)](https://ropensci.r-universe.dev/rixpress)
[![Docs](https://img.shields.io/badge/docs-release-blue.svg)](https://docs.ropensci.org/rixpress/)
[![Status at rOpenSci Software Peer
Review](https://badges.ropensci.org/706_status.svg)](https://github.com/ropensci/software-review/issues/706)

If you want to watch a 2-Minute video introduction, click the image
below:

<a href="https://www.youtube.com/watch?v=a1eNG9TFZ_o" target="_blank" rel="noopener noreferrer">
<img src="https://raw.githubusercontent.com/ropensci/rixpress/refs/heads/main/video_thumbnail.png" alt="Video Thumbnail" style="width:100%; max-width:560px; height:auto; display:block; margin:0 auto;">
</a>

`rixpress` streamlines creation of *micropipelines* (small-to-medium,
single–machine analytic pipelines) by expressing a pipeline in idiomatic
R while delegating build orchestration, dependency management, and
multi-language execution to the ‘Nix’ build system. It is inspired by
the user experience of the `{targets}` package and builds on the `{rix}`
package to obtain fully reproducible development/runtime environments.

Key ideas:

- Define pipeline *derivations* with concise `rxp_*()` helper functions.
- Seamlessly mix R, Python, Julia, Quarto / Markdown steps.
- Reuse hermetic environments (pinned system + language dependencies)
  defined via `{rix}` and a `default.nix`.
- Visualize and inspect the DAG; selectively read, load, or copy
  outputs.
- Export/import build artifacts to speed up CI or share results.

Here is what a basic pipeline looks like:

``` r
library(rixpress)

list(
  rxp_r_file(
    mtcars,
    'mtcars.csv',
    \(x) (read.csv(file = x, sep = "|"))
  ),

  rxp_r(
    mtcars_am,
    filter(mtcars, am == 1)
  ),

  rxp_r(
    mtcars_head,
    head(mtcars_am)
  ),

  rxp_r(
    mtcars_tail,
    tail(mtcars_head)
  ),

  rxp_r(
    mtcars_mpg,
    select(mtcars_tail, mpg)
  ),

  rxp_qmd(
    page,
    "page.qmd"
  )
) |>
  rxp_populate()
```

## Contents

- [Motivation](#motivation)
- [Feature Highlights](#feature-highlights)
- [Organizing with Sub-Pipelines](#organizing-with-sub-pipelines)
- [Examples](#examples)
- [Getting started](#getting-started)
- [Visualization](#visualization)
- [Exporting / Importing Artifacts](#exporting--importing-artifacts)
- [Contributing](#contributing)
- [Scope](#scope)
- [Acknowledgements](#acknowledgements)

## Motivation

Reproducibility involves two intertwined concerns: (1) the *environment*
(system libraries, compilers, language interpreters, packages) and (2)
the *execution graph* of analytic steps with their inputs and outputs.
There are many tools to address both of these concerns, but
orchestrating them is not always easy.

Another developing trend, in my opinion, is that data science is
becoming increasingly **polyglot**: teams rarely restrict themselves to
a single language. Python dominates machine learning, R excels at
statistical modeling and visualization, and Julia offers
high-performance numerics with a syntax that feels familiar to both
communities. Analysts and researchers often need to combine strengths
from all three, moving data and results fluidly across tools.

This trend makes a unifying foundation essential. Without it, people
waste time stitching together environments, dealing with dependency
conflicts, or struggling to reproduce results on different machines.
**Nix** provides that foundation by declaratively describing
environments which include system libraries, compilers, interpreters,
and packages in a way that is reproducible down to exact versions and
build instructions.

The `{rix}` package brings this power into the R ecosystem. It makes it
easy to generate declarative, date-pinned Nix expressions that work
consistently across systems. This expressions can then be used to build
reproducible development environments that include programming
languages, packages and other tools. `{rixpress}` addresses the next
part, the execution graph. With a concise R API, it lets users describe
analytic steps, inputs, and outputs, while delegating execution to Nix
for guaranteed determinism.

Together, this means data scientists can orchestrate polyglot pipelines
that combine R, Python, and Julia seamlessly, with environments and
workflows that are portable, reproducible, and future-proof.

If you are interested in a Python-first port of `{rixpress}`, check out
[ryxpress](https://github.com/b-rodrigues/ryxpress/tree/main).

## Feature Highlights

- Uniform `rxp_*()` constructors for files, R functions, Python / Julia
  code, and Quarto documents.
- Cross-language object exchange via serialization helpers.
- Pipelines as compositions of pure functions: derivations are
  content-addressed and cached in the Nix store.
- **Sub-pipelines**: Organize large projects with `rxp_pipeline()` for
  logical grouping and colour-coded visualization.
- DAG visualization (e.g. `rxp_ggdag()`, `rxp_visnetwork()`) with
  pipeline-aware colouring.
- Target introspection (`rxp_read()`, `rxp_load()`, `rxp_copy()`,
  `rxp_trace()`).
- Artifact export/import for CI acceleration.

## Organizing with Sub-Pipelines

For larger projects, you can organize derivations into named
**sub-pipelines** using `rxp_pipeline()`. Each sub-pipeline can have its
own colour for visual distinction:

``` r
# Define sub-pipelines in separate files
# pipelines/01_data_prep.R returns: list(rxp_r(...), rxp_r(...))
# pipelines/02_analysis.R returns: list(rxp_r(...), rxp_r(...))

# Master script combines them with colours
pipe_prep <- rxp_pipeline("Data Prep", "pipelines/01_data_prep.R", color = "#E69F00")
pipe_analysis <- rxp_pipeline("Analysis", "pipelines/02_analysis.R", color = "#56B4E9")

rxp_populate(list(pipe_prep, pipe_analysis))
```

When visualized with `rxp_ggdag()`, nodes show: - **Fill colour**:
Derivation type (R = blue, Python = yellow, etc.) - **Border colour**:
Pipeline group (the colour you specify)

This dual encoding lets you see both *what* each step does and *which
logical group* it belongs to. See `vignette("sub-pipelines")` for
details.

## Examples

Here is what a basic pipeline looks like:

``` r
library(rixpress)

list(
  rxp_r_file(
    mtcars,
    'mtcars.csv',
    \(x) (read.csv(file = x, sep = "|"))
  ),

  rxp_r(
    mtcars_am,
    filter(mtcars, am == 1)
  ),

  rxp_r(
    mtcars_head,
    head(mtcars_am)
  ),

  rxp_r(
    mtcars_tail,
    tail(mtcars_head)
  ),

  rxp_r(
    mtcars_mpg,
    select(mtcars_tail, mpg)
  ),

  rxp_qmd(
    page,
    "page.qmd"
  )
) |>
  rxp_populate()
```

Running `rxp_populate()` generates a `pipeline.nix` file, which contains
the build instructions for all derivations and final outputs expressed
as Nix code. You can define derivations that run Python or Julia code,
and objects can be exchanged between R and Python by using `rxp_py2r()`
and `rxp_r2py()`, or by serializing to a common format such as JSON. By
default, calling `rxp_populate()` also builds the pipeline, but it’s
possible to only generate the `pipeline.nix` file and build the pipeline
later using:

``` r
rxp_make()
```

The build process assumes the presence of a `default.nix` file that
defines the computational environment the pipeline runs in; this file
can be generated with the {rix} package. The `default.nix` typically
defines an environment with R and required R packages (and optionally
Python/Julia and their packages), Quarto, and any necessary system-level
dependencies pinned to a specific date to ensure reproducibility.

In the example above, the first derivation reads `mtcars.csv` (in the
example it’s pipe-separated, i.e. a `.psv` file). Each output (for
example, `mtcars`, `mtcars_am`, `mtcars_head`, `mtcars_tail`,
`mtcars_mpg`, `page`) is built by Nix within the environment defined by
`default.nix`. Concretely, {rix} makes using Nix as a package manager
easier for R users, and {rixpress} makes it easy to use Nix as a build
automation tool.

And this is what a polyglot pipeline, using both R and Python, looks
like:

``` r
library(rixpress)

list(
  rxp_py_file(
    name = mtcars_pl,
    path = "data/mtcars.csv",
    read_function = "lambda x: polars.read_csv(x, separator='|')"
  ),

  rxp_py(
    name = mtcars_pl_am,
    expr = "mtcars_pl.filter(polars.col('am') == 1)",
    user_functions = "functions.py",
    encoder = "serialize_to_json",
  ),

  rxp_r(
    name = mtcars_head,
    expr = my_head(mtcars_pl_am),
    user_functions = "functions.R",
    decoder = "jsonlite::fromJSON"
  ),

  rxp_r(
    name = mtcars_mpg,
    expr = dplyr::select(mtcars_head, mpg)
  )
) |>
  rxp_populate(project_path = ".", build = FALSE)
```

Because the pipeline is built using Nix, outputs are stored in the Nix
store under `/nix/store/`. To make working with these outputs easier,
`{rixpress}` provides several helper functions:

- `rxp_read("mtcars_mpg")` — read the content of `mtcars_mpg` into R
  (the return value depends on the derivation type: an R object, a file
  path, etc.);
- `rxp_load("mtcars_mpg")` — load objects from the result into the
  global environment;
- `rxp_copy("page")` — copy outputs (e.g. a generated document) from the
  Nix store into the current working directory so you can open or
  inspect them there.

Python objects will be converted into their equivalent R objects if
`{reticulate}` is available in the environment. For complex outputs such
as documents (for example the Quarto document `page` above),
`rxp_read("page")` returns the output file path; you can then open it
with `browseURL()` or copy it into your working directory with
`rxp_copy()`.

`{rixpress}` is flexible; please consult the examples repository for
many different patterns and complete demos:

<https://github.com/b-rodrigues/rixpress_demos/tree/master>

## Getting started

While `{rixpress}` is a regular R package, there is little point in
using it without having Nix installed. If you are not familiar with Nix,
I recommend you first start by checking out `{rix}`, which will teach
you how to use Nix to set up reproducible development environments for
your R projects!

We recommend you check out the following vignettes to get started:

- [Introductory
  concepts](https://docs.ropensci.org/rixpress/articles/intro-concepts.html)
- [Core rixpress Functions and
  Usage](https://docs.ropensci.org/rixpress/articles/core-functions.html)
- [Tutorial](https://docs.ropensci.org/rixpress/articles/tutorial.html)

## Visualization

It is possible to visually inspect the pipeline:

``` r
# ggdag-based static plot
rxp_ggdag()
```

<figure>

<img src="https://raw.githubusercontent.com/ropensci/rixpress/refs/heads/main/dag.png" alt="DAG" />
<figcaption aria-hidden="true">

DAG
</figcaption>

</figure>

``` r
# Interactive network (visNetwork)
rxp_visnetwork()
```

To create an artifact suitable for CI, use:

``` r
rxp_dag_for_ci()
```

This will export the directed acyclic graph of the pipeline as a `dot`
file, making it possible to visualize in CI (using
[`stacked-dag`](https://hackage.haskell.org/package/stacked-dag) for
example).

## Exporting / Importing Artifacts

Speed up continuous integration or move a cache between machines:

``` r
# Export build products
rxp_export_artifacts()

# Import elsewhere before building
rxp_import_artifacts()
```

## Contributing

Pull requests are welcome. If you’re unsure whether to open one, feel
free to open an issue first to discuss your idea. For contributor
guidelines, see CONTRIBUTING.md.

If you plan to contribute documentation or vignettes, please:

- Add runnable, minimal examples.
- Prefer small datasets and short-running examples.
- Document required system dependencies for the example (through a
  `default.nix`).

Please note that this package is released with a [Contributor Code of
Conduct](https://ropensci.org/code-of-conduct/). By contributing to this
project, you agree to abide by its terms.

## Scope

Please refer to the `vignette("scope")` to learn more about what
`{rixpress}` will and will not support.

## Acknowledgements

- Inspired by the design and ergonomics of `{targets}`.
- Built atop the reproducibility foundations provided by `{rix}` and the
  ‘Nix’ ecosystem.
- rOpenSci peer review contributors and reviewers for feedback that
  shaped the interface.

## ----eval = FALSE-------------------------------------------------------------
# library(chronicler)
# 
# # Create a recorded version of sqrt
# r_sqrt <- record(sqrt)
# 
# # Success case
# result <- r_sqrt(4)
# result$value
# #> Just
# #> [1] 2
# 
# # Failure case: sqrt(-1) produces a warning "NaNs produced"
# # With default strict=2, warnings are treated as failures
# result <- r_sqrt(-1)
# result$value
# #> Nothing

## ----eval = FALSE-------------------------------------------------------------
# library(rixpress)
# 
# list(
#   rxp_r(
#     name = result,
#     expr = r_sqrt(-1),  # This produces a `Nothing` value, not an error!
#     user_functions = "functions.R"
#   )
# ) |> rxp_populate(build = FALSE)

## ----eval = FALSE-------------------------------------------------------------
# # Build the pipeline - chronicle status is checked automatically!
# rxp_make()

## ----eval = FALSE-------------------------------------------------------------
# # functions.R
# library(chronicler)
# 
# r_filter <- record(dplyr::filter)
# r_pull <- record(dplyr::pull)
# r_sqrt <- record(sqrt)
# r_mean <- record(mean)

## ----eval = FALSE-------------------------------------------------------------
# library(rixpress)
# 
# list(
#   # Read data (not a chronicle)
#   rxp_r_file(
#     name = mtcars,
#     path = "data/mtcars.csv",
#     read_function = \(x) read.csv(file = x, sep = "|")
#   ),
# 
#   # Filter using chronicler - SUCCESS
#   rxp_r(
#     name = filtered_mtcars,
#     expr = mtcars |> r_filter(am == 1),
#     user_functions = "functions.R"
#   ),
# 
#   # Pull column - SUCCESS
#   rxp_r(
#     name = mtcars_mpg,
#     expr = filtered_mtcars |> bind_record(r_pull, mpg),
#     user_functions = "functions.R"
#   ),
# 
#   # Compute mean - SUCCESS
#   rxp_r(
#     name = mean_mpg,
#     expr = mtcars_mpg |> bind_record(r_mean),
#     user_functions = "functions.R"
#   ),
# 
#   # Intentional failure: sqrt(-1) - NOTHING
#   rxp_r(
#     name = sqrt_of_negative,
#     expr = r_sqrt(-1),
#     user_functions = "functions.R"
#   ),
# 
#   # Downstream of Nothing - also NOTHING
#   rxp_r(
#     name = downstream_of_nothing,
#     expr = sqrt_of_negative |> bind_record(r_mean),
#     user_functions = "functions.R"
#   )
# ) |>
#   rxp_populate(build = FALSE)

## ----eval = FALSE-------------------------------------------------------------
# rxp_make()
# 
# # You can also manually check chronicles at any time:
# # rxp_check_chronicles()

## ----eval = FALSE-------------------------------------------------------------
# rxp_read("sqrt_of_negative")
# #> Warning message:
# #> Derivation 'sqrt_of_negative' contains a chronicle with Nothing value!
# #>   Use chronicler::read_log() on this object for details.

## ----eval = FALSE-------------------------------------------------------------
# result <- rxp_read("sqrt_of_negative")
# read_log(result)

## ----eval = FALSE-------------------------------------------------------------
# library(rixpress)
# 
# rxp_init()

## ----eval = FALSE-------------------------------------------------------------
# # This script defines the default environment the pipeline runs in.
# # Add the required packages to execute the code necessary for each derivation.
# # If you want to create visual representations of the pipeline, consider adding
# # `{visNetwork}` and `{ggdag}` to the list of R packages.
# library(rix)
# 
# # Define execution environment
# rix(
#   date = NULL,
#   r_pkgs = NULL,
#   py_conf = NULL,
#   git_pkgs = list(
#     "package_name" = "rixpress",
#     "repo_url" = "https://github.com/ropensci/rixpress",
#     "commit" = "HEAD"
#   ),
#   ide = "none",
#   project_path = "."
# )

## ----eval = FALSE-------------------------------------------------------------
# library(rix)
# 
# # Define execution environment
# rix(
#   date = "2025-04-11",
#   r_pkgs = c("dplyr", "igraph"),
#   git_pkgs = list(
#     package_name = "rixpress",
#     repo_url = "https://github.com/ropensci/rixpress",
#     commit = "HEAD"
#   ),
#   ide = "rstudio",
#   project_path = ".",
#   overwrite = TRUE
# )

## ----eval = FALSE-------------------------------------------------------------
# library(rixpress)
# library(igraph)
# 
# list(
#   rxp_r_file(
#     name = NULL,
#     path = NULL,
#     read_function = \(x) read.csv(file = x, sep = ",")
#   ),
#   rxp_r(
#     name = NULL,
#     expr = NULL
#   )
# ) |> rxp_populate(build = TRUE)

## ----eval = FALSE-------------------------------------------------------------
# library(rixpress)
# library(igraph)
# 
# list(
#   rxp_r_file(
#     name = mtcars,
#     path = 'data/mtcars.csv',
#     read_function = \(x) (read.csv(file = x, sep = "|"))
#   ),
# 
#   rxp_r(
#     name = filtered_mtcars,
#     expr = filter(mtcars, am == 1)
#   )
# ) |> rxp_populate(build = TRUE)

## ----eval = FALSE-------------------------------------------------------------
# rxp_read("filtered_mtcars")

## ----eval = FALSE-------------------------------------------------------------
# filtered_mtcars <- rxp_read("filtered_mtcars")

## ----eval = FALSE-------------------------------------------------------------
# library(rixpress)
# library(igraph)
# 
# list(
#   rxp_r_file(
#     name = mtcars,
#     path = 'data/mtcars.csv',
#     read_function = \(x) (read.csv(file = x, sep = "|"))
#   ),
# 
#   rxp_r(
#     name = filtered_mtcars,
#     expr = filter(mtcars, am == 1)
#   ),
# 
#   rxp_r(
#     name = mtcars_mpg,
#     expr = select(filtered_mtcars, mpg)
#   )
# ) |>
#   rxp_populate(build = TRUE)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval = FALSE-------------------------------------------------------------
# library(rix)
# 
# rix(
#   date = "2025-03-31",
#   r_pkgs = c("dplyr", "igraph", "reticulate", "quarto"),
#   git_pkgs = list(
#     package_name = "rixpress",
#     repo_url = "https://github.com/ropensci/rixpress",
#     commit = "HEAD"
#   ),
#   py_conf = list(
#     py_version = "3.12",
#     py_pkgs = c("pandas", "polars", "pyarrow")
#   ),
#   ide = "none",
#   project_path = ".",
#   overwrite = TRUE
# )

## ----eval = FALSE-------------------------------------------------------------
# library(rixpress)
# library(igraph)
# 
# list(
#   rxp_py_file(
#     name = mtcars_pl,
#     path = 'data/mtcars.csv',
#     read_function = "lambda x: pl.read_csv(x, separator='|')"
#   ),
# 
#   rxp_py(
#     # reticulate doesn't support polars DFs yet, so need to convert
#     # first to pandas DF
#     name = mtcars_pl_am,
#     expr = "mtcars_pl.filter(pl.col('am') == 1).to_pandas()"
#   ),
# 
#   rxp_py2r(
#     name = mtcars_am,
#     expr = mtcars_pl_am
#   ),
# 
#   rxp_r(
#     name = mtcars_head,
#     expr = my_head(mtcars_am),
#     user_functions = "functions.R"
#   ),
# 
#   rxp_r2py(
#     name = mtcars_head_py,
#     expr = mtcars_head
#   ),
# 
#   rxp_py(
#     name = mtcars_tail_py,
#     expr = 'mtcars_head_py.tail()'
#   ),
# 
#   rxp_py2r(
#     name = mtcars_tail,
#     expr = mtcars_tail_py
#   ),
# 
#   rxp_r(
#     name = mtcars_mpg,
#     expr = dplyr::select(mtcars_tail, mpg)
#   ),
# 
#   rxp_qmd(
#     name = page,
#     qmd_file = "my_doc/page.qmd",
#     additional_files = c("my_doc/content.qmd", "my_doc/images")
#   )
# ) |>
#   rxp_populate(
#     project_path = ".",
#     py_imports = c(polars = "import polars as pl")
#   )

## ----eval = FALSE-------------------------------------------------------------
# py_imports = c(pillow = "from PIL import Image")

## ----eval = FALSE-------------------------------------------------------------
# library(rixpress)
# library(igraph)
# 
# list(
#   rxp_py_file(
#     name = mtcars_pl,
#     path = "data/mtcars.csv",
#     read_function = "lambda x: pl.read_csv(x, separator='|')"
#   ),
# 
#   rxp_py(
#     name = mtcars_pl_am,
#     expr = "mtcars_pl.filter(pl.col('am') == 1)",
#     user_functions = "functions.py",
#     encoder = "serialize_to_json",
#   ),
# 
#   rxp_r(
#     name = mtcars_head,
#     expr = my_head(mtcars_pl_am),
#     user_functions = "functions.R",
#     decoder = "jsonlite::fromJSON"
#   ),
# 
#   rxp_r(
#     name = mtcars_mpg,
#     expr = dplyr::select(mtcars_head, mpg)
#   )
# ) |>
#   rxp_populate(
#     project_path = ".",
#     py_imports = c(polars = "import polars as pl")
#   )
# 
# # Plot DAG for CI
# rxp_dag_for_ci()

## ----eval = FALSE-------------------------------------------------------------
# rxp_qmd(
#   name = page,
#   qmd_file = "my_doc/page.qmd",
#   additional_files = c("my_doc/content.qmd", "my_doc/images"),
#   args = "--to typst"
# )

## ----eval = FALSE-------------------------------------------------------------
# rix(
#   date = "2025-03-31",
#   r_pkgs = c("dplyr", "igraph", "reticulate", "quarto"),
#   system_pkgs = "typst",
#   git_pkgs = list(...

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----sub-pipeline-1, eval = FALSE---------------------------------------------
# # Data Preparation Sub-Pipeline
# # pipelines/01_data_prep.R
# library(rixpress)
# 
# list(
#   rxp_r(name = raw_mtcars, expr = mtcars),
#   rxp_r(name = clean_mtcars, expr = dplyr::filter(raw_mtcars, am == 1)),
#   rxp_r(name = selected_mtcars, expr = dplyr::select(clean_mtcars, mpg, cyl, hp, wt))
# )

## ----sub-pipeline-2, eval = FALSE---------------------------------------------
# # Analysis Sub-Pipeline
# # pipelines/02_analysis.R
# library(rixpress)
# 
# list(
#   rxp_r(name = summary_stats, expr = summary(selected_mtcars)),
#   rxp_r(name = mpg_model, expr = lm(mpg ~ hp + wt, data = selected_mtcars)),
#   rxp_r(name = model_coefs, expr = coef(mpg_model))
# )

## ----master-script, eval = FALSE----------------------------------------------
# # gen-pipeline.R
# library(rixpress)
# 
# # Create named pipelines with colours by pointing to the files
# pipe_data_prep <- rxp_pipeline(
#   name = "Data Preparation",
#   path = "pipelines/01_data_prep.R",
#   color = "#E69F00"
# )
# 
# pipe_analysis <- rxp_pipeline(
#   name = "Statistical Analysis",
#   path = "pipelines/02_analysis.R",
#   color = "#56B4E9"
# )
# 
# # Build combined pipeline
# rxp_populate(list(pipe_data_prep, pipe_analysis), project_path = ".", build = TRUE)

## ----color-modes, eval = FALSE------------------------------------------------
# # Dual encoding: fill = type, border = pipeline (default when pipelines are defined)
# rxp_ggdag(color_by = "pipeline")
# 
# # Colour entirely by derivation type (rxp_r, rxp_py, etc.) - original behaviour
# rxp_ggdag(color_by = "type")

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval = FALSE-------------------------------------------------------------
# library(rixpress)
# 
# list(
#   rxp_r_file(
#     name = mtcars,
#     path = 'data/mtcars.csv',
#     read_function = \(x) (read.csv(file = x, sep = "|"))
#   ),
# ...

## ----eval = FALSE-------------------------------------------------------------
# library(rixpress)
# 
# list(
#   rxp_r_file(
#     name = mtcars,
#     path = 'https://raw.githubusercontent.com/b-rodrigues/rixpress_demos/refs/heads/master/basic_r/data/mtcars.csv',
#     read_function = \(x) (read.csv(file = x, sep = "|"))
#   ),
# ...

## ----eval = FALSE-------------------------------------------------------------
# library(rixpress)
# 
# list(
#   rxp_r_file(
#     name = mtcars_r,
#     path = 'data',
#     read_function = \(x) {
#       (readr::read_delim(list.files(x, full.names = TRUE), delim = '|'))
#     }
#   )
# ) |>
#   rxp_populate(project_path = ".")

## ----eval = FALSE-------------------------------------------------------------
# library(rixpress)
# 
# list(
#   rxp_py_file(
#     name = mtcars_py,
#     path = 'data',
#     read_function = "read_many_csvs",
#     user_functions = "functions.py"
#   )
# ) |>
#   rxp_populate(project_path = ".")

## ----eval = FALSE-------------------------------------------------------------
# library(rixpress)
# 
# list(
#   rxp_py_file(
#     name = gdf,
#     # We provide the directory 'data' to ensure all shapefile components are copied.
#     path = 'data',
#     # The read_function must use the hardcoded path within the build environment.
#     read_function = "lambda x: geopandas.read_file('input_folder/data/oceans.shp', driver='ESRI Shapefile')"
#   ),
# 
#   rxp_py(
#     name = sa,
#     expr = "gdf.loc[gdf['Oceans'] == 'South Atlantic Ocean']['geometry'].loc[0]"
#   )
# ) |>
#   rxp_populate(project_path = ".")

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval = FALSE-------------------------------------------------------------
# library(rix)
# 
# rix(
#   date = "2025-09-04",
#   r_pkgs = c(
#     "arrow",
#     "dplyr",
#     "tidyr",
#     "ggplot2",
#     "hexbin"
#   ),
#   git_pkgs = list(
#     list(
#       package_name = "rix",
#       repo_url = "https://github.com/ropensci/rix/",
#       commit = "HEAD"
#     ),
#     list(
#       package_name = "rixpress",
#       repo_url = "https://github.com/ropensci/rixpress",
#       commit = "HEAD"
#     )
#   ),
#   jl_conf = list(
#     jl_version = "1.10",
#     jl_pkgs = c(
#       "Arrow",
#       "DataFrames",
#       "SparseArrays",
#       "LinearAlgebra"
#     )
#   ),
#   ide = "none",
#   project_path = ".",
#   overwrite = TRUE
# )

## ----eval = FALSE-------------------------------------------------------------
# library(rixpress)
# library(igraph)
# 
# list(
#   rxp_jl(d_size, '150'),
# 
#   rxp_jl(
#     data,
#     "0.1randn(d_size,d_size) + reshape( \
#     cholesky(gridlaplacian(d_size,d_size) + 0.003I) \\ randn(d_size*d_size), \
#     d_size, \
#     d_size \
#     )",
#     user_functions = "functions.jl"
#   ),
# 
#   rxp_jl(
#     laplace_df,
#     'DataFrame(data, :auto)',
#     encoder = 'arrow_write',
#     user_functions = "functions.jl"
#   ),
# 
#   rxp_r(
#     laplace_long_df,
#     prepare_data(laplace_df),
#     decoder = 'read_ipc_file',
#     user_functions = "functions.R"
#   ),
# 
#   rxp_r(
#     gg,
#     make_gg(laplace_long_df)
#   )
# 
# ) |>
#   rxp_populate(build = TRUE)

## ----eval = FALSE-------------------------------------------------------------
# prepare_data <- function(laplace){
#   laplace_df |>
#     mutate(
#       x_id = row_number()
#     ) |>
#     tidyr::pivot_longer(-x_id, names_to = "y_id", values_to = "z") |>
#     mutate(
#       y_id = gsub("x", "", y_id),
#       y_id = as.numeric(y_id)
#     )
# }
# 
# make_gg <- function(laplace_long_df){
#   laplace_long_df |>
#     ggplot(aes(x = x_id, y = y_id, z = z)) +
#     stat_summary_hex(fun = function(x) mean(x), bins = 45)  +
#     scale_fill_viridis_c(option = 12) +
#     theme_void() +
#     theme(legend.position = "none") +
#     labs(subtitle = "hexagonal 2-d heatmap of laplacian matrix")
# }
# 
# save_gg <- function(path, gg){
#   ggsave("gg.png", gg)
# }

# Constants for language configuration
# Constants for language configuration
LANG_CONFIG <- list(
  R = list(
    source_cmd = "source('%s')",
    derivation_type = "rxp_r",
    output_ext = "RDS"
  ),
  Py = list(
    source_cmd = "exec(open('%s').read())",
    derivation_type = "rxp_py",
    output_ext = "pickle"
  ),
  Jl = list(
    source_cmd = "include('%s')",
    derivation_type = "rxp_jl",
    output_ext = "jld2"
  )
)

extract_python_version <- function(nix_file, project_path = ".") {
  file_path <- file.path(project_path, nix_file)

  if (!file.exists(file_path)) {
    warning(paste("File not found:", file_path, "- using python3 as fallback"))
    return("python3")
  }

  lines <- readLines(file_path)

  # Find the pyconf block
  pyconf_start <- grep("^\\s*pyconf\\s*=", lines)

  if (length(pyconf_start) == 0) {
    # No pyconf found, return default
    return("python3")
  }

  pyconf_start <- pyconf_start[1]

  # Find the end of the pyconf block (closing };)
  end_idxs <- grep("^\\s*\\};\\s*$", lines)
  pyconf_end <- end_idxs[end_idxs > pyconf_start][1]

  if (is.na(pyconf_end)) {
    warning("Could not find end of pyconf block - using python3 as fallback")
    return("python3")
  }

  # Extract the pyconf block
  pyconf_lines <- lines[pyconf_start:pyconf_end]
  pyconf_text <- paste(pyconf_lines, collapse = "\n")

  # Look for python<version>Packages pattern
  # Matches: python39Packages, python310Packages, python311Packages, etc.
  pattern <- "python(\\d+)Packages"
  matches <- regmatches(
    pyconf_text,
    gregexpr(pattern, pyconf_text, perl = TRUE)
  )[[1]]

  if (length(matches) == 0) {
    # No specific version found, use default
    return("python3")
  }

  # Extract the version number from the first match
  version_match <- regmatches(matches[1], regexec(pattern, matches[1]))[[1]]
  version_num <- version_match[2]

  # Return pythonXYZ format (e.g., python312, python313)
  return(paste0("python", version_num))
}


#' Clean User Functions Vector
#'
#' Removes empty character entries from the provided `user_functions` vector.
#'
#' @param user_functions Character vector of user-defined function script paths.
#' @return Character vector without empty strings.
#' @noRd
clean_user_functions <- function(user_functions) {
  if (is.null(user_functions)) {
    return(character(0))
  }
  user_functions[nzchar(user_functions)]
}

#' Build Code Import/Source Commands for User Functions
#'
#' Creates language-specific commands to load user-defined function scripts.
#'
#' @param user_functions Character vector of script file names.
#' @param lang Language string, `"R"`, `"Py"`, or `"Jl"`.
#' @return A string of import/source statements.
#' @noRd
build_user_code_cmd <- function(user_functions, lang) {
  if (length(user_functions) == 0) {
    return("")
  }

  if (!lang %in% names(LANG_CONFIG)) {
    stop("Unsupported lang: ", lang)
  }

  files <- file.path("input_folder", user_functions)
  commands <- sprintf(LANG_CONFIG[[lang]]$source_cmd, files)

  paste(commands, collapse = "\n")
}

#' Build Environment Variable Export Commands
#'
#' Generates shell export commands for setting environment variables before
#' running the build script.
#'
#' @param env_var Named list of environment variables.
#' @return A string of `export` commands with line breaks, or `""` if none.
#' @noRd
build_env_exports <- function(env_var) {
  if (is.null(env_var) || length(env_var) == 0) {
    return("")
  }

  exports <- sprintf("export %s=%s", names(env_var), env_var)
  paste0(paste(exports, collapse = "\n"), "\n")
}

#' Build Nix Src Part
#'
#' Creates the `src` attribute for the Nix derivation. Handles both remote URLs
#' (using `nix-prefetch-url`) and local paths (with optional user function
#' scripts).
#'
#' @param path Character path to file or directory.
#' @param user_functions Character vector of additional files to include.
#' @return A string with the `src` Nix expression.
#' @noRd
build_src_part <- function(path, user_functions = character(0)) {
  if (is_remote_url(path)) {
    if (length(user_functions) > 0) {
      return(build_hybrid_src(path, user_functions))
    } else {
      return(build_remote_src(path))
    }
  } else {
    return(build_local_src(path, user_functions))
  }
}

#' Build Hybrid Source Configuration for Remote URL + Local User Functions
#' @param url Character URL
#' @param user_functions Character vector of additional files
#' @return Character Nix expression
#' @noRd
build_hybrid_src <- function(url, user_functions) {
  hash <- tryCatch(
    system(paste("nix-prefetch-url", shQuote(url)), intern = TRUE),
    error = function(e) stop("Failed to run nix-prefetch-url: ", e$message)
  )

  if (length(hash) == 0 || hash == "") {
    stop("nix-prefetch-url did not return a hash for URL: ", url)
  }

  remote_filename <- basename(url)

  # Create the local fileset for user functions
  user_files_list <- paste0("./", user_functions, collapse = " ")
  local_src_var <- sprintf(
    "localSrc = defaultPkgs.lib.fileset.toSource {\n        root = ./.;\n        fileset = defaultPkgs.lib.fileset.unions [ %s ];\n      }",
    user_files_list
  )

  # Create a single robust copy command for all user functions
  user_files_cp <- "cp -r ${localSrc}/* $out/"

  sprintf(
    "(let\n      %s;\n    in defaultPkgs.runCommand \"combined-src\" {} ''\n      mkdir -p $out\n      cp ${defaultPkgs.fetchurl {\n        url = \"%s\";\n        sha256 = \"%s\";\n      }} $out/%s\n      %s\n    '')",
    local_src_var,
    url,
    trimws(hash[1]),
    remote_filename,
    user_files_cp
  )
}

#' Check If Path Is a Remote URL
#' @param path Character path
#' @return Logical
#' @noRd
is_remote_url <- function(path) {
  grepl("^https?://", path)
}

#' Build Remote Source Configuration
#' @param url Character URL
#' @return Character Nix expression
#' @noRd
build_remote_src <- function(url) {
  hash <- tryCatch(
    system(paste("nix-prefetch-url", shQuote(url)), intern = TRUE),
    error = function(e) stop("Failed to run nix-prefetch-url: ", e$message)
  )

  if (length(hash) == 0 || hash == "") {
    stop("nix-prefetch-url did not return a hash for URL: ", url)
  }

  sprintf(
    "defaultPkgs.fetchurl {\n      url = \"%s\";\n      sha256 = \"%s\";\n    }",
    url,
    trimws(hash[1])
  )
}

#' Build Local Source Configuration
#' @param path Character path
#' @param user_functions Character vector of additional files
#' @return Character Nix expression
#' @noRd
build_local_src <- function(path, user_functions) {
  all_files <- c(path, user_functions)
  all_files <- all_files[nzchar(all_files)]
  fileset_nix <- paste0("./", all_files, collapse = " ")

  sprintf(
    "defaultPkgs.lib.fileset.toSource {\n      root = ./.;\n      fileset = defaultPkgs.lib.fileset.unions [ %s ];\n    }",
    fileset_nix
  )
}

#' Sanitize Nix Environment String
#'
#' Produces a base identifier string by replacing invalid characters and
#' stripping suffixes.
#'
#' @param nix_env Character path to a Nix environment file (e.g., `"default.nix"`).
#' @return Sanitized base string.
#' @noRd
sanitize_nix_env <- function(nix_env) {
  # Use basename to handle relative paths (e.g., "../../default.nix" -> "default.nix")
  base <- gsub("[^a-zA-Z0-9]", "_", basename(nix_env))
  sub("_nix$", "", base)
}

#' Build Language-Specific Execution Commands
#' @param lang Language string
#' @param read_func String representing the function to call
#' @param user_code Source/import statements for user functions
#' @param out_name Name of the output object
#' @param rel_path Relative path to input file
#' @param serialize_str Language-specific serializer identifier or NULL for default
#' @return Character vector of commands
#' @noRd
build_language_commands <- function(
  lang,
  read_func,
  user_code,
  out_name,
  rel_path,
  serialize_str = NULL
) {
  user_lines <- if (nzchar(user_code)) {
    strsplit(user_code, "\n")[[1]]
  } else {
    character(0)
  }

  switch(
    lang,
    R = {
      ser_fun <- if (is.null(serialize_str)) "saveRDS" else serialize_str
      ser_call <- sprintf("%s(data, '%s')", ser_fun, out_name)
      c(
        "Rscript -e \"",
        "source('libraries.R')",
        user_lines,
        sprintf("data <- do.call(%s, list('%s'))", read_func, rel_path),
        paste0(ser_call, "\"")
      )
    },
    Py = {
      if (is.null(serialize_str)) {
        ser_lines <- c(
          sprintf("with open('%s', 'wb') as f:", out_name),
          "    pickle.dump(data, f)"
        )
      } else {
        ser_lines <- c(sprintf("%s(data, '%s')", serialize_str, out_name))
      }
      c(
        "python -c \"",
        "exec(open('libraries.py').read())",
        user_lines,
        sprintf("file_path = '%s'", rel_path),
        sprintf("data = eval('%s')(file_path)", read_func),
        ser_lines,
        "\""
      )
    },
    Jl = {
      # Escape double quotes within -e string
      user_lines_jl <- gsub("'", "\\\"", user_lines, fixed = TRUE)
      read_line <- sprintf("data = %s(\\\"%s\\\")", read_func, rel_path)
      ser_line <- if (is.null(serialize_str)) {
        paste0(
          "using Serialization; ",
          "io = open(\\\"",
          out_name,
          "\\\", \\\"w\\\"); ",
          "serialize(io, data); ",
          "close(io)"
        )
      } else {
        sprintf("%s(data, \\\"%s\\\")", serialize_str, out_name)
      }
      c(
        "julia -e \"",
        "if isfile(\\\"libraries.jl\\\"); include(\\\"libraries.jl\\\"); end;",
        user_lines_jl,
        read_line,
        ser_line,
        "\""
      )
    },
    stop("Unsupported lang: ", lang)
  )
}

#' Build build_phase Command
#'
#' Constructs the build-phase shell command for R, Python, or Julia.
#'
#' @param lang `"R"`, `"Py"`, or `"Jl"`.
#' @param read_func String representing the function to call for reading data.
#' @param user_code Source/import statements for user functions.
#' @param out_name Name of the output object (RDS/pickle file).
#' @param path Input path (file or folder).
#' @param user_functions Character vector of user function files.
#' @param serialize_str Language-specific serializer identifier or NULL for default.
#' @return A string with the build phase commands.
#' @noRd
build_phase <- function(
  lang,
  read_func,
  user_code,
  out_name,
  path,
  user_functions = character(0),
  serialize_str = NULL
) {
  if (is_remote_url(path)) {
    rel_name <- basename(path)
    rel_path <- rel_name

    if (length(user_functions) > 0) {
      # Hybrid case: everything is in $src, just organize it
      copy_line <- paste(
        "cp $src/* .",
        "mkdir -p input_folder",
        paste(
          sprintf("cp %s input_folder/", basename(user_functions)),
          collapse = "\n"
        ),
        sep = "\n"
      )
    } else {
      # Pure remote case
      copy_line <- paste0("cp -r $src ", rel_path)
    }
  } else {
    # Local case
    rel_name <- path
    rel_path <- file.path("input_folder", rel_name)
    copy_line <- "cp -r $src input_folder"
  }

  lang_commands <- build_language_commands(
    lang,
    read_func,
    user_code,
    out_name,
    rel_path,
    serialize_str
  )
  all_commands <- c(copy_line, lang_commands)

  paste(all_commands, collapse = "\n")
}

#' Process Read Function for Different Languages
#' @param read_function Function or character
#' @param lang Language string
#' @param parent_env Environment from calling function for proper substitution
#' @return Character string
#' @noRd
process_read_function <- function(
  read_function,
  lang,
  parent_env = parent.frame()
) {
  switch(
    lang,
    R = {
      # Get the actual expression from the parent environment
      read_expr <- substitute(read_function, parent_env)

      # Convert to string, handling both symbols and function definitions
      if (is.symbol(read_expr)) {
        # Simple function name like readRDS
        as.character(read_expr)
      } else {
        # Function definition or call - need to deparse properly
        func_str <- deparse1(read_expr)
        # Join multi-line deparsed functions and clean up
        func_str <- paste(func_str, collapse = " ")
        # Replace double quotes with single quotes for Nix compatibility
        gsub("\"", "'", func_str, fixed = TRUE)
      }
    },
    Py = {
      if (!is.character(read_function) || length(read_function) != 1) {
        stop("For Python, read_function must be a single string")
      }
      gsub("'", "\\'", read_function, fixed = TRUE)
    },
    Jl = {
      read_expr <- substitute(read_function, parent_env)
      if (is.symbol(read_expr)) {
        as.character(read_expr)
      } else {
        func_str <- deparse1(read_expr)
        func_str <- paste(func_str, collapse = " ")
        gsub("\"", "", func_str, fixed = TRUE)
      }
    },
    stop("Unsupported lang: ", lang)
  )
}

#' Process Serialize Function for Different Languages
#'
#' Mirrors the behaviour of rxp_r/rxp_py/rxp_jl:
#' - R: supports bare symbol, character, or function; defaults to saveRDS
#' - Python: character function name, defaults to pickle.dump via with-open
#' - Julia: character function name, defaults to Serialization.serialize
#'
#' @param encoder Language-specific serializer (see above)
#' @param lang Language string
#' @param parent_env Environment for proper substitution (R only)
#' @return For R: function name string; For Py/Jl: NULL (use default) or function name string
#' @noRd
process_encoder <- function(
  encoder,
  lang,
  parent_env = parent.frame()
) {
  switch(
    lang,
    R = {
      serialize_expr <- substitute(encoder, parent_env)
      if (identical(serialize_expr, quote(NULL))) {
        "saveRDS"
      } else if (is.character(serialize_expr)) {
        serialize_expr
      } else {
        deparse1(serialize_expr)
      }
    },
    Py = {
      if (is.null(encoder)) {
        NULL
      } else {
        if (!is.character(encoder) || length(encoder) != 1) {
          stop(
            "For Python, encoder must be a single character string or NULL"
          )
        }
        encoder
      }
    },
    Jl = {
      if (is.null(encoder)) {
        NULL
      } else {
        if (!is.character(encoder) || length(encoder) != 1) {
          stop(
            "For Julia, encoder must be a single character string or NULL"
          )
        }
        encoder
      }
    },
    stop("Unsupported lang: ", lang)
  )
}

#' Generic Nix Expression Builder for R, Python, and Julia Data Readers
#'
#' You should not call it directly, but instead use one of `rxp_r_file()`,
#' `rxp_py_file()` or `rxp_jl_file()`.
#'
#' Creates a Nix derivation that reads a file or folder of data using R,
#' Python, or Julia. Handles user-defined functions, environment variables, and Nix
#' environment specification.
#'
#' @param lang `"R"`, `"Py"` or `"Jl"`.
#' @param name Symbol, the name of the derivation.
#' @param path Character, the file path to include (e.g., "data/mtcars.shp") or
#'   a folder path (e.g., "data"). See details.
#' @param read_function Function, an R function to read the data, taking one
#'   argument (the path). This can be a user-defined function that is made available
#'   using `user_functions`. See details.
#' @param user_functions Character vector, user-defined functions to include.
#'   This should be a script (or scripts) containing user-defined functions
#'   to include during the build process for this derivation. It is recommended
#'   to use one script per function, and only include the required script(s) in
#'   the derivation.
#' @param nix_env Character, path to the Nix environment file, default is
#'   "default.nix".
#' @param env_var List, defaults to NULL. A named list of environment variables
#'   to set before running the R script, e.g., c(VAR = "hello"). Each entry will
#'   be added as an export statement in the build phase.
#' @param encoder Function/character, defaults to NULL.
#'   A language-specific serializer to write the loaded object to disk.
#'   - R: function/symbol/character (e.g., `qs::qsave`) taking `(object, path)`. Defaults to `saveRDS`.
#'   - Python: character name of a function taking `(object, path)`. Defaults to using `pickle.dump`.
#'   - Julia: character name of a function taking `(object, path)`. Defaults to using `Serialization.serialize`.
#' @return An object of class `rxp_derivation`.
#' @keywords internal
rxp_file <- function(
  lang,
  name,
  path,
  read_function,
  user_functions = "",
  nix_env = "default.nix",
  env_var = NULL,
  encoder = NULL
) {
  out_name <- deparse1(substitute(name))
  path <- gsub("/+$", "", path)
  user_functions <- clean_user_functions(user_functions)
  read_func_str <- process_read_function(read_function, lang, environment())
  serialize_str <- process_encoder(
    encoder,
    lang,
    environment()
  )

  # Build components
  user_code <- build_user_code_cmd(user_functions, lang)
  env_exports <- build_env_exports(env_var)
  bp <- build_phase(
    lang,
    read_func_str,
    user_code,
    out_name,
    path,
    user_functions,
    serialize_str
  )

  # Add environment exports if present
  if (nzchar(env_exports)) {
    bp <- paste(env_exports, bp, sep = "\n")
  }

  # Build source and base identifiers
  src_part <- build_src_part(path, user_functions)
  base <- sanitize_nix_env(nix_env)

  derivation_type <- switch(
    lang,
    "R" = "rxp_r",
    "Py" = "rxp_py",
    "Jl" = "rxp_jl",
    stop("Unknown derivation type for lang: ", lang)
  )

  snippet <- make_derivation_snippet(
    out_name = out_name,
    src_snippet = sprintf("    src = %s;\n", src_part),
    base = base,
    build_phase = bp,
    derivation_type = derivation_type
  )

  # For display/debugging purposes, expose the effective serializer
  effective_serializer <- if (is.null(serialize_str)) {
    switch(
      lang,
      R = "saveRDS",
      Py = "pickle.dump",
      Jl = "Serialization.serialize"
    )
  } else {
    serialize_str
  }

  create_rxp_derivation(
    out_name,
    snippet,
    lang,
    user_functions,
    nix_env,
    env_var,
    encoder = effective_serializer
  )
}

#' Create rxp_derivation Object
#' @param out_name Character output name
#' @param snippet Character Nix snippet
#' @param lang Character language
#' @param user_functions Character vector
#' @param nix_env Character nix environment
#' @param env_var Named list of environment variables
#' @param encoder Character serializer identifier (for display)
#' @return rxp_derivation object
#' @noRd
create_rxp_derivation <- function(
  out_name,
  snippet,
  lang,
  user_functions,
  nix_env,
  env_var,
  encoder = NULL
) {
  structure(
    list(
      name = out_name,
      snippet = snippet,
      type = paste0("rxp_", tolower(lang)),
      additional_files = "",
      user_functions = user_functions,
      nix_env = nix_env,
      env_var = env_var,
      encoder = encoder
    ),
    class = "rxp_derivation"
  )
}

#' Creates a Nix Expression That Reads In a File (or Folder of Data) Using R
#'
#' @family derivations
#' @return An object of class `rxp_derivation`.
#' @inheritDotParams rxp_file name:encoder
#' @details The basic usage is to provide a path to a file, and the function
#'   to read it. For example: `rxp_r_file(mtcars, path = "data/mtcars.csv", read_function = read.csv)`.
#'   It is also possible instead to point to a folder that contains many
#'   files that should all be read at once, for example:
#'   `rxp_r_file(many_csvs, path = "data", read_function = \(x)(readr::read_csv(list.files(x, full.names = TRUE, pattern = ".csv$"))))`.
#'   See the `vignette("importing-data")` vignette for more detailed examples.
#' @export
rxp_r_file <- function(...) rxp_file("R", ...)

#' Creates a Nix Expression That Reads In a File (or Folder of Data) Using Python
#'
#' @family derivations
#' @return An object of class `rxp_derivation`.
#' @inheritDotParams rxp_file name:encoder
#' @details The basic usage is to provide a path to a file, and the function
#'   to read it. For example: `rxp_r_file(mtcars, path = "data/mtcars.csv", read_function = read.csv)`.
#'   It is also possible instead to point to a folder that contains many
#'   files that should all be read at once, for example:
#'   `rxp_r_file(many_csvs, path = "data", read_function = \(x)(readr::read_csv(list.files(x, full.names = TRUE, pattern = ".csv$"))))`
#'   See the `vignette("importing-data")` vignette for more detailed examples.
#' @export
rxp_py_file <- function(...) rxp_file("Py", ...)

#' Creates a Nix Expression That Reads In a File (or Folder of Data) Using Julia
#'
#' @family derivations
#' @return An object of class `rxp_derivation`.
#' @inheritDotParams rxp_file name:encoder
#' @details The basic usage is to provide a path to a file, and the function
#'   to read it. For example: `rxp_r_file(mtcars, path = "data/mtcars.csv", read_function = read.csv)`.
#'   It is also possible instead to point to a folder that contains many
#'   files that should all be read at once, for example:
#'   `rxp_r_file(many_csvs, path = "data", read_function = \(x)(readr::read_csv(list.files(x, full.names = TRUE, pattern = ".csv$"))))`.
#'   See the `vignette("importing-data")` vignette for more detailed examples.
#' @export
rxp_jl_file <- function(...) rxp_file("Jl", ...)

#' Generate the Nix Derivation Snippet for Python-R Object Transfer.
#'
#' This function constructs the `build_phase` and Nix derivation snippet
#' based on the given parameters.
#'
#' @param out_name Character, name of the derivation.
#' @param expr_str Character, name of the object being transferred.
#' @param nix_env Character, path to the Nix environment file.
#' @param direction Character, either "py2r" (Python to R) or "r2py" (R to
#'   Python).
#' @param project_path Character, path to the project root.
#' @return A list with elements: `name`, `snippet`, `type`, `additional_files`,
#'   `nix_env`.
#' @noRd
rxp_common_setup <- function(
  out_name,
  expr_str,
  nix_env,
  direction,
  project_path = "."
) {
  expr_str <- gsub("\"", "'", expr_str)
  base <- sanitize_nix_env(nix_env)

  r_command <- build_transfer_command(out_name, expr_str, direction)
  build_phase <- build_reticulate_phase(r_command, nix_env, project_path)

  snippet <- make_derivation_snippet(
    out_name = out_name,
    src_snippet = "",
    base = base,
    build_phase = build_phase,
    derivation_type = "rxp_r"
  )

  structure(
    list(
      name = out_name,
      snippet = snippet,
      type = paste0("rxp_", direction),
      additional_files = "",
      nix_env = nix_env
    ),
    class = "rxp_derivation"
  )
}


#' Build Transfer Command for py2r or r2py
#' @param out_name Character output name
#' @param expr_str Character expression string
#' @param direction Character direction
#' @return Character command
#' @noRd
build_transfer_command <- function(out_name, expr_str, direction) {
  switch(
    direction,
    py2r = sprintf(
      "         %s <- reticulate::py_load_object('${%s}/%s', pickle = 'pickle', convert = TRUE)\n         saveRDS(%s, '%s')",
      out_name,
      expr_str,
      expr_str,
      out_name,
      out_name
    ),
    r2py = sprintf(
      "         %s <- readRDS('${%s}/%s')\n         reticulate::py_save_object(%s, '%s', pickle = 'pickle')",
      expr_str,
      expr_str,
      expr_str,
      expr_str,
      out_name
    ),
    stop("Invalid direction. Use 'py2r' or 'r2py'.")
  )
}

#' Build Reticulate Build Phase
#' @param r_command Character R command
#' @param nix_env Character nix environment file
#' @param project_path Character project path
#' @return Character build phase
#' @noRd
build_reticulate_phase <- function(
  r_command,
  nix_env = "default.nix",
  project_path = "."
) {
  # Extract the Python version from the nix file
  python_version <- extract_python_version(nix_env, project_path)

  sprintf(
    "export RETICULATE_PYTHON=${defaultPkgs.%s}/bin/python\n       Rscript -e \"\n         source('libraries.R')\n%s\"",
    python_version,
    r_command
  )
}


#' Transfer Python Object into an R Session
#'
#' @family interop functions
#' @param name Symbol, name of the derivation.
#' @param expr Symbol, Python object to be loaded into R.
#' @param nix_env Character, path to the Nix environment file, default is
#'   "default.nix".
#' @param project_path Character, path to the project root, default is ".".
#' @details `rxp_py2r(my_obj, my_python_object)` loads a serialized Python
#'   object and saves it as an RDS file using `reticulate::py_load_object()`.
#' @return An object of class `rxp_derivation`.
#' @export
rxp_py2r <- function(name, expr, nix_env = "default.nix", project_path = ".") {
  out_name <- deparse1(substitute(name))
  expr_str <- deparse1(substitute(expr))
  rxp_common_setup(out_name, expr_str, nix_env, "py2r", project_path)
}


#' Transfer R Object into a Python Session
#'
#' @family interop functions
#' @param name Symbol, name of the derivation.
#' @param expr Symbol, R object to be saved into a Python pickle.
#' @param nix_env Character, path to the Nix environment file, default is
#'   "default.nix".
#' @param project_path Character, path to the project root, default is ".".
#' @details `rxp_r2py(my_obj, my_r_object)` saves an R object to a Python pickle
#'   using `reticulate::py_save_object()`.
#' @return An object of class `rxp_derivation`.
#' @export
rxp_r2py <- function(name, expr, nix_env = "default.nix", project_path = ".") {
  out_name <- deparse1(substitute(name))
  expr_str <- deparse1(substitute(expr))
  rxp_common_setup(out_name, expr_str, nix_env, "r2py", project_path)
}

test_that("rxp_r: generates correct list", {
  d1 <- rxp_r(mtcars_am, dplyr::filter(mtcars, am == 1))
  testthat::expect_equal(
    d1,
    structure(
      list(
        "name" = "mtcars_am",
        "snippet" = '  mtcars_am = makeRDerivation {\n    name = \"mtcars_am\";\n    buildInputs = defaultBuildInputs;\n    configurePhase = defaultConfigurePhase;\n    buildPhase = \'\'\n      Rscript -e \"\n        source(\'libraries.R\')\n        # RIXPRESS_LOAD_DEPENDENCIES_HERE:mtcars_am\n        mtcars_am <- dplyr::filter(mtcars, am == 1)\n        saveRDS(mtcars_am, \'mtcars_am\')\"\n    \'\';\n  };',
        "type" = "rxp_r",
        "additional_files" = "",
        "user_functions" = "",
        "nix_env" = "default.nix",
        "encoder" = "saveRDS",
        "decoder" = "readRDS",
        "env_var" = NULL,
        "noop_build" = FALSE
      ),
      class = "rxp_derivation"
    )
  )
})

test_that("rxp_py: generates correct list", {
  d1 <- rxp_py(mtcars_pl_am, "mtcars_pl.filter(pl.col('am') == 1)")
  testthat::expect_equal(
    d1,
    structure(
      list(
        "name" = "mtcars_pl_am",
        "snippet" = '  mtcars_pl_am = makePyDerivation {\n    name = \"mtcars_pl_am\";\n    buildInputs = defaultBuildInputs;\n    configurePhase = defaultConfigurePhase;\n    buildPhase = \'\'\n      python -c "\nexec(open(\'libraries.py\').read())\n# RIXPRESS_PY_LOAD_DEPENDENCIES_HERE:mtcars_pl_am\nexec(\'mtcars_pl_am = mtcars_pl.filter(pl.col(\\\'am\\\') == 1)\')\nwith open(\'mtcars_pl_am\', \'wb\') as f: pickle.dump(globals()[\'mtcars_pl_am\'], f)\n"\n    \'\';\n  };',
        "type" = "rxp_py",
        "additional_files" = "",
        "user_functions" = "",
        "nix_env" = "default.nix",
        "encoder" = "with open('mtcars_pl_am', 'wb') as f: pickle.dump(globals()['mtcars_pl_am'], f)",
        "decoder" = "pickle.load",
        "env_var" = NULL,
        "noop_build" = FALSE
      ),
      class = "rxp_derivation"
    )
  )
})

test_that("rxp_py: custom serialization functions work", {
  d1 <- rxp_py(
    mtcars_pl_am,
    "mtcars_pl.filter(pl.col('am') == 1)",
    encoder = "custom_save",
    decoder = "custom_load"
  )

  # Test the entire object
  testthat::expect_equal(
    d1,
    structure(
      list(
        "name" = "mtcars_pl_am",
        "snippet" = '  mtcars_pl_am = makePyDerivation {\n    name = \"mtcars_pl_am\";\n    buildInputs = defaultBuildInputs;\n    configurePhase = defaultConfigurePhase;\n    buildPhase = \'\'\n      python -c "\nexec(open(\'libraries.py\').read())\n# RIXPRESS_PY_LOAD_DEPENDENCIES_HERE:mtcars_pl_am\nexec(\'mtcars_pl_am = mtcars_pl.filter(pl.col(\\\'am\\\') == 1)\')\ncustom_save(globals()[\'mtcars_pl_am\'], \'mtcars_pl_am\')\n"\n    \'\';\n  };',
        "type" = "rxp_py",
        "additional_files" = "",
        "user_functions" = "",
        "nix_env" = "default.nix",
        "encoder" = "custom_save(globals()['mtcars_pl_am'], 'mtcars_pl_am')",
        "decoder" = "custom_load",
        "env_var" = NULL,
        "noop_build" = FALSE
      ),
      class = "rxp_derivation"
    )
  )
})

test_that("rxp_qmd: generates correct list", {
  # Create a temporary qmd file for testing
  qmd_file <- tempfile(fileext = ".qmd")
  on.exit(unlink(qmd_file))
  writeLines(
    "---\ntitle: Test\n---\n\nThis is a test quarto document with rxp_read(\"test_data\").",
    qmd_file
  )

  d1 <- suppressWarnings(rxp_qmd(
    report,
    qmd_file,
    additional_files = "images",
    args = "--to pdf"
  ))

  testthat::expect_equal(
    d1,
    structure(
      list(
        "name" = "report",
        "snippet" = paste0(
          '  report = defaultPkgs.stdenv.mkDerivation {\n    name = "report";\n    src = defaultPkgs.lib.fileset.toSource {\n      root = ./.;\n      fileset = defaultPkgs.lib.fileset.unions [ ./',
          qmd_file,
          ' ./images ];\n    };\n    buildInputs = defaultBuildInputs;\n    configurePhase = defaultConfigurePhase;\n    buildPhase = \'\'\n      mkdir home\n      export HOME=$PWD/home\n      export RETICULATE_PYTHON=${defaultPkgs.python3}/bin/python\n\n      substituteInPlace ',
          qmd_file,
          ' --replace-fail \'rxp_read("test_data")\' \'rxp_read("${test_data}")\'\n      quarto render ',
          qmd_file,
          ' --to pdf --output-dir $out\n    \'\';\n  };'
        ),
        "type" = "rxp_qmd",
        "qmd_file" = qmd_file,
        "additional_files" = "images",
        "nix_env" = "default.nix",
        "args" = "--to pdf",
        "env_var" = NULL,
        "noop_build" = FALSE
      ),
      class = "rxp_derivation"
    )
  )
})

test_that("rxp_r_file: single file reading works", {
  # Create a temporary CSV for testing
  csv_file <- tempfile(fileext = ".csv")
  on.exit(unlink(csv_file))

  write.csv(mtcars[1:5, ], csv_file, row.names = FALSE)

  d1 <- rxp_r_file(mtcars_data, path = csv_file, read_function = read.csv)

  # Create a subset of d1 with only the fields we want to test
  d1_subset <- list(
    "name" = d1$name,
    "type" = d1$type,
    "path" = csv_file,
    "read_function" = read.csv,
    "nix_env" = d1$nix_env
  )
  class(d1_subset) <- "rxp_derivation"

  testthat::expect_equal(
    d1_subset,
    structure(
      list(
        "name" = "mtcars_data",
        "type" = "rxp_r",
        "path" = csv_file,
        "read_function" = read.csv,
        "nix_env" = "default.nix"
      ),
      class = "rxp_derivation"
    )
  )
})

test_that("rxp_r_file: folder reading works", {
  # Create a temporary folder with data for testing
  test_dir <- tempdir()
  test_data_dir <- file.path(test_dir, "test_data")
  dir.create(test_data_dir, showWarnings = FALSE)
  # Cleanup
  on.exit(unlink(test_data_dir, recursive = TRUE))

  write.csv(
    mtcars[1:5, ],
    file.path(test_data_dir, "data1.csv"),
    row.names = FALSE
  )
  write.csv(
    mtcars[6:10, ],
    file.path(test_data_dir, "data2.csv"),
    row.names = FALSE
  )

  d1 <- rxp_r_file(
    all_data,
    path = test_data_dir,
    read_function = function(x) {
      lapply(list.files(x, full.names = TRUE, pattern = "\\.csv$"), read.csv)
    },
  )

  # Create a subset of d1 with only the fields we want to test
  d1_subset <- list(
    "name" = d1$name,
    "type" = d1$type,
    "path" = test_data_dir,
    "read_function" = function(x) {
      lapply(list.files(x, full.names = TRUE, pattern = "\\.csv$"), read.csv)
    },
    "nix_env" = d1$nix_env
  )
  class(d1_subset) <- "rxp_derivation"

  # Compare function bodies instead of the functions themselves
  testthat::expect_equal(
    body(d1_subset$read_function),
    body(
      function(x) {
        lapply(list.files(x, full.names = TRUE, pattern = "\\.csv$"), read.csv)
      }
    )
  )

  # Remove read_function for direct comparison
  d1_subset$read_function <- NULL
  expected <- structure(
    list(
      "name" = "all_data",
      "type" = "rxp_r",
      "path" = test_data_dir,
      "nix_env" = "default.nix"
    ),
    class = "rxp_derivation"
  )

  testthat::expect_equal(d1_subset, expected)
})

test_that("rxp_py_file: basic functionality works", {
  # Create a temporary CSV for testing
  csv_file <- tempfile(fileext = ".csv")
  on.exit(unlink(csv_file))
  write.csv(mtcars[1:5, ], csv_file, row.names = FALSE)

  d1 <- rxp_py_file(
    mtcars_data,
    path = csv_file,
    read_function = "pandas.read_csv"
  )

  # Create a subset of d1 with only the fields we want to test
  d1_subset <- list(
    "name" = d1$name,
    "type" = d1$type,
    "path" = csv_file,
    "read_function" = "pandas.read_csv",
    "nix_env" = d1$nix_env
  )
  class(d1_subset) <- "rxp_derivation"

  testthat::expect_equal(
    d1_subset,
    structure(
      list(
        "name" = "mtcars_data",
        "type" = "rxp_py",
        "path" = csv_file,
        "read_function" = "pandas.read_csv",
        "nix_env" = "default.nix"
      ),
      class = "rxp_derivation"
    )
  )
})

# Mock reticulate package for testing
mockReticulate <- function() {
  if (!exists("reticulate", envir = globalenv())) {
    reticulate <- new.env()
    assign("reticulate", reticulate, envir = globalenv())
  }
}

test_that("rxp_py2r: generates correct list", {
  mockReticulate()

  d1 <- suppressWarnings(rxp_py2r(r_data, py_data))

  # Create a subset of d1 with only the fields we want to test
  d1_subset <- list(
    "name" = d1$name,
    "type" = d1$type,
    "nix_env" = d1$nix_env
  )
  class(d1_subset) <- "rxp_derivation"

  testthat::expect_equal(
    d1_subset,
    structure(
      list(
        "name" = "r_data",
        "type" = "rxp_py2r",
        "nix_env" = "default.nix"
      ),
      class = "rxp_derivation"
    )
  )
})

test_that("rxp_r2py: generates correct list", {
  mockReticulate()

  d1 <- suppressWarnings(rxp_r2py(py_data, r_data))

  # Test the entire object
  testthat::expect_equal(
    d1,
    structure(
      list(
        "name" = "py_data",
        "snippet" = '  py_data = makeRDerivation {\n    name = "py_data";\n    buildInputs = defaultBuildInputs;\n    configurePhase = defaultConfigurePhase;\n    buildPhase = \'\'\n      export RETICULATE_PYTHON=${defaultPkgs.python3}/bin/python\n       Rscript -e "\n         source(\'libraries.R\')\n         r_data <- readRDS(\'${r_data}/r_data\')\n         reticulate::py_save_object(r_data, \'py_data\', pickle = \'pickle\')"\n    \'\';\n  };',
        "type" = "rxp_r2py",
        "additional_files" = "",
        "nix_env" = "default.nix"
      ),
      class = "rxp_derivation"
    )
  )
})

test_that("print.derivation: outputs expected format", {
  d1 <- rxp_r(mtcars_am, dplyr::filter(mtcars, am == 1))

  # Capture the output of print.derivation
  output <- capture.output(print(d1))

  # Create the expected output
  expected_output <- c(
    "Name: mtcars_am",
    "Type: rxp_r",
    "Serialize function: saveRDS",
    "Unserialize function: readRDS"
  )

  # Check that all expected lines are in the output
  testthat::expect_equal(
    all(sapply(expected_output, function(line) any(grepl(line, output)))),
    TRUE
  )
})

test_that("rxp_r: with additional files", {
  d1 <- rxp_r(
    mtcars_am,
    dplyr::filter(mtcars, am == 1),
    additional_files = "data.csv",
    user_functions = "functions.R"
  )

  # Test the entire object
  testthat::expect_equal(
    d1,
    structure(
      list(
        "name" = "mtcars_am",
        "snippet" = '  mtcars_am = makeRDerivation {\n    name = "mtcars_am";\n     src = defaultPkgs.lib.fileset.toSource {\n      root = ./.;\n      fileset = defaultPkgs.lib.fileset.unions [ ./data.csv ./functions.R ];\n    };\n    buildInputs = defaultBuildInputs;\n    configurePhase = defaultConfigurePhase;\n    buildPhase = \'\'\n      cp -r $src/* .\n      Rscript -e "\n        source(\'libraries.R\')\n        source(\'functions.R\')\n        # RIXPRESS_LOAD_DEPENDENCIES_HERE:mtcars_am\n        mtcars_am <- dplyr::filter(mtcars, am == 1)\n        saveRDS(mtcars_am, \'mtcars_am\')"\n    \'\';\n  };',
        "type" = "rxp_r",
        "additional_files" = "data.csv",
        "user_functions" = "functions.R",
        "nix_env" = "default.nix",
        "encoder" = "saveRDS",
        "decoder" = "readRDS",
        "env_var" = NULL,
        "noop_build" = FALSE
      ),
      class = "rxp_derivation"
    )
  )
})

test_that("rxp_py: with additional files", {
  d1 <- rxp_py(
    mtcars_pl_am,
    "mtcars_pl.filter(pl.col('am') == 1)",
    additional_files = "data.csv",
    user_functions = "functions.py"
  )

  # Test the entire object
  testthat::expect_equal(
    d1,
    structure(
      list(
        "name" = "mtcars_pl_am",
        "snippet" = '  mtcars_pl_am = makePyDerivation {\n    name = "mtcars_pl_am";\n     src = defaultPkgs.lib.fileset.toSource {\n      root = ./.;\n      fileset = defaultPkgs.lib.fileset.unions [ ./data.csv ./functions.py ];\n    };\n    buildInputs = defaultBuildInputs;\n    configurePhase = defaultConfigurePhase;\n    buildPhase = \'\'\n      cp -r $src/* .\n      python -c "\nexec(open(\'libraries.py\').read())\nexec(open(\'functions.py\').read())\n# RIXPRESS_PY_LOAD_DEPENDENCIES_HERE:mtcars_pl_am\nexec(\'mtcars_pl_am = mtcars_pl.filter(pl.col(\\\'am\\\') == 1)\')\nwith open(\'mtcars_pl_am\', \'wb\') as f: pickle.dump(globals()[\'mtcars_pl_am\'], f)\n"\n    \'\';\n  };',
        "type" = "rxp_py",
        "additional_files" = "data.csv",
        "user_functions" = "functions.py",
        "nix_env" = "default.nix",
        "encoder" = "with open('mtcars_pl_am', 'wb') as f: pickle.dump(globals()['mtcars_pl_am'], f)",
        "decoder" = "pickle.load",
        "env_var" = NULL,
        "noop_build" = FALSE
      ),
      class = "rxp_derivation"
    )
  )
})

test_that("rxp_r: with env_var parameter", {
  d1 <- rxp_r(
    mtcars_am,
    dplyr::filter(mtcars, am == 1),
    env_var = c(MY_VAR = "test_value", ANOTHER_VAR = "123")
  )

  # Test the entire object
  testthat::expect_equal(
    d1,
    structure(
      list(
        "name" = "mtcars_am",
        "snippet" = '  mtcars_am = makeRDerivation {\n    name = "mtcars_am";\n    buildInputs = defaultBuildInputs;\n    configurePhase = defaultConfigurePhase;\n    buildPhase = \'\'\n      export MY_VAR=test_value\n      export ANOTHER_VAR=123\n      Rscript -e "\n        source(\'libraries.R\')\n        # RIXPRESS_LOAD_DEPENDENCIES_HERE:mtcars_am\n        mtcars_am <- dplyr::filter(mtcars, am == 1)\n        saveRDS(mtcars_am, \'mtcars_am\')"\n    \'\';\n  };',
        "type" = "rxp_r",
        "additional_files" = "",
        "user_functions" = "",
        "nix_env" = "default.nix",
        "encoder" = "saveRDS",
        "decoder" = "readRDS",
        "env_var" = c(MY_VAR = "test_value", ANOTHER_VAR = "123"),
        "noop_build" = FALSE
      ),
      class = "rxp_derivation"
    )
  )
})

test_that("rxp_common_setup: handles invalid direction", {
  testthat::expect_error(
    rxp_common_setup(
      "test_name",
      "test_expr",
      "default.nix",
      "invalid_direction"
    ),
    "Invalid direction. Use 'py2r' or 'r2py'."
  )
})

test_that("rxp_py: with env_var parameter", {
  d1 <- rxp_py(
    mtcars_pl_am,
    "mtcars_pl.filter(pl.col('am') == 1)",
    env_var = c(PYTHON_ENV = "production", DEBUG = "0")
  )

  # Test the entire object
  testthat::expect_equal(
    d1,
    structure(
      list(
        "name" = "mtcars_pl_am",
        "snippet" = '  mtcars_pl_am = makePyDerivation {\n    name = "mtcars_pl_am";\n    buildInputs = defaultBuildInputs;\n    configurePhase = defaultConfigurePhase;\n    buildPhase = \'\'\n      export PYTHON_ENV=production\n      export DEBUG=0\n      python -c "\nexec(open(\'libraries.py\').read())\n# RIXPRESS_PY_LOAD_DEPENDENCIES_HERE:mtcars_pl_am\nexec(\'mtcars_pl_am = mtcars_pl.filter(pl.col(\\\'am\\\') == 1)\')\nwith open(\'mtcars_pl_am\', \'wb\') as f: pickle.dump(globals()[\'mtcars_pl_am\'], f)\n"\n    \'\';\n  };',
        "type" = "rxp_py",
        "additional_files" = "",
        "user_functions" = "",
        "nix_env" = "default.nix",
        "encoder" = "with open('mtcars_pl_am', 'wb') as f: pickle.dump(globals()['mtcars_pl_am'], f)",
        "decoder" = "pickle.load",
        "env_var" = c(PYTHON_ENV = "production", DEBUG = "0"),
        "noop_build" = FALSE
      ),
      class = "rxp_derivation"
    )
  )
})

test_that("rxp_qmd: with env_var parameter", {
  # Create a temporary qmd file for testing
  qmd_file <- tempfile(fileext = ".qmd")
  on.exit(unlink(qmd_file))

  writeLines(
    "---\ntitle: Test\n---\n\nThis is a test quarto document with rxp_read(\"test_data\").",
    qmd_file
  )

  d1 <- suppressWarnings(rxp_qmd(
    report,
    qmd_file,
    additional_files = "images",
    args = "--to pdf",
    env_var = c(QUARTO_PROFILE = "production", QUARTO_RENDER_TOKEN = "abc123")
  ))

  testthat::expect_equal(
    d1,
    structure(
      list(
        "name" = "report",
        "snippet" = paste0(
          '  report = defaultPkgs.stdenv.mkDerivation {\n    name = "report";\n    src = defaultPkgs.lib.fileset.toSource {\n      root = ./.;\n      fileset = defaultPkgs.lib.fileset.unions [ ./',
          qmd_file,
          ' ./images ];\n    };\n    buildInputs = defaultBuildInputs;\n    configurePhase = defaultConfigurePhase;\n    buildPhase = \'\'\n      mkdir home\n      export HOME=$PWD/home\n      export RETICULATE_PYTHON=${defaultPkgs.python3}/bin/python\n      export QUARTO_PROFILE=production\n      export QUARTO_RENDER_TOKEN=abc123\n\n      substituteInPlace ',
          qmd_file,
          ' --replace-fail \'rxp_read("test_data")\' \'rxp_read("${test_data}")\'\n      quarto render ',
          qmd_file,
          ' --to pdf --output-dir $out\n    \'\';\n  };'
        ),
        "type" = "rxp_qmd",
        "qmd_file" = qmd_file,
        "additional_files" = "images",
        "nix_env" = "default.nix",
        "args" = "--to pdf",
        "env_var" = c(
          QUARTO_PROFILE = "production",
          QUARTO_RENDER_TOKEN = "abc123"
        ),
        "noop_build" = FALSE
      ),
      class = "rxp_derivation"
    )
  )
})

test_that("rxp_r_file: with env_var parameter", {
  # Create a temporary CSV for testing
  csv_file <- tempfile(fileext = ".csv")
  on.exit(unlink(csv_file))

  write.csv(mtcars[1:5, ], csv_file, row.names = FALSE)

  d1 <- rxp_r_file(
    mtcars_data,
    path = csv_file,
    read_function = read.csv,
    env_var = c(R_DATA_DIR = "/path/to/data", R_DEBUG = "TRUE")
  )

  # Test the entire object
  testthat::expect_equal(
    d1$env_var,
    c(R_DATA_DIR = "/path/to/data", R_DEBUG = "TRUE")
  )
})

test_that("rxp_py_file: with env_var parameter", {
  # Create a temporary CSV for testing
  csv_file <- tempfile(fileext = ".csv")
  on.exit(unlink(csv_file))

  write.csv(mtcars[1:5, ], csv_file, row.names = FALSE)

  d1 <- rxp_py_file(
    mtcars_data,
    path = csv_file,
    read_function = "pandas.read_csv",
    env_var = c(PYTHONPATH = "/custom/modules", PYTHON_DEBUG = "1")
  )

  # Test the entire object
  testthat::expect_equal(
    d1$env_var,
    c(PYTHONPATH = "/custom/modules", PYTHON_DEBUG = "1")
  )
})

test_that("rxp_rmd: with env_var parameter", {
  # Create a temporary Rmd file for testing
  rmd_file <- tempfile(fileext = ".Rmd")
  on.exit(unlink(rmd_file))

  writeLines(
    "---\ntitle: Test\n---\n\nThis is a test R Markdown document with rxp_read(\"test_data\").",
    rmd_file
  )

  d1 <- suppressWarnings(rxp_rmd(
    report,
    rmd_file,
    additional_files = "images",
    env_var = c(
      RSTUDIO_PANDOC = "/usr/local/bin/pandoc",
      R_LIBS_USER = "/custom/r/libs"
    )
  ))

  # Test the entire object
  testthat::expect_equal(
    d1,
    structure(
      list(
        "name" = "report",
        "snippet" = paste0(
          '  report = defaultPkgs.stdenv.mkDerivation {\n    name = "report";\n    src = defaultPkgs.lib.fileset.toSource {\n      root = ./.;\n      fileset = defaultPkgs.lib.fileset.unions [ ./',
          rmd_file,
          ' ./images ];\n    };\n    buildInputs = defaultBuildInputs;\n    configurePhase = defaultConfigurePhase;\n    buildPhase = \'\'\n      mkdir home\n      export HOME=$PWD/home\n      export RETICULATE_PYTHON=${defaultPkgs.python3}/bin/python\n      export RSTUDIO_PANDOC=/usr/local/bin/pandoc\n      export R_LIBS_USER=/custom/r/libs\n\n      substituteInPlace ',
          rmd_file,
          ' --replace-fail \'rxp_read("test_data")\' \'rxp_read("${test_data}")\'\n      Rscript -e "rmd_file <- \'',
          rmd_file,
          '\'; rmarkdown::render(input = file.path(\'$PWD\', rmd_file), output_dir = \'$out\')"\n    \'\';\n  };'
        ),
        "type" = "rxp_rmd",
        "rmd_file" = rmd_file,
        "additional_files" = "images",
        "nix_env" = "default.nix",
        "params" = NULL,
        "env_var" = c(
          RSTUDIO_PANDOC = "/usr/local/bin/pandoc",
          R_LIBS_USER = "/custom/r/libs"
        ),
        "noop_build" = FALSE
      ),
      class = "rxp_derivation"
    )
  )
})

#' Create a Named Pipeline of Derivations
#'
#' Groups multiple derivations into a named pipeline for organizational
#' purposes. This allows you to structure large projects into logical
#' sub-pipelines (e.g., "ETL", "Model", "Report") that are visually
#' distinguished in DAG visualizations.
#'
#' @family pipeline functions
#' @param name Character, the name of the pipeline (e.g., "ETL", "Model").
#' @param path Character path to an R script returning a list of derivations, OR
#'   a list of derivation objects created by `rxp_r()`, `rxp_py()`, etc.
#' @param color Character, optional. A CSS color name (e.g., "darkorange") or
#'   hex code (e.g., "#FF5733") to use for this pipeline's nodes in DAG
#'   visualizations. If NULL, a default color will be assigned.
#' @param ... Additional arguments (currently unused, reserved for future use).
#'
#' @return An object of class `rxp_pipeline` containing the derivations with
#'   pipeline metadata attached.
#'
#' @details
#' The `rxp_pipeline()` function is used to organize derivations into logical
#' groups. When passed to `rxp_populate()`, the derivations are flattened but
#' retain their group and color metadata, which is then used in DAG
#' visualizations (`rxp_visnetwork()` and `rxp_ggdag()`) to distinguish
#' different parts of your workflow.
#'
#' This pattern enables a "Master Script" workflow where you can define
#' sub-pipelines in separate R scripts that each return a list of derivations.
#' You then pass the paths to these scripts to `rxp_pipeline()`:
#'
#' @examples
#' \dontrun{
#'   # Define derivations in separate scripts
#'   # pipelines/01_etl.R returns: list(rxp_r(...), rxp_r(...))
#'   # pipelines/02_model.R returns: list(rxp_r(...), rxp_r(...))
#'
#'   # Master script (run.R):
#'
#'   # Create named pipelines with colors by pointing to the files
#'   pipe_etl <- rxp_pipeline("ETL", "pipelines/01_etl.R", color = "darkorange")
#'   pipe_model <- rxp_pipeline("Model", "pipelines/02_model.R", color = "dodgerblue")
#'
#'   # Build the combined pipeline
#'   rxp_populate(list(pipe_etl, pipe_model))
#'   rxp_make()
#'
#'   # Visualize - ETL nodes will be orange, Model nodes will be blue
#'   rxp_visnetwork()
#' }
#'
#' @export
rxp_pipeline <- function(name, path, color = NULL, ...) {
  # Validate inputs

  if (!is.character(name) || length(name) != 1 || nchar(name) == 0) {
    stop("'name' must be a non-empty character string")
  }

  derivs <- NULL

  if (is.character(path)) {
    if (length(path) != 1) {
      stop("'path' must be a single file path")
    }
    if (!file.exists(path)) {
      stop("Pipeline file not found: ", path)
    }
    # Source the file in a new environment to avoid side effects
    # and capture the return value (result of last expression)
    res <- source(path, local = new.env())
    derivs <- res$value
  } else if (inherits(path, "rxp_derivation")) {
    stop(
      "'path' must be a list of derivation objects or a file path, not a single derivation."
    )
  } else if (is.list(path)) {
    # Accept a direct list for testing or specialized use cases
    derivs <- path
  } else {
    stop("'path' must be a file path (character) or a list of derivations")
  }

  if (!is.list(derivs)) {
    stop(
      "The pipeline script must return a list of derivations, but returned: ",
      class(derivs)[1]
    )
  }

  # Validate that all elements are derivation objects
  for (i in seq_along(derivs)) {
    if (!inherits(derivs[[i]], "rxp_derivation")) {
      stop(
        "Element ",
        i,
        " of pipeline derivations is not an rxp_derivation object. ",
        "All elements must be created by rxp_r(), rxp_py(), etc."
      )
    }
  }

  # Validate color if provided
  if (!is.null(color)) {
    if (!is.character(color) || length(color) != 1) {
      stop(
        "'color' must be a single character string (CSS color name or hex code)"
      )
    }
  }

  # Attach metadata to each derivation
  derivs <- lapply(derivs, function(deriv) {
    deriv$pipeline_group <- name
    deriv$pipeline_color <- color
    deriv
  })

  # Create the pipeline object
  structure(
    list(
      name = name,
      color = color,
      derivs = derivs
    ),
    class = c("rxp_pipeline", "list")
  )
}


#' Flatten a list of derivations and pipelines
#'
#' Takes a potentially nested list containing both `rxp_derivation` objects
#' and `rxp_pipeline` objects and returns a flat list of derivations with
#' metadata preserved.
#'
#' @param derivs A list that may contain `rxp_derivation` objects,
#'   `rxp_pipeline` objects, or a mix of both.
#' @return A flat list of `rxp_derivation` objects.
#' @keywords internal
flatten_derivations <- function(derivs) {
  result <- list()

  for (item in derivs) {
    if (inherits(item, "rxp_pipeline")) {
      # It's a pipeline - extract its derivations (already have metadata attached)
      result <- c(result, item$derivs)
    } else if (inherits(item, "rxp_derivation")) {
      # It's a plain derivation - add default metadata if missing
      if (is.null(item$pipeline_group)) {
        item$pipeline_group <- "default"
      }
      if (is.null(item$pipeline_color)) {
        item$pipeline_color <- NULL # Will be assigned a default in visualization
      }
      result <- c(result, list(item))
    } else if (is.list(item)) {
      # Recursively flatten nested lists
      result <- c(result, flatten_derivations(item))
    } else {
      stop(
        "Invalid element in derivations list. Expected rxp_derivation or rxp_pipeline, got: ",
        class(item)[1]
      )
    }
  }

  result
}


#' Print Method for rxp_pipeline Objects
#'
#' @param x An object of class "rxp_pipeline"
#' @param ... Additional arguments passed to print methods
#' @return Nothing, prints a summary of the pipeline object to the console.
#' @export
print.rxp_pipeline <- function(x, ...) {
  cat("rixpress pipeline:", x$name, "\n")
  if (!is.null(x$color)) {
    cat("  Color:", x$color, "\n")
  }
  cat("  Derivations:", length(x$derivs), "\n")
  if (length(x$derivs) > 0) {
    deriv_names <- sapply(x$derivs, function(d) d$name)
    cat("    -", paste(deriv_names, collapse = "\n    - "), "\n")
  }
  invisible(x)
}

test_that("rxp_pipeline creates valid pipeline object", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("rix")

  # Create derivations
  d1 <- rxp_r(test1, 1 + 1)
  d2 <- rxp_r(test2, 2 + 2)

  # Create a pipeline
  pipe <- rxp_pipeline("TestPipe", list(d1, d2), color = "#FF0000")

  # Check structure
  expect_s3_class(pipe, "rxp_pipeline")
  expect_equal(pipe$name, "TestPipe")
  expect_equal(pipe$color, "#FF0000")
  expect_length(pipe$derivs, 2)

  # Check that metadata is attached to derivations
  expect_equal(pipe$derivs[[1]]$pipeline_group, "TestPipe")
  expect_equal(pipe$derivs[[1]]$pipeline_color, "#FF0000")
  expect_equal(pipe$derivs[[2]]$pipeline_group, "TestPipe")
  expect_equal(pipe$derivs[[2]]$pipeline_color, "#FF0000")
})

test_that("rxp_pipeline validates inputs", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("rix")

  d1 <- rxp_r(test1, 1 + 1)

  # Empty name should error
  expect_error(rxp_pipeline("", list(d1)), "non-empty character string")

  # Single derivation (not in list) should error with helpful message
  expect_error(
    rxp_pipeline("Pipe", d1),
    "'path' must be a list of derivation objects or a file path"
  )

  # Non-derivation in list should error
  expect_error(
    rxp_pipeline("Pipe", list(d1, "not a deriv")),
    "not an rxp_derivation object"
  )
})

test_that("rxp_pipeline sources from file", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("rix")

  # Create a temporary R script defining a pipeline
  temp_script <- tempfile(fileext = ".R")
  writeLines(
    'library(rixpress)
     list(
       rxp_r(name = d1, expr = 1 + 1),
       rxp_r(name = d2, expr = 2 + 2)
     )',
    temp_script
  )
  on.exit(unlink(temp_script))

  # Create pipeline from file
  pipe <- rxp_pipeline("FilePipe", temp_script, color = "green")

  expect_s3_class(pipe, "rxp_pipeline")
  expect_equal(pipe$name, "FilePipe")
  expect_length(pipe$derivs, 2)
  expect_equal(pipe$derivs[[1]]$name, "d1")
})

test_that("flatten_derivations works correctly", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("rix")

  # Create derivations
  d1 <- rxp_r(etl1, 1)
  d2 <- rxp_r(etl2, 2)
  d3 <- rxp_r(model1, 3)
  d4 <- rxp_r(model2, 4)

  # Create pipelines
  pipe_etl <- rxp_pipeline("ETL", list(d1, d2), color = "orange")
  pipe_model <- rxp_pipeline("Model", list(d3, d4), color = "blue")

  # Flatten
  result <- flatten_derivations(list(pipe_etl, pipe_model))

  # Check result
  expect_length(result, 4)
  expect_equal(result[[1]]$name, "etl1")
  expect_equal(result[[1]]$pipeline_group, "ETL")
  expect_equal(result[[1]]$pipeline_color, "orange")
  expect_equal(result[[3]]$name, "model1")
  expect_equal(result[[3]]$pipeline_group, "Model")
  expect_equal(result[[3]]$pipeline_color, "blue")
})

test_that("flatten_derivations handles flat lists with defaults", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("rix")

  # Create plain derivations (no pipeline)
  d1 <- rxp_r(plain1, 1)
  d2 <- rxp_r(plain2, 2)

  # Flatten
  result <- flatten_derivations(list(d1, d2))

  # Check that defaults are applied
  expect_length(result, 2)
  expect_equal(result[[1]]$pipeline_group, "default")
  expect_null(result[[1]]$pipeline_color)
})

test_that("rxp_populate handles rxp_pipeline objects", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("rix")

  # Create temporary directory
  temp_dir <- tempfile()
  dir.create(temp_dir)
  on.exit(unlink(temp_dir, recursive = TRUE), add = TRUE)

  rix::rix(
    date = "2025-04-11",
    r_pkgs = "dplyr",
    git_pkgs = list(
      package_name = "rixpress",
      repo_url = "https://github.com/ropensci/rixpress",
      commit = "HEAD"
    ),
    ide = "none",
    project_path = temp_dir,
    overwrite = TRUE
  )

  setwd(temp_dir)

  # Create derivations and pipelines
  d1 <- rxp_r(etl_clean, 1 + 1)
  d2 <- rxp_r(model_train, etl_clean + 1)

  pipe_etl <- rxp_pipeline("Data Prep", list(d1), color = "#E69F00")
  pipe_model <- rxp_pipeline("Modeling", list(d2), color = "#56B4E9")

  # Populate
  rxp_populate(list(pipe_etl, pipe_model), project_path = ".", build = FALSE)

  # Check that dag.json contains pipeline metadata
  dag_json <- jsonlite::fromJSON("_rixpress/dag.json")

  expect_true("pipeline_group" %in% names(dag_json$derivations))
  expect_true("pipeline_color" %in% names(dag_json$derivations))

  # Check specific values (handle both list and vector formats from jsonlite)
  etl_idx <- which(dag_json$derivations$deriv_name == "etl_clean")
  model_idx <- which(dag_json$derivations$deriv_name == "model_train")

  # Use unlist to handle potential list format
  etl_group <- unlist(dag_json$derivations$pipeline_group[etl_idx])
  etl_color <- unlist(dag_json$derivations$pipeline_color[etl_idx])
  model_group <- unlist(dag_json$derivations$pipeline_group[model_idx])
  model_color <- unlist(dag_json$derivations$pipeline_color[model_idx])

  expect_equal(etl_group, "Data Prep")
  expect_equal(etl_color, "#E69F00")
  expect_equal(model_group, "Modeling")
  expect_equal(model_color, "#56B4E9")
})

test_that("get_nodes_edges reads pipeline metadata from dag.json", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("rix")

  # Create a temporary dag.json with pipeline metadata
  temp_dir <- tempfile()
  dir.create(temp_dir, recursive = TRUE)
  on.exit(unlink(temp_dir, recursive = TRUE), add = TRUE)

  dag_content <- list(
    derivations = data.frame(
      deriv_name = c("node1", "node2"),
      depends = I(list(character(0), "node1")),
      decoder = c("rds", "rds"),
      type = c("rxp_r", "rxp_r"),
      noop_build = c(FALSE, FALSE),
      pipeline_group = c("Pipeline A", "Pipeline B"),
      pipeline_color = c("#FF0000", "#00FF00"),
      stringsAsFactors = FALSE
    )
  )

  dag_path <- file.path(temp_dir, "dag.json")
  jsonlite::write_json(dag_content, dag_path, auto_unbox = TRUE)

  # Read with get_nodes_edges
  result <- get_nodes_edges(path_dag = dag_path)

  # Check that pipeline metadata is present
  expect_true("pipeline_group" %in% names(result$nodes))
  expect_true("pipeline_color" %in% names(result$nodes))

  expect_equal(result$nodes$pipeline_group[1], "Pipeline A")
  expect_equal(result$nodes$pipeline_group[2], "Pipeline B")
  expect_equal(result$nodes$pipeline_color[1], "#FF0000")
  expect_equal(result$nodes$pipeline_color[2], "#00FF00")
})

test_that("print.rxp_pipeline works", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("rix")

  d1 <- rxp_r(test1, 1)
  pipe <- rxp_pipeline("MyPipe", list(d1), color = "red")

  # Capture print output
  output <- capture.output(print(pipe))

  expect_true(any(grepl("MyPipe", output)))
  expect_true(any(grepl("red", output)))
  expect_true(any(grepl("test1", output)))
})

#' Generate Nix Pipeline Code
#'
#' @family pipeline functions
#' @param derivs A list of derivation objects, where each object is a list of
#'   five elements:
#'   - *name*, name of the derivation,
#'   - *snippet*, the nix code snippet to build this derivation,
#'   - *type*, can be R, Python or Quarto,
#'   - *additional_files*, character vector of paths to files to make available to build sandbox,
#'   - *nix_env*, path to Nix environment to build this derivation.
#'   A single deriv is the output of `rxp_r()`, `rxp_qmd()` or `rxp_py()`
#'   function.
#'
#' @param project_path Path to root of project, defaults to ".".
#'
#' @param build Logical, defaults to FALSE. Should the pipeline get built right
#'   after being generated? When FALSE, use `rxp_make()` to build the pipeline
#'   at a later stage.
#'
#' @param py_imports Named character vector of Python import rewrites. Names are
#'   the base modules that rixpress auto-imports as "import name", and values
#'   are the desired import lines. For example: c(numpy = "import numpy as np",
#'   xgboost = "from xgboost import XGBClassifier"). Each entry is applied by
#'   replacing "import name" with the provided string across generated
#'   _rixpress Python library files.
#'
#' @param ... Further arguments passed down to methods. Use `max-jobs` and
#'   `cores` to set parallelism during build. See the documentation of
#'   `rxp_make()` for more details.
#'
#' @return Nothing, writes a file called `pipeline.nix` with the Nix code to
#'   build the pipeline, as well as folder called _rixpress with required
#'   internal files.
#'
#' @details
#' This function generates a `pipeline.nix` file based on a list of derivation
#' objects. Each derivation defines a build step, and `rxp_populate()` chains these
#' steps and handles the serialization and conversion of Python objects into R
#' objects (or vice-versa). Derivations are created with `rxp_r()`, `rxp_py()`
#' and so on. By default, the pipeline is also immediately built after being
#' generated, but the build process can be postponed by setting `build` to
#' FALSE. In this case, the pipeline can then be built using `rxp_make()` at
#' a later stage.
#' The generated `pipeline.nix` expression includes:
#' - the required imports of environments, typically `default.nix` files generated by
#'   the `rix` package;
#' - correct handling of interdependencies of the different derivations;
#' - serialization and deserialization of both R and Python objects, and conversion
#'   between them when objects are passed from one language to another;
#' - correct loading of R and Python packages, or extra functions needed to build
#'   specific targets
#'
#' The `_rixpress` folder contains:
#' - R, Python or Julia scripts to load the required packages that need to be
#' available to the pipeline.
#' - a JSON file with the DAG of the pipeline, used for visualisation, and to
#' allow `rxp_populate()` to generate the right dependencies between derivations.
#' - `.rds` files with build logs, required for `rxp_inspect()` and `rxp_gc()`.
#' See `vignette("debugging")` for more details.
#'
#' Inline Python import adjustments
#' In some cases, due to the automatic handling of Python packages, users might
#' want to change import statements. By default if, say, `pandas` is needed to
#' build a derivation, it will be imported with `import pandas`. However, Python
#' programmers typically use `import pandas as pd`. You can either:
#' - use `py_imports` to rewrite these automatically during population, or
#' - use `adjust_import()` and `add_import()` for advanced/manual control.
#' See `vignette("polyglot")` for more details.
#'
#' @examples
#' \dontrun{
#' # Create derivation objects
#' d1 <- rxp_r(mtcars_am, filter(mtcars, am == 1))
#' d2 <- rxp_r(mtcars_head, head(mtcars_am))
#' list_derivs <- list(d1, d2)
#'
#' # Generate and build in one go
#' rxp_populate(derivs = list_derivs, project_path = ".", build = TRUE)
#'
#' # Or only populate, with inline Python import adjustments
#' rxp_populate(
#'   derivs = list_derivs,
#'   project_path = ".",
#'   build = FALSE,
#'   py_imports = c(pandas = "import pandas as pd")
#' )
#' # Then later:
#' rxp_make()
#' }
#' @export
rxp_populate <- function(
  derivs,
  project_path = ".",
  build = FALSE,
  py_imports = NULL,
  ...
) {
  # Flatten any rxp_pipeline objects to get a flat list of derivations

  # This supports hierarchical organization while preserving metadata
  derivs <- flatten_derivations(derivs)

  rxp_write_dag(
    derivs,
    output_file = file.path(project_path, "_rixpress", "dag.json")
  )

  # Read back the DAG to get the final no-op flags
  dag <- jsonlite::read_json(file.path(project_path, "_rixpress", "dag.json"))

  # Update derivation snippets based on final no-op flags
  for (i in seq_along(derivs)) {
    dag_entry <- Find(
      function(x) x$deriv_name[1] == derivs[[i]]$name,
      dag$derivations
    )
    if (!is.null(dag_entry) && isTRUE(dag_entry$noop_build[[1]])) {
      # Only update if it wasn't already a no-op
      if (!isTRUE(derivs[[i]]$noop_build)) {
        # Just update the fields directly
        derivs[[i]]$noop_build <- TRUE
        derivs[[i]]$snippet <- sprintf(
          "  %s = defaultPkgs.runCommand \"%s\" {} \"\n    mkdir -p $out\n    echo 'Build skipped for %s (cascading no-op)' > $out/NOOPBUILD\n  \";",
          derivs[[i]]$name,
          derivs[[i]]$name,
          derivs[[i]]$name
        )
      }
    }
  }

  # Need to combine nix envs and additional files into a
  # list of two elements, "nix_env" and "additional_files"
  # which list all the unique combinations
  nix_expressions_and_additional_files <- lapply(
    derivs,
    function(d) {
      list(
        "nix_env" = d$nix_env,
        "additional_files" = d$additional_files,
        "type" = d$type
      )
    }
  )
  # Drop quarto objects, as these are handled separately
  nix_expressions_and_additional_files <- lapply(derivs, function(d) {
    if (d$type == "rxp_qmd" || d$type == "rxp_rmd") {
      d$additional_files <- ""
    }
    list(
      nix_env = d$nix_env,
      additional_files = d$additional_files,
      type = d$type
    )
  })

  flat_list <- list(
    nix_env = sapply(
      X = nix_expressions_and_additional_files,
      FUN = `[[`,
      "nix_env",
      USE.NAMES = FALSE
    ),
    additional_files = sapply(
      X = nix_expressions_and_additional_files,
      FUN = `[[`,
      "additional_files",
      USE.NAMES = FALSE
    )
  )

  nix_env_all <- flat_list$nix_env
  add_files_all <- flat_list$additional_files

  unique_env <- unique(nix_env_all)

  additional_files_combined <- lapply(
    unique_env,
    function(env) {
      idx <- which(nix_env_all == env)
      files <- unlist(add_files_all[idx])
      files <- files[!is.na(files) & files != ""]
      if (length(files) == 0) {
        return("")
      }
      unique(files)
    }
  )

  result <- list(
    nix_env = unique_env,
    additional_files = additional_files_combined
  )

  suppressWarnings(
    for (i in seq_along(result$nix_env)) {
      generate_libraries_from_nix(
        result$nix_env[i],
        result$additional_files[[i]],
        project_path = project_path
      )
    }
  )

  # Apply inline Python import adjustments, if provided.
  if (!is.null(py_imports)) {
    if (!is.character(py_imports) || is.null(names(py_imports))) {
      stop(
        "py_imports must be a named character vector, e.g. c(numpy = 'import numpy as np')."
      )
    }
    for (mod in names(py_imports)) {
      desired <- unname(py_imports[[mod]])
      old <- paste0("import ", mod)
      adjust_import(
        old_import = old,
        new_import = desired,
        project_path = project_path
      )
    }
  }

  # Finalize pipeline
  flat_pipeline <- gen_flat_pipeline(derivs)

  pipeline <- gen_pipeline(
    dag_file = file.path(paste0(project_path, "/_rixpress/dag.json")),
    flat_pipeline = flat_pipeline
  )

  writeLines(pipeline, file.path(project_path, "pipeline.nix"))

  if (build) {
    rxp_make(...)
  }
}


#' Parse Nix Environment Element of a Derivation Object
#' @param derivs A list of derivation objects, where each object is a list of
#'   five elements:
#'     - `$name`, character, name of the derivation
#'     - `$snippet`, character, the nix code snippet to build this derivation
#'     - `$type`, character, can be R, Python or Quarto
#'     - `$additional_files`, character vector of paths to files to make
#'        available to build sandbox
#'     - `$nix_env`, character, path to Nix environment to build this derivation
#'   Typically, these objects are created by a function like `rxp_r`.
#' @noRd
parse_nix_envs <- function(derivs) {
  # Add required elements
  # base name of libraries file
  derivs <- lapply(
    derivs,
    function(d) {
      d$base_name <- sub(
        "_nix$",
        "",
        gsub("[^a-zA-Z0-9]", "_", basename(d$nix_env))
      )
      d
    }
  )
  # path to libraries file
  derivs <- lapply(
    derivs,
    function(d) {
      d$library <- list.files("_rixpress", pattern = d$base_name)
      d$library_in_sandbox <- gsub(paste0(d$base_name, "_"), "", d$library)
      list(
        "nix_env" = d$nix_env,
        "base_name" = d$base_name,
        "library" = d$library,
        "library_in_sandbox" = d$library_in_sandbox
      )
    }
  )

  derivs <- unique(derivs)

  generate_configurePhase <- function(d) {
    # Compute the configure_phases_str
    configure_phases_str <- paste0(
      d$base_name,
      "ConfigurePhase = ''\n    ",
      paste0(
        "cp ${./_rixpress/",
        unlist(d$library),
        "} ",
        unlist(d$library_in_sandbox),
        collapse = "\n    "
      ),
      "\n    mkdir -p $out  ",
      "\n    mkdir -p .julia_depot  ",
      "\n    export JULIA_DEPOT_PATH=$PWD/.julia_depot  ",
      "\n    export HOME_PATH=$PWD\n  ",
      "'';\n  "
    )

    # Create the individual lines
    lines <- c(
      paste0(d$base_name, " = import ./", d$nix_env, ";"),
      paste0(d$base_name, "Pkgs = ", d$base_name, ".pkgs;"),
      paste0(d$base_name, "Shell = ", d$base_name, ".shell;"),
      paste0(d$base_name, "BuildInputs = ", d$base_name, "Shell.buildInputs;"),
      configure_phases_str
    )

    # Combine all lines into a single string with newline separators
    paste(lines, collapse = "\n  ")
  }

  nix_lines <- character(0)
  for (d in seq_along(derivs)) {
    current_lines <- generate_configurePhase(derivs[[d]])
    nix_lines <- c(nix_lines, current_lines)
  }

  paste(nix_lines, collapse = "\n\n  ")
}

#' Generate Flat Pipeline Boilerplate for pipeline.nix
#' @param derivs A list of derivation objects, where each object is a list of
#'   five elements:
#'     - `$name`, character, name of the derivation
#'     - `$snippet`, character, the nix code snippet to build this derivation
#'     - `$type`, character, can be R, Python or Quarto
#'     - `$additional_files`, character vector of paths to files to make
#'        available to build sandbox
#'     - `$nix_env`, character, path to the Nix environment to build this derivation
#'   A single deriv is the output of `rxp_r()`, `rxp_qmd()` or `rxp_py()`
#'   function.
#' @noRd
gen_flat_pipeline <- function(derivs) {
  derivation_texts <- vapply(
    derivs,
    function(d) d$snippet,
    FUN.VALUE = character(1)
  )
  derivations_code <- paste(derivation_texts, collapse = "\n\n")

  deriv_names <- vapply(derivs, function(d) d$name, character(1))
  names_line <- paste(deriv_names, collapse = " ")

  nix_envs <- parse_nix_envs(derivs)

  # Determine required functions
  types <- vapply(derivs, function(d) d$type, character(1))
  need_r <- get_need_r(types)
  need_py <- get_need_py(types)
  need_jl <- get_need_jl(types)

  # Build function definitions
  function_defs <- ""
  if (need_r) {
    function_defs <- paste0(
      function_defs,
      "\n  # Function to create R derivations
  makeRDerivation = { name, buildInputs, configurePhase, buildPhase, src ? null }:
    defaultPkgs.stdenv.mkDerivation {
      inherit name src;
      dontUnpack = true;
      inherit buildInputs configurePhase buildPhase;
      installPhase = ''
        cp ${name} $out/
      '';
    };"
    )
  }
  if (need_py) {
    function_defs <- paste0(
      function_defs,
      "\n  # Function to create Python derivations
  makePyDerivation = { name, buildInputs, configurePhase, buildPhase, src ? null }:
    let
      pickleFile = \"${name}\";
    in
      defaultPkgs.stdenv.mkDerivation {
        inherit name src;
        dontUnpack = true;
        buildInputs = buildInputs;
        inherit configurePhase buildPhase;
        installPhase = ''
          cp ${pickleFile} $out
        '';
      };"
    )
  }
  if (need_jl) {
    function_defs <- paste0(
      function_defs,
      "\n  # Function to create Julia derivations
  makeJlDerivation = { name, buildInputs, configurePhase, buildPhase, src ? null }:
    defaultPkgs.stdenv.mkDerivation {
      inherit name src;
      dontUnpack = true;
      buildInputs = buildInputs;
      inherit configurePhase buildPhase;
      installPhase = ''
        cp ${name} $out/
      '';
    };"
    )
  }

  # Generate Nix code
  pipeline_nix <- sprintf(
    'let
  %s%s

  # Define all derivations
%s

  # Generic default target that builds all derivations
  allDerivations = defaultPkgs.symlinkJoin {
    name = "all-derivations";
    paths = with builtins; attrValues { inherit %s; };
  };

in
{
  inherit %s;
  default = allDerivations;
}
',
    nix_envs,
    function_defs,
    paste0("  ", derivations_code),
    names_line,
    names_line
  )

  strsplit(pipeline_nix, split = "\n")[[1]]
}

# Escape regex special characters in a literal string
#' @noRd
escape_regex <- function(x) {
  gsub("([][{}()+*^$|\\\\.?])", "\\\\\\1", x)
}

#' Finalise a Flat Pipeline
#' @param dag_file A json file giving the names and relationships between derivations.
#' @param flat_pipeline A flat pipeline, output of `gen_flat_elements()`.
#' @noRd
gen_pipeline <- function(dag_file, flat_pipeline) {
  dag <- jsonlite::read_json(dag_file)
  pipeline_str <- paste(flat_pipeline, collapse = "\n")

  for (d in dag$derivations) {
    if (
      length(d$depends) == 0 ||
        d$type %in% c("rxp_qmd", "rxp_rmd", "rxp_py2r", "rxp_r2py")
    ) {
      next
    }
    deriv_name <- as.character(d$deriv_name[1])
    deps <- d$depends
    type <- d$type[1]
    decoder <- d$decoder

    # Helper function to get the unserialize function for a specific dependency
    get_unserialize_func_for_dep <- function(
      dep_name,
      decoder,
      type
    ) {
      if (is.null(decoder) || length(decoder) == 0) {
        # Use default based on type
        return(switch(
          type,
          "rxp_r" = "readRDS",
          "rxp_py" = "pickle.load",
          "rxp_jl" = "Serialization.deserialize",
          "readRDS"
        ))
      }

      # Check if decoder is a list (from JSON)
      if (is.list(decoder)) {
        # Check if it has names (named list/vector case)
        func_names <- names(decoder)
        if (!is.null(func_names) && length(func_names) > 0) {
          # It's a named list - look up the specific dependency
          if (dep_name %in% func_names) {
            return(as.character(decoder[[dep_name]]))
          } else {
            # Dependency not in the named list, use default
            return(switch(
              type,
              "rxp_r" = "readRDS",
              "rxp_py" = "pickle.load",
              "rxp_jl" = "Serialization.deserialize",
              "readRDS"
            ))
          }
        } else {
          # It's a single value in a list
          return(as.character(decoder[[1]]))
        }
      } else {
        # It's a single string value
        return(as.character(decoder[1]))
      }
    }

    # Build load lines per type
    if (type == "rxp_r") {
      base_placeholder <- "# RIXPRESS_LOAD_DEPENDENCIES_HERE"
      load_lines <- vapply(
        deps,
        function(dep) {
          func <- get_unserialize_func_for_dep(dep, decoder, type)
          sprintf("%s <- %s('${%s}/%s')", dep, func, dep, dep)
        },
        character(1)
      )
    } else if (type == "rxp_py") {
      base_placeholder <- "# RIXPRESS_PY_LOAD_DEPENDENCIES_HERE"
      load_lines <- vapply(
        deps,
        function(dep) {
          func <- get_unserialize_func_for_dep(dep, decoder, type)
          sprintf(
            "with open('${%s}/%s', 'rb') as f: %s = %s(f)",
            dep,
            dep,
            dep,
            func
          )
        },
        character(1)
      )
    } else if (type == "rxp_jl") {
      base_placeholder <- "# RIXPRESS_JL_LOAD_DEPENDENCIES_HERE"
      load_lines <- vapply(
        deps,
        function(dep) {
          func <- get_unserialize_func_for_dep(dep, decoder, type)
          sprintf(
            "%s = open(\\\\\\\"%s\\\\\\\", \\\\\\\"r\\\\\\\") do io; %s(io); end",
            dep,
            paste0("${", dep, "}/", dep),
            func
          )
        },
        character(1)
      )
    } else {
      next
    }

    # Name-scoped placeholder pattern; preserve indentation
    specific_placeholder <- paste0(base_placeholder, ":", deriv_name)
    pattern <- paste0(
      "(?m)^([ \\t]*)",
      escape_regex(specific_placeholder),
      "\\s*$"
    )

    # Prefix every injected line with the captured indentation
    replacement_block <- paste(load_lines, collapse = "\n")
    replacement_block <- gsub("\n", "\n\\1", replacement_block, fixed = TRUE)
    replacement <- paste0("\\1", replacement_block)

    pipeline_str <- sub(pattern, replacement, pipeline_str, perl = TRUE)
  }

  strsplit(pipeline_str, "\n")[[1]]
}


#' Generate an R or Py Script with Library Calls from a default.nix File
#'
#' @param nix_env Nix environment where the derivation runs
#' @param additional_files Character vector, additional files to include. These
#'   are the files that contain custom functions required for this derivation.
#' @param project_path	Path to root of project, typically "."
#' @return A script to load the libraries inside of derivations.
#' @noRd
generate_libraries_from_nix <- function(
  nix_env,
  additional_files = "",
  project_path
) {
  generate_r_libraries_from_nix(
    nix_env,
    additional_files,
    project_path
  )
  generate_py_libraries_from_nix(
    nix_env,
    additional_files,
    project_path
  )
  generate_jl_libraries_from_nix(
    nix_env,
    additional_files,
    project_path
  )
}

#' @noRd
get_need_r <- function(types) {
  any(
    types %in%
      c("rxp_r", "rxp_r_file", "rxp_rmd", "rxp_qmd", "rxp_py2r", "rxp_r2py")
  )
}

#' @noRd
get_need_py <- function(types) {
  any(types %in% c("rxp_py", "rxp_py_file"))
}

#' @noRd
get_need_jl <- function(types) {
  any(types %in% c("rxp_jl"))
}

test_that("rxp_populate generates expected assets", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("rix")

  # Create temporary directory
  temp_dir <- tempfile()
  dir.create(temp_dir)
  on.exit(unlink(temp_dir, recursive = TRUE), add = TRUE)

  rix::rix(
    date = "2025-04-11",
    r_pkgs = "dplyr",
    py_conf = list(
      py_version = "3.13",
      py_pkgs = "xgboost"
    ),
    git_pkgs = list(
      package_name = "rixpress",
      repo_url = "https://github.com/ropensci/rixpress",
      commit = "HEAD"
    ),
    ide = "rstudio",
    project_path = temp_dir,
    overwrite = TRUE
  )

  derivs <- list(
    rxp_py(
      name = "mdl",
      expr = "XGBClassifier()"
    )
  )

  rxp_populate(
    derivs,
    project_path = temp_dir,
    py_imports = c(
      xgboost = "from xgboost import XGBClassifier"
    )
  )

  result <- readLines(paste0(temp_dir, "/_rixpress/default_libraries.py"))

  snapshot_gen_population_pipeline <- function(result) {
    tfile <- tempfile(pattern = "default_libraries", fileext = ".py")
    writeLines(result, tfile)
    tfile
  }

  testthat::expect_snapshot_file(
    path = snapshot_gen_population_pipeline(result),
    name = "default_libraries.py"
  )
})

test_that("rixpress: helper functions get_need_r/py work", {
  types_r_only <- c("rxp_r", "rxp_r_file")
  types_py_only <- c("rxp_py", "rxp_py_file")
  types_quarto <- c("rxp_qmd")
  types_transfer <- c("rxp_py2r", "rxp_r2py")
  types_mixed <- c("rxp_r", "rxp_py", "rxp_qmd", "rxp_py2r")

  expect_true(get_need_r(types_r_only))
  expect_false(get_need_py(types_r_only))

  expect_false(get_need_r(types_py_only))
  expect_true(get_need_py(types_py_only))

  expect_true(get_need_r(types_quarto))
  expect_false(get_need_py(types_quarto))

  expect_true(get_need_r(types_transfer))
  expect_false(get_need_py(types_transfer)) # Transfer functions are R-based

  expect_true(get_need_r(types_mixed))
  expect_true(get_need_py(types_mixed))
})

test_that("rixpress: gen_flat_pipeline", {
  testthat::skip_on_cran()
  d1 <- rxp_r(mtcars_am, dplyr::filter(mtcars, am == 1))
  d2 <- rxp_r(mtcars_head, head(mtcars_am))
  derivs <- list(d1, d2)

  flat_pipeline <- gen_flat_pipeline(derivs)

  snapshot_gen_flat_pipeline <- function(derivs, flat_pipeline) {
    tfile <- tempfile(pattern = "flat_pipeline", fileext = ".nix")
    writeLines(flat_pipeline, tfile)
    tfile
  }

  testthat::expect_snapshot_file(
    path = snapshot_gen_flat_pipeline(derivs, flat_pipeline),
    name = "flat_pipeline.nix"
  )

  dag_out <- tempfile(pattern = "dag", fileext = ".json")
  rxp_write_dag(derivs, output_file = dag_out)

  snapshot_gen_pipeline <- function(dag_out, flat_pipeline) {
    pipeline <- gen_pipeline(
      dag_file = dag_out,
      flat_pipeline = flat_pipeline
    )

    tfile <- tempfile(pattern = "pipeline", fileext = ".nix")
    writeLines(pipeline, tfile)
    tfile
  }

  testthat::expect_snapshot_file(
    path = snapshot_gen_pipeline(dag_out, flat_pipeline),
    name = "pipeline.nix"
  )
})

test_that("rixpress: several environments", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("rix")

  path_tmpdir <- "tempdir"
  dir.create(path_tmpdir)
  on.exit(unlink(path_tmpdir, recursive = TRUE), add = TRUE, after = TRUE)

  path_to_default.nix <- paste0(path_tmpdir, "/default.nix")
  path_to_default2.nix <- paste0(path_tmpdir, "/default2.nix")

  rix::rix(date = "2025-03-10", r_pkgs = "dplyr", project_path = path_tmpdir)
  file.rename(
    from = path_to_default.nix,
    to = path_to_default2.nix
  )

  d1 <- rxp_r(mtcars_am, dplyr::filter(mtcars, am == 1))
  d2 <- rxp_r(mtcars_head, head(mtcars_am), nix_env = path_to_default2.nix)
  derivs <- list(d1, d2)

  flat_pipeline <- gen_flat_pipeline(derivs)

  snapshot_gen_flat_pipeline <- function(derivs, flat_pipeline) {
    tfile <- tempfile(pattern = "flat_pipeline", fileext = ".nix")
    writeLines(flat_pipeline, tfile)
    tfile
  }

  testthat::expect_snapshot_file(
    path = snapshot_gen_flat_pipeline(derivs, flat_pipeline),
    name = "flat_pipeline_several_envs.nix"
  )

  dag_out <- tempfile(pattern = "dag", fileext = ".json")
  rxp_write_dag(derivs, output_file = dag_out)

  snapshot_gen_pipeline <- function(dag_out, flat_pipeline) {
    pipeline <- gen_pipeline(
      dag_file = dag_out,
      flat_pipeline = flat_pipeline
    )

    tfile <- tempfile(pattern = "pipeline", fileext = ".nix")
    writeLines(pipeline, tfile)
    tfile
  }

  testthat::expect_snapshot_file(
    path = snapshot_gen_pipeline(dag_out, flat_pipeline),
    name = "pipeline_several_envs.nix"
  )
})

test_that("parse_nix_envs handles different environments correctly", {
  # Create temporary directory for test files
  temp_dir <- tempfile()
  dir.create(temp_dir)
  on.exit(unlink(temp_dir, recursive = TRUE), add = TRUE)

  # Create _rixpress directory
  rixpress_dir <- file.path(temp_dir, "_rixpress")
  dir.create(rixpress_dir)

  # Create mock library files
  writeLines("# R libraries", file.path(rixpress_dir, "default_r_libraries.R"))
  writeLines(
    "# Python libraries",
    file.path(rixpress_dir, "py_env_py_libraries.py")
  )

  # Set working directory to temp_dir for the test
  old_wd <- getwd()
  setwd(temp_dir)
  on.exit(setwd(old_wd), add = TRUE)

  # Create test derivations with different environments
  d1 <- list(
    name = "test_r",
    snippet = "test_r_snippet",
    type = "rxp_r",
    additional_files = "",
    nix_env = "default.nix",
    base_name = "default"
  )

  d2 <- list(
    name = "test_py",
    snippet = "test_py_snippet",
    type = "rxp_py",
    additional_files = "",
    nix_env = "data-raw/py2r_example/py-env.nix",
    base_name = "py_env"
  )

  # Test the function
  result <- parse_nix_envs(list(d1, d2))

  # Check that the result contains expected strings for both environments
  expect_true(grepl("default = import ./default.nix;", result))
  expect_true(grepl(
    "py_env = import ./data-raw/py2r_example/py-env.nix;",
    result
  ))
  expect_true(grepl("defaultConfigurePhase", result))
  expect_true(grepl("py_envConfigurePhase", result))
})

test_that("rixpress correctly handles dependencies", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("rix")

  # Create temporary directory
  temp_dir <- tempfile()
  dir.create(temp_dir)
  on.exit(unlink(temp_dir, recursive = TRUE), add = TRUE)

  setwd(temp_dir)

  rix::rix(
    date = "2025-04-11",
    r_pkgs = "dplyr",
    git_pkgs = list(
      package_name = "rixpress",
      repo_url = "https://github.com/ropensci/rixpress",
      commit = "HEAD"
    ),
    ide = "rstudio",
    project_path = ".",
    overwrite = TRUE
  )

  # Create actual derivations
  d1 <- rxp_r(data_prep, mtcars)
  d2 <- rxp_r(analysis, summary(data_prep), nix_env = "default.nix")
  d3 <- rxp_py(
    py_process,
    "pd.DataFrame({'a': [1, 2, 3]})",
    nix_env = "default.nix"
  )

  # Create derivations list
  derivs <- list(d1, d2, d3)

  rxp_populate(derivs, project_path = ".", build = FALSE)

  result <- readLines("pipeline.nix")

  # Create snapshot of the pipeline
  snapshot_pipeline <- function(result) {
    tfile <- tempfile(pattern = "pipeline_with_deps", fileext = ".nix")
    writeLines(result, tfile)
    tfile
  }

  testthat::expect_snapshot_file(
    path = snapshot_pipeline(result),
    name = "pipeline_with_dependencies.nix"
  )
})

#' @title Copy Derivations From the Nix Store to Current Working Directory
#' @family utilities
#' @description When Nix builds a derivation, its output is saved in the
#'   Nix store located under `/nix/store/`. Even though you can import the
#'   derivations into the current R session using `rxp_read()` or `rxp_load()`,
#'   it can be useful to copy the outputs to the current working directory. This
#'   is especially useful for Quarto documents, where there can be more than one
#'   input, as is the case for `html` output.
#' @param derivation_name The name of the derivation to copy. If empty, then all
#'   the derivations are copied.
#' @param dir_mode Character, default "0755". POSIX permission mode to apply to
#'   directories under the copied output (including the top-level output
#'   directory).
#' @param file_mode Character, default "0644". POSIX permission mode to apply to
#'   files under the copied output.
#' @return Nothing, the contents of the Nix store are copied to the current
#'   working directory.
#' @examples
#' \dontrun{
#'   # Copy all derivations to the current working directory
#'   rxp_copy()
#'
#'   # Copy a specific derivation
#'   rxp_copy("mtcars")
#'
#'   # Copy with custom permissions (e.g., make scripts executable)
#'   rxp_copy("my_deriv", dir_mode = "0755", file_mode = "0644")
#'
#'   # Copy a Quarto document output with multiple files
#'   rxp_copy("my_quarto_doc")
#' }
#' @export
rxp_copy <- function(
  derivation_name = NULL,
  dir_mode = "0755",
  file_mode = "0644"
) {
  # Validate permission modes: accept 3 or 4 octal digits as characters
  valid_mode <- function(x) is.character(x) && grepl("^[0-7]{3,4}$", x)
  if (!valid_mode(dir_mode)) {
    stop(
      "Invalid dir_mode: provide a character octal like \"0755\" or \"755\"."
    )
  }
  if (!valid_mode(file_mode)) {
    stop(
      "Invalid file_mode: provide a character octal like \"0644\" or \"644\"."
    )
  }

  build_log <- rxp_inspect()
  output_dir <- file.path(getwd(), "pipeline-output")

  if (isFALSE(dir.exists(output_dir))) {
    dir.create(path = output_dir, recursive = TRUE)
  }

  if (
    !is.null(derivation_name) && !(derivation_name %in% build_log$derivation)
  ) {
    stop(
      "No derivation ",
      derivation_name,
      " found in the Nix store.\n  Did you build the pipeline with rxp_make()?"
    )
  }

  # Helper to apply permissions safely (POSIX only)
  apply_permissions <- function(root_dir, dir_mode, file_mode) {
    if (!dir.exists(root_dir)) {
      return(invisible(NULL))
    }

    # Directories (include root_dir itself)
    dirs <- unique(c(
      root_dir,
      list.dirs(root_dir, recursive = TRUE, full.names = TRUE)
    ))
    if (length(dirs)) {
      try(
        suppressWarnings(Sys.chmod(dirs, mode = dir_mode, use_umask = FALSE)),
        silent = TRUE
      )
    }

    # Files
    files <- list.files(
      root_dir,
      recursive = TRUE,
      full.names = TRUE,
      include.dirs = FALSE
    )
    if (length(files)) {
      try(
        suppressWarnings(Sys.chmod(files, mode = file_mode, use_umask = FALSE)),
        silent = TRUE
      )
    }
  }

  status <- 0L

  if (!is.null(derivation_name)) {
    derivation_path <- subset(
      build_log,
      subset = derivation == derivation_name
    )$path
    path_to_deriv <- list.files(
      derivation_path,
      full.names = TRUE
    )
    file.copy(from = path_to_deriv, to = output_dir, recursive = TRUE)
    apply_permissions(output_dir, dir_mode = dir_mode, file_mode = file_mode)
    status <- 1L
  } else {
    derivation_name <- "all-derivations"
    derivation_path <- subset(
      build_log,
      subset = derivation == derivation_name
    )$path
    files <- list.files(derivation_path, full.names = TRUE)
    file.copy(from = files, to = output_dir, recursive = TRUE)
    apply_permissions(output_dir, dir_mode = dir_mode, file_mode = file_mode)
    status <- 1L
  }

  if (status == 1L) {
    message("Copy successful, check out ", output_dir)
  } else {
    stop("Copy unsuccessful: did you build the pipeline?")
  }
}

test_that("rxp_copy: test for failure with non-existent derivation", {
  testthat::skip_if_not_installed("mockery")
  # Create a temporary mock of build_log.rds
  mock_build_log <- data.frame(
    derivation = c("existing_derivation1", "existing_derivation2"),
    build_success = c(TRUE, TRUE),
    path = c("/nix/store/path1", "/nix/store/path2"),
    output = I(list("output1", "output2"))
  )

  # Create temp directory structure for the test
  temp_dir <- tempdir()
  rixpress_dir <- file.path(temp_dir, "_rixpress")
  dir.create(rixpress_dir, recursive = TRUE, showWarnings = FALSE)

  # Save mock build log
  saveRDS(mock_build_log, file.path(rixpress_dir, "build_log.rds"))

  # Temporarily modify working directory
  old_wd <- getwd()
  setwd(temp_dir)

  # Mock rxp_inspect to return our mock data
  mockery::stub(rxp_copy, "rxp_inspect", mock_build_log)

  # Test that rxp_copy errors for a non-existent derivation
  expect_error(
    rxp_copy("non_existent_derivation"),
    "No derivation non_existent_derivation found in the Nix store"
  )

  # Clean up
  setwd(old_wd)
})

#' Generate a DAG From a List of Derivations
#'
#' Creates a JSON representation of a directed acyclic graph (DAG)
#' based on dependencies between derivations. Is automatically called
#' by `rxp_populate()`.
#'
#' @family ci utilities
#' @param rxp_list A list of derivations.
#' @param output_file Path to the output JSON file.
#'   Defaults to "_rixpress/dag.json".
#' @importFrom jsonlite write_json
#' @return Nothing, writes a JSON file representing the DAG.
#' @examples
#' \dontrun{
#'   rxp_write_dag(rxp_list)
#' }
#' @export
rxp_write_dag <- function(rxp_list, output_file = "_rixpress/dag.json") {
  # Validate inputs
  if (!is.list(rxp_list)) {
    stop("rxp_list must be a list of derivations")
  }

  # Create output directory if it doesn't exist
  dir.create(dirname(output_file), recursive = TRUE, showWarnings = FALSE)

  # Extract all derivation names for dependency checking
  all_derivs_names <- sapply(rxp_list, function(x) (x$name))

  n <- length(rxp_list)
  dag <- vector("list", n)

  # First pass: Process each derivation to extract dependencies
  for (i in seq_along(rxp_list)) {
    deriv <- rxp_list[[i]]
    name <- deriv$name
    type <- deriv$type
    decoder <- deriv$decoder
    noop_build <- if (is.null(deriv$noop_build)) FALSE else deriv$noop_build

    # Extract dependencies based on derivation type
    deps <- extract_dependencies(
      deriv,
      type,
      name,
      all_derivs_names
    )

    # Extract pipeline metadata (defaults to "default" group if not set)
    pipeline_group <- if (!is.null(deriv$pipeline_group)) {
      deriv$pipeline_group
    } else {
      "default"
    }
    pipeline_color <- deriv$pipeline_color # Can be NULL

    # Add the derivation to the DAG
    dag[[i]] <- list(
      deriv_name = name,
      depends = deps,
      decoder = decoder,
      type = type,
      noop_build = noop_build,
      pipeline_group = pipeline_group,
      pipeline_color = pipeline_color
    )
  }

  # Second pass: Propagate no-op builds to dependent derivations
  dag <- propagate_noop_builds(dag)

  # Write the DAG to a JSON file
  jsonlite::write_json(list(derivations = dag), output_file, pretty = TRUE)

  # Return the output file path when running tests
  if (identical(Sys.getenv("TESTTHAT"), "true")) {
    output_file
  }
}

#' Propagate No-Op Builds to Dependent Derivations
#'
#' @param dag A list of derivation nodes with dependencies
#' @return Updated DAG with propagated no-op flags
#' @importFrom stats setNames
#' @noRd
propagate_noop_builds <- function(dag) {
  # Create a lookup for easier access
  dag_lookup <- setNames(dag, sapply(dag, function(x) x$deriv_name))

  # Keep iterating until no more changes occur
  changed <- TRUE
  while (changed) {
    changed <- FALSE

    for (i in seq_along(dag)) {
      # Skip if already marked as no-op
      if (dag[[i]]$noop_build) {
        next
      }

      # Check if any dependencies are no-op
      deps <- dag[[i]]$depends
      if (length(deps) > 0) {
        dep_noop_flags <- sapply(deps, function(dep_name) {
          dep_node <- dag_lookup[[dep_name]]
          if (is.null(dep_node)) {
            warning("Dependency '", dep_name, "' not found in DAG")
            return(FALSE)
          }
          dep_node$noop_build
        })

        # If any dependency is no-op, mark this derivation as no-op too
        if (any(dep_noop_flags)) {
          dag[[i]]$noop_build <- TRUE
          dag_lookup[[dag[[i]]$deriv_name]]$noop_build <- TRUE
          changed <- TRUE

          message(
            "Derivation '",
            dag[[i]]$deriv_name,
            "' marked as no-op due to no-op dependency: ",
            paste(names(dep_noop_flags)[dep_noop_flags], collapse = ", ")
          )
        }
      }
    }
  }

  dag
}

#' Extract Dependencies for a Derivation
#'
#' @param deriv The derivation object
#' @param type The type of derivation
#' @param name The name of the derivation
#' @param all_derivs_names All derivation names
#' @return A character vector of dependency names
#' @noRd
extract_dependencies <- function(
  deriv,
  type,
  name,
  all_derivs_names
) {
  # Extract all potential dependencies based on type
  all_deps <- if (type %in% c("rxp_r", "rxp_py2r", "rxp_r2py")) {
    extract_r_dependencies(deriv, name, all_derivs_names)
  } else if (type %in% c("rxp_qmd", "rxp_rmd")) {
    extract_markdown_dependencies(deriv, type, name, all_derivs_names)
  } else if (type %in% c("rxp_py", "rxp_r2py")) {
    extract_python_dependencies(deriv, name, all_derivs_names)
  } else if (type == "rxp_jl") {
    extract_julia_dependencies(deriv, name, all_derivs_names)
  } else {
    stop("Unknown derivation type: ", type)
  }

  # Return all dependencies except the current derivation itself
  setdiff(all_deps, name)
}

#' Extract Dependencies from R Script Derivations
#'
#' @param deriv The derivation object
#' @param name The name of the derivation
#' @param all_derivs_names All derivation names
#' @return A character vector of dependency names
#' @noRd
extract_r_dependencies <- function(deriv, name, all_derivs_names) {
  snippet <- deriv$snippet

  # Extract the content inside the Rscript -e quotes (allowing for multiple lines)
  m <- regexec('Rscript -e \\"([\\s\\S]*?)\\"', snippet, perl = TRUE)
  match <- regmatches(snippet, m)

  # Filter out the current derivation from potential dependencies
  derivs_to_consider <- Filter(
    function(x) `!=`(x, name),
    all_derivs_names
  )

  # Check which derivations are referenced in the script
  deps <- sapply(
    derivs_to_consider,
    function(name) any(grepl(paste0("\\b", name, "\\b"), unlist(match)))
  )

  # Only keep dependencies that are TRUE
  deps <- Filter(isTRUE, deps)
  names(deps)
}

#' Extract Dependencies from Markdown (Quarto/RMarkdown) Derivations
#'
#' @param deriv The derivation object
#' @param type The type of derivation ("rxp_qmd" or "rxp_rmd")
#' @param name The name of the derivation
#' @param all_derivs_names All derivation names
#' @return A character vector of dependency names
#' @noRd
extract_markdown_dependencies <- function(deriv, type, name, all_derivs_names) {
  # Determine file path and extension based on type
  if (type == "rxp_qmd") {
    # Try both .qmd and .Qmd extensions
    doc_file <- deriv$qmd_file
    if (!file.exists(doc_file)) {
      doc_file <- paste0(name, ".Qmd")
      if (!file.exists(doc_file)) {
        stop("Quarto file not found: ", name, ".qmd or ", name, ".Qmd")
      }
    }
  } else {
    # rxp_rmd
    # Try .rmd and .Rmd extensions
    doc_file <- deriv$rmd_file
    if (!file.exists(doc_file)) {
      doc_file <- paste0(name, ".Rmd")
      if (!file.exists(doc_file)) {
        stop("R Markdown file not found: ", name, ".rmd or ", name, ".Rmd")
      }
    }
  }

  # Read the document file
  doc_content <- readLines(doc_file, warn = FALSE)
  doc_text <- paste(doc_content, collapse = "\n")

  # Extract R code chunks (between ```{r} and ```)
  chunk_pattern <- "```\\{r\\}[\\s\\S]*?```"
  chunks <- regmatches(
    doc_text,
    gregexpr(chunk_pattern, doc_text, perl = TRUE)
  )[[1]]

  # Process each chunk to find rxp_read() or rxp_load() calls
  deps <- character(0)
  for (chunk in chunks) {
    # Remove the ```{r} and ``` delimiters
    code <- sub("```\\{r\\}\\s*", "", chunk)
    code <- sub("```\\s*$", "", code)

    # Match rxp_read("name") or rxp_load("name"), with or without rixpress::
    pattern <- "(rixpress::)?rxp_(read|load)\\s*\\(\\s*['\"](\\w+)['\"]\\s*\\)"
    matches <- regmatches(code, gregexpr(pattern, code, perl = TRUE))[[1]]

    # Extract the dependency names (group 3 in the pattern)
    if (length(matches) > 0) {
      dep_names <- vapply(
        matches,
        function(m) {
          regmatches(m, regexec(pattern, m, perl = TRUE))[[1]][4]
        },
        character(1)
      )
      deps <- union(deps, dep_names)
    }
  }

  # Return all potential dependencies (filtering happens in extract_dependencies)
  deps
}

#' Extract Dependencies from Python Script Derivations
#'
#' @param deriv The derivation object
#' @param name The name of the derivation
#' @param all_derivs_names All derivation names
#' @return A character vector of dependency names
#' @noRd
extract_python_dependencies <- function(deriv, name, all_derivs_names) {
  snippet <- deriv$snippet

  # Extract the content inside the python -c quotes (allowing for multiple lines)
  m <- regexec('python -c \\"([\\s\\S]*?)\\"', snippet, perl = TRUE)
  match <- regmatches(snippet, m)

  # Filter out the current derivation from potential dependencies
  derivs_to_consider <- Filter(
    function(x) `!=`(x, name),
    all_derivs_names
  )

  # Check which derivations are referenced in the script
  deps <- sapply(
    derivs_to_consider,
    function(name) any(grepl(paste0("\\b", name, "\\b"), unlist(match)))
  )

  # Only keep dependencies that are TRUE
  deps <- Filter(isTRUE, deps)
  names(deps)
}

#' Extract Dependencies from Julia Script Derivations
#'
#' @param deriv The derivation object
#' @param name The name of the derivation
#' @param all_derivs_names All derivation names
#' @return A character vector of dependency names
#' @noRd
extract_julia_dependencies <- function(deriv, name, all_derivs_names) {
  snippet <- deriv$snippet

  # Extract everything between `julia -e "` and the following `'';` (dot‐all mode)
  pat <- '(?s)julia\\s+-e\\s+"(.+?)"\\s*\'\';'
  m <- regexec(pat, snippet, perl = TRUE)
  captures <- regmatches(snippet, m)

  if (length(captures) > 0 && length(captures[[1]]) >= 2) {
    match <- captures[[1]][2]
  } else {
    match <- ""
  }

  # Filter out the current derivation from potential dependencies
  derivs_to_consider <- Filter(
    function(x) `!=`(x, name),
    all_derivs_names
  )

  # Check which derivations are referenced in the script
  deps <- sapply(
    derivs_to_consider,
    function(name) any(grepl(paste0("\\b", name, "\\b"), match))
  )

  # Only keep dependencies that are TRUE
  deps <- Filter(isTRUE, deps)
  names(deps)
}

# Suppress R CMD check notes for non-standard evaluation
utils::globalVariables("color_group")

#' @title Export DAG of Pipeline and Prepare It for Rendering on CI

#' @family ci utilities
#' @description This function generates a DOT file representation of the
#'   pipeline DAG, suitable for visualization, potentially on CI platforms. It
#'   is called by `rxp_ga()`.
#' @param nodes_and_edges List, output of `get_nodes_edges()`. Defaults to
#'   calling `get_nodes_edges()`.
#' @param output_file Character, the path where the DOT file should be saved.
#'   Defaults to `"_rixpress/dag.dot"`. The directory will be created if it
#'   doesn't exist.
#' @return Nothing, writes the DOT file to the specified `output_file`.
#' @importFrom igraph graph_from_data_frame V set_vertex_attr delete_vertex_attr
#'   write_graph
#' @examples
#' \dontrun{
#'   # Generate the default _rixpress/dag.dot
#'   rxp_dag_for_ci()
#'
#' }
#' @export
rxp_dag_for_ci <- function(
  nodes_and_edges = get_nodes_edges(),
  output_file = "_rixpress/dag.dot"
) {
  edges <- nodes_and_edges$edges
  edges <- edges[, c("from", "to")]

  dag_obj <- igraph::graph_from_data_frame(
    edges,
    directed = TRUE,
  )

  dag_obj <- igraph::set_vertex_attr(
    dag_obj,
    "label",
    value = igraph::V(dag_obj)$name
  )

  dag_obj <- igraph::delete_vertex_attr(dag_obj, "name")

  igraph::write_graph(dag_obj, file = output_file, format = "dot")
}

#' @title Prepare Data for Plotting the DAG of the Pipeline
#' @param path_dag Character, defaults to `"_rixpress/dag.json"`.
#' @return A list of two datasets, `nodes` and `edges`.
#' @examples \dontrun{
#'   get_nodes_edges()
#' }
#' @importFrom jsonlite fromJSON
#' @noRd
get_nodes_edges <- function(path_dag = "_rixpress/dag.json") {
  # Check if the DAG file exists
  if (!file.exists(path_dag)) {
    stop("dag.json missing! Did you run 'rxp_populate()'?")
  }

  json_data <- fromJSON(path_dag)

  derivations <- json_data$derivations

  # Extract node information including pipeline metadata
  n <- derivations[, c("deriv_name", "type")]
  nn <- unnest_all_columns(n)

  # Get the number of rows (after unnesting)
  n_rows <- nrow(nn)

  # Extract pipeline_group (default to "default" if not present or null)
  if (
    "pipeline_group" %in%
      names(derivations) &&
      !is.null(derivations$pipeline_group)
  ) {
    pipeline_group <- vapply(
      derivations$pipeline_group,
      function(x) {
        if (is.null(x) || length(x) == 0 || (length(x) == 1 && is.na(x))) {
          "default"
        } else {
          as.character(x)
        }
      },
      character(1)
    )
  } else {
    pipeline_group <- rep("default", n_rows)
  }

  # Extract pipeline_color (can be NULL)
  if (
    "pipeline_color" %in%
      names(derivations) &&
      !is.null(derivations$pipeline_color)
  ) {
    pipeline_color <- vapply(
      derivations$pipeline_color,
      function(x) {
        if (is.null(x) || length(x) == 0 || (length(x) == 1 && is.na(x))) {
          NA_character_
        } else {
          as.character(x)
        }
      },
      character(1)
    )
  } else {
    pipeline_color <- rep(NA_character_, n_rows)
  }

  # Ensure lengths match (after unnesting, there might be differences)
  if (length(pipeline_group) != n_rows) {
    pipeline_group <- rep("default", n_rows)
  }
  if (length(pipeline_color) != n_rows) {
    pipeline_color <- rep(NA_character_, n_rows)
  }

  nodes_df <- data.frame(
    "id" = nn$deriv_name,
    "label" = nn$deriv_name,
    "group" = nn$type,
    "pipeline_group" = pipeline_group,
    "pipeline_color" = pipeline_color,
    stringsAsFactors = FALSE
  )
  nodes <- unique(nodes_df)

  e <- derivations[, c("deriv_name", "depends")]
  ee <- unnest_all_columns(e)
  edges <- data.frame(
    "from" = ee$depends,
    "to" = ee$deriv_name,
    "arrows" = "to",
    stringsAsFactors = FALSE
  )

  list(
    "nodes" = nodes,
    "edges" = edges
  )
}


#' @title Create a Directed Acyclic Graph (DAG) Representing the Pipeline
#'   Using `{ggplot2}`
#' @family visualisation functions
#' @description Uses `{ggdag}` to generate the plot. `{ggdag}` is a soft
#'   dependency of `{rixpress}` so you need to install it to use this
#'   function. When derivations are organized into pipelines using
#'   `rxp_pipeline()`, nodes use a dual-encoding approach: the interior fill
#'   shows the derivation type (R, Python, etc.) while a thick border shows
#'   the pipeline group colour.
#' @param nodes_and_edges List, output of `get_nodes_edges()`.
#' @param color_by Character, either "pipeline" (default) or "type".
#'   When "pipeline", nodes show type as fill colour and pipeline as border.
#'   When "type", nodes are coloured entirely by derivation type (rxp_r, rxp_py, etc.).
#' @param colour_by Character, alias for `color_by`.
#' @return A `{ggplot2}` object.
#' @examples \dontrun{
#'   rxp_ggdag()  # Dual encoding: fill = type, border = pipeline
#'   rxp_ggdag(colour_by = "type")  # Color entirely by derivation type
#' }
#' @export
rxp_ggdag <- function(
  nodes_and_edges = get_nodes_edges(),
  color_by = c("pipeline", "type"),
  colour_by = NULL
) {
  if (!requireNamespace("ggdag", quietly = TRUE)) {
    stop("You need to install {ggdag} to use this feature.")
  }

  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("You need to install {ggplot2} to use this feature.")
  }

  if (!is.null(colour_by)) {
    color_by <- colour_by
  }

  color_by <- match.arg(color_by)

  nodes <- nodes_and_edges$nodes
  nodes$name <- nodes$id

  # Default color palette for pipeline groups
  default_pipeline_colors <- c(
    "#E69F00",
    "#56B4E9",
    "#009E73",
    "#F0E442",
    "#0072B2",
    "#D55E00",
    "#CC79A7",
    "#999999"
  )

  # Check if we have any non-default pipeline groups
  has_pipelines <- "pipeline_group" %in%
    names(nodes) &&
    any(nodes$pipeline_group != "default")

  if (color_by == "pipeline" && has_pipelines) {
    # Color by pipeline group
    unique_groups <- unique(nodes$pipeline_group)

    # Create a color mapping for groups
    group_colors <- c()
    color_idx <- 1
    for (grp in unique_groups) {
      # Get the first color specified for this group (if any)
      grp_nodes <- nodes[nodes$pipeline_group == grp, ]
      explicit_color <- grp_nodes$pipeline_color[
        !is.na(grp_nodes$pipeline_color)
      ][1]

      if (!is.na(explicit_color)) {
        group_colors[grp] <- explicit_color
      } else {
        # Assign a default color
        group_colors[grp] <- default_pipeline_colors[
          ((color_idx - 1) %% length(default_pipeline_colors)) + 1
        ]
        color_idx <- color_idx + 1
      }
    }

    # Use pipeline_group for coloring
    nodes$color_group <- nodes$pipeline_group
    nodes <- nodes[, c("name", "color_group", "group")] # Keep type for shape

    edges <- nodes_and_edges$edges
    edges$name <- edges$from
    edges <- edges[, c("name", "to")]

    dag_df <- ggdag::as_tidy_dagitty(edges) |>
      merge(nodes, by = "name")

    # Derivation type colors for node fill (interior)
    type_colors <- c(
      "rxp_r" = "#246ABF",
      "rxp_r2py" = "#FFD343",
      "rxp_py" = "#FFD343",
      "rxp_py2r" = "#246ABF",
      "rxp_jl" = "#9558b2",
      "rxp_qmd" = "#4F789E",
      "rxp_rmd" = "#4F789E"
    )

    # Scale for node fill (type)
    rxp_fill_scale <- ggplot2::scale_fill_manual(
      values = type_colors,
      name = "Type"
    )

    # Scale for node border/stroke (pipeline)
    rxp_colour_scale <- ggplot2::scale_colour_manual(
      values = group_colors,
      name = "Pipeline"
    )

    # Use derivation type for shapes
    rxp_shapes <- ggplot2::scale_shape_manual(
      values = c(
        "rxp_r" = 23,
        "rxp_r2py" = 23,
        "rxp_py" = 23,
        "rxp_py2r" = 23,
        "rxp_jl" = 23,
        "rxp_qmd" = 22,
        "rxp_rmd" = 22
      ),
      name = "Type"
    )

    p <- ggplot2::ggplot(
      dag_df,
      ggplot2::aes(x = x, y = y, xend = xend, yend = yend)
    ) +
      ggdag::geom_dag_edges() +
      rxp_fill_scale +
      rxp_colour_scale +
      rxp_shapes +
      # Nodes: fill = type color, colour (border) = pipeline color, thick stroke
      ggdag::geom_dag_node(
        ggplot2::aes(fill = group, colour = color_group, shape = group),
        size = 20,
        stroke = 2.5
      ) +
      ggdag::geom_dag_text(
        ggplot2::aes(label = name),
        col = "black",
        nudge_y = -.3
      ) +
      ggdag::theme_dag()

    p
  } else {
    # Color by derivation type (original behavior)
    nodes <- nodes[, c("name", "group")]

    edges <- nodes_and_edges$edges
    edges$name <- edges$from
    edges <- edges[, c("name", "to")]

    dag_df <- ggdag::as_tidy_dagitty(edges) |>
      merge(nodes, by = "name")

    rxp_scale <- ggplot2::scale_fill_manual(
      values = c(
        "rxp_r" = "#246ABF",
        "rxp_r2py" = "#FFD343",
        "rxp_py" = "#FFD343",
        "rxp_py2r" = "#246ABF",
        "rxp_jl" = "#9558b2",
        "rxp_qmd" = "#4F789E",
        "rxp_rmd" = "#4F789E"
      ),
      name = "Type"
    )

    rxp_shapes <- ggplot2::scale_shape_manual(
      values = c(
        "rxp_r" = 23,
        "rxp_r2py" = 23,
        "rxp_py" = 23,
        "rxp_py2r" = 23,
        "rxp_jl" = 23,
        "rxp_qmd" = 22,
        "rxp_rmd" = 22
      ),
      name = "Type"
    )

    ggplot2::ggplot(
      dag_df,
      ggplot2::aes(x = x, y = y, xend = xend, yend = yend)
    ) +
      ggdag::geom_dag_edges() +
      rxp_scale +
      rxp_shapes +
      ggdag::geom_dag_node(ggplot2::aes(fill = group, shape = group)) +
      ggdag::geom_dag_text(
        ggplot2::aes(label = name),
        col = "black",
        nudge_y = -.3
      ) +
      ggdag::theme_dag()
  }
}

#' @title Create a Directed Acyclic Graph (DAG) Representing the Pipeline
#'   Using `{visNetwork}`
#' @family visualisation functions
#' @description Uses `{visNetwork}` to generate the plot. `{visNetwork}` is a
#'   soft dependency of `{rixpress}` so you need to install it to use this
#'   function. When derivations are organized into pipelines using
#'   `rxp_pipeline()`, nodes use a dual-encoding approach: the interior fill
#'   shows the derivation type (R, Python, etc.) while the border shows
#'   the pipeline group colour.
#' @param nodes_and_edges List, output of `get_nodes_edges()`.
#' @param color_by Character, either "pipeline" (default) or "type".
#'   When "pipeline", nodes show type as fill colour and pipeline as border.
#'   When "type", nodes are colored by their derivation type (rxp_r, rxp_py, etc.).
#' @param colour_by Character, alias for `color_by`.
#' @return Nothing, this function opens a new tab in your browser with
#'   the DAG generated using `{visNetwork}`.
#' @examples \dontrun{
#'   rxp_visnetwork()
#'   rxp_visnetwork(colour_by = "type")  # Color by derivation type instead
#' }
#' @export
#' @importFrom jsonlite fromJSON
rxp_visnetwork <- function(
  nodes_and_edges = get_nodes_edges(),
  color_by = c("pipeline", "type"),
  colour_by = NULL
) {
  if (!requireNamespace("visNetwork", quietly = TRUE)) {
    stop("You need to install {visNetwork} to use this feature.")
  }

  if (!is.null(colour_by)) {
    color_by <- colour_by
  }

  color_by <- match.arg(color_by)

  nodes <- nodes_and_edges$nodes
  edges <- nodes_and_edges$edges

  # Default color palette for pipeline groups (used when no color specified)
  default_pipeline_colors <- c(
    "#E69F00",
    "#56B4E9",
    "#009E73",
    "#F0E442",
    "#0072B2",
    "#D55E00",
    "#CC79A7",
    "#999999"
  )

  # Derivation type colors (original colors)
  type_colors <- c(
    "rxp_r" = "#246ABF",
    "rxp_jl" = "#9558b2",
    "rxp_r2py" = "#FFD343",
    "rxp_py" = "#FFD343",
    "rxp_py2r" = "#246ABF",
    "rxp_qmd" = "#4F789E",
    "rxp_rmd" = "#4F789E"
  )

  # Derivation type shapes
  type_shapes <- c(
    "rxp_r" = "diamond",
    "rxp_jl" = "diamond",
    "rxp_r2py" = "triangleDown",
    "rxp_py" = "diamond",
    "rxp_py2r" = "triangleDown",
    "rxp_qmd" = "box",
    "rxp_rmd" = "box"
  )

  # Check if we have any non-default pipeline groups
  has_pipelines <- "pipeline_group" %in%
    names(nodes) &&
    any(nodes$pipeline_group != "default")

  if (color_by == "pipeline" && has_pipelines) {
    # Color by pipeline group
    unique_groups <- unique(nodes$pipeline_group)

    # Create a color mapping for groups without explicit colors
    group_colors <- list()
    color_idx <- 1
    for (grp in unique_groups) {
      # Get the first color specified for this group (if any)
      grp_nodes <- nodes[nodes$pipeline_group == grp, ]
      explicit_color <- grp_nodes$pipeline_color[
        !is.na(grp_nodes$pipeline_color)
      ][1]

      if (!is.na(explicit_color)) {
        group_colors[[grp]] <- explicit_color
      } else {
        # Assign a default color
        group_colors[[grp]] <- default_pipeline_colors[
          ((color_idx - 1) %% length(default_pipeline_colors)) + 1
        ]
        color_idx <- color_idx + 1
      }
    }

    # Assign colors to nodes based on their pipeline group and derivation type
    nodes$color <- lapply(
      seq_len(nrow(nodes)),
      function(i) {
        pipeline_grp <- nodes$pipeline_group[i]
        deriv_type <- nodes$group[i]

        border_col <- group_colors[[pipeline_grp]]
        # Fallback for type color if not found
        bg_col <- if (deriv_type %in% names(type_colors)) {
          type_colors[[deriv_type]]
        } else {
          "#97C2FC" # Standard visNetwork blue
        }

        list(
          background = bg_col,
          border = border_col,
          highlight = list(
            background = bg_col,
            border = border_col
          ),
          hover = list(
            background = bg_col,
            border = border_col
          )
        )
      }
    )

    # Set border width to make it visible
    nodes$borderWidth <- 3

    # Use derivation type for shape
    nodes$shape <- vapply(
      nodes$group,
      function(t) {
        if (t %in% names(type_shapes)) type_shapes[[t]] else "dot"
      },
      character(1)
    )

    # Change group to pipeline_group for legend
    nodes$group <- nodes$pipeline_group

    # Build the network with dynamic groups
    net <- visNetwork::visNetwork(nodes, edges)

    for (grp in unique_groups) {
      net <- visNetwork::visGroups(
        net,
        groupname = grp,
        color = group_colors[[grp]],
        font = list(align = "top", size = 14),
        size = 20
      )
    }

    net <- net |>
      visNetwork::visLegend(
        position = "right",
        main = "Pipeline Groups"
      ) |>
      visNetwork::visExport(
        type = "png",
        name = "export-network",
        float = "left",
        label = "Save DAG",
        background = "white",
        style = ""
      )

    return(net)
  } else {
    # Color by derivation type (original behavior)
    visNetwork::visNetwork(nodes, edges) |>
      visNetwork::visGroups(
        groupname = "rxp_r",
        shape = "diamond",
        color = "#246ABF",
        font = list(align = "top", size = 14),
        size = 20
      ) |>
      visNetwork::visGroups(
        groupname = "rxp_jl",
        shape = "diamond",
        color = "#9558b2",
        font = list(align = "top", size = 14),
        size = 20
      ) |>
      visNetwork::visGroups(
        groupname = "rxp_r2py",
        shape = "triangleDown",
        color = "#FFD343",
        font = list(align = "top", size = 14),
        size = 10
      ) |>
      visNetwork::visGroups(
        groupname = "rxp_py",
        shape = "diamond",
        color = "#FFD343",
        font = list(align = "top", size = 14),
        size = 20
      ) |>
      visNetwork::visGroups(
        groupname = "rxp_py2r",
        shape = "triangleDown",
        color = "#246ABF",
        font = list(align = "top", size = 14),
        size = 10
      ) |>
      visNetwork::visGroups(
        groupname = "rxp_qmd",
        shape = "box",
        color = "#4F789E",
        font = list(align = "top", size = 40)
      ) |>
      visNetwork::visGroups(
        groupname = "rxp_rmd",
        shape = "box",
        color = "#4F789E",
        font = list(align = "top", size = 40)
      ) |>
      visNetwork::visLegend(
        position = "right",
        main = "Derivation Types"
      ) |>
      visNetwork::visExport(
        type = "png",
        name = "export-network",
        float = "left",
        label = "Save DAG",
        background = "white",
        style = ""
      )
  }
}

#' @noRd
unnest_all_columns <- function(df) {
  # Identify list-columns
  list_cols <- sapply(df, is.list)

  # If no list-columns, return original data frame
  if (!any(list_cols)) {
    return(df)
  }

  # Initialize result with non-list columns
  result <- df[!list_cols]

  # Process each list-column
  for (col_name in names(df)[list_cols]) {
    col_data <- df[[col_name]]

    # Determine the type of the first non-NULL element
    first_non_null <- Filter(Negate(is.null), col_data)[[1]]

    if (is.data.frame(first_non_null)) {
      # Unnest list of data frames
      unnested <- do.call(
        rbind,
        lapply(col_data, function(x) {
          if (is.null(x)) {
            # Replace NULL with NA row
            as.data.frame(matrix(
              NA,
              nrow = 1,
              ncol = ncol(first_non_null),
              dimnames = list(NULL, names(first_non_null))
            ))
          } else {
            x
          }
        })
      )
      # Combine with result
      result <- cbind(result, unnested)
    } else if (is.atomic(first_non_null)) {
      # Unnest list of atomic vectors
      lengths <- sapply(col_data, length)
      repeated_rows <- rep(1:nrow(df), lengths)
      values <- unlist(col_data)
      result <- result[repeated_rows, , drop = FALSE]
      result[[col_name]] <- values
    } else {
      warning(paste("Column", col_name, "contains unsupported data types."))
    }
  }

  result
}

test_that("get_nodes_edges processes dag.json correctly", {
  testthat::skip_on_cran()
  test_dag_path <- "test-data/dag.json"

  actual_output <- get_nodes_edges(path_dag = test_dag_path)

  expected_output <- list(
    nodes = data.frame(
      id = c(
        "mtcars",
        "mtcars_am",
        "mtcars_head",
        "mtcars_tail",
        "mtcars_mpg",
        "page"
      ),
      label = c(
        "mtcars",
        "mtcars_am",
        "mtcars_head",
        "mtcars_tail",
        "mtcars_mpg",
        "page"
      ),
      group = c("rxp_r", "rxp_r", "rxp_r", "rxp_r", "rxp_r", "rxp_qmd"),
      pipeline_group = rep("default", 6),
      pipeline_color = rep(NA_character_, 6),
      stringsAsFactors = FALSE
    ),
    edges = data.frame(
      from = c(
        "mtcars",
        "mtcars_am",
        "mtcars_head",
        "mtcars_tail",
        "mtcars_head",
        "mtcars_tail",
        "mtcars_mpg"
      ),
      to = c(
        "mtcars_am",
        "mtcars_head",
        "mtcars_tail",
        "mtcars_mpg",
        "page",
        "page",
        "page"
      ),
      arrows = rep("to", 7),
      stringsAsFactors = FALSE
    )
  )

  expect_equal(actual_output, expected_output)
})

test_that("rxp_dag_for_ci generates correct .dot file", {
  testthat::skip_on_cran()
  test_dag_path <- "test-data/dag.json"

  nodes_and_edges <- get_nodes_edges(path_dag = test_dag_path)

  snapshot_rxp_dag_for_ci <- function(nodes_and_edges) {
    tfile <- tempfile(pattern = "dag_ci", fileext = ".dot")

    rxp_dag_for_ci(nodes_and_edges = nodes_and_edges, output_file = tfile)

    tfile
  }

  expect_snapshot_file(
    path = snapshot_rxp_dag_for_ci(nodes_and_edges),
    name = "rxp_dag_for_ci.dot"
  )
})

test_that("get_nodes_edges errors if dag.json is missing", {
  # Use a path that is highly unlikely to exist
  non_existent_path <- "a/very/unlikely/path/to/dag.json"

  # Expect the specific error message
  expect_error(
    get_nodes_edges(path_dag = non_existent_path),
    regexp = "dag\\.json missing! Did you run 'rxp_populate\\(\\)'\\?",
    fixed = FALSE # Use regexp matching because of special characters
  )
})

test_that("unnest_all_columns correctly unnests data frames", {
  # Create a test data frame with a list column containing data frames
  test_df <- data.frame(
    id = 1:2,
    stringsAsFactors = FALSE
  )
  test_df$nested <- list(
    data.frame(
      colA = c("a", "b"),
      colB = c("x", "y"),
      stringsAsFactors = FALSE
    ),
    data.frame(colA = c("c", "d"), colB = c("z", "w"), stringsAsFactors = FALSE)
  )

  result <- unnest_all_columns(test_df)

  expected <- data.frame(
    id = c(1, 2, 1, 2),
    colA = c("a", "b", "c", "d"),
    colB = c("x", "y", "z", "w"),
    stringsAsFactors = FALSE
  )

  expect_equal(result, expected)
})

#' Garbage Collect Rixpress Build Artifacts and Logs
#'
#' This function performs garbage collection on Nix store paths and build log files
#' generated by rixpress. It can operate in two modes: full garbage collection
#' (when \code{keep_since = NULL}) or targeted deletion based on log file age.
#'
#' @param keep_since Date or character string (YYYY-MM-DD format). If provided,
#'   only build logs older than this date will be targeted for deletion, along
#'   with their associated Nix store paths. If \code{NULL}, performs a full
#'   Nix garbage collection. Default is \code{NULL}.
#' @param project_path Character string specifying the path to the project
#'   directory containing the \code{_rixpress} folder with build logs.
#'   Default is \code{"."} (current directory).
#' @param dry_run Logical. If \code{TRUE}, shows what would be deleted without
#'   actually performing any deletions. Useful for previewing the cleanup
#'   operation. Default is \code{FALSE}.
#' @param timeout_sec Numeric. Timeout in seconds for individual Nix commands.
#'   Also used for concurrency lock expiration. Default is 300 seconds.
#' @param verbose Logical. If \code{TRUE}, provides detailed output including
#'   full paths, command outputs, and diagnostic information about references
#'   preventing deletion. Default is \code{FALSE}.
#' @param ask Logical. If \code{TRUE}, ask for user confirmation before
#'   performing deleting artifacts. Default is \code{TRUE}.
#'
#' @return Invisibly returns a list with cleanup summary information:
#' \itemize{
#'   \item \code{kept}: Vector of build log filenames that were kept
#'   \item \code{deleted}: Vector of build log filenames targeted for deletion
#'   \item \code{protected}: Number of store paths protected via GC roots (date-based mode)
#'   \item \code{deleted_count}: Number of store paths successfully deleted
#'   \item \code{failed_count}: Number of store paths that failed to delete
#'   \item \code{referenced_count}: Number of store paths skipped due to references
#'   \item \code{log_files_deleted}: Number of build log files successfully deleted
#'   \item \code{log_files_failed}: Number of build log files that failed to delete
#'   \item \code{dry_run_details}: List of detailed information when dry_run = TRUE
#' }
#'
#' @details
#' The function operates in two modes:
#'
#' \strong{Full Garbage Collection Mode} (\code{keep_since = NULL}):
#' \itemize{
#'   \item Runs \code{nix-store --gc} to delete all unreferenced store paths
#'   \item Does not delete any build log files
#'   \item Suitable for complete cleanup of unused Nix store paths
#' }
#'
#' \strong{Targeted Deletion Mode} (\code{keep_since} specified):
#' \itemize{
#'   \item Identifies build logs older than the specified date
#'   \item Extracts store paths from old logs using \code{rxp_inspect()}
#'   \item Protects recent store paths by creating temporary GC roots
#'   \item Attempts to delete old store paths individually using \code{nix-store --delete}
#'   \item Deletes the corresponding build log \code{.json} files from \code{_rixpress/}
#'   \item Handles referenced paths gracefully (paths that cannot be deleted due to dependencies)
#' }
#'
#' \strong{Concurrency Safety:}
#' The function uses a lock file mechanism to prevent multiple instances from
#' running simultaneously, which could interfere with each other's GC root management.
#'
#' \strong{Reference Handling:}
#' Some store paths may not be deletable because they are still referenced by:
#' \itemize{
#'   \item User or system profile generations
#'   \item Active Nix shell environments
#'   \item Result symlinks in project directories
#'   \item Other store paths that depend on them
#' }
#' These paths are reported but not considered errors.
#'
#' @examples
#' \dontrun{
#' # Preview what would be deleted (dry run)
#' rxp_gc(keep_since = "2025-08-01", dry_run = TRUE, verbose = TRUE)
#'
#' # Delete artifacts from builds older than August 1st, 2025
#' rxp_gc(keep_since = "2025-08-01")
#'
#' # Full garbage collection of all unreferenced store paths
#' rxp_gc()
#'
#' # Clean up artifacts older than 30 days ago
#' rxp_gc(keep_since = Sys.Date() - 30)
#' }
#'
#' @seealso
#' \code{\link{rxp_list_logs}}, \code{\link{rxp_inspect}}
#'
#' @family utilities
#' @export
rxp_gc <- function(
  keep_since = NULL,
  project_path = ".",
  dry_run = FALSE,
  timeout_sec = 300,
  verbose = FALSE,
  ask = TRUE
) {
  nix_bin <- Sys.which("nix-store")
  if (!nzchar(nix_bin)) {
    stop("nix-store not found on PATH. Install Nix or adjust PATH.")
  }

  safe_system2 <- function(command, args, timeout = timeout_sec) {
    out <- tryCatch(
      system2(command, args, stdout = TRUE, stderr = TRUE, timeout = timeout),
      error = function(e) {
        if (grepl("timeout", conditionMessage(e), ignore.case = TRUE)) {
          stop("Command '", command, "' timed out after ", timeout, " seconds.")
        }
        stop("Command '", command, "' failed: ", conditionMessage(e))
      }
    )
    status <- attr(out, "status")
    if (!is.null(status) && status != 0) {
      stop(
        "Command '",
        command,
        " ",
        paste(args, collapse = " "),
        "' failed (exit ",
        status,
        ").\n",
        paste(out, collapse = "\n")
      )
    }
    out
  }

  validate_store_paths <- function(paths) {
    if (length(paths) == 0) {
      return(character(0))
    }
    paths <- unique(paths[nzchar(paths)])
    paths <- paths[grepl("^/nix/store/[a-z0-9]{32}-", paths)]
    # Only return paths that actually exist on the filesystem
    paths[file.exists(paths) | dir.exists(paths)]
  }

  extract_which_log <- function(filename) {
    # Expect: build_log_YYYYMMDD_HHMMSS_*.rds -> returns HHMMSS
    m <- regexec("build_log_[0-9]{8}_([0-9]{6})_", filename)
    r <- regmatches(filename, m)[[1]]
    if (length(r) >= 2) r[2] else NA_character_
  }

  get_paths_from_logs <- function(filenames) {
    # Returns named list: filename -> character() of store paths
    out <- setNames(vector("list", length(filenames)), filenames)
    for (i in seq_along(filenames)) {
      fn <- filenames[i]
      wl <- extract_which_log(fn)
      if (is.na(wl)) {
        warning("Could not parse which_log from filename: ", fn)
        out[[fn]] <- character(0)
        next
      }
      insp <- tryCatch(
        suppressMessages(rxp_inspect(which_log = wl)),
        error = function(e) NULL
      )
      if (is.null(insp) || !is.data.frame(insp)) {
        warning("rxp_inspect failed for ", fn)
        out[[fn]] <- character(0)
        next
      }
      if (!("path" %in% names(insp))) {
        warning("rxp_inspect returned no 'path' column for ", fn)
        out[[fn]] <- character(0)
        next
      }
      out[[fn]] <- validate_store_paths(insp$path)
    }
    out
  }

  # ---- Concurrency lock -----------------------------------------------------
  lock_file <- file.path(tempdir(), "rixpress_gc.lock")
  if (file.exists(lock_file)) {
    # try to detect live process on Unix
    ok_to_remove <- TRUE
    pid <- NA_integer_
    tstamp <- NA_character_
    info <- try(readLines(lock_file, warn = FALSE), silent = TRUE)
    if (!inherits(info, "try-error") && length(info) >= 2) {
      pid <- suppressWarnings(as.integer(info[1]))
      tstamp <- info[2]
      alive <- FALSE
      if (.Platform$OS.type == "unix" && !is.na(pid)) {
        ps_out <- try(
          system2("ps", c("-p", pid), stdout = TRUE, stderr = FALSE),
          silent = TRUE
        )
        if (!inherits(ps_out, "try-error")) {
          alive <- any(grepl(paste0("\\b", pid, "\\b"), ps_out))
        }
      }
      if (alive) {
        stop(
          "Another rxp_gc process appears to be running (PID: ",
          pid,
          "). ",
          "If not, remove the lock: ",
          lock_file
        )
      }
      # Remove if stale (older than timeout)
      ts <- suppressWarnings(as.POSIXct(tstamp))
      if (
        !is.na(ts) && difftime(Sys.time(), ts, units = "secs") <= timeout_sec
      ) {
        ok_to_remove <- FALSE
      }
    }
    if (ok_to_remove) {
      unlink(lock_file)
      message("Removed stale lock file.")
    } else {
      stop("Lock file exists and is recent: ", lock_file)
    }
  }
  writeLines(c(as.character(Sys.getpid()), as.character(Sys.time())), lock_file)
  on.exit(if (file.exists(lock_file)) unlink(lock_file), add = TRUE)

  if (!is.null(keep_since)) {
    if (inherits(keep_since, "character")) {
      keep_since <- tryCatch(as.Date(keep_since), error = function(e) NA)
    }
    if (!inherits(keep_since, "Date") || is.na(keep_since)) {
      stop("Invalid 'keep_since'. Use a Date or 'YYYY-MM-DD' string.")
    }
  }

  project_path <- normalizePath(project_path, mustWork = TRUE)
  all_logs <- rxp_list_logs(project_path)
  if (
    !is.data.frame(all_logs) ||
      !all(c("filename", "modification_time") %in% names(all_logs))
  ) {
    stop(
      "rxp_list_logs() must return a data.frame with columns 'filename' and 'modification_time'."
    )
  }
  if (nrow(all_logs) == 0) {
    message("No build logs found. Nothing to do.")
    return(invisible(NULL))
  }

  # Partition logs
  if (is.null(keep_since)) {
    logs_to_keep <- all_logs
    logs_to_delete <- all_logs[0, ]
  } else {
    logs_to_keep <- all_logs[
      as.Date(all_logs$modification_time) >= keep_since,
      ,
      drop = FALSE
    ]
    logs_to_delete <- all_logs[
      as.Date(all_logs$modification_time) < keep_since,
      ,
      drop = FALSE
    ]
  }

  # Gather paths for keep/delete (via rxp_inspect)
  keep_paths_by_log <- if (nrow(logs_to_keep)) {
    get_paths_from_logs(logs_to_keep$filename)
  } else {
    list()
  }
  delete_paths_by_log <- if (nrow(logs_to_delete)) {
    get_paths_from_logs(logs_to_delete$filename)
  } else {
    list()
  }

  keep_paths_all <- validate_store_paths(unique(unlist(
    keep_paths_by_log,
    use.names = FALSE
  )))
  delete_paths_all <- validate_store_paths(unique(unlist(
    delete_paths_by_log,
    use.names = FALSE
  )))

  summary_info <- list(
    kept = logs_to_keep$filename,
    deleted = logs_to_delete$filename,
    protected = 0L,
    dry_run_details = NULL
  )

  if (!is.null(keep_since) && dry_run) {
    message("--- DRY RUN --- No changes will be made. ---")

    # 1) List logs that would be deleted
    message("Logs that would be deleted (", nrow(logs_to_delete), "):")
    if (nrow(logs_to_delete)) {
      cat("  ", logs_to_delete$filename, sep = "\n  ", "\n")
    } else {
      cat("  (none)\n")
    }

    # 2) For each such log, print its paths (via rxp_inspect) as 'path  output'
    details <- list()
    if (length(delete_paths_by_log)) {
      message("\nArtifacts per log (from rxp_inspect):")
      for (fn in names(delete_paths_by_log)) {
        wl <- extract_which_log(fn)
        insp <- tryCatch(
          suppressMessages(rxp_inspect(which_log = wl)),
          error = function(e) NULL
        )
        cat("\n== ", fn, " ==\n", sep = "")
        if (is.null(insp) || !is.data.frame(insp)) {
          cat("  (rxp_inspect unavailable)\n")
          details[[fn]] <- data.frame(
            path = character(0),
            output = character(0)
          )
          next
        }
        cols <- intersect(c("path", "output"), names(insp))
        df <- unique(insp[cols])
        if (!nrow(df)) {
          cat("  (no paths)\n")
        } else {
          # Print two columns: path and output
          for (j in seq_len(nrow(df))) {
            p <- if ("path" %in% cols) df$path[j] else ""
            o <- if (
              "output" %in% cols && !is.na(df$output[j]) && nzchar(df$output[j])
            ) {
              paste0("  ", df$output[j])
            } else {
              ""
            }
            cat("  ", p, o, "\n", sep = "")
          }
        }
        # keep a clean DF with both columns for return
        if (!("path" %in% names(df))) {
          df$path <- character(nrow(df))
        }
        if (!("output" %in% names(df))) {
          df$output <- character(nrow(df))
        }
        details[[fn]] <- df[, c("path", "output"), drop = FALSE]
      }
    } else {
      message("\n(no artifacts found under logs_to_delete)")
    }

    # 3) Aggregate candidate store paths to delete
    if (length(delete_paths_all)) {
      # Filter to only existing paths for more accurate reporting
      existing_delete_paths <- delete_paths_all[
        file.exists(delete_paths_all) | dir.exists(delete_paths_all)
      ]
      missing_paths <- setdiff(delete_paths_all, existing_delete_paths)

      message(
        "\nAggregate store paths targeted for deletion (deduplicated): ",
        length(delete_paths_all),
        " total, ",
        length(existing_delete_paths),
        " existing, ",
        length(missing_paths),
        " already missing"
      )

      if (length(existing_delete_paths) > 0) {
        message("\nExisting paths that would be deleted:")
        cat("  ", existing_delete_paths, sep = "\n  ", "\n")
      }

      if (length(missing_paths) > 0) {
        message("\nPaths already missing (will be skipped):")
        cat("  ", missing_paths, sep = "\n  ", "\n")
      }

      if (verbose && length(existing_delete_paths) > 0) {
        message("The paths listed above would be deleted in a real run.")
      }
    } else {
      message("\nNo valid Nix store paths were found under the logs to delete.")
    }

    summary_info$dry_run_details <- details

    # Also show which log files would be deleted
    if (nrow(logs_to_delete) > 0) {
      message("\nBuild log files that would be deleted:")
      for (i in seq_len(nrow(logs_to_delete))) {
        log_file <- logs_to_delete$filename[i]
        log_path <- file.path(project_path, "_rixpress", log_file)
        exists_indicator <- if (file.exists(log_path)) "[OK]" else "[X]"
        message("  ", exists_indicator, " ", log_file)
      }
    }

    return(invisible(summary_info))
  }

  if (is.null(keep_since) && dry_run) {
    message(
      "--- DRY RUN --- Would run 'nix-store --gc' (delete all unreferenced store paths). ---"
    )
    if (verbose) {
      message(
        "(Tip: for an approximate preview, run 'nix-collect-garbage -n' from a shell.)"
      )
    }
    return(invisible(summary_info))
  }

  if (is.null(keep_since)) {
    # Full GC mode
    if (
      ask &&
        !utils::askYesNo(
          "Run full Nix garbage collection (delete all unreferenced artifacts)?",
          default = FALSE
        )
    ) {
      message("Operation cancelled.")
      return(invisible(NULL))
    }

    # Run full GC
    message("Running Nix garbage collector...")
    gc_out <- safe_system2(nix_bin, "--gc")
    if (length(gc_out)) {
      if (verbose) {
        cat(gc_out, sep = "\n")
      } else {
        rel <- gc_out[grepl(
          "freed|removing|deleting",
          gc_out,
          ignore.case = TRUE
        )]
        if (length(rel)) {
          message("GC summary (last up to 10 lines):")
          cat(utils::tail(rel, 10), sep = "\n")
        }
      }
    }
    message("Garbage collection complete.")
    return(invisible(summary_info))
  }

  # Targeted deletion mode
  if (nrow(logs_to_delete) == 0) {
    message(
      "No build logs older than ",
      format(keep_since, "%Y-%m-%d"),
      " found. Nothing to do."
    )
    return(invisible(NULL))
  }

  if (length(delete_paths_all) == 0) {
    message(
      "No valid store paths found in logs older than ",
      format(keep_since, "%Y-%m-%d"),
      ". Nothing to delete."
    )
    return(invisible(NULL))
  }

  prompt <- paste0(
    "This will permanently delete ",
    length(delete_paths_all),
    " store paths from ",
    nrow(logs_to_delete),
    " build(s) older than ",
    format(keep_since, "%Y-%m-%d"),
    ". Continue?"
  )
  if (ask && !isTRUE(utils::askYesNo(prompt, default = FALSE))) {
    message("Operation cancelled.")
    return(invisible(NULL))
  }

  # Protect recent artifacts (date-based mode only)
  temp_gcroots_dir <- NULL
  if (length(keep_paths_all)) {
    temp_gcroots_dir <- tempfile("rixpress-gc-")
    dir.create(temp_gcroots_dir)
    on.exit(
      if (!is.null(temp_gcroots_dir)) {
        unlink(temp_gcroots_dir, recursive = TRUE, force = TRUE)
      },
      add = TRUE
    )

    message(
      "Protecting ",
      length(keep_paths_all),
      " recent artifacts via GC roots..."
    )
    protected <- 0L
    for (i in seq_along(keep_paths_all)) {
      link_path <- file.path(temp_gcroots_dir, paste0("root-", i))
      # Create/register an indirect root so Nix tracks it under gcroots/auto
      tryCatch(
        {
          safe_system2(
            nix_bin,
            c("--add-root", link_path, "--indirect", keep_paths_all[i])
          )
          protected <- protected + 1L
        },
        error = function(e) {
          warning(
            "Failed to add GC root for ",
            keep_paths_all[i],
            ": ",
            conditionMessage(e)
          )
        }
      )
    }
    if (protected == 0L) {
      stop("Failed to protect any store paths. Aborting.")
    }
    summary_info$protected <- protected
  }

  # Delete specific store paths
  message("Deleting ", length(delete_paths_all), " targeted store paths...")

  # Filter to only existing paths
  existing_paths <- delete_paths_all[
    file.exists(delete_paths_all) | dir.exists(delete_paths_all)
  ]
  missing_paths <- setdiff(delete_paths_all, existing_paths)

  if (length(missing_paths) > 0) {
    message("Skipping ", length(missing_paths), " paths that no longer exist.")
    if (verbose) {
      message("Missing paths:")
      cat("  ", missing_paths, sep = "\n  ", "\n")
    }
  }

  if (length(existing_paths) == 0) {
    message("No existing paths to delete. All targeted paths are already gone.")
    return(invisible(summary_info))
  }

  message("Proceeding to delete ", length(existing_paths), " existing paths...")

  # Try to delete paths individually to handle references gracefully
  total_deleted <- 0L
  failed_paths <- character(0)
  referenced_paths <- character(0)

  for (i in seq_along(existing_paths)) {
    path <- existing_paths[i]

    # Double-check path still exists
    if (!file.exists(path) && !dir.exists(path)) {
      message(
        "  [",
        i,
        "/",
        length(existing_paths),
        "] Skipping ",
        basename(path),
        " (already gone)"
      )
      next
    }

    message(
      "  [",
      i,
      "/",
      length(existing_paths),
      "] Attempting to delete ",
      basename(path),
      "..."
    )

    tryCatch(
      {
        delete_out <- system2(
          nix_bin,
          c("--delete", path),
          stdout = TRUE,
          stderr = TRUE
        )

        status <- attr(delete_out, "status")
        if (is.null(status) || status == 0) {
          total_deleted <- total_deleted + 1L
          message("    [OK] Successfully deleted")
          if (verbose && length(delete_out)) {
            cat("    ", delete_out, sep = "\n    ", "\n")
          }
        } else {
          # Check if it's a "still alive" error (has references)
          if (
            any(grepl(
              "still alive|Cannot delete",
              delete_out,
              ignore.case = TRUE
            ))
          ) {
            referenced_paths <- c(referenced_paths, path)
            message("    [!] Skipped (still referenced)")
            if (verbose) {
              message("    Details: ", paste(delete_out, collapse = " "))
            }
          } else {
            failed_paths <- c(failed_paths, path)
            message("    [X] Failed to delete")
            if (verbose) {
              cat("    ", delete_out, sep = "\n    ", "\n")
            }
          }
        }
      },
      error = function(e) {
        failed_paths <<- c(failed_paths, path)
        message("    [X] Error: ", conditionMessage(e))
      }
    )
  }

  # Summary reporting
  message("\nDeletion summary:")
  message("  Successfully deleted: ", total_deleted, " paths")
  message("  Skipped (still referenced): ", length(referenced_paths), " paths")
  message("  Failed (other errors): ", length(failed_paths), " paths")

  if (length(referenced_paths) > 0 && verbose) {
    message("\nReferenced paths (cannot delete):")
    for (path in referenced_paths) {
      message("  ", basename(path))

      # Show GC roots
      tryCatch(
        {
          roots <- system2(
            nix_bin,
            c("--query", "--roots", path),
            stdout = TRUE,
            stderr = FALSE
          )
          if (length(roots) > 0) {
            message("    GC roots: ", paste(roots, collapse = ", "))
          } else {
            message("    GC roots: (none found)")
          }
        },
        error = function(e) {
          message("    GC roots: (query failed)")
        }
      )

      # Show referrers (what depends on this path)
      tryCatch(
        {
          refs <- system2(
            nix_bin,
            c("--query", "--referrers", path),
            stdout = TRUE,
            stderr = FALSE
          )
          if (length(refs) > 0) {
            message(
              "    Referenced by: ",
              paste(basename(refs), collapse = ", ")
            )
          } else {
            message("    Referenced by: (none)")
          }
        },
        error = function(e) {
          message("    Referenced by: (query failed)")
        }
      )
    }

    message("\nTip: To investigate further, you can run:")
    message(
      "  nix-store --query --roots <path>  # Show what GC roots reference this path"
    )
    message(
      "  nix-store --query --referrers <path>  # Show what store paths depend on this"
    )
    message("  ls -la /nix/var/nix/gcroots/  # Browse all GC roots")
  }

  if (length(failed_paths) > 0 && verbose) {
    message("\nFailed paths:")
    cat("  ", basename(failed_paths), sep = "\n  ", "\n")
  }

  message(
    "\nStore path deletion complete. Successfully deleted ",
    total_deleted,
    " of ",
    length(delete_paths_all),
    " targeted paths."
  )
  summary_info$deleted_count <- total_deleted
  summary_info$failed_count <- length(failed_paths)
  summary_info$referenced_count <- length(referenced_paths)

  # Delete old build log files
  if (nrow(logs_to_delete) > 0) {
    message("\nDeleting old build log files...")
    log_files_deleted <- 0L
    log_files_failed <- character(0)

    for (i in seq_len(nrow(logs_to_delete))) {
      log_file <- logs_to_delete$filename[i]
      log_path <- file.path(project_path, "_rixpress", log_file)

      message(
        "  [",
        i,
        "/",
        nrow(logs_to_delete),
        "] Deleting ",
        log_file,
        "..."
      )

      if (!file.exists(log_path)) {
        message("    [!] File not found (already deleted?)")
        next
      }

      tryCatch(
        {
          unlink(log_path)
          if (!file.exists(log_path)) {
            log_files_deleted <- log_files_deleted + 1L
            message("    [OK] Successfully deleted")
          } else {
            log_files_failed <- c(log_files_failed, log_file)
            message("    [X] Failed to delete (file still exists)")
          }
        },
        error = function(e) {
          log_files_failed <<- c(log_files_failed, log_file)
          message("    [X] Error: ", conditionMessage(e))
        }
      )
    }

    message("\nBuild log deletion summary:")
    message("  Successfully deleted: ", log_files_deleted, " files")
    message("  Failed: ", length(log_files_failed), " files")

    if (length(log_files_failed) > 0 && verbose) {
      message("\nFailed to delete log files:")
      cat("  ", log_files_failed, sep = "\n  ", "\n")
    }

    summary_info$log_files_deleted <- log_files_deleted
    summary_info$log_files_failed <- length(log_files_failed)
  }

  message("\nCleanup complete!")
  invisible(summary_info)
}

#' Extract Packages from a Specified Block in a default.nix File
#'
#' @param nix_file Path to the default.nix file
#' @param project_path Path to root of project, typically "."
#' @param block_name Name of the block to parse (e.g., "rpkgs" or "pyconf")
#' @param transform Function to transform package names (default: identity)
#' @return List of package names, or NULL if the block is not found
#' @noRd
parse_packages <- function(
  nix_file,
  project_path,
  block_name,
  transform = identity
) {
  file_path <- file.path(project_path, nix_file)
  if (!file.exists(file_path)) {
    warning(paste("File not found:", file_path))
    return(NULL)
  }

  lines <- readLines(file_path)

  # Find the start of the block
  # (e.g., "rpkgs = builtins.attrValues {" or "pyconf = ...")
  start_pattern <- paste0(
    "^\\s*",
    block_name,
    "\\s*=\\s*(?:",
    "builtins\\.attrValues\\s*\\{", # matches rpkgs (or pyconf) = builtins.attrValues {
    "|", # or, matches Julia blocks
    "pkgs\\.julia(?:[-_\\.][A-Za-z0-9]+)*\\.withPackages\\s*\\[",
    ")"
  )

  start_idx <- grep(start_pattern, lines)
  if (length(start_idx) == 0) {
    return(NULL)
  }
  start_idx <- start_idx[1]

  # Find the end of the block ("};" or "];" or "} ++ [ ... ];")
  # We look for a closing brace/bracket followed by optional ++ list, ending with ;
  end_idxs <- grep("^\\s*(\\}|\\])(\\s*\\+\\+.*\\])?;\\s*$", lines, perl = TRUE)
  end_idx <- end_idxs[end_idxs > start_idx][1]

  if (is.na(end_idx)) {
    stop(paste("Could not find the end of the", block_name, "block"))
  }

  # Extract lines. If the end line contains packages (e.g. in ++ list), include it.
  end_line <- lines[end_idx]
  if (grepl("\\+\\+", end_line)) {
    block_lines <- lines[(start_idx + 1):end_idx]
  } else {
    block_lines <- lines[(start_idx + 1):(end_idx - 1)]
  }

  block_lines <- gsub("#.*", "", block_lines) # Remove comments
  block_lines <- trimws(block_lines) # Remove whitespace
  block_lines <- block_lines[block_lines != ""] # Remove empty lines

  # Remove inherit statements
  # (e.g., "inherit (pkgs.rPackages)" or "inherit (pkgs.python312Packages)")
  inherit_pattern <- "inherit \\(pkgs\\.[a-zA-Z0-9]+\\)"
  block_lines <- gsub(inherit_pattern, "", block_lines)

  # Clean up structural tokens
  # Remove semicolons, braces, brackets, equals sign, builtins, ++, etc.
  block_lines <- gsub(";", " ", block_lines, fixed = TRUE)
  block_lines <- gsub("=", " ", block_lines, fixed = TRUE)
  block_lines <- gsub("{", " ", block_lines, fixed = TRUE)
  block_lines <- gsub("}", " ", block_lines, fixed = TRUE)
  block_lines <- gsub("[", " ", block_lines, fixed = TRUE)
  block_lines <- gsub("]", " ", block_lines, fixed = TRUE)
  block_lines <- gsub("+", " ", block_lines, fixed = TRUE)
  block_lines <- gsub("builtins\\.attrValues", " ", block_lines)

  # Split into package names and apply transformation
  packages <- unlist(strsplit(paste(block_lines, collapse = " "), "\\s+"))
  packages <- packages[packages != ""]

  transform(packages)
}

#' Extract R Packages Defined via pkgs.rPackages.buildRPackage from Git Sources
#'
#' @param nix_file Path to the default.nix file
#' @param project_path Path to root of project, typically "."
#' @param transform Function to transform package names (default: identity)
#' @return List of R package names from git, or NULL if none are found
#' @noRd
parse_rpkgs_git <- function(
  nix_file,
  project_path,
  transform = identity
) {
  file_path <- file.path(project_path, nix_file)
  if (!file.exists(file_path)) {
    warning(paste("File not found:", file_path))
    return(NULL)
  }

  lines <- readLines(file_path)
  git_packages <- c()

  # Pattern to find the start of a buildRPackage definition
  # e.g., my_pkg = (pkgs.rPackages.buildRPackage {
  build_r_pkg_start_pattern <- "^\\s*([a-zA-Z0-9_.-]+)\\s*=\\s*\\(pkgs\\.rPackages\\.buildRPackage\\s*\\{"
  start_indices <- grep(build_r_pkg_start_pattern, lines)

  if (length(start_indices) == 0) {
    return(NULL)
  }

  # Pattern for the end of these blocks is typically "});"
  end_block_pattern <- "^\\s*\\}\\);"

  for (start_idx in start_indices) {
    # Find the corresponding end of this specific block
    potential_end_indices <- grep(
      end_block_pattern,
      lines[(start_idx + 1):length(lines)],
      fixed = FALSE
    )
    if (length(potential_end_indices) == 0) {
      next # No proper end found for this block, skip
    }
    # The index is relative to lines[(start_idx + 1):length(lines)], so adjust
    end_idx <- start_idx + potential_end_indices[1]

    block_lines <- lines[(start_idx + 1):(end_idx - 1)]

    # Check if this block contains a src = pkgs.fetchgit (or similar)
    # Covers fetchgit, fetchFromGitHub, fetchFromGitLab
    fetch_pattern <- "^\\s*src\\s*=\\s*pkgs\\.(fetchgit|fetchFromGitHub|fetchFromGitLab)\\s*\\{"
    if (any(grepl(fetch_pattern, block_lines))) {
      # If it's a git source, look for the 'name' attribute
      name_pattern <- "^\\s*name\\s*=\\s*\"([a-zA-Z0-9_.-]+)\"\\s*;"
      name_lines_match <- grep(name_pattern, block_lines, value = TRUE)

      if (length(name_lines_match) > 0) {
        # Extract the package name (the part in quotes)
        pkg_name <- sub(name_pattern, "\\1", name_lines_match[1])
        git_packages <- c(git_packages, pkg_name)
      }
    }
  }

  if (length(git_packages) == 0) {
    return(NULL)
  }

  transform(unique(git_packages))
}


#' Generate a Script with Import Statements and Optional Additional Content
#'
#' @param packages List of package names
#' @param additional_files Character vector of additional files to include
#' @param outfile Path to the output file
#' @param import_formatter Function to format import statements
#' @param additional_file_pattern Regex pattern to identify additional files
#' @return Writes a script to the specified outfile and returns the path
#' @noRd
generate_libraries_script <- function(
  packages,
  additional_files,
  outfile,
  import_formatter,
  additional_file_pattern
) {
  # Generate import statements for each package
  import_lines <- sapply(packages, import_formatter)

  # Filter additional files based on pattern (only if pattern is not NULL)
  if (!is.null(additional_file_pattern)) {
    additional_scripts <- Filter(
      function(x) grepl(additional_file_pattern, x),
      additional_files
    )

    # Combine import statements with additional content
    if (length(additional_scripts) > 0) {
      list_additional_content <- lapply(additional_scripts, readLines)
      additional_content <- Reduce(append, list_additional_content)
      output <- append(import_lines, additional_content)
    } else {
      output <- import_lines
    }
  } else {
    output <- import_lines
  }

  # Create output directory if it doesn't exist
  output_dir <- dirname(outfile)
  if (!dir.exists(output_dir)) {
    dir.create(output_dir, recursive = TRUE)
  }

  # Write the output to file
  writeLines(output, outfile)

  return(outfile)
}

#' @noRd
transform_r <- function(packages) {
  gsub("_", ".", packages) # in nixpkgs: data_table, but in CRAN data.table
}

#' @noRd
adjust_r_packages <- identity # No additional adjustments for R

#' @noRd
import_formatter_r <- function(package) {
  paste0("library(", package, ")") # R import format
}

#' @noRd
adjust_py_packages <- function(packages) {
  # Only needed for interactive work
  packages <- packages[!(packages %in% c("pip", "ipykernel"))]
  # Pickle is needed to import pickle’d objects
  packages <- sort(c("pickle", packages))
  # Scikit is a common package where its name is different
  # when importing
  packages <- gsub("scikit-learn", "sklearn", packages)
  packages
}

#' @noRd
transform_py <- function(packages) {
  packages # No transformation by default
}

#' @noRd
import_formatter_py <- function(package) {
  paste0("import ", package)
}

#' @noRd
transform_jl <- function(packages) {
  packages # No transformation by default
}

#' @noRd
adjust_jl_packages <- function(packages) {
  # Serialization is needed to import pickle’d objects
  packages <- sort(c("Serialization", packages))
  # Julia Packages are passed as strings in the nix expression
  packages <- gsub("\"", "", packages)
  packages
}

#' @noRd
import_formatter_jl <- function(package) {
  paste0("using ", package)
}

#' Configuration for Supported Languages
#'
#' This list defines the configuration for each supported language,
#' including block names, file patterns, and function references.
#' @noRd
language_configs <- list(
  "R" = list(
    block_name = "rpkgs",
    transform_func = transform_r,
    adjust_func = adjust_r_packages,
    import_formatter = import_formatter_r,
    additional_file_pattern = NULL, # Changed from "functions\\.[Rr]"
    extension = "R",
    parse_git = TRUE
  ),
  "Python" = list(
    block_name = "pyconf",
    transform_func = transform_py,
    adjust_func = adjust_py_packages,
    import_formatter = import_formatter_py,
    additional_file_pattern = NULL, # Changed from "functions\\.[py]"
    extension = "py",
    parse_git = FALSE
  ),
  "Julia" = list(
    block_name = "jlconf",
    transform_func = transform_jl,
    adjust_func = adjust_jl_packages,
    import_formatter = import_formatter_jl,
    additional_file_pattern = NULL, # Changed from "functions\\.[jl]"
    extension = "jl",
    parse_git = FALSE
  )
)

#' Generate a Script with Import Statements from a default.nix File
#'
#' @param nix_file Path to the default.nix file
#' @param additional_files Character vector of additional files to include
#' @param project_path Path to root of project, typically "."
#' @param language Language to generate the script for ("R", "Python", or "Julia")
#' @return A script file path for the specified language,
#'   or NULL if no packages are found
#' @noRd
generate_r_or_py_libraries_from_nix <- function(
  nix_file,
  additional_files = "",
  project_path,
  language
) {
  # Validate language parameter
  if (!language %in% names(language_configs)) {
    stop(paste(
      "Unsupported language:",
      language,
      "- Must be one of:",
      paste(names(language_configs), collapse = ", ")
    ))
  }

  # Get language configuration
  config <- language_configs[[language]]
  all_parsed_packages <- c()

  # Parse packages from the main block
  packages_from_block <- parse_packages(
    nix_file = nix_file,
    project_path = project_path,
    block_name = config$block_name,
    transform = config$transform_func
  )
  if (!is.null(packages_from_block)) {
    all_parsed_packages <- c(all_parsed_packages, packages_from_block)
  }

  # Parse packages from git sources (R only)
  if (config$parse_git && language == "R") {
    packages_from_git <- parse_rpkgs_git(
      nix_file = nix_file,
      project_path = project_path,
      transform = config$transform_func
    )
    if (!is.null(packages_from_git)) {
      all_parsed_packages <- c(all_parsed_packages, packages_from_git)
    }
  }

  # If no packages found, return NULL
  if (length(all_parsed_packages) == 0) {
    return(NULL)
  }

  # Process packages
  packages <- unique(all_parsed_packages)
  packages <- config$adjust_func(packages)
  packages <- sort(packages)

  # Create output filename (use basename to handle relative paths)
  nix_file_name <- gsub("[^a-zA-Z0-9]", "_", basename(nix_file))
  nix_file_name <- sub("_nix$", "", nix_file_name)
  outfile_dir <- file.path(project_path, "_rixpress")
  outfile <- file.path(
    outfile_dir,
    paste0(nix_file_name, "_libraries.", config$extension)
  )

  # Generate the script
  generate_libraries_script(
    packages,
    additional_files,
    outfile,
    config$import_formatter,
    config$additional_file_pattern
  )

  return(outfile)
}

#' Generate Language-Specific Library Scripts from a Nix File
#'
#' These are convenience wrappers around generate_r_or_py_libraries_from_nix
#' for specific languages.
#' Generate R library script from a Nix file
#' @param nix_file Path to the default.nix file
#' @param additional_files Character vector of additional files to include
#' @param project_path Path to root of project, typically "."
#' @return Path to the generated script, or NULL if no packages found
#' @noRd
generate_r_libraries_from_nix <- function(
  nix_file,
  additional_files = "",
  project_path = "."
) {
  generate_r_or_py_libraries_from_nix(
    nix_file,
    additional_files,
    project_path,
    language = "R"
  )
}

#' Generate Python library script from a Nix file
#'
#' @param nix_file Path to the default.nix file
#' @param additional_files Character vector of additional files to include
#' @param project_path Path to root of project, typically "."
#' @return Path to the generated script, or NULL if no packages found
#' @noRd
generate_py_libraries_from_nix <- function(
  nix_file,
  additional_files = "",
  project_path = "."
) {
  generate_r_or_py_libraries_from_nix(
    nix_file,
    additional_files,
    project_path,
    language = "Python"
  )
}

#' Generate Julia library script from a Nix file
#' @param nix_file Path to the default.nix file
#' @param additional_files Character vector of additional files to include
#' @param project_path Path to root of project, typically "."
#' @return Path to the generated script, or NULL if no packages found
#' @noRd
generate_jl_libraries_from_nix <- function(
  nix_file,
  additional_files = "",
  project_path = "."
) {
  generate_r_or_py_libraries_from_nix(
    nix_file,
    additional_files,
    project_path,
    language = "Julia"
  )
}

#' Adjust Python Import Statements
#'
#' When calling `rxp_populate()`, a file containing Python import statements is
#' automatically generated inside the `_rixpress` folder. For example, if the
#' `numpy` package is needed, the file will include a line like
#' `"import numpy"`. However, Python programmers often write
#' `"import numpy as np"` instead.
#'
#' In some cases, the correct import statement is entirely different. For
#' example, for the `pillow` package, the generated file will contain
#' `"import pillow"`, which is incorrect—Python code should import from the
#' `PIL` namespace instead, e.g., `"from PIL import Image"`.
#'
#' Because these adjustments cannot be automated reliably, the `adjust_import()`
#' function allows you to search and replace import statements programmatically.
#' It reads each file in the `_rixpress` folder, performs the replacement, and
#' writes the modified content back to the file.
#' @family python import
#' @param old_import A character string representing the import statement to be
#'   replaced. For example, `"import pillow"`.
#' @param new_import A character string representing the new import statement to
#'   replace with. For example, `"from PIL import Image"`.
#' @param project_path Path to root of project, typically ".".
#'
#' @return No return value; the function performs in-place
#'   modifications of the files.
#'
#' @examples
#' \dontrun{
#' adjust_import("import pillow", "from PIL import Image")
#' adjust_import("import pillow", "from PIL import Image", project_path = "path/to/project")
#' }
#' @export
adjust_import <- function(old_import, new_import, project_path = ".") {
  # Validate inputs
  if (!is.character(old_import) || length(old_import) != 1) {
    stop("old_import must be a single character string")
  }
  if (!is.character(new_import) || length(new_import) != 1) {
    stop("new_import must be a single character string")
  }

  # Check if _rixpress folder exists
  rixpress_folder <- file.path(project_path, "_rixpress")
  if (!dir.exists(rixpress_folder)) {
    warning(paste("_rixpress folder not found in", project_path))
    return(invisible(NULL))
  }

  # Get all files in the _rixpress folder
  files <- list.files(
    path = rixpress_folder,
    full.names = TRUE,
    recursive = TRUE
  )

  # Process each file
  for (file in files) {
    content <- readLines(file, warn = FALSE)
    new_content <- gsub(old_import, new_import, content, fixed = TRUE)

    # Only write if content changed
    if (!identical(content, new_content)) {
      writeLines(new_content, con = file)
    }
  }

  invisible(NULL)
}

#' Add an Import Statement to Python Files in the _rixpress Folder Matching a
#' Nix Environment Name
#'
#' This function appends a specified import statement to the end of each Python
#' file within the `_rixpress` folder and its subdirectories, but only for files
#' whose base name matches the provided Nix environment.
#'
#' @family python import
#' @param import_statement A character string representing the import statement
#'   to be added. For example, `"import numpy as np"`.
#' @param nix_env A character string naming the Nix environment file (e.g.
#'   `"default.nix"` or `"py-env.nix"` or similar).
#' @param project_path Path to root of project, typically ".".
#'
#' @return No return value; the function performs in-place modifications of the
#'   files.
#'
#' @examples
#' \dontrun{
#' add_import("import numpy as np", "default.nix")
#' add_import("import numpy as np", "default.nix", project_path = "path/to/project")
#' }
#' @export
add_import <- function(import_statement, nix_env, project_path = ".") {
  # Validate inputs
  if (!is.character(import_statement) || length(import_statement) != 1) {
    stop("import_statement must be a single character string")
  }
  if (!is.character(nix_env) || length(nix_env) != 1) {
    stop("nix_env must be a single character string, e.g. 'default.nix'.")
  }

  # Check if _rixpress folder exists
  rixpress_folder <- file.path(project_path, "_rixpress")
  if (!dir.exists(rixpress_folder)) {
    warning(paste("_rixpress folder not found in", project_path))
    return(invisible(NULL))
  }

  # Extract base name from nix_env
  base_name <- sub("\\.nix$", "", nix_env)
  if (identical(base_name, nix_env)) {
    warning(
      "Provided nix_env did not end with '.nix'; using entire string as base name."
    )
  }

  # Construct regex to match Python files starting with the base name
  # e.g. ^default.*\.[pP]y$
  file_pattern <- paste0(
    "^",
    base_name,
    ".*\\.[pP]y$"
  )

  # List only .py/.Py files whose names match the base_name prefix
  files <- list.files(
    path = rixpress_folder,
    pattern = file_pattern,
    full.names = TRUE,
    recursive = TRUE
  )

  if (length(files) == 0) {
    return(invisible(NULL))
  }

  # Loop through each matching Python file
  for (file in files) {
    content <- readLines(file, warn = FALSE)

    # Escape special regex characters in the import statement
    escaped_import <- gsub(
      "([.*+?^${}()|\\[\\]\\\\])",
      "\\\\\\1",
      import_statement,
      perl = TRUE
    )

    # Check if import statement already exists
    if (!any(grepl(paste0("^", escaped_import, "$"), content))) {
      new_content <- c(content, import_statement)
      writeLines(new_content, con = file)
    }
  }

  invisible(NULL)
}

test_that("generate_r_libraries_from_nix: generate R script by parsing default.nix", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("rix")
  path_tmpdir <- tempdir()

  dir.create(paste0(path_tmpdir, "/", "_rixpress"))

  on.exit(unlink(path_tmpdir, recursive = TRUE), add = TRUE, after = TRUE)

  rix::rix(
    date = "2025-03-10",
    r_pkgs = c("dplyr", "tidyr", "stringi"),
    project_path = path_tmpdir,
    overwrite = TRUE
  )

  generate_r_libraries_from_nix(
    nix_file = "default.nix",
    project_path = file.path(path_tmpdir, "/")
  )

  library_lines <- readLines(file.path(
    path_tmpdir,
    "_rixpress",
    "default_libraries.R"
  ))

  testthat::expect_equal(
    library_lines,
    c("library(dplyr)", "library(stringi)", "library(tidyr)")
  )
})


test_that("generate_py_libraries_from_nix: generate Py script by parsing default.nix", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("rix")

  path_tmpdir <- tempdir()

  dir.create(paste0(path_tmpdir, "/", "_rixpress"))

  on.exit(unlink(path_tmpdir, recursive = TRUE), add = TRUE, after = TRUE)

  rix::rix(
    date = "2025-03-10",
    r_pkgs = "reticulate",
    py_conf = list(
      py_version = "3.12",
      py_pkgs = c("polars", "great-tables")
    ),
    project_path = path_tmpdir,
    overwrite = TRUE
  )

  generate_py_libraries_from_nix(
    nix_file = "default.nix",
    project_path = file.path(path_tmpdir, "/")
  )

  library_lines <- readLines(file.path(
    path_tmpdir,
    "_rixpress",
    "default_libraries.py"
  ))

  testthat::expect_equal(
    library_lines,
    c("import great-tables", "import pickle", "import polars")
  )
})

# Helper function to create a temporary Nix file
create_temp_nix <- function(content, project_path) {
  nix_file_path <- file.path(project_path, "temp_default.nix")
  writeLines(content, nix_file_path)
  return("temp_default.nix")
}

test_that("parse_packages handles various formats and edge cases", {
  path_tmpdir <- tempdir()
  dir.create(path_tmpdir, showWarnings = FALSE)
  on.exit(unlink(path_tmpdir, recursive = TRUE), add = TRUE, after = TRUE)

  nix_content1 <- c(
    "let",
    "  rpkgs = builtins.attrValues {",
    "    inherit (pkgs.rPackages);",
    "    tidyr",
    "    dplyr;",
    "  };",
    "in {}"
  )
  nix_file1 <- create_temp_nix(nix_content1, path_tmpdir)
  on.exit(unlink(nix_file1))

  pkgs1 <- parse_packages(nix_file1, path_tmpdir, "rpkgs", transform_r)
  expect_equal(sort(pkgs1), c("dplyr", "tidyr"))

  nix_content2 <- c(
    "let",
    "  pyconf = builtins.attrValues {",
    "    inherit (pkgs.python312Packages);",
    "    pandas",
    "    numpy",
    "    scikit-learn;",
    "  };",
    "in {}"
  )
  nix_file2 <- create_temp_nix(nix_content2, path_tmpdir)
  on.exit(unlink(nix_file2))

  pkgs2 <- parse_packages(nix_file2, path_tmpdir, "pyconf")
  expect_equal(sort(pkgs2), c("numpy", "pandas", "scikit-learn"))

  pkgs3 <- parse_packages(nix_file1, path_tmpdir, "nonexistent_block")
  expect_null(pkgs3)

  nix_content4 <- c(
    "let",
    "  rpkgs = builtins.attrValues {",
    "    dplyr"
    # Missing "};"
  )
  nix_file4 <- create_temp_nix(nix_content4, path_tmpdir)
  on.exit(unlink(nix_file4))

  expect_error(
    parse_packages(nix_file4, path_tmpdir, "rpkgs"),
    "Could not find the end of the rpkgs block"
  )

  nix_content5 <- c(
    "let",
    "  rpkgs = builtins.attrValues {",
    "    inherit (pkgs.rPackages);",
    "  };",
    "in {}"
  )
  nix_file5 <- create_temp_nix(nix_content5, path_tmpdir)
  on.exit(unlink(nix_file5))

  pkgs5 <- parse_packages(nix_file5, path_tmpdir, "rpkgs")
  expect_equal(pkgs5, character(0))
})

test_that("parse_packages handles pyconf with git packages (} ++ [ ... ];)", {
  path_tmpdir <- tempdir()
  dir.create(path_tmpdir, showWarnings = FALSE)
  on.exit(unlink(path_tmpdir, recursive = TRUE), add = TRUE, after = TRUE)

  nix_content <- c(
    'let',
    ' pkgs = import (fetchTarball "https://github.com/rstats-on-nix/nixpkgs/archive/2025-04-29.tar.gz") {};',
    ' ',
    '  rpkgs = builtins.attrValues {',
    '    inherit (pkgs.rPackages) ',
    '      dplyr',
    '      ggplot2',
    '      reticulate',
    '      yardstick;',
    '  };',
    ' ',
    '    rix = (pkgs.rPackages.buildRPackage {',
    '      name = "rix";',
    '      src = pkgs.fetchgit {',
    '        url = "https://github.com/ropensci/rix/";',
    '        rev = "HEAD";',
    '        sha256 = "sha256-4cn9/BxpGljoS65FcsYrXAI2wMKzOLYvjevsct1/+Ik=";',
    '      };',
    '      propagatedBuildInputs = builtins.attrValues {',
    '        inherit (pkgs.rPackages) ',
    '          codetools',
    '          curl',
    '          jsonlite',
    '          sys;',
    '      };',
    '    });',
    '',
    '    rixpress = (pkgs.rPackages.buildRPackage {',
    '      name = "rixpress";',
    '      src = pkgs.fetchgit {',
    '        url = "https://github.com/ropensci/rixpress";',
    '        rev = "HEAD";',
    '        sha256 = "sha256-JmArZXrLsjj68e8Xo75lQ3j3i9oP5vsNnc+x/IgvspI=";',
    '      };',
    '      propagatedBuildInputs = builtins.attrValues {',
    '        inherit (pkgs.rPackages) ',
    '          cli',
    '          igraph',
    '          jsonlite',
    '          processx;',
    '      };',
    '    });',
    '   ',
    '',
    '    ryxpress = (pkgs.python313Packages.buildPythonPackage {',
    '      pname = "ryxpress";',
    '      version = "HEAD-git";',
    '      src = pkgs.fetchgit {',
    '        url = "https://github.com/b-rodrigues/ryxpress";',
    '        rev = "HEAD";',
    '        sha256 = "sha256-agDRYWC4wV3Oum8KhS78ZXCXPo/4X1OmNER+WWnbxBw=";',
    '      };',
    '      pyproject = true;',
    '      build-system = [ pkgs.python313Packages.setuptools ];',
    '      doCheck = false;',
    '      propagatedBuildInputs = [ ];',
    '    });',
    ' ',
    '  pyconf = builtins.attrValues {',
    '    inherit (pkgs.python313Packages) ',
    '      pip',
    '      ipykernel',
    '      numpy',
    '      pandas',
    '      scikit-learn',
    '      xgboost;',
    '  } ++ [ ryxpress ];',
    '   ',
    '  system_packages = builtins.attrValues {',
    '    inherit (pkgs) ',
    '      glibcLocales',
    '      nix',
    '      python313',
    '      R;',
    '  };',
    '  ',
    '  shell = pkgs.mkShell {',
    '    LOCALE_ARCHIVE = if pkgs.stdenv.hostPlatform.system == "x86_64-linux" then "${pkgs.glibcLocales}/lib/locale/locale-archive" else "";',
    '    LANG = "en_US.UTF-8";',
    '    LC_ALL = "en_US.UTF-8";',
    '    LC_TIME = "en_US.UTF-8";',
    '    LC_MONETARY = "en_US.UTF-8";',
    '    LC_PAPER = "en_US.UTF-8";',
    '    LC_MEASUREMENT = "en_US.UTF-8";',
    '    RETICULATE_PYTHON = "${pkgs.python313}/bin/python";',
    '',
    '    buildInputs = [ rix rixpress rpkgs pyconf system_packages ];',
    '    ',
    '  }; ',
    'in',
    '  {',
    '    inherit pkgs shell;',
    '  }'
  )

  nix_file <- create_temp_nix(nix_content, path_tmpdir)
  pkgs <- parse_packages(nix_file, path_tmpdir, "pyconf")

  expected_pkgs <- c(
    "pip",
    "ipykernel",
    "numpy",
    "pandas",
    "scikit-learn",
    "xgboost",
    "ryxpress"
  )
  expect_equal(sort(pkgs), sort(expected_pkgs))
})

test_that("parse_packages handles mixed python (git) and julia blocks", {
  path_tmpdir <- tempdir()
  dir.create(path_tmpdir, showWarnings = FALSE)
  on.exit(unlink(path_tmpdir, recursive = TRUE), add = TRUE, after = TRUE)

  nix_content <- c(
    'let',
    ' pkgs = import (fetchTarball "https://github.com/rstats-on-nix/nixpkgs/archive/2026-01-19.tar.gz") {};',
    ' ',
    '     ryxpress = (pkgs.python313Packages.buildPythonPackage {',
    '       pname = "ryxpress";',
    '       version = "HEAD-git";',
    '       src = pkgs.fetchgit {',
    '         url = "https://github.com/b-rodrigues/ryxpress";',
    '         rev = "HEAD";',
    '         sha256 = "sha256-agDRYWC4wV3Oum8KhS78ZXCXPo/4X1OmNER+WWnbxBw=";',
    '       };',
    '       pyproject = true;',
    '       build-system = [ pkgs.python313Packages.setuptools ];',
    '       doCheck = false;',
    '       propagatedBuildInputs = [ ];',
    '     });',
    ' ',
    '   pyconf = builtins.attrValues {',
    '     inherit (pkgs.python313Packages) ',
    '       pip',
    '       ipykernel',
    '       pandas',
    '       pyarrow',
    '       scikit-learn',
    '       xgboost;',
    '   } ++ [ ryxpress ];',
    '  ',
    '   jlconf = pkgs.julia-lts.withPackages [ ',
    '       "Arrow"',
    '       "DataFrames"',
    '       "Distributions"',
    '       "Random"',
    '   ];',
    '   ',
    'in {}'
  )

  nix_file <- create_temp_nix(nix_content, path_tmpdir)
  pkgs_py <- parse_packages(nix_file, path_tmpdir, "pyconf")

  # Should ONLY contain python packages + ryxpress, NOT julia packages
  expected_py <- c(
    "pip",
    "ipykernel",
    "pandas",
    "pyarrow",
    "scikit-learn",
    "xgboost",
    "ryxpress"
  )
  expect_equal(sort(pkgs_py), sort(expected_py))
})

test_that("generate_libraries_script handles additional files", {
  path_tmpdir <- tempdir()
  dir.create(path_tmpdir, showWarnings = FALSE)
  on.exit(unlink(path_tmpdir, recursive = TRUE), add = TRUE, after = TRUE)

  # Setup additional files
  func_r_path <- file.path(path_tmpdir, "functions.R")
  func_py_path <- file.path(path_tmpdir, "functions.py")
  other_file_path <- file.path(path_tmpdir, "data.csv")
  writeLines("my_r_func <- function() {}", func_r_path)
  writeLines("def my_py_func(): pass", func_py_path)
  writeLines("a,b\n1,2", other_file_path)

  # Case 1: R script with functions.R
  r_outfile1 <- file.path(path_tmpdir, "lib1.R")
  generate_libraries_script(
    packages = c("dplyr"),
    additional_files = c(func_r_path, other_file_path), # Use full paths
    outfile = r_outfile1,
    import_formatter = import_formatter_r,
    additional_file_pattern = "functions\\.[Rr]"
  )
  lines1 <- readLines(r_outfile1)
  expect_equal(lines1, c("library(dplyr)", "my_r_func <- function() {}"))

  # Case 2: Python script with functions.py
  py_outfile1 <- file.path(path_tmpdir, "lib1.py")
  generate_libraries_script(
    packages = c("pandas"),
    additional_files = c(func_py_path, other_file_path), # Use full paths
    outfile = py_outfile1,
    import_formatter = import_formatter_py,
    additional_file_pattern = "functions\\.py"
  )
  lines2 <- readLines(py_outfile1)
  expect_equal(lines2, c("import pandas", "def my_py_func(): pass"))

  # Case 3: No matching additional files
  r_outfile2 <- file.path(path_tmpdir, "lib2.R")
  generate_libraries_script(
    packages = c("tidyr"),
    additional_files = c(other_file_path), # Use full path
    outfile = r_outfile2,
    import_formatter = import_formatter_r,
    additional_file_pattern = "functions\\.[Rr]"
  )
  lines3 <- readLines(r_outfile2)
  expect_equal(lines3, "library(tidyr)")

  # Case 4: Empty additional files
  r_outfile3 <- file.path(path_tmpdir, "lib3.R")
  generate_libraries_script(
    packages = c("stringr"),
    additional_files = "",
    outfile = r_outfile3,
    import_formatter = import_formatter_r,
    additional_file_pattern = "functions\\.[Rr]"
  )
  lines4 <- readLines(r_outfile3)
  expect_equal(lines4, "library(stringr)")
})

test_that("adjust_py_packages works correctly", {
  pkgs_in <- c("pandas", "pip", "numpy", "scikit-learn", "ipykernel")
  pkgs_out <- adjust_py_packages(pkgs_in)
  expect_equal(sort(pkgs_out), c("numpy", "pandas", "pickle", "sklearn"))

  pkgs_in2 <- c("tensorflow")
  pkgs_out2 <- adjust_py_packages(pkgs_in2)
  expect_equal(sort(pkgs_out2), c("pickle", "tensorflow"))

  pkgs_in3 <- character(0)
  pkgs_out3 <- adjust_py_packages(pkgs_in3)
  expect_equal(sort(pkgs_out3), c("pickle"))
})

test_that("transform_r works correctly", {
  expect_equal(transform_r("data_table"), "data.table")
  expect_equal(transform_r("ggplot2"), "ggplot2")
  expect_equal(
    transform_r(c("data_table", "stringi")),
    c("data.table", "stringi")
  )
})

test_that("adjust_import modifies files correctly", {
  path_tmpdir <- tempdir()
  path_rixpress <- file.path(path_tmpdir, "_rixpress")
  dir.create(path_rixpress, recursive = TRUE)
  on.exit(unlink(path_tmpdir, recursive = TRUE), add = TRUE, after = TRUE)

  # Create dummy files
  file1_path <- file.path(path_rixpress, "lib1.py")
  file2_path <- file.path(path_rixpress, "lib2.py")
  file3_path <- file.path(path_rixpress, "other.txt")
  writeLines(c("import pandas", "import numpy"), file1_path)
  writeLines(c("import os", "import pandas"), file2_path)
  writeLines(c("some text"), file3_path)

  # Perform adjustment (need to temporarily set working directory)
  old_wd <- getwd()
  setwd(path_tmpdir)
  adjust_import("import pandas", "import pandas as pd")
  setwd(old_wd)

  # Check file contents
  lines1 <- readLines(file1_path)
  lines2 <- readLines(file2_path)
  lines3 <- readLines(file3_path)

  expect_equal(lines1, c("import pandas as pd", "import numpy"))
  expect_equal(lines2, c("import os", "import pandas as pd"))
  expect_equal(lines3, c("some text")) # Should not change other files

  # Test non-matching import
  setwd(path_tmpdir)
  adjust_import("import non_existent", "import something_else")
  setwd(old_wd)

  lines1_after <- readLines(file1_path)
  expect_equal(lines1_after, c("import pandas as pd", "import numpy")) # No change
})

#' Helper Function to Generate a Nix Derivation Snippet
#'
#' @param out_name Character, name of the derivation
#' @param src_snippet Character, the src part of the derivation
#' @param base Character, base name for buildInputs and configurePhase
#' @param build_phase Character, the build phase commands
#' @param derivation_type Character, one of "rxp_r", "rxp_py", "rxp_jl", "rxp_qmd", "rxp_rmd"
#' @param noop_build Logical, defaults to FALSE. If TRUE, the derivation
#'   produces a no-op build (a stub output with no actual build steps). Any
#'   downstream derivations depending on a no-op build will themselves also
#'   become no-op builds.
#' @return Character string with the formatted Nix derivation
#' @noRd
make_derivation_snippet <- function(
  out_name,
  src_snippet,
  base,
  build_phase,
  derivation_type,
  noop_build = FALSE
) {
  # If noop_build is TRUE, create a dummy derivation that just creates an empty output
  if (noop_build) {
    return(sprintf(
      "  %s = defaultPkgs.runCommand \"%s\" {} \"\n    mkdir -p $out\n    echo 'Build skipped for %s' > $out/NOOPBUILD\n  \";",
      out_name,
      out_name,
      out_name
    ))
  }

  # Normal build logic for when noop_build = FALSE
  # Determine the derivation function based on type
  derivation_func <- switch(
    derivation_type,
    "rxp_r" = "makeRDerivation",
    "rxp_py" = "makePyDerivation",
    "rxp_jl" = "makeJlDerivation",
    "rxp_qmd" = "defaultPkgs.stdenv.mkDerivation",
    "rxp_rmd" = "defaultPkgs.stdenv.mkDerivation",
    stop("Unknown derivation type: ", derivation_type)
  )

  # Format the build phase with appropriate indentation
  formatted_build_phase <- if (derivation_type %in% c("rxp_qmd", "rxp_rmd")) {
    paste0("\n", build_phase, "\n    ")
  } else {
    paste0("\n      ", build_phase, "\n    ")
  }

  # Generate the snippet
  sprintf(
    "  %s = %s {\n    name = \"%s\";\n%s    buildInputs = %sBuildInputs;\n    configurePhase = %sConfigurePhase;\n    buildPhase = ''%s'';\n  };",
    out_name,
    derivation_func,
    out_name,
    src_snippet,
    base,
    base,
    formatted_build_phase
  )
}

#' Create a Nix Expression Running an R Function
#' @family derivations
#' @param name Symbol, name of the derivation.
#' @param expr R code to generate the expression. Ideally it should be a call
#'   to a pure function, or a piped expression. Multi-line expressions are not
#'   supported.
#' @param additional_files Character vector, additional files to include
#'   during the build process. For example, if a function expects a certain
#'   file to be available, this is where you should include it.
#' @param user_functions Character vector, user-defined functions to include.
#'   This should be a script (or scripts) containing user-defined functions
#'   to include during the build process for this derivation. It is recommended
#'   to use one script per function, and only include the required script(s) in
#'   the derivation.
#' @param nix_env Character, path to the Nix environment file, default is
#'   "default.nix".
#' @param encoder Function or character defaults to NULL. A function used to
#'   encode (serialize) objects for transfer between derivations. It must accept two
#'   arguments: the object to encode (first), and the target file path
#'   (second). If your function has a different signature, wrap it to match this
#'   interface. By default, `saveRDS()` is used, but this may yield unexpected
#'   results, especially for complex objects like machine learning models. For
#'   instance, for `{keras}` models, use `keras::save_model_hdf5()` to capture
#'   the full model (architecture, weights, training config, optimiser state,
#'   etc.). See `vignette("encoding-decoding")` for more details.
#' @param decoder Function, character, or named vector/list,
#'   defaults to NULL. Can be:
#'   - A single function/string to decode (unserialize) all upstream objects (e.g., `readRDS`)
#'   - A named vector/list where names are upstream dependency names and values
#'     are their specific decoding functions (e.g.,
#'     `c(mtcars_tail = "qs::qread", mtcars_head = "read.csv")`)
#'   By default, `readRDS()` is used. See `vignette("encoding-decoding")` for more details.
#' @param env_var Character vector, defaults to NULL. A named vector of
#'   environment variables to set before running the R script, e.g.,
#'   `c("CMDSTAN" = "${defaultPkgs.cmdstan}/opt/cmdstan)"`.
#'   Each entry will be added as an export statement in the build phase.
#' @param noop_build Logical, defaults to FALSE. If TRUE, the derivation
#'   produces a no-op build (a stub output with no actual build steps). Any
#'   downstream derivations depending on a no-op build will themselves also
#'   become no-op builds.
#' @details At a basic level, `rxp_r(mtcars_am, filter(mtcars, am == 1))` is
#'   equivalent to `mtcars_am <- filter(mtcars, am == 1)`. `rxp_r()` generates the
#'   required Nix boilerplate to output a so-called "derivation" in Nix jargon.
#'   A Nix derivation is a recipe that defines how to create an output (in this
#'   case `mtcars_am`) including its dependencies, build steps, and output
#'   paths.
#' @return An object of class derivation which inherits from lists.
#' @examples \dontrun{
#'   # Basic usage
#'   rxp_r(name = filtered_mtcars, expr = filter(mtcars, am == 1))
#'
#'   # Skip building this derivation
#'   rxp_r(
#'     name = turtles,
#'     expr = occurrence(species, geometry = atlantic),
#'     noop_build = TRUE
#'   )
#'
#'   # Serialize object using qs
#'   rxp_r(
#'    name = filtered_mtcars,
#'    expr = filter(mtcars, am == 1),
#'    encoder = qs::qsave
#'   )
#'   # Unerialize using qs::qread in the next derivation
#'   rxp_r(
#'    name = mtcars_mpg,
#'    expr = select(filtered_mtcars, mpg),
#'    decoder = qs::qread
#'   )
#' }
#' @export
rxp_r <- function(
  name,
  expr,
  additional_files = "",
  user_functions = "",
  nix_env = "default.nix",
  encoder = NULL,
  decoder = NULL,
  env_var = NULL,
  noop_build = FALSE
) {
  out_name <- deparse1(substitute(name))
  expr_str <- deparse1(substitute(expr))
  expr_str <- gsub("\"", "'", expr_str) # Replace " with ' for Nix
  expr_str <- gsub("$", "\\$", expr_str, fixed = TRUE) # Replace $ with \$ for Nix

  # Capture without evaluating promises; supports bare symbols (qs::qsave)
  # and character literals ("qs::qsave") without loading packages now.
  serialize_expr <- substitute(encoder)
  if (identical(serialize_expr, quote(NULL))) {
    serialize_str <- "saveRDS"
  } else if (is.character(serialize_expr)) {
    # User passed a character literal; use it as-is (no quotes in final code)
    serialize_str <- serialize_expr
  } else {
    serialize_str <- deparse1(serialize_expr)
  }

  # Handle decoder - can be single value or named vector/list
  unserialize_expr <- substitute(decoder)
  if (identical(unserialize_expr, quote(NULL))) {
    unserialize_str <- "readRDS"
  } else {
    # Check if it's a call to c() with named arguments
    if (
      is.call(unserialize_expr) && identical(unserialize_expr[[1]], quote(c))
    ) {
      # It's a c() call - check if it has names
      call_names <- names(unserialize_expr)
      if (!is.null(call_names) && any(nzchar(call_names[-1]))) {
        # Has named arguments (skip first element which is the function name 'c')
        # Extract as a named list to preserve names in JSON
        unserialize_str <- list()
        for (i in 2:length(unserialize_expr)) {
          if (nzchar(call_names[i])) {
            # Get the value as a string
            val <- unserialize_expr[[i]]
            if (is.character(val)) {
              unserialize_str[[call_names[i]]] <- val
            } else {
              unserialize_str[[call_names[i]]] <- as.character(val)
            }
          }
        }
        # Convert to a named list explicitly
        unserialize_str <- as.list(unserialize_str)
      } else {
        # No names, treat as single value
        unserialize_str <- deparse1(unserialize_expr)
      }
    } else if (is.character(unserialize_expr)) {
      # Direct character value
      unserialize_str <- unserialize_expr
    } else {
      # Try to evaluate it to see if it's a pre-existing named vector
      tryCatch(
        {
          unserialize_val <- eval(unserialize_expr, envir = parent.frame())
          if (
            !is.null(names(unserialize_val)) &&
              length(names(unserialize_val)) > 0
          ) {
            # Convert named vector to named list for JSON preservation
            unserialize_str <- as.list(setNames(
              unserialize_val,
              names(unserialize_val)
            ))
          } else {
            # Single value
            if (is.character(unserialize_val)) {
              unserialize_str <- unserialize_val
            } else {
              unserialize_str <- deparse1(unserialize_expr)
            }
          }
        },
        error = function(e) {
          # If evaluation fails, treat as a symbol
          unserialize_str <- deparse1(unserialize_expr)
        }
      )
    }
  }

  # Generate environment variable export statements if env_var is provided
  env_exports <- ""
  if (!is.null(env_var)) {
    env_exports <- paste(
      vapply(
        names(env_var),
        function(var_name) {
          sprintf("export %s=%s", var_name, env_var[[var_name]])
        },
        character(1)
      ),
      collapse = "\n      "
    )
    if (env_exports != "") {
      env_exports <- paste0(env_exports, "\n      ")
    }
  }

  # Prepare the fileset for src
  all_files <- c(additional_files, user_functions)
  fileset_parts <- all_files[nzchar(all_files)]

  # If there are any source files, copy them all into the build directory
  copy_cmd <- if (length(fileset_parts) > 0) {
    "cp -r $src/* .\n      "
  } else {
    ""
  }

  # Generate source commands for user_functions
  source_cmd <- ""
  if (!is.null(user_functions) && length(user_functions) > 0) {
    user_functions_clean <- user_functions[nzchar(user_functions)]
    if (length(user_functions_clean) > 0) {
      source_lines <- vapply(
        user_functions_clean,
        function(f) sprintf("source('%s')", f),
        character(1)
      )
      source_cmd <- paste0(
        paste(source_lines, collapse = "\n        "),
        "\n        "
      )
    }
  }

  # If you adopted name-scoped placeholders, keep using it here
  unique_placeholder <- sprintf(
    "# RIXPRESS_LOAD_DEPENDENCIES_HERE:%s",
    out_name
  )

  build_phase <- sprintf(
    "%s%sRscript -e \"\n        source('libraries.R')\n        %s%s\n        %s <- %s\n        %s(%s, '%s')\"",
    env_exports,
    copy_cmd,
    source_cmd,
    unique_placeholder,
    out_name,
    expr_str,
    serialize_str,
    out_name,
    out_name
  )

  # Derive base from nix_env (use basename to handle relative paths)
  base <- gsub("[^a-zA-Z0-9]", "_", basename(nix_env))
  base <- sub("_nix$", "", base)

  if (length(fileset_parts) > 0) {
    fileset_nix <- paste0("./", fileset_parts, collapse = " ")
    src_snippet <- sprintf(
      "     src = defaultPkgs.lib.fileset.toSource {\n      root = ./.;\n      fileset = defaultPkgs.lib.fileset.unions [ %s ];\n    };\n",
      fileset_nix
    )
  } else {
    src_snippet <- ""
  }

  snippet <- make_derivation_snippet(
    out_name = out_name,
    src_snippet = src_snippet,
    base = base,
    build_phase = build_phase,
    derivation_type = "rxp_r",
    noop_build = noop_build
  )

  list(
    name = out_name,
    snippet = snippet,
    type = "rxp_r",
    additional_files = additional_files,
    user_functions = user_functions,
    nix_env = nix_env,
    encoder = serialize_str,
    decoder = unserialize_str,
    env_var = env_var,
    noop_build = noop_build
  ) |>
    structure(class = "rxp_derivation")
}

#' Create a Nix Expression Running a Python Function
#'
#' @family derivations
#' @param name Symbol, name of the derivation.
#' @param expr Character, Python code to generate the expression. Ideally it
#'   should be a call to a pure function. Multi-line expressions are not
#'   supported.
#' @param additional_files Character vector, additional files to include
#'   during the build process. For example, if a function expects a certain
#'   file to be available, this is where you should include it.
#' @param user_functions Character vector, user-defined functions to include.
#'   This should be a script (or scripts) containing user-defined functions
#'   to include during the build process for this derivation. It is recommended
#'   to use one script per function, and only include the required script(s) in
#'   the derivation.
#' @param nix_env Character, path to the Nix environment file, default is
#'   "default.nix".
#' @param encoder Character, defaults to NULL. The name of the Python
#'   function used to serialize the object. It must accept two arguments: the
#'   object to serialize (first), and the target file path (second). If NULL,
#'   the default behaviour uses `pickle.dump`. Define this function in
#'   `functions.py`. See `vignette("encoding-decoding")` for more details.
#' @param decoder Character or named vector/list, defaults to NULL. Can be:
#'   - A single string for the Python function to unserialize all upstream objects
#'   - A named vector/list where names are upstream dependency names and values
#'     are their specific unserialize functions
#'   If NULL, the default uses `pickle.load`. See `vignette("encoding-decoding")` for more details.
#' @param env_var Character vector, defaults to NULL. A named vector of
#'   environment variables
#'   before running the Python script, e.g., c(PYTHONPATH = "/path/to/modules").
#'   Each entry will be added as an export statement in the build phase.
#' @param noop_build Logical, defaults to FALSE. If TRUE, the derivation
#'   produces a no-op build (a stub output with no actual build steps). Any
#'   downstream derivations depending on a no-op build will themselves also
#'   become no-op builds.
#' @details At a basic level,
#'   `rxp_py(mtcars_am, "mtcars.filter(polars.col('am') == 1).to_pandas()")`
#'    is equivalent to
#'   `mtcars_am = mtcars.filter(polars.col('am') == 1).to_pandas()`. `rxp_py()`
#'   generates the required Nix boilerplate to output a so-called "derivation"
#'   in Nix jargon. A Nix derivation is a recipe that defines how to create an
#'   output (in this case `mtcars_am`) including its dependencies, build steps,
#'   and output paths.
#' @return An object of class derivation which inherits from lists.
#' @examples
#' \dontrun{
#'   rxp_py(
#'     mtcars_pl_am,
#'     expr = "mtcars_pl.filter(polars.col('am') == 1).to_pandas()"
#'   )
#'
#'   # Skip building this derivation
#'   rxp_py(
#'     data_prep,
#'     expr = "preprocess_data(raw_data)",
#'     noop_build = TRUE
#'   )
#'
#'   # Custom serialization
#'   rxp_py(
#'     mtcars_pl_am,
#'     expr = "mtcars_pl.filter(polars.col('am') == 1).to_pandas()",
#'     user_functions = "functions.py",
#'     encoder = "serialize_model",
#'     additional_files = "some_required_file.bin")
#' }
#' @export
rxp_py <- function(
  name,
  expr,
  additional_files = "",
  user_functions = "",
  nix_env = "default.nix",
  encoder = NULL,
  decoder = NULL,
  env_var = NULL,
  noop_build = FALSE
) {
  out_name <- deparse1(substitute(name))
  expr <- gsub("'", "\\'", expr, fixed = TRUE)

  # Handle encoder for the build_phase
  if (is.null(encoder)) {
    serialize_str <- sprintf(
      "with open('%s', 'wb') as f: pickle.dump(globals()['%s'], f)",
      out_name,
      out_name
    )
  } else {
    if (!is.character(encoder)) {
      stop("encoder must be a character string or NULL")
    }
    serialize_str <- sprintf(
      "%s(globals()['%s'], '%s')",
      encoder,
      out_name,
      out_name
    )
  }

  # Handle decoder - can be single value or named vector/list
  unserialize_expr <- substitute(decoder)
  if (identical(unserialize_expr, quote(NULL))) {
    unserialize_str <- "pickle.load"
  } else {
    # Check if it's a call to c() with named arguments
    if (
      is.call(unserialize_expr) && identical(unserialize_expr[[1]], quote(c))
    ) {
      # It's a c() call - check if it has names
      call_names <- names(unserialize_expr)
      if (!is.null(call_names) && any(nzchar(call_names[-1]))) {
        # Has named arguments (skip first element which is the function name 'c')
        # Extract as a named list to preserve names in JSON
        unserialize_str <- list()
        for (i in 2:length(unserialize_expr)) {
          if (nzchar(call_names[i])) {
            # Get the value as a string
            val <- unserialize_expr[[i]]
            if (is.character(val)) {
              unserialize_str[[call_names[i]]] <- val
            } else {
              unserialize_str[[call_names[i]]] <- as.character(val)
            }
          }
        }
        # Convert to a named list explicitly
        unserialize_str <- as.list(unserialize_str)
      } else {
        # No names, treat as single value
        if (length(unserialize_expr) > 1) {
          # It's c() with multiple unnamed values - take the first
          val <- unserialize_expr[[2]]
          unserialize_str <- if (is.character(val)) val else as.character(val)
        } else {
          unserialize_str <- deparse1(unserialize_expr)
        }
      }
    } else if (is.character(unserialize_expr)) {
      # Direct character value
      unserialize_str <- unserialize_expr
    } else {
      # Try to evaluate it to see if it's a pre-existing named vector
      tryCatch(
        {
          unserialize_val <- eval(unserialize_expr, envir = parent.frame())
          if (
            !is.null(names(unserialize_val)) &&
              length(names(unserialize_val)) > 0
          ) {
            # Convert named vector to named list for JSON preservation
            unserialize_str <- as.list(setNames(
              unserialize_val,
              names(unserialize_val)
            ))
          } else {
            # Single value
            if (is.character(unserialize_val)) {
              unserialize_str <- unserialize_val
            } else {
              unserialize_str <- as.character(unserialize_val)
            }
          }
        },
        error = function(e) {
          # If evaluation fails, treat as a symbol and convert to string
          unserialize_str <- deparse1(unserialize_expr)
        }
      )
    }
  }

  # Generate environment variable export statements if env_var is provided
  env_exports <- ""
  if (!is.null(env_var)) {
    env_exports <- paste(
      vapply(
        names(env_var),
        function(var) sprintf("export %s=%s", var, env_var[[var]]),
        character(1)
      ),
      collapse = "\n      "
    )
    if (env_exports != "") env_exports <- paste0(env_exports, "\n      ")
  }

  # Prepare the fileset for src
  all_files <- c(additional_files, user_functions)
  fileset_parts <- all_files[nzchar(all_files)]

  # If there are any source files, copy them all into the build directory
  copy_cmd <- if (length(fileset_parts) > 0) {
    "cp -r $src/* .\n      "
  } else {
    ""
  }

  # Generate import commands for user_functions
  user_import_cmd <- ""
  if (!is.null(user_functions) && any(nzchar(user_functions))) {
    user_functions_clean <- user_functions[nzchar(user_functions)]
    import_lines <- vapply(
      user_functions_clean,
      function(f) sprintf("exec(open('%s').read())", f),
      character(1)
    )
    user_import_cmd <- paste0(paste(import_lines, collapse = "\n"), "\n")
  }

  # Unique placeholder per derivation
  unique_placeholder <- sprintf(
    "# RIXPRESS_PY_LOAD_DEPENDENCIES_HERE:%s",
    out_name
  )

  # Construct build_phase including cp commands then python execution
  build_phase <- paste0(
    env_exports,
    copy_cmd,
    "python -c \"\n",
    "exec(open('libraries.py').read())\n",
    user_import_cmd,
    unique_placeholder,
    "\n",
    "exec('",
    out_name,
    " = ",
    expr,
    "')\n",
    serialize_str,
    "\n",
    "\""
  )

  # Derive base from nix_env (use basename to handle relative paths)
  base <- gsub("[^a-zA-Z0-9]", "_", basename(nix_env))
  base <- sub("_nix$", "", base)

  # Prepare the src snippet with all files
  if (length(fileset_parts) > 0) {
    fileset_nix <- paste0("./", fileset_parts, collapse = " ")
    src_snippet <- sprintf(
      "     src = defaultPkgs.lib.fileset.toSource {\n      root = ./.;\n      fileset = defaultPkgs.lib.fileset.unions [ %s ];\n    };\n",
      fileset_nix
    )
  } else {
    src_snippet <- ""
  }

  # Generate the Nix snippet
  snippet <- make_derivation_snippet(
    out_name = out_name,
    src_snippet = src_snippet,
    base = base,
    build_phase = build_phase,
    derivation_type = "rxp_py",
    noop_build = noop_build
  )

  list(
    name = out_name,
    snippet = snippet,
    type = "rxp_py",
    additional_files = additional_files,
    user_functions = user_functions,
    nix_env = nix_env,
    encoder = serialize_str,
    decoder = unserialize_str,
    env_var = env_var,
    noop_build = noop_build
  ) |>
    structure(class = "rxp_derivation")
}

#' Create a Nix Expression Running a Julia Function
#'
#' @param name Symbol, name of the derivation.
#' @param expr Character, Julia code to generate the expression. Ideally it
#'   should be a call to a pure function. Multi-line expressions are not
#'   supported.
#' @param additional_files Character vector, additional files to include
#'   during the build process. For example, if a function expects a certain
#'   file to be available, this is where you should include it.
#' @param user_functions Character vector, user-defined functions to include.
#'   This should be a script (or scripts) containing user-defined functions
#'   to include during the build process for this derivation. It is recommended
#'   to use one script per function, and only include the required script(s) in
#'   the derivation.
#' @param nix_env Character, path to the Nix environment file, default is
#'   "default.nix".
#' @param encoder Character, defaults to NULL. The name of the Julia
#'   function used to serialize the object. It must accept two arguments: the
#'   object to serialize (first), and the target file path (second). If NULL,
#'   the default behaviour uses the built‐in `Serialization.serialize` API. Define
#'   any custom serializer in `functions.jl`. See `vignette("encoding-decoding")` for more details.
#' @param decoder Character or named vector/list, defaults to NULL. Can be:
#'   - A single string for the Julia function to unserialize all upstream objects
#'   - A named vector/list where names are upstream dependency names and values
#'     are their specific unserialize functions
#'   If NULL, the default is `Serialization.deserialize`. See `vignette("encoding-decoding")` for more details.
#' @param env_var Character vector, defaults to NULL. A named vector of
#'   environment variables to set before running the Julia script, e.g.,
#'   `c("JULIA_DEPOT_PATH" = "/path/to/depot")`. Each entry will be added as
#'   an `export` statement in the build phase.
#' @param noop_build Logical, defaults to FALSE. If TRUE, the derivation
#'   produces a no-op build (a stub output with no actual build steps). Any
#'   downstream derivations depending on a no-op build will themselves also
#'   become no-op builds.
#' @details At a basic level,
#'   `rxp_jl(filtered_data, "filter(df, :col .> 10)")` is equivalent to
#'   `filtered_data = filter(df, :col .> 10)` in Julia. `rxp_jl()` generates the
#'   required Nix boilerplate to output a so‐called "derivation" in Nix jargon.
#'   A Nix derivation is a recipe that defines how to create an output (in this
#'   case `filtered_data`) including its dependencies, build steps, and output
#'   paths.
#' @return An object of class derivation which inherits from lists.
#' @examples
#' \dontrun{
#' # Basic usage, no custom serializer
#' rxp_jl(
#'   name = filtered_df,
#'   expr = "filter(df, :col .> 10)"
#' )
#'
#' # Skip building this derivation
#' rxp_jl(
#'   name = model_result,
#'   expr = "train_model(data)",
#'   noop_build = TRUE
#' )
#'
#' # Custom serialization: assume `save_my_obj(obj, path)` is defined in functions.jl
#' rxp_jl(
#'   name = model_output,
#'   expr = "train_model(data)",
#'   encoder = "save_my_obj",
#'   user_functions = "functions.jl"
#' )
#' }
#' @family derivations
#' @export
rxp_jl <- function(
  name,
  expr,
  additional_files = "",
  user_functions = "",
  nix_env = "default.nix",
  encoder = NULL,
  decoder = NULL,
  env_var = NULL,
  noop_build = FALSE
) {
  out_name <- deparse1(substitute(name))
  # Escape double quotes for Julia one-liner
  expr_escaped <- gsub("\"", "\\\\\"", expr)

  # Determine which serialize function to call
  if (is.null(encoder)) {
    # Default: use built-in Serialization.serialize
    serialize_str <- paste0(
      "using Serialization; ",
      "io = open(\\\"",
      out_name,
      "\\\", \\\"w\\\"); ",
      "serialize(io, ",
      out_name,
      "); ",
      "close(io)"
    )
  } else {
    if (!is.character(encoder) || length(encoder) != 1) {
      stop("encoder must be a single character string or NULL")
    }
    serialize_str <- sprintf(
      "%s(%s, \\\"%s\\\")",
      encoder,
      out_name,
      out_name
    )
  }

  # Handle decoder - can be single value or named vector/list
  unserialize_expr <- substitute(decoder)
  if (identical(unserialize_expr, quote(NULL))) {
    unserialize_str <- "Serialization.deserialize"
  } else {
    # Check if it's a call to c() with named arguments
    if (
      is.call(unserialize_expr) && identical(unserialize_expr[[1]], quote(c))
    ) {
      # It's a c() call - check if it has names
      call_names <- names(unserialize_expr)
      if (!is.null(call_names) && any(nzchar(call_names[-1]))) {
        # Has named arguments (skip first element which is the function name 'c')
        # Extract as a named list to preserve names in JSON
        unserialize_str <- list()
        for (i in 2:length(unserialize_expr)) {
          if (nzchar(call_names[i])) {
            # Get the value as a string
            val <- unserialize_expr[[i]]
            if (is.character(val)) {
              unserialize_str[[call_names[i]]] <- val
            } else {
              unserialize_str[[call_names[i]]] <- as.character(val)
            }
          }
        }
        # Convert to a named list explicitly
        unserialize_str <- as.list(unserialize_str)
      } else {
        # No names, treat as single value
        if (length(unserialize_expr) > 1) {
          # It's c() with multiple unnamed values - take the first
          val <- unserialize_expr[[2]]
          unserialize_str <- if (is.character(val)) val else as.character(val)
        } else {
          unserialize_str <- deparse1(unserialize_expr)
        }
      }
    } else if (is.character(unserialize_expr)) {
      # Direct character value
      unserialize_str <- unserialize_expr
    } else {
      # Try to evaluate it to see if it's a pre-existing named vector
      tryCatch(
        {
          unserialize_val <- eval(unserialize_expr, envir = parent.frame())
          if (
            !is.null(names(unserialize_val)) &&
              length(names(unserialize_val)) > 0
          ) {
            # Convert named vector to named list for JSON preservation
            unserialize_str <- as.list(setNames(
              unserialize_val,
              names(unserialize_val)
            ))
          } else {
            # Single value
            if (is.character(unserialize_val)) {
              unserialize_str <- unserialize_val
            } else {
              unserialize_str <- as.character(unserialize_val)
            }
          }
        },
        error = function(e) {
          # If evaluation fails, treat as a symbol and convert to string
          unserialize_str <- deparse1(unserialize_expr)
        }
      )
    }
  }

  # Generate environment variable export statements if env_var is provided
  env_exports <- ""
  if (!is.null(env_var)) {
    env_exports <- paste(
      vapply(
        names(env_var),
        function(var) sprintf("export %s=%s", var, env_var[[var]]),
        character(1)
      ),
      collapse = "\n      "
    )
    if (nzchar(env_exports)) {
      env_exports <- paste0(env_exports, "\n      ")
    }
  }

  # Prepare the fileset for src
  all_files <- c(additional_files, user_functions)
  fileset_parts <- all_files[nzchar(all_files)]

  # If there are any source files, copy them all into the build directory
  copy_cmd <- if (length(fileset_parts) > 0) {
    "cp -r $src/* .\n      "
  } else {
    ""
  }

  # Generate include commands for user_functions
  user_include_cmd <- ""
  if (!is.null(user_functions) && any(nzchar(user_functions))) {
    user_functions_clean <- user_functions[nzchar(user_functions)]
    include_lines <- vapply(
      user_functions_clean,
      function(f) sprintf("include(\\\"%s\\\")", f),
      character(1)
    )
    user_include_cmd <- paste0(paste(include_lines, collapse = "; "), "; ")
  }

  # Unique placeholder per derivation (line-only, no trailing semicolon)
  unique_placeholder <- sprintf(
    "# RIXPRESS_JL_LOAD_DEPENDENCIES_HERE:%s",
    out_name
  )

  # Ensure there is a newline after user_include_cmd if it is non-empty
  user_include_cmd_nl <- if (nzchar(user_include_cmd)) {
    paste0(user_include_cmd, "\n")
  } else {
    ""
  }

  # Construct the Julia build phase: include libraries.jl if present,
  # include user_functions, run expression, then serialize
  build_phase <- paste0(
    env_exports,
    copy_cmd,
    "julia -e \"\n",
    "if isfile(\\\"libraries.jl\\\"); include(\\\"libraries.jl\\\"); end;\n",
    user_include_cmd_nl,
    unique_placeholder,
    "\n",
    out_name,
    " = ",
    expr_escaped,
    "; ",
    serialize_str,
    "\n",
    "\""
  )

  # Derive base variable from nix_env (use basename to handle relative paths)
  base <- gsub("[^a-zA-Z0-9]", "_", basename(nix_env))
  base <- sub("_nix$", "", base)

  # Prepare src snippet with all relevant files
  if (length(fileset_parts) > 0) {
    fileset_nix <- paste0("./", fileset_parts, collapse = " ")
    src_snippet <- sprintf(
      "    src = defaultPkgs.lib.fileset.toSource {\n      root = ./.;\n      fileset = defaultPkgs.lib.fileset.unions [ %s ];\n    };\n",
      fileset_nix
    )
  } else {
    src_snippet <- ""
  }

  # Assemble the Nix-derivation snippet
  snippet <- make_derivation_snippet(
    out_name = out_name,
    src_snippet = src_snippet,
    base = base,
    build_phase = build_phase,
    derivation_type = "rxp_jl",
    noop_build = noop_build
  )

  list(
    name = out_name,
    snippet = snippet,
    type = "rxp_jl",
    additional_files = additional_files,
    user_functions = user_functions,
    nix_env = nix_env,
    encoder = if (is.null(encoder)) {
      "Serialization.serialize"
    } else {
      encoder
    },
    decoder = unserialize_str,
    env_var = env_var,
    noop_build = noop_build
  ) |>
    structure(class = "rxp_derivation")
}


#' Render a Quarto Document as a Nix Derivation
#'
#' @family derivations
#' @param name Symbol, derivation name.
#' @param qmd_file Character, path to .qmd file.
#' @param additional_files Character vector, additional files to include, for
#'   example a folder containing images to include in the Quarto document.
#' @param nix_env Character, path to the Nix environment file, default is
#'   "default.nix".
#' @param args A character of additional arguments to be passed directly to the
#'   `quarto` command.
#' @param env_var List, defaults to NULL. A named list of environment variables
#'   to set before running the Quarto render command, e.g., c(QUARTO_PROFILE =
#'   "production"). Each entry will be added as an export statement in the build
#'   phase.
#' @param noop_build Logical, defaults to FALSE. If TRUE, the derivation
#'   produces a no-op build (a stub output with no actual build steps). Any
#'   downstream derivations depending on a no-op build will themselves also
#'   become no-op builds.
#' @details To include built derivations in the document,
#'   `rxp_read("derivation_name")` should be put in the .qmd file.
#' @return An object of class derivation which inherits from lists.
#' @examples
#' \dontrun{
#'   # Compile a .qmd file to a pdf using typst
#'   # `images` is a folder containing images to include in the Quarto doc
#'   rxp_qmd(
#'     name = report,
#'     qmd_file = "report.qmd",
#'     additional_files = "images",
#'     args = "--to typst"
#'   )
#'
#'   # Skip building this derivation
#'   rxp_qmd(
#'     name = draft_report,
#'     qmd_file = "draft.qmd",
#'     noop_build = TRUE
#'   )
#' }
#' @export
rxp_qmd <- function(
  name,
  qmd_file,
  additional_files = "",
  nix_env = "default.nix",
  args = "",
  env_var = NULL,
  noop_build = FALSE
) {
  out_name <- deparse1(substitute(name))

  content <- readLines(qmd_file, warn = FALSE)
  content_str <- paste(content, collapse = "\n")

  # Helper function to extract actual matches (including namespace info)
  extract_actual_matches <- function(func_name, quote_char) {
    # Pattern that matches both bare and namespaced versions
    if (quote_char == '"') {
      pattern <- sprintf('((?:rixpress::)?%s)\\("([^"]+)"\\)', func_name)
    } else {
      pattern <- sprintf("((?:rixpress::)?%s)\\('([^']+)'\\)", func_name)
    }

    matches <- gregexpr(pattern, content_str)
    full_matches <- regmatches(content_str, matches)[[1]]

    if (length(full_matches) == 0) {
      return(data.frame())
    }

    results <- data.frame(
      full_match = character(0),
      func_call = character(0),
      path = character(0),
      quote_char = character(0),
      stringsAsFactors = FALSE
    )

    for (match in full_matches) {
      if (quote_char == '"') {
        parts <- regmatches(
          match,
          regexec(
            sprintf('((?:rixpress::)?%s)\\("([^"]+)"\\)', func_name),
            match
          )
        )[[1]]
      } else {
        parts <- regmatches(
          match,
          regexec(
            sprintf("((?:rixpress::)?%s)\\('([^']+)'\\)", func_name),
            match
          )
        )[[1]]
      }

      if (length(parts) == 3) {
        results <- rbind(
          results,
          data.frame(
            full_match = parts[1],
            func_call = parts[2],
            path = parts[3],
            quote_char = quote_char,
            stringsAsFactors = FALSE
          )
        )
      }
    }

    results
  }

  # Extract all actual matches
  read_matches_double <- extract_actual_matches('rxp_read', '"')
  read_matches_single <- extract_actual_matches('rxp_read', "'")
  load_matches_double <- extract_actual_matches('rxp_load', '"')
  load_matches_single <- extract_actual_matches('rxp_load', "'")

  # Combine all matches
  all_matches <- rbind(
    read_matches_double,
    read_matches_single,
    load_matches_double,
    load_matches_single
  )

  # Get unique paths for environment variables
  all_refs <- unique(all_matches$path)

  # Generate substitution commands based on actual matches
  sub_cmds <- character(0)

  if (nrow(all_matches) > 0) {
    for (i in 1:nrow(all_matches)) {
      match <- all_matches[i, ]

      # Build the search pattern
      if (match$quote_char == '"') {
        search_pattern <- sprintf('%s("%s")', match$func_call, match$path)
      } else {
        search_pattern <- sprintf("%s('%s')", match$func_call, match$path)
      }

      # Build replacement based on function type
      is_load <- grepl("rxp_load", match$func_call)

      # Determine the correct function name to use in replacement
      # If original had namespace, preserve it for rxp_read
      if (grepl("rixpress::", match$func_call)) {
        rxp_read_func <- "rixpress::rxp_read"
      } else {
        rxp_read_func <- "rxp_read"
      }

      if (is_load) {
        replacement <- sprintf(
          '%s <- %s("${%s}")',
          match$path,
          rxp_read_func,
          match$path
        )
      } else {
        replacement <- sprintf('%s("${%s}")', rxp_read_func, match$path)
      }

      cmd <- sprintf(
        "substituteInPlace %s --replace-fail '%s' '%s'",
        qmd_file,
        search_pattern,
        replacement
      )

      sub_cmds <- c(sub_cmds, cmd)
    }
  }

  # Generate environment variable export statements if env_var is provided
  env_exports <- ""
  if (!is.null(env_var)) {
    env_exports <- paste(
      vapply(
        names(env_var),
        function(var_name) {
          sprintf("      export %s=%s", var_name, env_var[[var_name]])
        },
        character(1)
      ),
      collapse = "\n"
    )
    if (env_exports != "") {
      env_exports <- paste0(env_exports, "\n")
    }
  }

  # Extract Python version from nix_env
  python_version <- extract_python_version(nix_env, project_path = ".")

  build_phase <- paste(
    "      mkdir home",
    "      export HOME=$PWD/home",
    sprintf(
      "      export RETICULATE_PYTHON=${defaultPkgs.%s}/bin/python",
      python_version
    ),
    env_exports,
    if (length(sub_cmds) > 0) {
      paste("      ", sub_cmds, sep = "", collapse = "\n")
    } else {
      ""
    },
    sprintf("      quarto render %s %s --output-dir $out", qmd_file, args),
    sep = "\n"
  )

  # Prepare the fileset for src
  if (identical(additional_files, "")) {
    additional_files <- NULL
  }
  fileset_parts <- c(qmd_file, additional_files)
  fileset_nix <- paste0("./", fileset_parts, collapse = " ")

  # Derive base from nix_env (use basename to handle relative paths)
  base <- gsub("[^a-zA-Z0-9]", "_", basename(nix_env))
  base <- sub("_nix$", "", base)

  # Generate the Nix derivation snippet with updated buildInputs and configurePhase
  snippet <- make_derivation_snippet(
    out_name = out_name,
    src_snippet = sprintf(
      "    src = defaultPkgs.lib.fileset.toSource {\n      root = ./.;\n      fileset = defaultPkgs.lib.fileset.unions [ %s ];\n    };\n",
      fileset_nix
    ),
    base = base,
    build_phase = build_phase,
    derivation_type = "rxp_qmd",
    noop_build = noop_build
  )

  list(
    name = out_name,
    snippet = snippet,
    type = "rxp_qmd",
    qmd_file = qmd_file,
    additional_files = additional_files,
    nix_env = nix_env,
    args = args,
    env_var = env_var,
    noop_build = noop_build
  ) |>
    structure(class = "rxp_derivation")
}

#' Render an R Markdown Document as a Nix Derivation
#'
#' @family derivations
#' @param name Symbol, derivation name.
#' @param rmd_file Character, path to .Rmd file.
#' @param additional_files Character vector, additional files to include, for
#'   example a folder containing the pictures to include in the R Markdown
#'   document.
#' @param nix_env Character, path to the Nix environment file, default is
#'   "default.nix".
#' @param params List, parameters to pass to the R Markdown document. Default is
#'   NULL.
#' @param env_var List, defaults to NULL. A named list of environment variables
#'   to set before running the R Markdown render command, e.g., c(RSTUDIO_PANDOC
#'   = "/path/to/pandoc"). Each entry will be added as an export statement in
#'   the build phase.
#' @param noop_build Logical, defaults to FALSE. If TRUE, the derivation
#'   produces a no-op build (a stub output with no actual build steps). Any
#'   downstream derivations depending on a no-op build will themselves also
#'   become no-op builds.
#' @details To include objects built in the pipeline,
#'   `rxp_read("derivation_name")` should be put in the .Rmd file.
#' @return An object of class derivation which inherits from lists.
#' @examples
#' \dontrun{
#'   # Compile a .Rmd file to a pdf
#'   # `images` is a folder containing images to include in the R Markdown doc
#'   rxp_rmd(
#'     name = report,
#'     rmd_file = "report.Rmd",
#'     additional_files = "images"
#'   )
#'
#'   # Skip building this derivation
#'   rxp_rmd(
#'     name = draft_report,
#'     rmd_file = "draft.Rmd",
#'     noop_build = TRUE
#'   )
#' }
#' @export
rxp_rmd <- function(
  name,
  rmd_file,
  additional_files = "",
  nix_env = "default.nix",
  params = NULL,
  env_var = NULL,
  noop_build = FALSE
) {
  out_name <- deparse1(substitute(name))

  content <- readLines(rmd_file, warn = FALSE)
  content_str <- paste(content, collapse = "\n")

  # Extract unique rxp_read references
  matches <- gregexpr('rxp_read\\("([^"]+)"\\)', content_str)
  refs <- regmatches(content_str, matches)[[1]]
  refs <- sub('rxp_read\\("([^"]+)"\\)', '\\1', refs)
  refs <- unique(refs)

  # Generate substitution commands for each reference
  sub_cmds <- vapply(
    refs,
    function(ref) {
      sprintf(
        "substituteInPlace %s --replace-fail 'rxp_read(\"%s\")' 'rxp_read(\"${%s}\")'",
        rmd_file,
        ref,
        ref
      )
    },
    character(1)
  )

  # Prepare render arguments
  render_args <- "rmarkdown::render(input = file.path('$PWD', rmd_file), output_dir = '$out'"

  if (!is.null(params)) {
    params_str <- paste0(
      "list(",
      paste(
        mapply(
          function(name, value) sprintf("%s = %s", name, deparse1(value)),
          names(params),
          params
        ),
        collapse = ", "
      ),
      ")"
    )
    render_args <- paste0(render_args, sprintf(", params = %s", params_str))
  }

  render_args <- paste0(render_args, ")")

  # Generate environment variable export statements if env_var is provided
  env_exports <- ""
  if (!is.null(env_var)) {
    env_exports <- paste(
      vapply(
        names(env_var),
        function(var_name) {
          sprintf("      export %s=%s", var_name, env_var[[var_name]])
        },
        character(1)
      ),
      collapse = "\n"
    )
    if (env_exports != "") {
      env_exports <- paste0(env_exports, "\n")
    }
  }

  # Extract Python version from nix_env
  python_version <- extract_python_version(nix_env, project_path = ".")

  build_phase <- paste(
    "      mkdir home",
    "      export HOME=$PWD/home",
    sprintf(
      "      export RETICULATE_PYTHON=${defaultPkgs.%s}/bin/python",
      python_version
    ),
    env_exports,
    if (length(sub_cmds) > 0) {
      paste("      ", sub_cmds, sep = "", collapse = "\n")
    } else {
      ""
    },
    sprintf("      Rscript -e \"rmd_file <- '%s'; %s\"", rmd_file, render_args),
    sep = "\n"
  )

  # Prepare the fileset for src
  fileset_parts <- c(rmd_file, additional_files)
  fileset_nix <- paste0("./", fileset_parts, collapse = " ")

  # Derive base from nix_env (use basename to handle relative paths)
  base <- gsub("[^a-zA-Z0-9]", "_", basename(nix_env))
  base <- sub("_nix$", "", base)

  snippet <- make_derivation_snippet(
    out_name = out_name,
    src_snippet = sprintf(
      "    src = defaultPkgs.lib.fileset.toSource {\n      root = ./.;\n      fileset = defaultPkgs.lib.fileset.unions [ %s ];\n    };\n",
      fileset_nix
    ),
    base = base,
    build_phase = build_phase,
    derivation_type = "rxp_rmd",
    noop_build = noop_build
  )

  list(
    name = out_name,
    snippet = snippet,
    type = "rxp_rmd",
    rmd_file = rmd_file,
    additional_files = additional_files,
    nix_env = nix_env,
    params = params,
    env_var = env_var,
    noop_build = noop_build
  ) |>
    structure(class = "rxp_derivation")
}

#' Print Method for Derivation Objects
#' @param x An object of class "rxp_derivation"
#' @param ... Additional arguments passed to print methods
#' @return Nothing, prints a summary of the derivation object to the console.
#' @examples
#' \dontrun{
#' # d0 is a previously defined derivation
#'   print(d0)
#' }
#' @family utilities
#' @export
print.rxp_derivation <- function(x, ...) {
  cat("Name:", x$name, "\n")
  cat("Type:", x$type, "\n")
  cat("No-op Build:", x$noop_build, "\n")
  if ("encoder" %in% names(x)) {
    cat("Serialize function:", x$encoder, "\n")
  }
  if ("decoder" %in% names(x)) {
    cat("Unserialize function:", x$decoder, "\n")
  }
  if (x$type == "rxp_qmd") {
    cat("QMD file:", x$qmd_file, "\n")
  }
  if (x$type == "rxp_rmd") {
    cat("RMD file:", x$rmd_file, "\n")
  }
  cat(
    "Additional files:",
    if (
      is.null(x$additional_files) ||
        length(x$additional_files) == 0 ||
        all(x$additional_files == "")
    ) {
      "None"
    } else {
      paste(x$additional_files, collapse = ", ")
    },
    "\n"
  )
  if ("user_functions" %in% names(x)) {
    cat(
      "User functions:",
      if (
        is.null(x$user_functions) ||
          length(x$user_functions) == 0 ||
          all(x$user_functions == "")
      ) {
        "None"
      } else {
        paste(x$user_functions, collapse = ", ")
      },
      "\n"
    )
  }
  cat("Nix env:", x$nix_env, "\n")
  if ("env_var" %in% names(x)) {
    cat(
      "Env variables:",
      if (is.null(x$env_var)) {
        "None"
      } else {
        paste(names(x$env_var), x$env_var, sep = "=", collapse = ", ")
      },
      "\n"
    )
  }
}

testthat::skip_on_cran()

test_that("rxp_write_dag: generate correct json of dag R derivs only", {
  d1 <- rxp_r(mtcars_am, dplyr::filter(mtcars, am == 1))
  d2 <- rxp_r(mtcars_head, head(mtcars_am))
  derivs <- list(d1, d2)
  testthat::expect_snapshot_file(
    path = rxp_write_dag(derivs),
    name = "r-dag.json"
  )
})

test_that("rxp_write_dag: generate correct json of dag Py derivs only", {
  d1 <- rxp_py(mtcars_am, 'mtcars[mtcars["am"] == 1]')
  d2 <- rxp_py(mtcars_head, "mtcars_am.head()")
  derivs <- list(d1, d2)
  testthat::expect_snapshot_file(
    path = rxp_write_dag(derivs),
    name = "py-dag.json"
  )
})

test_that("rxp_write_dag: generate correct json of dag R derivs only, unsorted", {
  d1 <- rxp_r(mtcars_am, dplyr::filter(mtcars, am == 1))
  d2 <- rxp_r(mtcars_head, head(mtcars_am))
  derivs <- list(d2, d1)
  testthat::expect_snapshot_file(
    path = rxp_write_dag(derivs),
    name = "r-dag-unsorted.json"
  )
})

#' @title List All Available Build Logs
#' @family utilities
#' @description Returns a data frame with information about all build logs
#'     in the project's _rixpress directory.
#'
#' @param project_path Character, defaults to ".".
#'   Path to the root directory of the project.
#'
#' @return A data frame with log filenames, modification times, and file sizes.
#' @examples
#' \dontrun{
#'   # List all build logs in the current project
#'   logs <- rxp_list_logs()
#'
#'   # List logs from a specific project directory
#'   logs <- rxp_list_logs("path/to/project")
#' }
#' @export
rxp_list_logs <- function(project_path = ".") {
  rixpress_dir <- file.path(project_path, "_rixpress")

  if (!dir.exists(rixpress_dir)) {
    stop("_rixpress directory not found. Did you initialise the project?")
  }

  # Find all build log files (now JSON instead of RDS)
  log_files <- list.files(
    path = rixpress_dir,
    pattern = "build_log.*\\.json$",
    full.names = TRUE
  )

  if (length(log_files) == 0) {
    message("No build logs found in ", rixpress_dir)
    return(data.frame(
      filename = character(0),
      modification_time = character(0),
      size_kb = numeric(0)
    ))
  }

  # Get file information
  file_info <- file.info(log_files)

  # Create data frame with relevant information
  logs_df <- data.frame(
    filename = basename(log_files),
    modification_time = file_info$mtime,
    size_kb = round(file_info$size / 1024, 2)
  )

  # Sort by modification time (most recent first)
  logs_df <- logs_df[order(logs_df$modification_time, decreasing = TRUE), ]

  # Reset row names
  rownames(logs_df) <- NULL

  logs_df
}

#' @title Inspect the Build Result of a Pipeline
#' @family utilities
#' @description Returns a data frame with four columns:
#'     - derivation: the name of the derivation
#'     - build_success: whether the build was successful or not
#'     - path: the path of this derivation in the Nix store
#'     - output: the output, if this derivation was built successfully.
#'               Empty outputs mean that this derivation was not built
#'               successfully. Several outputs for a single derivation
#'               are possible.
#'   In the `derivation` column you will find an object called `all-derivations`.
#'   This object is generated automatically for internal purposes, and you can
#'   safely ignore it.
#'
#' @param project_path Character, defaults to ".".
#'   Path to the root directory of the project.
#' @param which_log Character, defaults to NULL. If NULL the most recent
#'   build log is used. If a string is provided, it's used as a
#'   regular expression to match against available log files.
#'
#' @return A data frame with derivation names, if their build was successful,
#'   their paths in the /nix/store, and their build outputs.
#' @examples
#' \dontrun{
#'   # Inspect the most recent build
#'   build_results <- rxp_inspect()
#'
#'   # Inspect a specific build log
#'   build_results <- rxp_inspect(which_log = "20250510")
#'
#'   # Check which derivations failed
#'   failed <- subset(build_results, !build_success)
#' }
#' @export
rxp_inspect <- function(project_path = ".", which_log = NULL) {
  rixpress_dir <- file.path(project_path, "_rixpress")

  # Get all available logs
  logs_df <- rxp_list_logs(project_path)

  if (nrow(logs_df) == 0) {
    stop("No build logs found, did you build the pipeline?")
  }

  if (is.null(which_log)) {
    # Use the most recent log
    log_path <- file.path(rixpress_dir, logs_df$filename[1])
  } else {
    # Find logs matching the pattern
    matches <- grep(which_log, logs_df$filename, value = TRUE)

    if (length(matches) == 0) {
      stop("No build logs found matching the pattern: ", which_log)
    }

    # Get the full path of the most recent matching log
    match_idx <- match(matches[1], logs_df$filename)
    log_path <- file.path(rixpress_dir, logs_df$filename[match_idx])
    message("Using log file: ", basename(log_path))
  }

  # Read JSON log and convert back into data frame
  log_data <- jsonlite::read_json(log_path, simplifyVector = TRUE)
  as.data.frame(log_data, stringsAsFactors = FALSE)
}

testthat::skip_on_cran()

test_that("rxp_inspect throws error when build log is missing", {
  expect_error(
    rxp_inspect(),
    "No build logs found, did you build the pipeline?"
  )
})

test_that("rxp_inspect correctly reads build log", {
  # Create a temporary directory for the test
  temp_dir <- tempdir()
  old_wd <- getwd()
  rixpress_dir <- file.path(temp_dir, "_rixpress")
  dir.create(rixpress_dir, recursive = TRUE, showWarnings = FALSE)

  # Change to temp directory
  setwd(temp_dir)
  on.exit(setwd(old_wd), add = TRUE)

  # Create mock build log data frame
  mock_build_log <- data.frame(
    derivation = c(
      "all-derivations",
      "mtcars",
      "mtcars_am",
      "mtcars_head",
      "mtcars_mpg",
      "mtcars_tail",
      "page"
    ),
    build_success = rep(TRUE, 7),
    path = c(
      "/nix/store/j77yah0z3jsnangrfg18mx5shpbv16bd-all-derivations",
      "/nix/store/8gk580ivgpd3hssdbasgl9mn9dnj6fjv-mtcars",
      "/nix/store/1709wy6ky5a408vaf8rxl04kdiyb5k0f-mtcars_am",
      "/nix/store/a9ahnfdx06cvny0f2gjvqvgrw81q4wf6-mtcars_head",
      "/nix/store/hf6ql2f5lci20zfkw1h5b5w10b4bsznr-mtcars_mpg",
      "/nix/store/96lnlag5bvdyph32wq27fjxw4pc6c85n-mtcars_tail",
      "/nix/store/lfzl996k7py3138wik49b8flklpa9dj3-page"
    ),
    output = c(
      "images, ....",
      "mtcars",
      "mtcars_am",
      "mtcars_head",
      "mtcars_mpg",
      "mtcars_tail",
      "images, ...."
    ),
    stringsAsFactors = FALSE
  )

  # Save mock build log as JSON
  jsonlite::write_json(
    mock_build_log,
    file.path(rixpress_dir, "build_log.json"),
    pretty = TRUE,
    auto_unbox = TRUE,
    dataframe = "rows"
  )

  # Test that rxp_inspect returns the correct data frame
  expected <- rxp_inspect()

  expect_equal(mock_build_log, expected)
})

#' Validate Verbose Parameter and Handle Backward Compatibility
#'
#' @param verbose The verbose parameter value to validate
#' @return A single non-negative integer
#' @keywords internal
#' @noRd
.rxp_validate_verbose <- function(verbose) {
  if (!is.numeric(verbose) || length(verbose) != 1) {
    stop("verbose must be a single numeric or integer value", call. = FALSE)
  }

  verbose <- as.integer(verbose)

  if (verbose < 0) {
    stop("rxp_make(): verbose must be a non-negative integer", call. = FALSE)
  }

  if (verbose > 5L) {
    warning(
      sprintf(
        "Argument 'verbose' (%d) exceeds the maximum of 5; using 5.",
        verbose
      ),
      call. = FALSE
    )
    verbose <- 5L
  }

  verbose
}

#' Parse Nix Build Output to Track Derivation Progress
#'
#' @param line Output line from nix-store
#' @param derivation_names Vector of derivation names to track
#' @return List with status info, or NULL if not a relevant line
#' @keywords internal
#' @noRd
.rxp_parse_build_line <- function(line, derivation_names) {
  # Handle final store paths (when derivations are already built)
  if (grepl("^/nix/store/.*-[^/]+$", line)) {
    for (deriv_name in derivation_names) {
      if (grepl(paste0("-", deriv_name, "$"), line)) {
        # Skip internal derivations from progress display
        if (deriv_name == "all-derivations") {
          return(NULL)
        }
        return(list(
          derivation = deriv_name,
          status = "completed",
          message = line,
          from_store = TRUE
        ))
      }
    }
  }

  # Handle dispatch: "building '/nix/store/...-derivation_name.drv'..."
  if (grepl("building '/nix/store/.*\\.drv'\\.\\.\\.", line)) {
    for (deriv_name in derivation_names) {
      if (grepl(paste0("-", deriv_name, "\\.drv"), line)) {
        # Skip internal derivations from progress display
        if (deriv_name == "all-derivations") {
          return(NULL)
        }
        return(list(
          derivation = deriv_name,
          status = "dispatched",
          message = line,
          from_store = FALSE
        ))
      }
    }
  }

  # Handle direct build failures: "error: builder for '/nix/store/...-derivation_name.drv' failed"
  if (grepl("error: builder for.*failed", line)) {
    for (deriv_name in derivation_names) {
      if (grepl(paste0("-", deriv_name, "\\.drv"), line)) {
        return(list(
          derivation = deriv_name,
          status = "errored",
          message = line,
          from_store = FALSE
        ))
      }
    }
  }

  # Handle dependency failures: "error: X dependencies of derivation '/nix/store/...-derivation_name.drv' failed"
  if (grepl("error: .* dependencies of derivation.*failed", line)) {
    for (deriv_name in derivation_names) {
      if (grepl(paste0("-", deriv_name, "\\.drv"), line)) {
        return(list(
          derivation = deriv_name,
          status = "errored",
          message = line,
          from_store = FALSE
        ))
      }
    }
  }

  # Capture actual R/Python/etc errors (these are the useful error messages)
  if (grepl("^Error in.*:|^Error:|^Fatal error", line)) {
    return(list(
      derivation = "RUNTIME_ERROR", # Special marker
      status = "runtime_error",
      message = line,
      from_store = FALSE
    ))
  }

  return(NULL)
}

#' Prepare nix-store Command Arguments
#'
#' @param max_jobs Integer, number of derivations to be built in parallel
#' @param cores Integer, number of cores a derivation can use during build
#' @param drv_paths Character vector of derivation paths
#' @param verbose Integer, verbosity level (0 = silent, >=1 = verbose)
#' @return Character vector of command arguments
#' @keywords internal
#' @noRd
.rxp_prepare_nix_store_args <- function(max_jobs, cores, drv_paths, verbose) {
  args <- c(
    "--realise",
    "--keep-going",
    "--max-jobs",
    max_jobs,
    "--cores",
    cores
  )

  # Add --verbose flags based on verbosity level
  if (verbose > 0) {
    verbose_flags <- rep("--verbose", verbose)
    args <- c(args[1], verbose_flags, args[-1])
  }

  # Add derivation paths at the end
  c(args, drv_paths)
}

#' Build Pipeline Using Nix
#'
#' Runs `nix-build` with a quiet flag, outputting to `_rixpress/result`.
#' @family pipeline functions
#' @param verbose Integer, defaults to 0L. Verbosity level: 0 = show progress
#'   indicators only, 1+ = show nix output with increasing verbosity. 0:
#'   "Progress only", 1: "Informational", 2: "Talkative", 3: "Chatty", 4: "Debug",
#'   5: "Vomit". Values higher than 5 are capped to 5.
#'   Each level adds one --verbose flag to nix-store command.
#' @param max_jobs Integer, number of derivations to be built in parallel.
#' @param cores Integer, number of cores a derivation can use during build.
#' @details
#' When the `{chronicler}` package is available, `rxp_make()` automatically
#' calls `rxp_check_chronicles()` after a successful build to check for
#' `Nothing` values in chronicle objects. This helps detect silent failures
#' in pipelines that use chronicler's `record()` decorated functions.
#' See `vignette("chronicler")` for more details.
#' @importFrom processx run
#' @importFrom utils capture.output
#' @return A character vector of paths to the built outputs.
#' @examples
#' \dontrun{
#'   # Build the pipeline with progress indicators (default)
#'   rxp_make()
#'
#'   # Build with verbose output and parallel execution
#'   rxp_make(verbose = 2, max_jobs = 4, cores = 2)
#'
#'   # Maximum verbosity
#'   rxp_make(verbose = 3)
#' }
#' @export
rxp_make <- function(verbose = 0L, max_jobs = 1, cores = 1) {
  # [existing validation code]
  verbose <- .rxp_validate_verbose(verbose)

  message("Build process started...\n")

  # [existing instantiation code]
  instantiate <- processx::run(
    command = "nix-instantiate",
    args = "pipeline.nix",
    error_on_status = FALSE,
    spinner = TRUE
  )

  if (instantiate$status != 0) {
    cat(instantiate$stderr)
    stop("nix-instantiate failed")
  }
  drv_paths <- strsplit(instantiate$stdout, "\n")[[1]]
  # Extract derivation names: everything after hash- and before .drv
  derivation_names <- gsub("^/nix/store/[^-]+-(.+)\\.drv$", "\\1", drv_paths)
  # Filter out only the internal "all-derivations"
  derivation_names <- derivation_names[derivation_names != "all-derivations"]

  # Initialize progress tracking
  build_status <- setNames(
    rep("pending", length(derivation_names)),
    derivation_names
  )
  error_messages <- setNames(
    rep(as.character(NA), length(derivation_names)),
    derivation_names
  )
  currently_building <- NULL # Track which derivation is currently building

  # Set up callbacks based on verbosity level
  if (verbose >= 1) {
    cb_stdout <- function(line, proc) cat(line, "\n")
    cb_stderr <- function(line, proc) cat(line, "\n")
  } else {
    # verbose == 0: show progress only
    cb_stdout <- function(line, proc) {
      # Skip the garbage collector warning
      if (grepl("warning.*garbage collector", line)) {
        return()
      }

      parsed <- .rxp_parse_build_line(line, derivation_names)
      if (!is.null(parsed)) {
        if (parsed$status == "completed" && parsed$from_store) {
          cat(paste0("\u2713 ", parsed$derivation, " built\n"))
          build_status[[parsed$derivation]] <<- "completed"
        }
      }
    }

    cb_stderr <- function(line, proc) {
      # Skip garbage collector warnings
      if (grepl("warning.*garbage collector", line)) {
        return()
      }

      parsed <- .rxp_parse_build_line(line, derivation_names)
      if (!is.null(parsed)) {
        if (parsed$status == "dispatched") {
          cat(paste0("+ ", parsed$derivation, " building\n"))
          build_status[[parsed$derivation]] <<- "dispatched"
          currently_building <<- parsed$derivation # Track current build
        } else if (parsed$status == "completed" && !parsed$from_store) {
          cat(paste0("\u2713 ", parsed$derivation, " built\n"))
          build_status[[parsed$derivation]] <<- "completed"
          currently_building <<- NULL
        } else if (parsed$status == "errored") {
          cat(paste0("\u2716 ", parsed$derivation, " errored\n"))
          build_status[[parsed$derivation]] <<- "errored"
          # Only store the Nix error message if we don't already have a runtime error
          if (is.na(error_messages[[parsed$derivation]])) {
            error_messages[[parsed$derivation]] <<- parsed$message
          }
          currently_building <<- NULL
        } else if (
          parsed$status == "runtime_error" && !is.null(currently_building)
        ) {
          # Associate runtime errors with currently building derivation
          # Runtime errors are more informative, so always store them
          error_messages[[currently_building]] <<- parsed$message
        }
      }
    }
  }

  # [rest of the existing build process code]
  nix_store_args <- .rxp_prepare_nix_store_args(
    max_jobs,
    cores,
    drv_paths,
    verbose
  )

  build_process <- processx::run(
    command = "nix-store",
    args = nix_store_args,
    stdout_line_callback = cb_stdout,
    stderr_line_callback = cb_stderr,
    error_on_status = FALSE,
    spinner = ifelse(verbose >= 1, TRUE, FALSE)
  )

  # Show final status for verbose = 0
  if (verbose == 0) {
    completed_count <- sum(build_status == "completed")
    errored_count <- sum(build_status == "errored")

    if (build_process$status == 0) {
      cat(paste0(
        "\u2713 pipeline completed [",
        completed_count,
        " completed, 0 errored]\n"
      ))
    } else {
      cat(paste0(
        "\u2716 errored pipeline [",
        completed_count,
        " completed, ",
        errored_count,
        " errored]\n"
      ))
    }
  }

  # [rest of existing build log creation code - unchanged]
  build_log <- lapply(drv_paths, function(drv_path) {
    output_result <- processx::run(
      command = "nix-store",
      args = c("-q", "--outputs", drv_path),
      error_on_status = FALSE
    )
    output_paths <- strsplit(output_result$stdout, "\n")[[1]]

    output_checks <- lapply(output_paths, function(path) {
      files <- list.files(path, all.files = FALSE)
      list(path = path, files = files, build_success = length(files) > 0)
    })

    deriv_name <- gsub("\\.drv$", "", gsub("^[^-]*-", "", drv_path))

    data.frame(
      derivation = deriv_name,
      build_success = vapply(output_checks, `[[`, logical(1), "build_success"),
      path = vapply(output_checks, `[[`, character(1), "path"),
      output = I(lapply(output_checks, `[[`, "files")),
      error_message = ifelse(
        deriv_name %in%
          names(error_messages) &&
          !is.na(error_messages[[deriv_name]]),
        error_messages[[deriv_name]],
        NA_character_
      ),
      stringsAsFactors = FALSE
    )
  })

  build_log <- do.call(rbind, build_log)

  # [rest of existing code unchanged...]
  timestamp <- format(Sys.time(), "%Y%m%d_%H%M%S")

  all_derivs_row <- which(build_log$derivation == "all-derivations")
  all_derivs_hash <- ""
  if (length(all_derivs_row) > 0) {
    path <- build_log$path[all_derivs_row[1]]
    all_derivs_hash <- gsub(
      "^/nix/store/([^-]+)-all-derivations.*$",
      "\\1",
      path
    )
    if (all_derivs_hash == path) {
      all_derivs_hash <- ""
    }
  }

  # choose filename with .json instead of .rds
  log_filename <- sprintf(
    "_rixpress/build_log_%s%s.json",
    timestamp,
    ifelse(all_derivs_hash != "", paste0("_", all_derivs_hash), "")
  )

  # save JSON instead of RDS
  jsonlite::write_json(
    x = build_log,
    path = log_filename,
    pretty = TRUE,
    auto_unbox = TRUE,
    dataframe = "rows" # each row as a JSON object
  )

  failures <- subset(build_log, subset = !build_success)
  if (nrow(failures) > 0) {
    warning(
      "Build failures:\n",
      paste(capture.output(print(failures)), collapse = "\n")
    )
  }

  if (build_process$status != 0) {
    if (verbose >= 1) {
      cat(build_process$stderr)
    }
    stop("Build failed! Check the log above for hints\nor run `rxp_inspect()`.")
  }

  if (build_process$status == 0) {
    message(
      "Build successful! Run `rxp_inspect()` for a summary.\n",
      "Use `rxp_read(\"derivation_name\")` to read objects or\n",
      "`rxp_load(\"derivation_name\")` to load them into the global environment."
    )

    # Automatically check for chronicle Nothing values if chronicler is available
    if (.rxp_has_chronicler()) {
      message(
        "\nThis pipeline uses {chronicler}. ",
        "Here is a summary of chronicle results:"
      )
      rxp_check_chronicles()
    }
  }

  invisible(build_log)
}

#' Export Nix Store Paths to an Archive
#'
#' Creates a single archive file containing the specified Nix store paths
#'   and their dependencies.
#' This archive can be transferred to another machine and imported into
#'   its Nix store.
#'
#' @family archive caching functions
#' @param archive_file Character, path to the archive, defaults to
#'   "_rixpress/pipeline-outputs.nar"
#' @param which_log Character or NULL, regex pattern to match a specific log file.
#'   If NULL (default), the most recent log file will be used.
#' @param project_path Character, defaults to ".".
#'   Path to the root directory of the project.
#' @return Nothing, creates an archive file at the specified location.
#' @examples
#' \dontrun{
#'   # Export the most recent build to the default location
#'   rxp_export_artifacts()
#'
#'   # Export a specific build to a custom location
#'   rxp_export_artifacts(
#'     archive_file = "my_archive.nar",
#'     which_log = "20250510"
#'   )
#' }
#' @export
rxp_export_artifacts <- function(
  archive_file = "_rixpress/pipeline_outputs.nar",
  which_log = NULL,
  project_path = "."
) {
  rixpress_dir <- file.path(project_path, "_rixpress")

  # Get all available logs
  logs_df <- rxp_list_logs(project_path)

  if (nrow(logs_df) == 0) {
    stop("No build logs found, did you build the pipeline?")
  }

  if (is.null(which_log)) {
    # Use the most recent log
    log_path <- file.path(rixpress_dir, logs_df$filename[1])
  } else {
    # Find logs matching the pattern
    matches <- grep(which_log, logs_df$filename, value = TRUE)

    if (length(matches) == 0) {
      stop("No build logs found matching the pattern: ", which_log)
    }

    # Get the full path of the most recent matching log
    match_idx <- match(matches[1], logs_df$filename)
    log_path <- file.path(rixpress_dir, logs_df$filename[match_idx])
    message("Using log file: ", basename(log_path))
  }

  message("Using build log: ", basename(log_path))

  # Read JSON instead of RDS
  build_log <- jsonlite::read_json(log_path, simplifyVector = TRUE)
  store_paths <- build_log$path

  message("Exporting store paths to ", archive_file)
  system2("nix-store", args = c("--export", store_paths), stdout = archive_file)
  message("Export completed")
}

#' Import Nix Store Paths from an Archive
#'
#' Imports the store paths contained in an archive file into the local Nix store.
#' Useful for transferring built outputs between machines.
#'
#' @family archive caching functions
#' @param archive_file Character, path to the archive, defaults to
#'   "_rixpress/pipeline-outputs.nar"
#' @return Nothing, imports the archive contents into the local Nix store.
#' @examples
#' \dontrun{
#'   # Import from the default archive location
#'   rxp_import_artifacts()
#'
#'   # Import from a custom archive file
#'   rxp_import_artifacts("path/to/my_archive.nar")
#' }
#' @export
rxp_import_artifacts <- function(
  archive_file = "_rixpress/pipeline_outputs.nar"
) {
  if (!is.character(archive_file) || length(archive_file) != 1) {
    stop("archive_file must be a single character string")
  }
  if (!file.exists(archive_file)) {
    stop("Archive file does not exist")
  }
  message("Importing store paths from ", archive_file)
  system2("nix-store", args = "--import", stdin = archive_file)
  message("Import completed")
}

test_that(".rxp_prepare_nix_store_args works with verbose = 0", {
  args <- .rxp_prepare_nix_store_args(
    max_jobs = 1,
    cores = 1,
    drv_paths = c("/nix/store/test.drv"),
    verbose = 0
  )

  # Should not contain --verbose flag
  expect_false("--verbose" %in% args)

  # Should contain other expected args
  expect_true("--realise" %in% args)
  expect_true("--keep-going" %in% args)
  expect_true("--max-jobs" %in% args)
  expect_true("--cores" %in% args)
})

test_that(".rxp_prepare_nix_store_args works with verbose = 1", {
  args <- .rxp_prepare_nix_store_args(
    max_jobs = 2,
    cores = 3,
    drv_paths = c("/nix/store/test.drv"),
    verbose = 1
  )

  # Should contain exactly one --verbose flag
  verbose_count <- sum(args == "--verbose")
  expect_equal(verbose_count, 1)

  # Should contain other expected args
  expect_true("--realise" %in% args)
  expect_true("--keep-going" %in% args)
  expect_equal(args[which(args == "--max-jobs") + 1], "2")
  expect_equal(args[which(args == "--cores") + 1], "3")
})

test_that(".rxp_prepare_nix_store_args works with verbose = 2", {
  args <- .rxp_prepare_nix_store_args(
    max_jobs = 1,
    cores = 1,
    drv_paths = c("/nix/store/test.drv"),
    verbose = 2
  )

  # Should contain exactly two --verbose flags
  verbose_count <- sum(args == "--verbose")
  expect_equal(verbose_count, 2)
})

test_that(".rxp_prepare_nix_store_args works with verbose = 3", {
  args <- .rxp_prepare_nix_store_args(
    max_jobs = 1,
    cores = 1,
    drv_paths = c("/nix/store/test.drv"),
    verbose = 3
  )

  # Should contain exactly three --verbose flags
  verbose_count <- sum(args == "--verbose")
  expect_equal(verbose_count, 3)
})

test_that("verbose parameter validation works", {
  # Test that negative verbose triggers an error
  expect_error(
    .rxp_validate_verbose(-1),
    "verbose must be a non-negative integer"
  )

  # Test that non-integer triggers an error
  expect_error(
    .rxp_validate_verbose("invalid"),
    "verbose must be a single numeric or integer value"
  )

  # Test that multiple values trigger an error
  expect_error(
    .rxp_validate_verbose(c(1, 2)),
    "verbose must be a single numeric or integer value"
  )

  # Test that valid integers work
  expect_equal(.rxp_validate_verbose(0), 0L)
  expect_equal(.rxp_validate_verbose(1), 1L)
  expect_equal(.rxp_validate_verbose(2L), 2L)
  expect_equal(.rxp_validate_verbose(3.0), 3L)
})

#' Check If chronicler Package Is Available
#'
#' @return Logical, TRUE if package is available
#' @noRd
.rxp_has_chronicler <- function() {
  requireNamespace("chronicler", quietly = TRUE) &&
    requireNamespace("maybe", quietly = TRUE)
}


#' Determine Chronicle Status: "success", "warning", or "nothing"
#'
#' @param x Object to check
#' @return Character: "success", "warning", "nothing", or NULL if not a chronicle
#' @noRd
.rxp_chronicle_state <- function(x) {
  if (!.rxp_has_chronicler()) {
    return(NULL)
  }

  if (!inherits(x, "chronicle")) {
    return(NULL)
  }

  if (maybe::is_nothing(x$value)) {
    return("nothing")
  }

  log_df <- x$log_df
  has_nok <- any(grepl("NOK", log_df$outcome))

  if (has_nok) {
    return("warning")
  }

  "success"
}


#' Get Detailed Chronicle Status Information
#'
#' @param x A chronicle object
#' @return List with status information, or NULL if not a chronicle
#' @noRd
.rxp_chronicle_status <- function(x) {
  if (!.rxp_has_chronicler()) {
    return(NULL)
  }

  if (!inherits(x, "chronicle")) {
    return(NULL)
  }

  state <- .rxp_chronicle_state(x)
  is_nothing <- (state == "nothing")

  # Extract log summary
  log_df <- x$log_df
  nok_ops <- log_df[grepl("NOK", log_df$outcome), , drop = FALSE]

  list(
    is_chronicle = TRUE,
    state = state,
    is_nothing = is_nothing,
    has_warnings = (state == "warning"),
    num_operations = nrow(log_df),
    num_failed = nrow(nok_ops),
    failed_functions = if (nrow(nok_ops) > 0) {
      nok_ops[["function"]]
    } else {
      character(0)
    },
    messages = if (nrow(nok_ops) > 0) nok_ops$message else character(0)
  )
}

#' Get the Display Symbol for a Chronicle State
#'
#' @param state Character: "success", "warning", or "nothing"
#' @return Character: the display symbol
#' @noRd
.rxp_chronicle_symbol <- function(state) {
  switch(
    state,
    "success" = "\u2713", # checkmark
    "warning" = "\u26A0", # warning sign
    "nothing" = "\u2716", # X mark
    "?" # fallback
  )
}

#' Format Chronicle Status Message for Display
#'
#' @param derivation_name Name of the derivation
#' @param status Output from .rxp_chronicle_status()
#' @return Formatted message string
#' @noRd
.rxp_format_chronicle_message <- function(derivation_name, status) {
  if (is.null(status)) {
    return(NULL)
  }

  symbol <- .rxp_chronicle_symbol(status$state)

  if (status$state == "success") {
    return(sprintf("%s %s (chronicle: OK)", symbol, derivation_name))
  }

  msg <- sprintf(
    "%s %s (chronicle: %s)",
    symbol,
    derivation_name,
    toupper(status$state)
  )

  if (length(status$failed_functions) > 0) {
    msg <- paste0(
      msg,
      "\n    Failed: ",
      paste(status$failed_functions, collapse = ", ")
    )
  }

  if (length(status$messages) > 0) {
    real_msgs <- status$messages[!is.na(status$messages)]
    real_msgs <- real_msgs[real_msgs != "Pipeline failed upstream"]
    if (length(real_msgs) > 0) {
      msg <- paste0(
        msg,
        "\n    Message: ",
        paste(real_msgs, collapse = "; ")
      )
    }
  }

  msg
}

#' Check Pipeline Outputs for Chronicle Status
#'
#' @family utilities
#' @description Scans all derivation outputs for chronicle objects and
#'   reports their status: success (Just, no warnings), warning (Just with
#'   warnings), or nothing (failed computation). Only active when chronicler
#'   is installed.
#' @param project_path Character, defaults to ".".
#'   Path to the root directory of the project.
#' @param which_log Character, defaults to NULL. If NULL the most recent
#'   build log is used. If a string is provided, it's used as a
#'   regular expression to match against available log files.
#' @return A data frame with columns: derivation, chronicle_state,
#'   num_operations, num_failed, failed_functions, messages.
#'   Returns NULL invisibly if chronicler is not installed or no
#'   chronicle objects are found.
#' @details
#' This function is useful when using the `{chronicler}` package in your
#' rixpress pipeline. Because chronicler catches errors and warnings,
#' returning `Nothing` values instead of failing, Nix builds will always
#' succeed. This function helps identify derivations that contain failed
#' computations.
#'
#' The function displays one of three symbols for each chronicle:
#' \itemize{
#'   \item \code{checkmark} Success: Just value, no warnings or errors
#'   \item \code{warning sign} Warning: Just value, but warnings were captured
#'   \item \code{X mark} Nothing: Failed computation, errors captured
#' }
#' @examples
#' \dontrun{
#'   # After building a pipeline with chronicler functions
#'   rxp_check_chronicles()
#'
#'   # Check a specific build log

#'   rxp_check_chronicles(which_log = "20250131")
#' }
#' @export
rxp_check_chronicles <- function(project_path = ".", which_log = NULL) {
  if (!.rxp_has_chronicler()) {
    message(
      "chronicler package not installed. Chronicle checking not available."
    )
    return(invisible(NULL))
  }

  build_log <- rxp_inspect(project_path = project_path, which_log = which_log)

  # Filter to successful builds only
  successful <- build_log[build_log$build_success, ]

  results <- lapply(seq_len(nrow(successful)), function(i) {
    deriv_name <- successful$derivation[i]

    # Skip internal derivations
    if (deriv_name == "all-derivations") {
      return(NULL)
    }

    # Try to read the derivation, suppressing messages/warnings
    # since we'll report them ourselves
    obj <- tryCatch(
      suppressMessages(suppressWarnings(
        rxp_read(deriv_name, which_log = which_log, project_path = project_path)
      )),
      error = function(e) NULL
    )

    if (is.null(obj)) {
      return(NULL)
    }

    status <- .rxp_chronicle_status(obj)

    if (is.null(status)) {
      return(NULL) # Not a chronicle object
    }

    data.frame(
      derivation = deriv_name,
      chronicle_state = status$state,
      num_operations = status$num_operations,
      num_failed = status$num_failed,
      failed_functions = I(list(status$failed_functions)),
      messages = I(list(status$messages)),
      stringsAsFactors = FALSE
    )
  })

  results <- results[!vapply(results, is.null, logical(1))]

  if (length(results) == 0) {
    message("No chronicle objects found in pipeline outputs.")
    return(invisible(NULL))
  }

  result_df <- do.call(rbind, results)

  # Display summary with symbols
  cat("Chronicle status:\n")
  for (i in seq_len(nrow(result_df))) {
    row <- result_df[i, ]
    status <- list(
      state = row$chronicle_state,
      failed_functions = row$failed_functions[[1]],
      messages = row$messages[[1]]
    )
    cat(.rxp_format_chronicle_message(row$derivation, status), "\n")
  }

  # Summary counts
  n_success <- sum(result_df$chronicle_state == "success")
  n_warning <- sum(result_df$chronicle_state == "warning")
  n_nothing <- sum(result_df$chronicle_state == "nothing")

  cat(sprintf(
    "\nSummary: %d success, %d with warnings, %d nothing\n",
    n_success,
    n_warning,
    n_nothing
  ))

  if (n_nothing > 0) {
    warning(
      sprintf("%d derivation(s) contain Nothing values!", n_nothing),
      call. = FALSE
    )
  }

  invisible(result_df)
}

# Tests for chronicler integration

test_that("chronicle state detection works when chronicler is available", {
  skip_if_not_installed("chronicler")
  skip_if_not_installed("maybe")

  # Success case
  r_sqrt <- chronicler::record(sqrt)
  success_result <- r_sqrt(4)
  expect_equal(rixpress:::.rxp_chronicle_state(success_result), "success")

  # Nothing case (default strict = 2 catches warnings as Nothing)
  nothing_result <- r_sqrt(-1)
  expect_equal(rixpress:::.rxp_chronicle_state(nothing_result), "nothing")
})

test_that("chronicle warning state detection works", {
  skip_if_not_installed("chronicler")
  skip_if_not_installed("maybe")

  # Warning case (strict = 1 allows warnings through but captures them in log)
  r_sqrt_strict1 <- chronicler::record(sqrt, strict = 1)
  # Suppress the warning in test output - we're testing the capture, not the warning itself
  warning_result <- suppressWarnings(r_sqrt_strict1(-1))

  # With strict = 1, warning is captured but value is still Just (not Nothing)
  # The log will contain NOK entries for warnings
  state <- rixpress:::.rxp_chronicle_state(warning_result)
  # Depending on chronicler behavior, this might be "success" or "warning"
  expect_true(state %in% c("success", "warning"))
})

test_that("status extraction includes correct details", {
  skip_if_not_installed("chronicler")
  skip_if_not_installed("maybe")

  r_sqrt <- chronicler::record(sqrt)
  result <- r_sqrt(-1)

  status <- rixpress:::.rxp_chronicle_status(result)
  expect_true(status$is_chronicle)
  expect_true(status$is_nothing)
  expect_equal(status$state, "nothing")
  expect_equal(status$num_failed, 1)
  expect_true("sqrt" %in% status$failed_functions)
})

test_that("non-chronicle objects return NULL", {
  # This should work even without chronicler installed
  expect_null(rixpress:::.rxp_chronicle_state(mtcars))
  expect_null(rixpress:::.rxp_chronicle_status(mtcars))
})

test_that("symbol function returns correct symbols", {
  expect_equal(rixpress:::.rxp_chronicle_symbol("success"), "\u2713")
  expect_equal(rixpress:::.rxp_chronicle_symbol("warning"), "\u26A0")
  expect_equal(rixpress:::.rxp_chronicle_symbol("nothing"), "\u2716")
  expect_equal(rixpress:::.rxp_chronicle_symbol("unknown"), "?")
})

test_that("format message function works correctly", {
  skip_if_not_installed("chronicler")
  skip_if_not_installed("maybe")

  r_sqrt <- chronicler::record(sqrt)
  result <- r_sqrt(-1)
  status <- rixpress:::.rxp_chronicle_status(result)

  msg <- rixpress:::.rxp_format_chronicle_message("test_deriv", status)
  expect_true(grepl("test_deriv", msg))
  expect_true(grepl("NOTHING", msg))
  expect_true(grepl("sqrt", msg))
})

test_that("has_chronicler function works", {
  # This just tests that the function runs without error
  result <- rixpress:::.rxp_has_chronicler()
  expect_type(result, "logical")
})

test_that("chronicle detection gracefully handles missing chronicler", {
  # When chronicler is not installed, functions should return NULL
  # We mock .rxp_has_chronicler to simulate missing chronicler
  skip_if_not_installed("mockery")

  mockery::stub(
    rixpress:::.rxp_chronicle_state,
    "rixpress:::.rxp_has_chronicler",
    function() FALSE
  )

  # After stubbing, should return NULL for any input
  expect_null(rixpress:::.rxp_chronicle_state(list()))
})

#' Core Setup Function for rxp_read() and rxp_load()
#' @param derivation_name The name of the derivation.
#' @param which_log Character, defaults to NULL. If NULL the most recent
#'   build log is used. If a string is provided, it's used as a
#'   regular expression to match against available log files.
#' @param project_path Character, defaults to ".".
#'   Path to the root directory of the project.
#' @details If a path to the store is passed, return it, otherwise
#'   returns the path to the object(s).
#' @return A character vector of paths to the matching files.
#' @noRd
rxp_read_load_setup <- function(
  derivation_name,
  which_log = NULL,
  project_path = "."
) {
  # if derivation_name is a full path to the /nix/store simply return it
  # this is useful for Quarto documents
  if (grepl("^/nix/store/", derivation_name)) {
    files <- list.files(derivation_name, full.names = TRUE)

    if (length(files) == 1) {
      return(files)
    } else {
      return(derivation_name)
    }
  }

  build_log <- rxp_inspect(project_path = project_path, which_log = which_log)

  derivation <- build_log[build_log$derivation == derivation_name, ]

  files <- unlist(derivation$output)

  if (length(files) == 0) {
    stop(paste(
      "No derivation called",
      derivation_name,
      "found. Run `rxp_inspect()` to check if",
      derivation_name,
      "was built successfully."
    ))
  }

  file_paths <- paste0(derivation$path, "/", files)

  # If any matching file is a .pickle, check if reticulate is installed
  if (
    any(grepl(".*\\.pickle$", file_paths)) &&
      !requireNamespace("reticulate", quietly = TRUE)
  ) {
    stop(
      "The 'reticulate' package is required to load .pickle files.\nPlease install it to use this feature."
    )
  }
  file_paths
}

#' Read Output of a Derivation
#' @family utilities
#' @description Reads the output of derivations in the current session,
#'   returns a path if reading directly is not possible.
#' @details When `derivation_name` points to a single R object,
#'   it gets read in the current session using `readRDS()`.
#'   If it's a Python object and `{reticulate}` is available,
#'   `reticulate::py_load_object()` is used. In case
#'   the derivation is pointing to several outputs (which can
#'   happen when building a Quarto document for example) or
#'   neither `readRDS()` nor `reticulate::py_load_object()`
#'   successfully read the object, the path to the object is
#'   returned instead.
#' @param derivation_name Character, the name of the derivation.
#' @param which_log Character, defaults to NULL. If NULL the most recent
#'   build log is used. If a string is provided, it's used as a
#'   regular expression to match against available log files.
#' @param project_path Character, defaults to ".".
#'   Path to the root directory of the project.
#' @return The derivation's output.
#' @examples \dontrun{
#'   mtcars <- rxp_read("mtcars")
#'
#'   # Read from a specific build log
#'   mtcars <- rxp_read("mtcars", which_log = "2025-05-10")
#' }
#' @export
rxp_read <- function(derivation_name, which_log = NULL, project_path = ".") {
  files <- rxp_read_load_setup(derivation_name, which_log, project_path)

  if (length(files) != 1) {
    return(files)
  }

  path <- files

  # Try RDS, then pickle (checking for reticulate), else return the path
  obj <- tryCatch(
    readRDS(path),
    error = function(e1) {
      if (!requireNamespace("reticulate", quietly = TRUE)) {
        message(
          "If you're trying to load a pickle'd object,\nyou need to install the '{reticulate}' package."
        )
        return(path)
      }
      tryCatch(
        reticulate::py_load_object(path),
        error = function(e2) path
      )
    }
  )

  # Check for chronicle status (only if chronicler is available)
  status <- .rxp_chronicle_status(obj)
  if (!is.null(status)) {
    if (status$state == "nothing") {
      warning(
        sprintf(
          "Derivation '%s' contains a chronicle with Nothing value!\n%s",
          derivation_name,
          "  Use chronicler::read_log() on this object for details."
        ),
        call. = FALSE
      )
    } else if (status$state == "warning") {
      message(
        sprintf(
          "Note: Derivation '%s' is a chronicle with captured warnings.\n%s",
          derivation_name,
          "  Use chronicler::read_log() on this object for details."
        )
      )
    }
  }

  obj
}

#' Load Output of a Derivation

#' @family utilities
#' @description Loads the output of derivations in the parent frame of the
#'   current session, returns a path if reading directly is not possible.
#' @details When `derivation_name` points to a single R object, it gets loaded
#'   in the current session using `assign(..., envir = parent.frame())`, which
#'   corresponds to the global environment in a regular interactive session. If
#'   you're trying to load a Python object and `{reticulate}` is available,
#'   `reticulate::py_load_object()` is used and then the object gets loaded into
#'   the global environment. In case the derivation is pointing to several
#'   outputs (which can happen when building a Quarto document for example) or
#'   loading fails, the path to the object is returned instead.
#' @param derivation_name Character, the name of the derivation.
#' @param which_log Character, defaults to NULL. If NULL the most recent
#'   build log is used. If a string is provided, it's used as a
#'   regular expression to match against available log files.
#' @param project_path Character, defaults to ".".
#'   Path to the root directory of the project.
#' @return Nothing, this function has the side effect of loading objects into
#'   the parent frame.

#' @examples
#' \dontrun{
#'   # Load an R object
#'   rxp_load("mtcars")
#'
#'   # Load a Python object
#'   rxp_load("my_python_model")
#'
#'   # Load from a specific build log
#'   rxp_load("mtcars", which_log = "2025-05-10")
#' }
#' @export
rxp_load <- function(derivation_name, which_log = NULL, project_path = ".") {
  files <- rxp_read_load_setup(derivation_name, which_log, project_path)
  if (length(files) != 1) {
    return(files)
  }
  path <- files

  # Attempt to read as RDS, else fall back to pickle (with reticulate check)
  value <- tryCatch(
    readRDS(path),
    error = function(e1) {
      # If it wasn't an RDS, try pickle
      if (!requireNamespace("reticulate", quietly = TRUE)) {
        message(
          "If you're trying to load a pickle'd object, you need to install the 'reticulate' package."
        )
        return(path)
      }
      tryCatch(
        reticulate::py_load_object(path),
        error = function(e2) {
          message("Failed to load file '", path, "': ", e2$message)
          return(path)
        }
      )
    }
  )

  # If we got back a path (i.e. both readers failed), just return it
  if (is.character(value) && length(value) == 1 && value == path) {
    message("Note: Returning file path instead of loaded object.")
    return(path)
  }

  # Check for chronicle status (only if chronicler is available)
  status <- .rxp_chronicle_status(value)
  if (!is.null(status)) {
    if (status$state == "nothing") {
      warning(
        sprintf(
          "Derivation '%s' contains a chronicle with Nothing value!\n%s",
          derivation_name,
          "  Use chronicler::read_log() on this object for details."
        ),
        call. = FALSE
      )
    } else if (status$state == "warning") {
      message(
        sprintf(
          "Note: Derivation '%s' is a chronicle with captured warnings.\n%s",
          derivation_name,
          "  Use chronicler::read_log() on this object for details."
        )
      )
    }
  }

  assign(derivation_name, value, envir = parent.frame())
  invisible(value)
}

#' Run a Pipeline on GitHub Actions
#' @family ci utilities
#' @details This function puts a `.yaml` file inside the `.github/workflows/`
#'   folder on the root of your project. This workflow file expects both
#'   scripts generated by `rxp_init()`, `gen-env.R` and `gen-pipeline.R` to be
#'   present. If that's not the case, edit the `.yaml` file accordingly. Build
#'   artifacts are archived and restored automatically between runs. Make sure
#'   to give read and write permissions to the GitHub Actions bot.
#' @return Nothing, copies file to a directory.
#' @export
#' @examples
#' \dontrun{
#'   rxp_ga()
#' }
rxp_ga <- function() {
  # is this being tested? If no, set the path to ".github/workflows"
  # if yes, set it to a temporary directory
  if (!identical(Sys.getenv("TESTTHAT"), "true")) {
    # Add an empty .gitignore file if there isn’t any

    if (file.exists(".gitignore")) {
      NULL
    } else {
      file.create(".gitignore")
    }

    path <- ".github/workflows"

    dir.create(path, recursive = TRUE)
  } else {
    path <- tempdir()
  }

  source <- system.file(
    file.path("extdata", "run-rxp-pipeline.yaml"),
    package = "rixpress",
    mustWork = TRUE
  )

  file.copy(source, path, overwrite = TRUE)

  # Generate dag.dot to view in CI
  rxp_dag_for_ci()

  if (identical(Sys.getenv("TESTTHAT"), "false")) {
    message("GitHub Actions workflow file saved to: ", path)
  }

  if (identical(Sys.getenv("TESTTHAT"), "true")) {
    paste0(path, "/run-rxp-pipeline.yaml")
  }
}

testthat::skip_on_cran()

test_that("rxp_ga creates expected GitHub Actions workflow file", {
  # Create a temporary directory for testing
  test_dir <- tempfile("rxp_ga_test")
  dir.create(test_dir)
  on.exit(unlink(test_dir, recursive = TRUE), add = TRUE)

  # Call rxp_ga
  workflow_file <- rxp_ga()

  # Capture file content for snapshot testing
  capture_file_content <- function(filepath, filename) {
    content <- readLines(filepath)
    output_file <- tempfile(
      pattern = paste0("rxp_ga_", filename),
      fileext = ".yaml"
    )
    writeLines(content, output_file)
    output_file
  }

  # Test the content of generated workflow file using snapshot
  testthat::expect_snapshot_file(
    path = capture_file_content(workflow_file, "workflow"),
    name = "rxp_ga_github_workflow.yaml"
  )
})

#' Trace Lineage of Derivations
#'
#' @family utilities
#' @param name Character, defaults to NULL. Name of the derivation to inspect.
#'   If NULL, the function prints the whole pipeline (inverted global view).
#' @param dag_file Character, defaults to "_rixpress/dag.json". Path to dag.json.
#' @param transitive Logical, defaults to TRUE. If TRUE, show transitive closure and
#'   mark transitive-only nodes with "*". If FALSE, show immediate neighbours only.
#' @param include_self Logical, defaults to FALSE. If TRUE, include `name` itself in the results.
#' @importFrom utils head
#' @return Invisibly, a named list mapping each inspected derivation name to a
#'   list with elements:
#'     - dependencies
#'     - reverse_dependencies
#'   The function also prints a tree representation to the console.
#' @export
rxp_trace <- function(
  name = NULL,
  dag_file = file.path("_rixpress", "dag.json"),
  transitive = TRUE,
  include_self = FALSE
) {
  if (!file.exists(dag_file)) {
    stop(
      "Could not find dag file at: ",
      dag_file,
      ". By default rxp_trace expects '_rixpress/dag.json'. ",
      "If your dag.json is elsewhere, pass dag_file explicitly."
    )
  }
  if (!requireNamespace("jsonlite", quietly = TRUE)) {
    stop(
      "Package 'jsonlite' is required to read dag.json. Install it and retry."
    )
  }

  load_dag <- function(path) {
    dag <- jsonlite::fromJSON(path, simplifyVector = FALSE)
    if (
      !is.list(dag) || is.null(dag$derivations) || length(dag$derivations) == 0
    ) {
      stop("Invalid dag.json: no derivations found.")
    }
    dag$derivations
  }
  derivs <- load_dag(dag_file)

  extract_name <- function(d) {
    dn <- d$deriv_name
    if (is.null(dn)) {
      return(NA_character_)
    }
    if (is.list(dn) || length(dn) > 1) {
      for (el in dn) {
        if (!is.null(el) && nzchar(as.character(el))) return(as.character(el))
      }
      return(NA_character_)
    }
    as.character(dn)
  }

  extract_color <- function(d) {
    pc <- d$pipeline_color
    if (is.null(pc) || length(pc) == 0 || is.na(pc)) {
      return(NA_character_)
    }
    # Handle list wrapping from jsonlite
    if (is.list(pc)) pc <- unlist(pc)
    as.character(pc)
  }

  all_names <- vapply(derivs, extract_name, character(1))
  all_colors <- vapply(derivs, extract_color, character(1))
  names(all_colors) <- all_names

  if (anyNA(all_names)) {
    stop("Found derivations with missing or unparsable names in dag.json.")
  }

  if (!is.null(name) && !name %in% all_names) {
    stop(
      "Derivation '",
      name,
      "' not found in dag.json (available: ",
      paste(head(all_names, 20), collapse = ", "),
      if (length(all_names) > 20) ", ..." else "",
      ")."
    )
  }

  # Helper to colorize text
  colorize_name <- function(nm) {
    col <- all_colors[nm]
    if (!is.na(col) && requireNamespace("cli", quietly = TRUE)) {
      tryCatch(
        cli::make_ansi_style(col)(nm),
        error = function(e) nm
      )
    } else {
      nm
    }
  }

  make_depends_map <- function(derivs, names) {
    out <- setNames(vector("list", length(names)), names)
    for (i in seq_along(derivs)) {
      deps <- derivs[[i]]$depends
      if (is.null(deps)) {
        deps <- character(0)
      } else {
        deps <- as.character(unlist(deps, use.names = FALSE))
        deps <- deps[nzchar(deps)]
      }
      # only keep DAG-internal deps, and drop self-loops
      deps <- deps[deps %in% names & deps != names[i]]
      out[[names[i]]] <- unique(deps)
    }
    out
  }

  depends_map <- make_depends_map(derivs, all_names)
  reverse_map <- lapply(all_names, function(n) character(0))
  names(reverse_map) <- all_names
  for (src in names(depends_map)) {
    for (dep in depends_map[[src]]) {
      reverse_map[[dep]] <- unique(c(reverse_map[[dep]], src))
    }
  }

  traverse <- function(start, graph) {
    visited <- character()
    stack <- graph[[start]]
    if (is.null(stack)) {
      stack <- character()
    }
    while (length(stack)) {
      node <- stack[[1]]
      stack <- stack[-1]
      if (node %in% visited) {
        next
      }
      visited <- c(visited, node)
      nb <- graph[[node]]
      if (is.null(nb)) {
        nb <- character()
      }
      stack <- c(stack, setdiff(nb, c(visited, stack)))
    }
    visited
  }

  marked_vec <- function(target, graph, transitive) {
    imm <- graph[[target]]
    if (is.null(imm)) {
      imm <- character()
    }
    if (!transitive) {
      return(unique(imm))
    }
    full <- traverse(target, graph)
    trans_only <- setdiff(full, imm)
    c(imm, paste0(trans_only, "*"))
  }

  print_single <- function(target) {
    cat("==== Lineage for:", colorize_name(target), "====\n")

    cat("Dependencies (ancestors):\n")
    visited <- character(0)
    rec_dep <- function(node, depth) {
      parents <- depends_map[[node]]
      if (is.null(parents) || length(parents) == 0) {
        if (depth == 0) {
          cat("  - <none>\n")
        }
        return()
      }
      for (p in parents) {
        cat(strrep("  ", depth + 1), "- ", colorize_name(p), sep = "")
        if (transitive && depth >= 1) cat("*")
        cat("\n")
        if (!(p %in% visited)) {
          visited <<- c(visited, p)
          rec_dep(p, depth + 1)
        }
      }
    }
    rec_dep(target, 0)

    cat("\nReverse dependencies (children):\n")
    visited <- character(0)
    rec_rev <- function(node, depth) {
      kids <- reverse_map[[node]]
      if (is.null(kids) || length(kids) == 0) {
        if (depth == 0) {
          cat("  - <none>\n")
        }
        return()
      }
      for (k in kids) {
        cat(strrep("  ", depth + 1), "- ", colorize_name(k), sep = "")
        if (transitive && depth >= 1) cat("*")
        cat("\n")
        if (!(k %in% visited)) {
          visited <<- c(visited, k)
          rec_rev(k, depth + 1)
        }
      }
    }
    rec_rev(target, 0)

    if (transitive) {
      cat("\nNote: '*' marks transitive dependencies (depth >= 2).\n\n")
    }
    invisible(NULL)
  }

  print_forest_once <- function(roots, graph, transitive) {
    visited <- character(0)
    rec <- function(node, depth) {
      cat(strrep("  ", depth), "- ", colorize_name(node), sep = "")
      if (transitive && depth >= 2) cat("*")
      cat("\n")

      if (node %in% visited) {
        return()
      }
      visited <<- c(visited, node)
      kids <- graph[[node]]
      if (is.null(kids) || length(kids) == 0) {
        return()
      }
      for (k in kids) {
        rec(k, depth + 1)
      }
    }
    for (r in roots) {
      rec(r, 0)
    }
    invisible(NULL)
  }

  sinks <- function() {
    no_children <- names(Filter(function(x) length(x) == 0, reverse_map))
    if (length(no_children)) {
      return(no_children)
    }
    outdeg <- vapply(reverse_map, length, integer(1))
    names(outdeg)[outdeg == min(outdeg)]
  }

  results <- setNames(
    lapply(all_names, function(nm) {
      deps <- marked_vec(nm, depends_map, transitive)
      rdeps <- marked_vec(nm, reverse_map, transitive)
      if (include_self) {
        deps <- unique(c(nm, deps))
        rdeps <- unique(c(nm, rdeps))
      }
      list(dependencies = deps, reverse_dependencies = rdeps)
    }),
    all_names
  )

  if (is.null(name)) {
    cat(
      "==== Pipeline dependency tree (outputs \u2192 inputs) ====\n"
    )
    for (r in sinks()) {
      print_forest_once(r, depends_map, transitive)
    }
    if (transitive) {
      cat("\nNote: '*' marks transitive dependencies (depth >= 2).\n\n")
    }
    invisible(results)
  } else {
    print_single(name)
    invisible(results[name])
  }
}

#' Initialize Rixpress Project
#'
#' @family utilities
#' @description
#' Generates `gen-env.R` and `gen-pipeline.R` scripts in the specified project
#'   directory, after asking the user for confirmation. If the user declines, no
#'   changes are made.
#'
#' @param project_path Character string specifying the project's path.
#' @param skip_prompt Logical. If TRUE, skips all confirmation prompts and proceeds
#'   with initialization, useful on continuous integration. Defaults to FALSE.
#'
#' @details
#' Creates (overwriting if they already exist):
#' - `gen-env.R`: Script to define an execution environment with `{rix}`.
#' - `gen-pipeline.R`: Defines a data pipeline with `{rixpress}`.
#'
#' @return Logical. Returns TRUE if initialization was successful, FALSE if the
#'   operation was cancelled by the user.
#'
#' @examples
#' # Default usage (will prompt before any action)
#' \dontrun{
#'   rxp_init()
#' }
#'
#' @export
rxp_init <- function(project_path = ".", skip_prompt = FALSE) {
  confirm <- function(question) {
    if (skip_prompt) {
      return(TRUE)
    }
    ans <- readline(paste0(question, " [y/n]: "))
    tolower(substr(ans, 1, 1)) == "y" # just pick the first letter and lower it
  }

  # Initial confirmation before any action
  if (!confirm(paste0("Initialize project at '", project_path, "'?"))) {
    message(
      "Operation cancelled by user. No files or directories were created."
    )
    return(invisible(FALSE))
  }

  # Ensure project_path exists, create it if it doesn't
  if (!dir.exists(project_path)) {
    dir.create(project_path, recursive = TRUE)
  }

  # Define file paths
  env_file <- file.path(project_path, "gen-env.R")
  pipeline_file <- file.path(project_path, "gen-pipeline.R")

  gen_env_lines <- c(
    "# This script defines the default environment the pipeline runs in.",
    "# Add the required packages to execute the code necessary for each derivation.",
    "# If you want to create visual representations of the pipeline, consider adding",
    "# `{visNetwork}` and `{ggdag}` to the list of R packages.",
    "library(rix)",
    "",
    "# Define execution environment",
    "rix(",
    "  date = NULL,",
    "  r_pkgs = NULL,",
    "  py_conf = NULL,",
    "  git_pkgs = list(",
    "    \"package_name\" = \"rixpress\",",
    "    \"repo_url\" = \"https://github.com/ropensci/rixpress\",",
    "    \"commit\" = \"4d5da094c2dc33d13f9b4966b9c502b328086714\",",
    "  ),",
    "  ide = \"none\",",
    "  project_path = \".\"",
    ")"
  )

  gen_pipeline_lines <- c(
    "library(rixpress)",
    "library(igraph)",
    "",
    "list(",
    "  rxp_r_file(",
    "    name = NULL,",
    "    path = NULL,",
    "    read_function = \\(x) read.csv(file = x, sep = \",\")",
    "  ),",
    "  rxp_r(",
    "    name = NULL,",
    "    expr = NULL",
    "  )",
    ") |>",
    "  rxp_populate(build = FALSE)"
  )

  writeLines(gen_env_lines, env_file)
  message("File ", env_file, " has been written.")
  writeLines(gen_pipeline_lines, pipeline_file)
  message("File ", pipeline_file, " has been written.")

  # Skip Git initialization when on CRAN, CI, or during testing
  if (!interactive()) {
    message(
      "Skipping Git initialization (non-interactive session, CRAN, CI, or test environment detected)."
    )
    return(invisible(TRUE))
  }

  if (confirm("Would you like to initialise a Git repository here?")) {
    if (!requireNamespace("usethis", quietly = TRUE)) {
      message(
        "The 'usethis' package is not installed. ",
        "Please install it and then run `usethis::use_git()` manually."
      )
    } else {
      usethis::use_git()
    }
  } else {
    message("Skipping Git initialization.")
  }

  invisible(TRUE)
}

test_that("rxp_init creates expected files", {
  testthat::skip_on_cran()
  # Create a temporary directory for testing
  test_dir <- tempfile("rxp_init_test")
  dir.create(test_dir)
  on.exit(unlink(test_dir, recursive = TRUE), add = TRUE)

  # Call rxp_init with skip_prompt=TRUE to avoid interactivity during testing
  rxp_init(project_path = test_dir, skip_prompt = TRUE)

  # Check if files were created
  env_file <- file.path(test_dir, "gen-env.R")
  pipeline_file <- file.path(test_dir, "gen-pipeline.R")

  expect_true(file.exists(env_file))
  expect_true(file.exists(pipeline_file))

  # Create snapshot function to capture file contents
  capture_file_content <- function(filepath, filename) {
    content <- readLines(filepath)
    output_file <- tempfile(
      pattern = paste0("rxp_init_", filename),
      fileext = ".R"
    )
    writeLines(content, output_file)
    output_file
  }

  # Test the content of generated files using snapshots
  testthat::expect_snapshot_file(
    path = capture_file_content(env_file, "env"),
    name = "rxp_init_gen-env.R"
  )

  testthat::expect_snapshot_file(
    path = capture_file_content(pipeline_file, "pipeline"),
    name = "rxp_init_gen-pipeline.R"
  )
})

---
title: "Introductory Concepts"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introductory Concepts}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

For a video version of this tutorial, [click here](https://youtu.be/bMiNfXJdgXg).

This vignette introduces key terminology and walks you through setting up a
simple pipeline using `{rixpress}`. Understanding the concepts presented here is
helpful but not mandatory. I recommend reading through this vignette, but don't
worry if you don't grasp everything immediately. You can try building a simple
pipeline by following the next vignette, `vignette("core-functions")`, and then
return to this one; things should become clearer after some hands-on experience!

Also, this package makes heavy use of its sister package `{rix}`, and I highly
recommend that you first get familiar with `Nix` by using `{rix}`. With `{rix}`,
you'll learn how to set up reproducible development environments for data
science, which you can use for interactive data analysis work. Then, if you want
to go one step further in the reproducibility
spectrum/continuum^[https://doi.org/10.1126/science.1213847], this is where
`{rixpress}` comes in.

## Definitions

In `Nix` terminology, a derivation is *a specification for running an executable
on precisely defined input files to repeatably produce output files at uniquely
determined file system paths.*
([source](https://nix.dev/manual/nix/2.25/language/derivations))

In simpler terms, a derivation is a recipe with precisely defined inputs, steps,
and a fixed output. This means that given identical inputs and build steps, the
exact same output will always be produced. To achieve this level of
reproducibility, several important measures must be taken:

- All inputs to a derivation must be explicitly declared.
- Inputs include not just data files, but also software dependencies,
  configuration flags, and environment variables, essentially anything necessary
  for the build process.
- The build process takes place in a *hermetic* sandbox to ensure the exact same
  output is always produced.

The next sections of this document explain these three points in more detail.

## Derivations

Here is an example of a *simple* `Nix` expression:

```nix
let
  pkgs = import (fetchTarball "https://github.com/rstats-on-nix/nixpkgs/archive/2025-04-11.tar.gz") {};
in
pkgs.stdenv.mkDerivation {
  name = "filtered_mtcars";
  buildInputs = [ pkgs.gawk ];
  dontUnpack = true;
  src = ./mtcars.csv;
  installPhase = ''
    mkdir -p $out
    awk -F',' 'NR==1 || $9=="1" { print }' $src > $out/filtered.csv
  '';
}
```

I won't go into details here, but what's important is that this code uses `awk`,
a common Unix data processing tool, to filter the `mtcars.csv` file to keep only
rows where the 9th column (the `am` column) equals 1. As you can see, a
significant amount of boilerplate code is required to perform this simple
operation. However, this approach is completely reproducible: the dependencies
are declared and pinned to a specific dated branch of our
`rstats-on-nix/nixpkgs` fork, and the only thing that could make this pipeline
fail (though it's a bit of a stretch to call this a *pipeline*) is if the
`mtcars.csv` file is not provided to it.

You could then add another step that uses `filtered.csv` as input and continue
processing it. If you label the above code as `f` and a subsequent chunk of `Nix`
code as `g`, then adding another step would essentially result in the following
computation: `mtcars |> f |> g`, where `f` and `g` are pure functions, and the
pipeline is thus a composition of pure functions.

`Nix` builds `filtered.csv` in two steps: it first generates a *derivation* from
this expression, and only then builds it. For clarity in this document, I'll
refer to code like the example above as a *derivation* rather than an
expression, to avoid confusion with the concept of *expression* in R.

The goal of `{rixpress}` is to help you write pipelines like `mtcars |> f |> g`
without needing to learn `Nix`, while still benefiting from its powerful
reproducibility features.

## Dependencies of derivations

`Nix` requires the dependencies of any derivation to be explicitly listed and
managed by `Nix` itself. If you're building output that requires, for example,
Quarto, then Quarto must be explicitly listed as an input, even if you already
have Quarto installed on your system. The same applies to Quarto's dependencies,
and all the dependencies of those dependencies, all the way down to the common
ancestor of all packages. With `Nix`, to run a linear regression with R, you
essentially need to build the entire universe of dependencies first.

In `Nix` terms, this complete set of packages and their dependencies are what its
author, Eelco Dolstra, refers to as *component closures*:

> The idea is to always deploy component closures: if we deploy a component,
> then we must also deploy its dependencies, their dependencies, and so on. That
> is, we must always deploy a set of components that is closed under the
> ''depends on'' relation. Since closures are self-contained, they are the units
> of complete software deployment. After all, if a set of components is not
> closed, it is not safe to deploy, since using them might cause other
> components to be referenced that are missing on the target system.

(*Nix: A Safe and Policy-Free System for Software Deployment*, Dolstra et al., 2004).

The figure below, from the same paper, illustrates this idea:

<figure>
<img src="https://raw.githubusercontent.com/ropensci/rixpress/refs/heads/main/vignettes/figure_4.png" alt="Figure 4 of Dolstra et al. (2004)" />
<figcaption aria-hidden="true">
Figure 4 of Dolstra et al. (2004)
</figcaption>
</figure>

In the figure, `subversion` depends on `openssl`, which itself depends on
`glibc`. Similarly, if you write a derivation that builds a data frame by
filtering `mtcars`, this derivation requires:

- An input file, such as `mtcars.csv`.
- R and potentially R packages like `{dplyr}`.
- All of R's dependencies and the dependencies of any R packages.
- The dependencies of those dependencies (all the way down).

All of these must be managed by `Nix`. If any dependency exists "outside" this
component closure and is only available on your machine, then the pipeline will
only *work on your machine* - defeating the purpose of reproducibility! (It
should be noted, however, that there are sometimes good reasons to have a
dependency that is not managed by `Nix`, in which case you might prefer to use
`{targets}` running inside a Nix shell instead of `{rixpress}`, but these
situations should be the exception rather than the rule).

`Nix` distinguishes between different types of dependencies (`buildInputs`,
`nativeBuildInputs`, `propagatedBuildInputs`, `propagatedNativeBuildInputs`),
but let's skip this concept, which is only relevant for packaging upstream
software, not for defining our pipelines. But if you're curious, read
[this](https://gist.github.com/b-rodrigues/c677b59126d05d43347ed9623ddd5b0c).

## The `Nix` store and hermetic builds

When building derivations, their outputs are saved into the *Nix store*.
Typically located at `/nix/store/`, this folder contains all the software and
build artifacts produced by `Nix`.

For example, if you write a derivation that computes the tail of a file named
`mtcars.csv`, once the derivation is built, its output would be stored at a path
like `/nix/store/81k4s9q652jlka0c36khpscnmr8wk7jb-mtcars_tail`. The long
cryptographic hash uniquely identifies the build output and is computed based on
the content of the derivation along with all its inputs and dependencies. This
ensures that the build is fully reproducible.

As a result, building the same derivation on two different machines will yield
the same cryptographic hash, and you can substitute the built artifact with the
derivation that generates it one-to-one. This is analogous to mathematics: if
you consider the function $f(x):= x^2$, then writing $f(2)$ or $4$ represents
the same value.

This mechanism is what makes it possible to import and export build artifacts
between pipelines to avoid having to rebuild everything from scratch on
different machines or on continuous integration platforms. `{rixpress}` has two
functions that allow this, called `rxp_export_artifacts()` and
`rxp_import_artifacts()`.

To ensure that building derivations always produces exactly the same outputs,
builds must occur in an isolated environment, often referred to as a sandbox.
This approach, known as a hermetic build process, ensures that the build is
unaffected by external factors or the state of the host system.

This isolation extends to environment variables as well. For example, R users
might set the variable `JAVA_HOME` to make R aware of where the Java runtime is
installed. However, if Java is required for a derivation, setting `JAVA_HOME`
outside of the sandbox has no effect; it must be explicitly set within the
sandbox. This isolation also means that if you need to access an API to download
data, it won't work because no Internet connection is allowed from within the
build sandbox.

This may seem very restrictive, but it makes perfect sense if your goal is to
achieve complete reproducibility. Consider a scenario where you need to use a
function `f()` to access an API to get data for your analysis. What guarantee do
you have that running `f()` today will yield the same result as running `f()` in
six months or a year? Will the API even still be online?

For true reproducibility, you should obtain the data from the API once, then
version and archive it, and continue using this archived data for your analysis
(and share it with anyone who might want to reproduce your study).

## Summary and conclusion

As explained at the beginning of this vignette, `Nix` generates a derivation from
a `Nix` expression through a process called instantiation. Writing a reproducible
pipeline with `Nix` directly would require writing very long and complex `Nix`
expressions. This is where `{rixpress}` comes in - it handles this complexity
for you.

During instantiation, `Nix` processes your declarations, resolves all inputs
(including source files, build scripts, and external dependencies), and computes
a unique cryptographic hash. This hash is derived from both the contents of your
derivation and its entire dependency graph, forming part of the derivation's
identity. This ensures that even the smallest change in your inputs will result
in a distinct derivation, guaranteeing reproducibility. To avoid confusion with
the concept of expression in R, throughout this documentation I refer to `Nix`
expressions as *derivations*.

Once instantiated, derivations can be built. During the build process, `Nix`
constructs an isolated, hermetic environment where only the explicitly declared
dependencies are available. This makes the build entirely deterministic, meaning
that identical inputs always produce identical outputs, regardless of the
machine or environment. This isolation improves reliability and facilitates
debugging and maintenance by eliminating external variables.

After a successful build, `Nix` stores the output in the `Nix` store (typically at
`/nix/store/`). For example, if you build a derivation that processes the
`mtcars.csv` file, the output might be saved under a unique path like
`/nix/store/81k4s9q652jlka0c36khpscnmr8wk7jb-mtcars_tail`. The cryptographic
hash is computed based on the derivation's inputs and build process. If
*anything* changes, the hash will be different. This is extremely precise - even
changing the separator in the `mtcars.csv` data set from `,` to `|` will result
in a different hash, even though the resulting `mtcars_tail` object might look
identical to us. From `Nix`'s perspective, they're different because one of the
inputs was different.

The key take-away is that `Nix` is a complex tool because it solves a complex
problem: ensuring complete reproducibility across different environments
and time. `{rixpress}` and `{rix}` are packages designed to make `Nix` more
accessible to R users, allowing you to benefit from `Nix`'s reproducibility
without having to learn all its complexities.

Now that you're familiar with the basic `Nix` concepts, let's move on to the next
vignette where you'll set up your first basic pipeline:
`vignette("core-functions")` (if you wish to learn more about the scope and roadmap
of the package, you can also take a look at `vignette("scope")` first).

---
title: "Using {cmdstanr} with {rixpress}"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Using cmdstanr with rixpress}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This vignette details how to effectively use the `{cmdstanr}` package within a
`{rixpress}` pipeline for Bayesian statistical modelling with Stan. For a general
introduction to `{rixpress}` and its core concepts, please refer to
`vignette("intro-concepts")` and `vignette("core-functions")`.

`{cmdstanr}` provides a user-friendly R interface to `cmdstan`, Stan's
command-line interface. While powerful, its reliance on external processes and
file system interactions requires careful handling within the hermetic build
environment of `{rixpress}`.

## Setting up the Environment

As with any `{rixpress}` pipeline, the first step is to define the execution
environment using `{rix}`:

```r
library(rix)

rix(
  date = "2025-04-29",
  r_pkgs = c("readr", "dplyr", "ggplot2"), # Add other R packages as needed
  system_pkgs = "cmdstan", # Crucial: include cmdstan as a system dependency
  git_pkgs = list(
    list(
      package_name = "cmdstanr",
      repo_url = "https://github.com/stan-dev/cmdstanr",
      commit = "79d37792d8e4ffcf3cf721b8d7ee4316a1234b0c" # Pin to a specific commit
    ),
    list(
      package_name = "rixpress",
      repo_url = "https://github.com/ropensci/rixpress",
      commit = "HEAD" # Or pin to a specific commit
    )
  ),
  ide = "none", # Or your preferred IDE
  project_path = ".",
  overwrite = TRUE
)
```

Key points in this environment definition:

-   `cmdstan` is included in `system_pkgs`. This makes the `cmdstan` executables
    available to the pipeline.
-   `{cmdstanr}` is installed from its GitHub repository, as it's not available
    on CRAN. Pinning to a specific commit is recommended for maximum
    reproducibility.

With the environment set up, we can define the pipeline:

## Setting up the pipeline

The Stan model code itself should reside in a `.stan` file. We use
`rxp_r_file()` to bring its contents into the pipeline as a character string.

```r
rxp_r_file(
  bayesian_linear_regression_model,
  "model.stan",
  readLines
)
```


Next, we define parameters and simulate some data for our model.

```r
  rxp_r(
    parameters,
    list(
      N = 100,
      alpha = 2,
      beta = -0.5,
      sigma = 1.e-1
    )
  ),
  rxp_r(
    x,
    rnorm(parameters$N, 0, 1)
  ),
  rxp_r(
    y,
    rnorm(
      n = parameters$N,
      mean = parameters$alpha + parameters$beta * x,
      sd = parameters$sigma
    )
  ),
  rxp_r(
    # Prepare the data list for cmdstanr
    inputs,
    list(N = parameters$N, x = x, y = y)
  ),
```

## Compiling and Sampling the Model

Interfacing with `cmdstan` from within `{rixpress}` requires a specific strategy
due to the hermetic nature of Nix sandboxes. We'll use a wrapper function to
handle model compilation and sampling within a single `rxp_r()` step.

First, let's define the wrapper function (e.g., in a `functions.R` file that
we'll include):

```r
# In functions.R
cmdstan_model_wrapper <- function(
  stan_string = NULL, # The Stan model code as a character string
  inputs,             # Data list for the model
  seed,               # Seed for reproducibility
  ...                 # Additional arguments for cmdstan_model or sample
) {
  # Create a temporary .stan file within the sandbox
  stan_file <- tempfile(pattern = "model_", fileext = ".stan")
  writeLines(stan_string, con = stan_file)

  # Compile the Stan model
  # cmdstanr will find cmdstan via the CMDSTAN environment variable
  model <- cmdstanr::cmdstan_model(
    stan_file = stan_file,
    ...
  )

  # Sample from the posterior
  fitted_model <- model$sample(
    data = inputs,
    seed = seed,
    ...
  )

  return(fitted_model)
}
```

Now, we use this wrapper in our pipeline:

```r
# ... (continuation of pipeline_steps list)
  rxp_r(
    model, # Target name for the fitted model object
    cmdstan_model_wrapper(
      stan_string = bayesian_linear_regression_model,
      inputs = inputs,
      seed = 22
    ),
    user_functions = "functions.R",
    encoder = "save_model",
    env_var = c("CMDSTAN" = "${defaultPkgs.cmdstan}/opt/cmdstan")
  )
```

**Explanation of the Wrapper Approach:**

1.  **`stan_string = bayesian_linear_regression_model`**: We pass the model code
    (read by `rxp_r_file`) as a string to our wrapper.
2.  **`writeLines(stan_string, con = stan_file)`**: Inside the wrapper, the Stan
    code is written to a temporary `.stan` file. This file exists *within the
    sandbox* of the current `rxp_r` step. This is crucial because
    `cmdstan_model` needs a file path. Attempting to pass the original
    `model.stan` path directly via `additional_files` to `cmdstan_model` can
    lead to permission or path issues when `cmdstan` tries to compile it from a
    different working directory or context.
3.  **`cmdstanr::cmdstan_model()`**: Compiles the model from the temporary
    `stan_file`.
4.  **`model$sample()`**: Samples from the compiled model.
5.  **Single Step**: Both compilation and sampling *must* happen within the same
    `rxp_r` step (and thus the same sandbox). This is because the `model` object
    returned by `cmdstan_model()` contains paths to the compiled executable. If
    these were separate steps, the paths from the compilation sandbox wouldn't
    be valid in the sampling sandbox.
6.  **`env_var = c("CMDSTAN" = "${defaultPkgs.cmdstan}/opt/cmdstan")`**: This
    sets the `CMDSTAN` environment variable within the sandbox for this specific
    step. `{cmdstanr}` uses this variable to locate the `cmdstan` installation.
    The `${defaultPkgs.cmdstan}` is a Nix interpolation that resolves to the
    path of the `cmdstan` package in the Nix store. If the environment providing
    `cmdstan` were named differently, for example `cmdstan-env.nix`, then you would
    need to use `${cmdstan_envPkgs.cmdstan}`.

### Custom Serialisation

`{cmdstanr}` provides a specific method for saving fitted model objects to
ensure all necessary components are preserved. We define a simple wrapper for
this to use with `{rixpress}`.

```r
save_model <- function(fitted_model, path, ...) {
  fitted_model$save_object(file = path, ...)
}
```

By specifying `encoder = "save_model"` in the `rxp_r()` call,
`{rixpress}` will use this function instead of the default `saveRDS()`. The fitted
model can then be read using `rxp_read("model")`, which will internally use
`readRDS()`.

## Summary

Using `{cmdstanr}` with `{rixpress}` involves these key considerations:

- Include `cmdstan` in `system_pkgs` and `{cmdstanr}` (from Git) in your
  `{rix}` environment definition.

- Read your `.stan` file into the pipeline using `rxp_r_file()`.

- Implement a wrapper function that:
   *   Takes the model code string and writes it to a temporary `.stan` file
       *inside the wrapper*.
   *   Calls `cmdstanr::cmdstan_model()` on this temporary file.
   *   Calls `model$sample()` to fit the model.
   *   Returns the fitted model object.

- Perform model compilation and sampling within the *same* `rxp_r()` call using
  the wrapper.

- Set the `CMDSTAN` environment variable for the `rxp_r()` step that runs the
  wrapper, pointing to the Nix store path of `cmdstan`.

- Use `{cmdstanr}`'s `$save_object()` method via a custom `encoder`
  for robust saving of the fitted model.

This approach ensures that `cmdstan` can operate correctly within the isolated
and reproducible environment provided by `{rixpress}` and Nix.

---
title: "Running Pipelines in CI"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Running Pipelines in CI}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This vignette demonstrates how to build a polyglot pipeline and assumes you've
read `vignette("polyglot")`.

## Running pipelines on GitHub Actions

Running pipelines on GitHub Actions is quite easy. First, run the `rxp_ga()`
function in your project's root. This will generate a GitHub Actions `.yaml`
file to run the pipeline on each push or pull request. Here are the different
steps that happen:

- if previous run artifacts exist, those are restored to avoid recomputing them
  using `rixpress::rxp_import_artifacts()`;
- required software is installed;
- the execution environment is generated and built;
- the `rstats-on-nix` cache is configured to decrease build times, see this
  [documentation](https://docs.ropensci.org/rix/articles/setting-up-linux-windows.html#using-the-determinate-systems-installer)
  (ignore the part about installing `Nix`);
- the pipeline is generated (and potentially built, depending on whether you set
  `build` to `FALSE` in the `rxp_populate()` call);
- the DAG gets printed;
- the pipeline is built: if you set `build` to `TRUE` previously, the build
  process is skipped anyway;
- the build artifacts and their paths are printed;
- the build artifacts are archived for reuse using
  `rixpress::rxp_export_artifacts()` for subsequent runs and are pushed to the
  `rixpress-runs` branch.

Let me explain how to view the DAG in CI. In an interactive
session, you only need to call `plot_dag()` to see a graphical representation of
the pipeline. But in CI, since there's no graphical interface, you need to
use a tool that allows you to represent the pipeline in text mode. One such tool is
the `stacked-dag` package for the Haskell programming language. It takes an
`igraph` object as a `.dot` file, and returns a textual representation of the
DAG. So, there's a step in the `.yaml` file used to run the pipeline in CI that
does exactly this:

```
- name: Check DAG if dag.dot exists and show it if yes
  run: |
    if [ -f dag.dot ]; then
      nix-shell --quiet -p haskellPackages.stacked-dag --run "stacked-dag dot _rixpress/dag.dot"
    else
      echo "dag.dot not found"
    fi
```

As you can see, `stacked-dag` processes the file from the `_rixpress/dag.dot`
folder. When calling `rxp_ga()`, the `rxp_dag_for_ci()` function is called
automatically to generate the `.dot` file and put it in the right spot.

Here is what this looks like:

<figure>
<img src="https://raw.githubusercontent.com/ropensci/rixpress/refs/heads/main/vignettes/stacked-dag.png" alt="Text representation of the DAG." />
<figcaption aria-hidden="true">
Text representation of the DAG.
</figcaption>
</figure>

---
title: "Scope and Roadmap"
output:
  rmarkdown::html_vignette:
    df_print: paged
vignette: >
  %\VignetteIndexEntry{Scope and Roadmap}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# Overview

`rixpress` focuses on “micropipelines”: pipelines executed on a single machine
for small-to-medium projects, with reproducible environments via Nix and a
simple, pragmatic user experience from R. This document clarifies what
`rixpress` is and is not, and lays out a short roadmap so users and contributors
can align expectations and proposals.

Key goals:

- Single-machine, small-to-medium pipelines (“micropipelines”).
- Reproducible builds pinned by Nix.
- Simple DAG wiring.
- Polyglot steps where R is first-class, with pragmatic Python/Julia interop.
- CI-friendly (e.g., GitHub Actions) without requiring distributed stacks.

# What `rixpress` is

- A thin, opinionated bridge between R and Nix for building pipelines locally.
- A way to define steps (R / Python / Julia / Quarto/Rmd) as Nix derivations.
- A reproducible mechanism to:
  - parse a DAG from the steps,
  - generate and build a Nix pipeline,
  - store artifacts in the Nix store, and
  - read/load/copy outputs back into R (or interoperate with Python/Julia).

- A “just enough” visualisation layer to inspect the DAG and understand
  dependencies, geared toward small-to-medium graphs.

# What `rixpress` is not

- A distributed workflow engine (no cluster schedulers, no multi-node
  orchestration).
- A long-running workflow daemon, task scheduler, or job server.
- A general-purpose ETL/ELT framework for very large datasets or streaming
  workloads.
- A feature-rich dashboarding / progress-tracking UI.
- A pluggable storage abstraction beyond the Nix store.
- An alternative DSL with complex branching semantics or dynamic targets (beyond
  what Nix + the current design support).

# Primary audience

- R users (and teams) who want strong reproducibility via Nix without leaving R.
- Python users that are not satisfied with current solutions for micropipelines,
  and who are ok with defining the pipeline as an R script.
- Projects that fit on a developer’s workstation or CI runner:
  - Research, analysis, reporting.
  - Model training/evaluation at modest scales.
  - Lightweight ETL on bounded datasets.
- Teams that occasionally mix R steps with Python/Julia, without adopting a
  heavyweight orchestration platform.

# In-scope features (current)

- Define R/Python/Julia/Quarto steps as derivations (`rxp_*()`).
- Generate and build a Nix pipeline (`rxp_populate()`, `rxp_make()`).
- Serialise/deserialise and basic interop (R↔Python/Julia) via common
  formats/APIs.
- Read/load/copy outputs from the Nix store (`rxp_read()`, `rxp_load()`,
  `rxp_copy()`).
- DAG generation and simple visual inspection (local and CI-friendly).
- CI support (GitHub Actions) with cache import/export helpers.

# Out-of-scope features (not planned)

- Cluster execution, distributed scheduling, or remote executors.
- Complex DSLs for branching, runtime expansion, or heavy dynamic targets.
- Pluggable storage backends other than the Nix store.
- High-frequency streaming pipelines, message brokers, or long-running services.
- Large-scale progress servers/dashboards.

But depending on `{rixpress}`'s success and outside contributions, these
features might be implemented sometime in the future.

# Roadmap

This roadmap lists “near-term”, “maybe later”, and “not planned” items to
clarify priorities. Timelines are indicative and may change.

## Near-term (next minor releases)

- Execution of several interconnected pipelines from a single script.

## Maybe later

- Visualisation: Mermaid-first DAG output (portable, text-first), with simple CI
  rendering. Keep a minimal interactive option for large DAGs.
- Lightweight progress summaries: per-derivation timestamps in build logs and
  simple summaries.
- Optional: a separate low-level Nix client package if deeper integrations are
  needed across `rix` and `rixpress`.

## Not planned

- Distributed compute, cluster backends, server-side schedulers.
- Alternative storage engines or backends beyond the Nix store.
- Full-featured dashboards or heavy progress monitoring services.

# How to propose new features

Before filing a feature request, please:

1. Read this scope and roadmap.
2. Check existing issues and discussions.
3. If your request is out-of-scope, consider whether it belongs in:
   - a separate package,
   - an optional companion tool, or
   - a PR to documentation (e.g., “how-to” under current scope).

When you file an issue, please:
- Explain your use case and scale (single machine, data sizes).
- Clarify why the feature belongs in `rixpress` vs. alternatives.
- Suggest a minimal interface that preserves simplicity and reproducibility.

# Related projects and inspiration

- `{targets}`, whose user-experience `{rixpress}` takes the most inspiration from.
- Micropipelines inspirations like Ploomber’s local-first approach (Python).
- Nix and nixpkgs for declarative, reproducible environments.
- `{rix}` for defining Nix environments from R.

# Links

- README: https://github.com/ropensci/rixpress#readme
- Website: https://docs.ropensci.org/rixpress/
- Demos: https://github.com/b-rodrigues/rixpress_demos

---
title: "Polyglot Pipelines and Literate Programming"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Polyglot Pipelines and Literate Programming}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This vignette demonstrates how to build a polyglot pipeline and assumes you've
read `vignette("core-functions")`.

For a video version of this vignette, [click here](https://youtu.be/LYtN1aOsTWQ).

You can find all the code of this example
[here](https://github.com/b-rodrigues/rixpress_demos/tree/master/r_python_quarto).
The built Quarto document can be viewed
[here](https://b-rodrigues.github.io/rixpress_demos/r_python_quarto/index.html)
(the pipeline in this vignette is a slightly simplified version). For the Rmd
version, look
[here](https://github.com/b-rodrigues/rixpress_demos/blob/master/r_python_rmd/Readme.md).

For various other examples of polyglot pipelines, check out the folder labeled
`python_r` in this [github
repository](https://github.com/b-rodrigues/rixpress_demos/).

## Analysing the mtcars dataset using R and Python

`{rixpress}` makes it easy to write polyglot (multilingual) data science
pipelines with derivations that run R or Python code. This vignette explains how
you can easily set up such a pipeline.

Let's assume that you only have `Nix` installed on your system, and no R
installation (this is the ideal scenario: if you plan to use `Nix` full-time for
your development environments, you shouldn't have a system-wide installation of
R).

Before installing R and R packages for your pipeline, install
[cachix](https://www.cachix.org/) and configure the `rstats-on-nix` cache. This
way, pre-compiled, binary packages will be used instead of being built from
source. Run the following line in a terminal:

```bash
nix-env -iA cachix -f https://cachix.org/api/v1/install
```

then use the cache:

```bash
cachix use rstats-on-nix
```

There might be a message telling you to add your user to a configuration file by
executing another command. If so, follow the instructions; you only need to do
this once per machine you want to use `{rixpress}` on. Many thanks to
[Cachix](https://www.cachix.org/) for sponsoring the `rstats-on-nix` cache!

Now that the cache is configured, it's time to bootstrap your development
environment. Run this line:

```
nix-shell --expr "$(curl -sl https://raw.githubusercontent.com/ropensci/rix/main/inst/extdata/default.nix)"
```

This will drop you into a temporary shell with R and both `{rix}` and
`{rixpress}` available. Simply start R by typing `R`, and load `{rixpress}` and
call `rxp_init()` which will generate two files, `gen-env.R` and
`gen-pipeline.R`. You can open `gen-env.R` in your favourite text editor and
define the execution environment there:

```{r, eval = FALSE}
library(rix)

rix(
  date = "2025-03-31",
  r_pkgs = c("dplyr", "igraph", "reticulate", "quarto"),
  git_pkgs = list(
    package_name = "rixpress",
    repo_url = "https://github.com/ropensci/rixpress",
    commit = "HEAD"
  ),
  py_conf = list(
    py_version = "3.12",
    py_pkgs = c("pandas", "polars", "pyarrow")
  ),
  ide = "none",
  project_path = ".",
  overwrite = TRUE
)
```

Notice the `py_conf` argument to `rix()`: this will install Python and the
listed Python packages in that environment. You'll notice that we add
`{reticulate}` to the list of R packages to install as well; this is primarily
for converting data between R and Python if you're not using a universal format
like JSON. Python build steps are executed in a standard Python shell and do not
require `{reticulate}` for Python code execution itself, so if you're only using
JSON to transfer data, `{reticulate}` is not required. If you prefer, you can
also use `uv` to manage Python and Python packages. While this is not a pure Nix
solution, it is still useful in cases you need a specific Python package that
might not be available through Nix, as not all PyPI packages are available
through Nix. In this case, refer to this
[section](https://docs.ropensci.org/rix/articles/installing-r-packages.html#installing-python-packages-not-available-via-nixpkgs-impure)
of the *Installing R and Python packages in a Nix environment* vignette from
`{rix}`.

Now that you defined the execution environment of the pipeline, you can run the
`gen-env.R` script, still from the temporary `Nix` shell by running
`source("gen-env.R")`. This will generate the required `default.nix`. Then, quit
R and the temporary shell (CTRL-D or `quit()` in R, `exit` in the terminal) and
then build the environment defined by the freshly generated `default.nix` by
typing `nix-build`. This will now build the execution environment of the
pipeline. You can use this environment to work on your project interactively as
usual. To learn more, check out [`{rix}`](https://docs.ropensci.org/rix/).

You can now edit the pipeline script in `gen-pipeline.R`:

```{r, eval = FALSE}
library(rixpress)
library(igraph)

list(
  rxp_py_file(
    name = mtcars_pl,
    path = 'data/mtcars.csv',
    read_function = "lambda x: pl.read_csv(x, separator='|')"
  ),

  rxp_py(
    # reticulate doesn't support polars DFs yet, so need to convert
    # first to pandas DF
    name = mtcars_pl_am,
    expr = "mtcars_pl.filter(pl.col('am') == 1).to_pandas()"
  ),

  rxp_py2r(
    name = mtcars_am,
    expr = mtcars_pl_am
  ),

  rxp_r(
    name = mtcars_head,
    expr = my_head(mtcars_am),
    user_functions = "functions.R"
  ),

  rxp_r2py(
    name = mtcars_head_py,
    expr = mtcars_head
  ),

  rxp_py(
    name = mtcars_tail_py,
    expr = 'mtcars_head_py.tail()'
  ),

  rxp_py2r(
    name = mtcars_tail,
    expr = mtcars_tail_py
  ),

  rxp_r(
    name = mtcars_mpg,
    expr = dplyr::select(mtcars_tail, mpg)
  ),

  rxp_qmd(
    name = page,
    qmd_file = "my_doc/page.qmd",
    additional_files = c("my_doc/content.qmd", "my_doc/images")
  )
) |>
  rxp_populate(
    project_path = ".",
    py_imports = c(polars = "import polars as pl")
  )
```

As you can see, it starts by reading in some data using the Python `polars`
package, and then converts it to an R data frame for further manipulation,
converts it back to a Python data frame and back to R. You'll notice that at
some point the *head* of the data is computed using a user-defined function
called `my_head()`. User-defined functions should all go into a script called
`functions.R` or `functions.py` and derivations that use them need to be aware
of them by setting the `user_functions` argument. If derivations need further
files to be available in the sandbox, these should be listed in the
`additional_files` argument. A main difference between `rxp_py()` and `rxp_r()`
is that Python code should be passed as a string, and not as an expression.

What's also import for Python is to define how packages should be imported. In
this case, I want `polars` to be imported using `import polars as pl`, so I need
to use the `py_imports` argument of `rxp_populate()`. It is possible to skip
this, but then you'd need to write the entire package name each time:
`polars.read_csv()`. This is sometimes mandatory, for example if you want to
import a package's submodule:

```{r, eval = FALSE}
py_imports = c(pillow = "from PIL import Image")
```

The package is called `pillow`, so `{rixpress}` will write the statement
as `import pillow`, but this will simply not work.

It is also possible to use `adjust_import()` after the creation of the
`pipeline.nix` but more importantly is `add_import()`. This is required in cases
where a built-in Python module needs to be loaded, such as `os`. Because the
`os` module is not listed in the required Python packages in `rix(..., py_conf =
...)` to create the execution environment, it won't get automatically loaded by
`rxp_populate()`. Because of this, if `os` is needed for the pipeline,
`add_import()` is how you can add it. The `vignette("importing-data")` show such
an example.

If you want to use JSON to transfer data between derivations, you should use
the `encoder` and `decoder` arguments respectively:

```{r, eval = FALSE}
library(rixpress)
library(igraph)

list(
  rxp_py_file(
    name = mtcars_pl,
    path = "data/mtcars.csv",
    read_function = "lambda x: pl.read_csv(x, separator='|')"
  ),

  rxp_py(
    name = mtcars_pl_am,
    expr = "mtcars_pl.filter(pl.col('am') == 1)",
    user_functions = "functions.py",
    encoder = "serialize_to_json",
  ),

  rxp_r(
    name = mtcars_head,
    expr = my_head(mtcars_pl_am),
    user_functions = "functions.R",
    decoder = "jsonlite::fromJSON"
  ),

  rxp_r(
    name = mtcars_mpg,
    expr = dplyr::select(mtcars_head, mpg)
  )
) |>
  rxp_populate(
    project_path = ".",
    py_imports = c(polars = "import polars as pl")
  )

# Plot DAG for CI
rxp_dag_for_ci()
```

The Python `serialize_to_json` function is defined in the `functions.py` script
and looks like this:

```
def serialize_to_json(pl_df, path):
    with open(path, 'w') as f:
        f.write(pl_df.write_json())
```

The `encoder` and `decoder` arguments can be used to
serialise objects using any function, for example `qs::save()` or machine
learning-specific functions for specific models, such as those from `xgboost`.

## Building a Quarto or R Markdown document

The last pipeline I want to discuss builds a Quarto document using `rxp_qmd()`
(use `rxp_rmd()` for an R Markdown document). Here again, the `additional_files`
argument is used to make the derivation aware of required files to build the
document. Here is what the source of the document looks like:

````text

---
title: "Loading Derivations Outputs in a Quarto Doc"
format:
  html:
    embed-resources: true
    toc: true
---

![Meme](images/meme.png)

Use `rxp_read()` to show object in the document:

```
#| eval: true

rixpress::rxp_read("mtcars_head")
```

```
#| eval: true

rixpress::rxp_read("mtcars_tail")
```

```
#| eval: true

rixpress::rxp_read("mtcars_mpg")
```

{{< include content.qmd >}}

```
#| eval: true

rixpress::rxp_read("mtcars_tail_py")
```

````

Just like in an interactive session, `rxp_read()` is used to retrieve the
objects from the store. See how I refer to the other document `content.qmd` and
the image `meme.png`.

If you want to add further arguments to the Quarto command line tool, you can
use the `args` argument:

```{r, eval = FALSE}
rxp_qmd(
  name = page,
  qmd_file = "my_doc/page.qmd",
  additional_files = c("my_doc/content.qmd", "my_doc/images"),
  args = "--to typst"
)
```

and don't forget to add `typst` to the list of system packages in the call to
`rix()`:

```{r, eval = FALSE}
rix(
  date = "2025-03-31",
  r_pkgs = c("dplyr", "igraph", "reticulate", "quarto"),
  system_pkgs = "typst",
  git_pkgs = list(...
```

For more examples, check out [rixpress_demos
repository](https://github.com/b-rodrigues/rixpress_demos). These examples
demonstrate additional features of `{rixpress}`, including:

- [Using the Python 'xgboost' library and transferring data to R](https://github.com/b-rodrigues/rixpress_demos/tree/master/r_py_xgboost)
- [Importing multiple files at once](https://github.com/b-rodrigues/rixpress_demos/tree/master/many_inputs_example)
- [Using multiple environments instead of a single `default.nix` file](https://github.com/b-rodrigues/rixpress_demos/tree/master/r_multi_envs)

and many others! Don’t hesitate to submit more examples as well!

---
title: "Polyglot Pipelines with Julia and R"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Polyglot Pipelines with Julia and R}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This vignette demonstrates how to build a polyglot pipeline using R and Julia.
It assumes you've read `vignette("core-functions")` and are familiar with the
basic concepts of `{rixpress}`.

For various examples of polyglot pipelines, check out the [rixpress_demos
repository](https://github.com/b-rodrigues/rixpress_demos/).

## Generating *waveshaders* data using Julia and plotting it using R

This example was provided by Moritz Schauer (@mschauer).
example adapted from: https://github.com/frankiethull/waveshaders/tree/main/experiments/simple_example

`{rixpress}` makes it easy to write polyglot (multilingual) data science
pipelines with derivations that run R or Julia code. This vignette explains how
you can easily set up such a pipeline.

### Setting up the environment

Start your project by calling `rixpress::rxp_init()` and edit `gen-env.R`.
You can open `gen-env.R` in your favourite text editor and define the execution
environment there. For a Julia and R polyglot pipeline, you'll need to add Julia
configuration:

```{r, eval = FALSE}
library(rix)

rix(
  date = "2025-09-04",
  r_pkgs = c(
    "arrow",
    "dplyr",
    "tidyr",
    "ggplot2",
    "hexbin"
  ),
  git_pkgs = list(
    list(
      package_name = "rix",
      repo_url = "https://github.com/ropensci/rix/",
      commit = "HEAD"
    ),
    list(
      package_name = "rixpress",
      repo_url = "https://github.com/ropensci/rixpress",
      commit = "HEAD"
    )
  ),
  jl_conf = list(
    jl_version = "1.10",
    jl_pkgs = c(
      "Arrow",
      "DataFrames",
      "SparseArrays",
      "LinearAlgebra"
    )
  ),
  ide = "none",
  project_path = ".",
  overwrite = TRUE
)
```

Notice the `jl_conf` argument to `rix()`: this will install Julia and the listed
Julia packages in that environment.

Now that you've defined the execution environment of the pipeline, you can run
the `gen-env.R` script, still from the temporary `Nix` shell by running
`source("gen-env.R")`. This will generate the required `default.nix`. Then, quit
R and the temporary shell (CTRL-D or `quit()` in R, `exit` in the terminal) and
then build the environment defined by the freshly generated `default.nix` by
typing `nix-build`. This will now build the execution environment of the
pipeline. You can use this environment to work on your project interactively as
usual. To learn more, check out [`{rix}`](https://docs.ropensci.org/rix/).

### Creating the pipeline

You can now edit the pipeline script in `gen-pipeline.R`. Here's an example of a
pipeline that uses both R and Julia:

```{r, eval = FALSE}
library(rixpress)
library(igraph)

list(
  rxp_jl(d_size, '150'),

  rxp_jl(
    data,
    "0.1randn(d_size,d_size) + reshape( \
    cholesky(gridlaplacian(d_size,d_size) + 0.003I) \\ randn(d_size*d_size), \
    d_size, \
    d_size \
    )",
    user_functions = "functions.jl"
  ),

  rxp_jl(
    laplace_df,
    'DataFrame(data, :auto)',
    encoder = 'arrow_write',
    user_functions = "functions.jl"
  ),

  rxp_r(
    laplace_long_df,
    prepare_data(laplace_df),
    decoder = 'read_ipc_file',
    user_functions = "functions.R"
  ),

  rxp_r(
    gg,
    make_gg(laplace_long_df)
  )

) |>
  rxp_populate(build = TRUE)
```

Let's break down what's happening in this pipeline:

- First, we define a simple variable called `d_size`, which equals `150`.
- Second, define a variable called `data`, which uses a function we define
  ourselves called `gridlaplacian` (described below).
- Then, we convert that data to a data frame and save it using Arrow. This step
  and the previous one are all executed in Julia.
- We import that data into R and prepare it for plotting.
- Finally, we plot the data using `{ggplot2}`.

### Helper functions

For this pipeline to work, we need to define some helper functions in both R and
Julia. Let's look at the Julia functions first:

```julia
# Define the precision matrix (inverse covariance matrix)
# for the Gaussian noise matrix. It approximately coincides
# with the Laplacian of the 2d grid or the graph representing
# the neighborhood relation of pixels in the picture,
# https://en.wikipedia.org/wiki/Laplacian_matrix

function gridlaplacian(m, n)
    S = sparse(0.0I, n*m, n*m)
    linear = LinearIndices((1:m, 1:n))
    for i in 1:m
        for j in 1:n
            for (i2, j2) in ((i + 1, j), (i, j + 1))
                if i2 <= m && j2 <= n
                    S[linear[i, j], linear[i2, j2]] -= 1
                    S[linear[i2, j2], linear[i, j]] -= 1
                    S[linear[i, j], linear[i, j]] += 1
                    S[linear[i2, j2], linear[i2, j2]] += 1
                end
            end
        end
    end
    return S
end

function arrow_write(df, path)
    Arrow.write(path, df)
end
```

And here are the R functions:

```{r, eval = FALSE}
prepare_data <- function(laplace){
  laplace_df |>
    mutate(
      x_id = row_number()
    ) |>
    tidyr::pivot_longer(-x_id, names_to = "y_id", values_to = "z") |>
    mutate(
      y_id = gsub("x", "", y_id),
      y_id = as.numeric(y_id)
    )
}

make_gg <- function(laplace_long_df){
  laplace_long_df |>
    ggplot(aes(x = x_id, y = y_id, z = z)) +
    stat_summary_hex(fun = function(x) mean(x), bins = 45)  +
    scale_fill_viridis_c(option = 12) +
    theme_void() +
    theme(legend.position = "none") +
    labs(subtitle = "hexagonal 2-d heatmap of laplacian matrix")
}

save_gg <- function(path, gg){
  ggsave("gg.png", gg)
}
```

These functions are used in the appropriate derivations using the
`user_functions` argument.

### Data transfer between R and Julia

When working with both R and Julia (or Python, for that matter), it's important
to understand how data is transferred between the two languages. In our example,
we're using Arrow to save the file in an interchangeable format. From Julia, we
save using `arrow_write()` a wrapper around `Arrow.write(path, df)`. Then, in R,
we can use `arrow::read_ipc_file()` to read the file back.

## Conclusion

`{rixpress}` makes it easy to combine the strengths of R and Julia in a single
pipeline. By using Nix to manage the environments and Arrow (or other formats)
for data transfer, you can create reproducible polyglot pipelines that leverage
the best of both languages.

For more examples and advanced usage, check out the [rixpress_demos repository](https://github.com/b-rodrigues/rixpress_demos).

---
title: "Encoding, Decoding, and Cross-Language Data Transfer"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Encoding, Decoding, and Cross-Language Data Transfer}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Introduction

Data pipelines in `{rixpress}` often require controlling how objects are stored
and restored, especially when dealing with:

1. Non-standard R objects (e.g., machine learning models, large tables).
2. Multiple file formats (CSV, `{qs}` compressed files, etc.).
3. Cross-language workflows mixing R and Python.

This vignette focuses on **encoding and decoding** in R, and on transferring
data between R and Python using `rxp_py2r()` and `rxp_r2py()`.

## Custom Encoding and Decoding in R

By default, `{rixpress}` uses `saveRDS()` and `readRDS()`. You can override this
to handle different formats or complex objects:

```r
library(rixpress)

# Encode output as CSV instead of RDS
d2 <- rxp_r(
  mtcars_head,
  my_head(mtcars_am, 100),
  user_functions = "my_head.R",
  nix_env = "default.nix",
  encoder = write.csv
)

# Encode as qs, decode input from CSV
d3 <- rxp_r(
  mtcars_tail,
  my_tail(mtcars_head),
  user_functions = "my_tail.R",
  nix_env = "default2.nix",
  encoder = qs::qsave,
  decoder = read.csv
)

# Decode multiple upstream objects with different decoders
d4 <- rxp_r(
  mtcars_mpg,
  full_join(mtcars_tail, mtcars_head),
  nix_env = "default2.nix",
  decoder = c(
    mtcars_tail = "qs::qread",
    mtcars_head = "read.csv"
  )
)
```

**Key points:**

- `encoder` controls how this step’s output is stored.
- `decoder` specifies how to read inputs from upstream derivations.
- You can assign different decoders per upstream object using a named vector.

As shown in the examples above, you can pass a function or a string
representation of the function to `encoder` and `decoder`.

By encoding the object in a cross-language format, it is possible to pass it
to another language. For example, read a csv file using Julia, encode it to Arrow
and read it back in R:

```r
library(rixpress)

list(
  rxp_jl_file(
    mtcars,
    # Assume here that mtcars.csv is separated by "|" instead of ","
    path = "data/mtcars.csv",
    read_function = "read_csv",
    user_functions = "functions.jl",
    encoder = "write_arrow"
    # read_csv and write_arrow are both
    # defined in the functions.jl script
    # and looks like this:

    #function write_arrow(df::DataFrame, filename::String)
    #    Arrow.write(filename, df)
    #end

    #function read_csv(path::String)
    #    df = CSV.read(path, DataFrame; delim="|")
    #return df
    #end

  ),

  rxp_r(
    mtcars2,
    select(mtcars, am, cyl, mpg),
    decoder = "read_feather"
  )
) |>
  rxp_populate()

```

You can find this example [here](https://github.com/b-rodrigues/rixpress_demos/tree/master/jl_input).
You can use the same approach to transfer data to Python (well, from and to any of the three
supported languages).

## Cross-Language Data Transfer: R ↔ Python

In the specific case of transferring objects (data, lists, vectors, arrays, etc.)
between R and Python, it also possible to use `{reticulate}`'s built-in
conversion by using `rxp_py2r()` and `rxp_r2py()`. These functions enable
seamless movement of objects between R and Python:

```r
library(rixpress)

# Python step producing pandas DataFrame
d1 <- rxp_py(
  name = mtcars_pl_am,
  expr = "mtcars_pl.filter(polars.col('am') == 1).to_pandas()"
)

# Transfer Python -> R
d2 <- rxp_py2r(
  name = mtcars_am,
  expr = mtcars_pl_am
)

# R step processing the data
d3 <- rxp_r(
  name = mtcars_head,
  expr = my_head(mtcars_am),
  user_functions = "functions.R"
)

# Transfer R -> Python
d3_1 <- rxp_r2py(
  name = mtcars_head_py,
  expr = mtcars_head
)
```

For this to work, you need to add `{reticulate}` to the pipeline's execution environment.

## Summary

- Use `encoder`/`decoder` for non-RDS objects (CSV, `{qs}`, Keras models) and to pass data
  to and from different languages.
- Explicitly set decoders per upstream object to avoid mismatches.
- Use `rxp_py2r()` and `rxp_r2py()` if you want to re-use `{reticulate}`'s bulit-in conversion
  (useful for more complex objects).

---
title: "Core Rixpress Functions and Usage"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Core rixpress Functions and Usage}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This vignette introduces the core functions required to build a `{rixpress}`
pipeline, but doesn't cover everything yet. It also
assumes that you've read `vignette("intro-concepts")`. In the next vignette
`vignette("tutorial")`, you'll learn how to set up a complete pipeline from
start to finish.

## Getting data into the pipeline

`{rixpress}` provides several functions to help you write derivations. These
functions typically start with the prefix `rxp_` and follow a similar structure.
The first step in any pipeline is usually to import data. To include data in a
`{rixpress}` pipeline, use `rxp_r_file()`:

```{r, eval = FALSE}
d0 <- rxp_r_file(
  name = mtcars,
  path = 'data/mtcars.csv',
  read_function = \(x) (read.csv(file = x, sep = "|"))
)
```

`rxp_r_file()`'s `read_function` argument requires an R function with a single
argument: the path to the file to be read. In this example, we assume the
columns in the `mtcars.csv` file are separated by the `|` symbol. We use an
anonymous function to set the correct separator and create a temporary function
with a single argument to read the file at `'data/mtcars.csv'`.

Important: This approach means that the `mtcars.csv` file **will be copied** to
the `Nix` store. This is essential to how `Nix` works.

Note that `rxp_r_file()` is quite flexible: it works with any function that
reads a file, regardless of the file type. The path to the file can also be a
URL. See the `vignette("importing-data")` for more details.

## Declaring build steps

Once the data is imported, we can start manipulating it. To generate a
derivation similar to the one described in `vignette("intro-concepts")`, but
using R and `{dplyr}` instead of `awk`, we would write:

```{r, eval = FALSE}
d1 <- rxp_r(
  name = filtered_mtcars,
  expr = dplyr::filter(mtcars, am == 1)
)
```

This syntax should be familiar to users of the `{targets}` package: similar to
the `tar_target()` function, you simply provide a name for the derivation and
the expression to generate it. That's all you need to write for `{rixpress}` to
generate all the required `Nix` code automatically.

To continue transforming the data, you only need to define a new derivation:

```{r, eval = FALSE}
d2 <- rxp_r(
  name = mtcars_mpg,
  expr = dplyr::select(filtered_mtcars, mpg)
)
```

Notice how the name of `d1` (`filtered_mtcars`) is used in `d2`: this is how
dependencies between derivations are defined.

## Generating the pipeline

Let's stop here and generate our pipeline. First, we need to define a list of
derivations:

```{r, eval = FALSE}
derivs <- list(d0, d1, d2)
```

and pass it to the `rxp_populate()` function:

```{r, eval = FALSE}
rxp_populate(derivs)
```

To make the code more concise, you can directly define the list and pass it to
`rxp_populate()` using the pipe operator `|>`:

```{r, eval = FALSE}
library(rixpress)

list(
  rxp_r_file(
    name = mtcars,
    path = 'data/mtcars.csv',
    read_function = \(x) (read.csv(file = x, sep = "|"))
  ),

  rxp_r(
    name = filtered_mtcars,
    expr = dplyr::filter(mtcars, am == 1)
  ),

  rxp_r(
    name = mtcars_mpg,
    expr = dplyr::select(filtered_mtcars, mpg)
  )
) |>
  rxp_populate()
```

Running `rxp_populate()` performs several actions:

- creates a folder called `_rixpress` in the project's root directory. This
  folder contains automatically generated files needed for the pipeline to build
  successfully.
- generates a file called `pipeline.nix`, which defines the entire pipeline in
  the `Nix` language.
- if `build = TRUE`, calls `rxp_make()` to build the pipeline.

However, if you try to run the code above, it will likely fail. This is because
a crucial piece is missing: the environment in which the pipeline must run!

## Defining a Reproducible Shell for Execution

Remember that the core purpose of using `Nix` is to ensure reproducibility by
forcing you to explicitly declare all dependencies. For our pipeline above, we
need to specify: Which version of R and which R packages should be used? The
pipeline uses `filter()` and `select()` from the `{dplyr}` package, so we must
declare these dependencies.

This is where the `{rix}` package comes in. `{rix}` allows you to define
reproducible development environments using simple R code. For example, we can
define an environment with R and `{dplyr}` like this:

```{r, eval = FALSE}
library(rix)

rix(
  date = "2025-04-11",
  r_pkgs = "dplyr",
  ide = "rstudio",
  project_path = ".",
  overwrite = TRUE
)
```

Running this code generates a `default.nix` file that can be built using `Nix`
by calling `nix-build`. This creates a development environment containing
RStudio, R, and `{dplyr}` as they existed on April 11, 2025. You can use this
environment for interactive data analysis just as you would with a standard
installation of RStudio, R, and `{dplyr}`. To learn more about `{rix}`, visit
[https://docs.ropensci.org/rix/](https://docs.ropensci.org/rix/).

The reproducible development environments generated by `{rix}` define all the
dependencies needed for your pipeline. To use this environment to build a
`{rixpress}` pipeline, you must also add `{rixpress}` to the list of packages in
the environment. Since `{rixpress}` is still under development, it must be
installed from GitHub. Here's how the complete environment setup script looks:

```{r, eval = FALSE}
library(rix)

# Define execution environment
rix(
  date = "2025-04-11",
  r_pkgs = "dplyr",
  git_pkgs = list(
    package_name = "rixpress",
    repo_url = "https://github.com/ropensci/rixpress",
    commit = "HEAD"
  ),
  ide = "rstudio",
  project_path = ".",
  overwrite = TRUE
)
```

In the next vignette, we'll learn how to use `{rix}` effectively to provide a
reproducible execution environment for our pipelines. For now, let's assume
that we've used the code above to generate our environment, which we can
also use for interactive data analysis.

We can go back to our pipeline to finalise it:

```{r, eval = FALSE}
library(rixpress)
# Define pipeline
list(
  rxp_r_file(
    name = mtcars,
    path = 'data/mtcars.csv',
    read_function = \(x) (read.csv(file = x, sep = "|"))
  ),

  rxp_r(
    name = filtered_mtcars,
    expr = dplyr::filter(mtcars, am == 1)
  ),

  rxp_r(
    name = mtcars_mpg,
    expr = dplyr::select(filtered_mtcars, mpg)
  )
) |>
  rxp_populate(project_path = ".")
```

I recommend always using two separate scripts:

- `gen-env.R`: Uses `{rix}` to define the execution environment
- `gen-pipeline.R`: Uses `{rixpress}` to define the reproducible analytical
  pipeline

You can quickly create these scripts using the `rxp_init()` function, which
generates both files with starter code to help you get started quickly.

## Optional steps before building the pipeline

### Graphical representation of the pipeline's DAG

It's often helpful to visualise your pipeline as a DAG (directed acyclic graph).
By default, the `build` argument of `rxp_populate()` is `FALSE`, so calling
this will not build the pipeline:

```r
rxp_populate(derivs)
```

This won't build the pipeline but will generate useful files, including a JSON
representation of the pipeline at `_rixpress/dag.json`. This process is quick
and allows you to visualise the graph using `rxp_visnetwork()`, which opens a
new tab in your web browser displaying the pipeline's DAG, generated using the
`{visNetwork}` package:

<video
  id="rixpress-video"
  controls
  autoplay
  loop
  muted
  style="max-width: 100%; height: auto; display: block;"
>
  <source
    src="https://github.com/ropensci/rixpress/raw/refs/heads/main/rxp_visnetwork.mp4"
    type="video/mp4"
  />
  <!-- Fallback message -->
  Your browser does not support the MP4 video format.
</video>

(This image shows the DAG of a more complex example pipeline.)

For static documents, you can use `rxp_ggdag()` which uses `{ggdag}` under
the hood:

<figure>
<img src="https://raw.githubusercontent.com/ropensci/rixpress/refs/heads/main/dag.png" alt="DAG" />
<figcaption aria-hidden="true">DAG</figcaption>
</figure>

You can also return the underlying `igraph` object to plot the DAG using other
tools:

```r
rxp_dag_for_ci()
```

which saves the `dag.dot` object in the project's `_rixpress/` folder.

After reviewing the DAG, you can build the pipeline by running `rxp_make()`
instead of modifying your original `rxp_populate()` call.

### Tracing the lineage of derivations

It is possible to also trace the lineage of individual derivations using
`rxp_trace()`. For example:

```r
rxp_trace("mtcars_mpg")
```

will return:

```r
==== Lineage for: mtcars_mpg ====
Dependencies (ancestors):
  - mtcars_head
    - mtcars_am*
      - mtcars*
  - mtcars_tail
    - mtcars_head*

Reverse dependencies (children):
          - page

Note: '*' marks transitive dependencies (depth >= 2).
```

This makes it quite easy to quickly double check whether derivations
were defined correctly. A `*` symbol next to a derivation's name indicates
it is a transitive dependency. Calling `rxp_trace()` without arguments
shows the whole graph:

```r
rxp_trace()
```

```r
==== Pipeline dependency tree (outputs → inputs) ====
- page
  - mtcars_head
    - mtcars_am*
      - mtcars*
  - mtcars_tail
    - mtcars_head*
  - mtcars_mpg
    - mtcars_head*
    - mtcars_tail*

Note: '*' marks transitive dependencies (depth >= 2).
```

We are now ready to actually build the artifacts. This is also quite
useful for debugging, as detailed in the `vignette("debugging")`.

## Building and inspecting outputs

When you run `gen-pipeline.R` (or execute its contents line-by-line), the
environment defined in `default.nix` is used (it's also possible to define
separate environments for different derivations, which we'll cover in a later
vignette).

By default, `rxp_populate()` doesn't build the pipeline, so to trigger the
build, you have to use `rxp_make()`:

```r
rxp_make()
```

You should see something like this:

```r
Build process started...

+ > mtcars building
+ > mtcars_am building
+ > mtcars_head building
+ > mtcars_tail building
+ > mtcars_mpg building
+ > page building
✓ mtcars built
✓ mtcars_am built
✓ mtcars_head built
✓ mtcars_mpg built
✓ mtcars_tail built
✓ page built
✓ pipeline completed [6 completed, 0 errored]
Build successful! Run `rxp_inspect()` for a summary.
Use `rxp_read("derivation_name")` to read objects or
`rxp_load("derivation_name")` to load them into the global environment.
```

Now you can follow these instructions:

1. Use `rxp_inspect()` to see where the outputs are located. This function is
   particularly useful if the pipeline fails, as it shows which derivations
   succeeded and which failed, and captures the error messages.
2. Use `rxp_read("mtcars_mpg")` to read the object into your current R session,
   or `rxp_load("mtcars_mpg")` to load it directly into your global environment.
3. Alternatively, use `rxp_copy("mtcars_mpg")` to create a folder called
   `pipeline-outputs` containing `mtcars_mpg` as an `.rds` file. If you call
   `rxp_copy()` without arguments, all pipeline outputs will be copied to this
   folder.

## No-op builds for individual derivations

You can disable building a specific derivation by setting its `noop_build`
parameter to `TRUE`. This creates a no-op build, a placeholder derivation that
performs no work:

```r
rxp_r(
  name = turtles,
  expr = occurrence(species, geometry = atlantic),
  noop_build = TRUE
)
```

Any derivations that depend on a no-op build will themselves also resolve to
no-op builds. This can be useful when prototyping or debugging a pipeline,
allowing you to skip expensive or unnecessary computations while keeping the
dependency graph intact. Further details are given in the vignette
`vignette("debugging")`.

## Caveats

There are some caveats that you need to be aware of when using `{rixpress}`.
Due to how `Nix` works, certain things are simply not possible:

- as mentioned in `vignette("intro-concepts")`, functions are executed in
  a hermetic sandbox. If they need access to an external resource, the build
  will fail. For example, if you use a function to get data from an
  API, you must first retrieve the data in a standard interactive R session, save it
  to disk, and then include it in the pipeline. The only exception
  to this is `rxp_r_file()`, which can download a file from a URL;
- if you functions need to access internal resources, use the `additional_files`
  argument of `rxp_r()` to include these resources into the build sandbox;
- all build artifacts will be saved in the `Nix` store, `/nix/store/`. If you
  are working with confidential data, make sure no one else can access the
  `/nix/store` path;
- if you have proprietary R packages, you will need to include them in the `Nix`
  shell. This is primarily a concern for `{rix}`, as it generates the execution
  environment. If you need help packaging your proprietary packages, please open
  an issue on the `{rix}` GitHub repository;
- multi-line expressions aren’t supported; write your derivations as single
  calls to pure functions.

## Conclusion

Now that you understand the basic, high-level concepts, let's move on to the
next vignette, `vignette("tutorial")`, where we'll learn how to set up a
pipeline from start to finish.

---
title: "Effective Debugging"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Effective Debugging}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Introduction

This vignette will guide you through the primary debugging workflow in `{rixpress}`, covering how to:

1.  Inspect the error messages from a failed build.
2.  Trace the dependency graph to find structural problems.
3.  Isolate specific parts of the pipeline for focused debugging.
4.  Access logs from previous builds to investigate regressions.

## The First Response to a Failed Build: `rxp_inspect()`

Imagine you have just run `rxp_make()` and are greeted with an error message in your console.

```r
Build process started...

+ > mtcars building
+ > mtcars_am building
+ > mtcars_head building
x mtcars_head errored
✓ mtcars built
✓ mtcars_am built
! pipeline completed [2 completed, 1 errored]
Build failed! Run `rxp_inspect()` for a summary.
```

The build has failed. Your immediate next step should always be to run
`rxp_inspect()`. By default, this function reads the most recent build log,
which in this case is the one from our failed run.

```{r, eval=FALSE}
rxp_inspect()
```

This will return a data frame summarizing the status of every derivation in the pipeline. Let's look at a hypothetical output:

```
       derivation build_success                                               path    output
1 all-derivations         FALSE /nix/store/j5...-all-derivations       mtcars_head
2       mtcars_am          TRUE /nix/store/a4...-mtcars_am                  mtcars_am
3     mtcars_head         FALSE                                      <NA>          <NA>
4          mtcars          TRUE /nix/store/b9...-mtcars                       mtcars
                                              error_message
1                                                      <NA>
2                                                      <NA>
3 Error: function 'headd' not found\nExecution halted\n
4                                                      <NA>
```

The two most important columns for debugging are `build_success` and `error_message`.

-   `build_success`: This `TRUE`/`FALSE` column immediately tells you which
    derivation failed. In our example, `mtcars_head` is the culprit.
-   `error_message`: This column contains the standard error output captured
    from the Nix build process. It provides the exact reason for the failure.
    Here, the message `"Error: function 'headd' not found"` points to a simple
    typo in our R code.

By pinpointing the specific derivation and providing the raw error message,
`rxp_inspect()` eliminates guesswork and directs you straight to the source of
the problem.

## Investigating Structural Issues with `rxp_trace()`

Sometimes, a pipeline fails not because of a typo in a single derivation, but
because of a logical error in how the derivations are connected. `rxp_trace()`
is the tool for diagnosing these structural issues. It reads the pipeline's
dependency graph (`dag.json`) and helps you answer questions like:

-   "What steps must run before this one?" (Dependencies)
-   "If I change this step, what other steps will be affected?" (Reverse Dependencies)

For instance, if `mtcars_mpg` is producing an unexpected result, you can trace its lineage:

```{r, eval=FALSE}
rxp_trace("mtcars_mpg")
```

This might return:

```
==== Lineage for: mtcars_mpg ====
Dependencies (ancestors):
  - filtered_mtcars
    - mtcars*

Reverse dependencies (children):
  - final_report

Note: '*' marks transitive dependencies (depth >= 2).
```

This output clearly shows that `mtcars_mpg` depends directly on
`filtered_mtcars` and indirectly (transitively) on `mtcars`. It also shows that
`final_report` depends on it. If you expected `mtcars_mpg` to depend on a
different intermediate object, this trace would immediately reveal the mistake
in your pipeline definition.

Calling `rxp_trace()` without any arguments will print the entire dependency
tree, which is useful for getting a high-level overview of your project's
structure.

You could instead plot the DAG using `rxp_ggdag()` for example, but if the project
is large, reading the DAG could be difficult. `rxp_trace()` should be more useful
in these cases.

## A Proactive Strategy: Isolating Derivations with `noop_build`

When debugging or prototyping, you often need to make frequent changes to an
early step in your pipeline. If a slow, computationally expensive derivation
depends on this changing step, your development cycle can become painfully slow.
Because Nix's caching is based on inputs, any change to an upstream step will
invalidate the cache for all downstream steps. Imagine a pipeline where you are
tuning a data preprocessing step, which is then followed by a lengthy model
training process:

```{r, eval=FALSE}
list(
  # We are actively changing the filter condition in this step
  rxp_r(
    name = preprocessed_data,
    expr = filter(raw_data, year > 2020)
  ),
  # This step takes hours to run
  rxp_r(
    name = expensive_model,
    expr = run_long_simulation(preprocessed_data)
  ),
  rxp_rmd(
    name = final_report,
    rmd_file = "report.Rmd" # Depends on expensive_model
  )
)

```

In this scenario, every time you adjust the filter() condition in
preprocessed_data, Nix correctly invalidates the cache for expensive_model. This
means the hours-long simulation will be re-triggered with every small change,
making it impossible to iterate quickly on the preprocessing logic. This is the
perfect use case for noop_build = TRUE. By applying it to the expensive
downstream step, you temporarily break the dependency chain:

```r
list(
  # We can now change this step as much as we want
  rxp_r(
    name = preprocessed_data,
    expr = filter(raw_data, year > 2020)
  ),
  # This and all downstream steps will be skipped
  rxp_r(
    name = expensive_model,
    expr = run_long_simulation(preprocessed_data),
    noop_build = TRUE
  ),
  rxp_rmd(
    name = final_report,
    rmd_file = "report.Rmd" # Also becomes a no-op
  )
)
```

Now, when you run `rxp_make()`, `preprocessed_data` will build as normal.
However, `expensive_model` will resolve to a no-op build, and because `final_report`
depends on it, it will also become a no-op. This allows you to rapidly iterate
on and validate the `preprocessed_data` logic in isolation, without waiting for
the simulation to run. Once you are satisfied with the preprocessing, simply
remove `noop_build = TRUE` to re-enable the full pipeline and run the expensive
model training with your finalized data.

## Historical Debugging: Going Back in Time

When iterating quickly, it might be useful to compare results to the ones
obtained from previous runs. It is possible to check results from previous
runs using the logs.

First, use `rxp_list_logs()` to see the build history:

```{r, eval=FALSE}
rxp_list_logs()
```

```
                                                        filename   modification_time size_kb
1 build_log_20250815_113000_a1b2c3d4e5f6g7h8i9j0k1l2m3n4o5p6.rds 2025-08-15 11:30:00    0.51
2 build_log_20250814_170000_z9y8x7w6v5u4t3s2r1q0p9o8n7m6l5k4.rds 2025-08-14 17:00:00    0.50
```

You can see a successful build from yesterday (`20250814`). To find out the
differences with today's results, you can inspect that specific log by providing
a unique part of its filename to `which_log`:

```{r, eval=FALSE}
# Inspect yesterday's successful build log
rxp_inspect(which_log = "20250814")
```

This allows you to compare yesterday's build summary with today's one.
Furthermore, you can use `rxp_read()` with `which_log` to load the *actual
artifact* from the previous run, which is invaluable for comparing data or model
outputs across different versions of your pipeline.

```{r, eval=FALSE}
# Load the output of `mtcars_head` from yesterday's build
old_head <- rxp_read("mtcars_head", which_log = "20250814")
```

## Conclusion

Debugging in `{rixpress}` is a systematic process supported by a powerful set of
tools. By following this workflow, you can efficiently resolve issues in your
pipelines:

- For runtime errors, start with `rxp_inspect()` to find the failed derivation and its error message.
- For logical or structural errors, use `rxp_trace()` to understand the dependencies.
- To speed up iteration, use `noop_build = TRUE` to isolate the part of the pipeline you are working on.
- For regressions, use `rxp_list_logs()` and the `which_log` argument to travel back in time and compare results.

---
title: "Organising Large Projects with Sub-Pipelines"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Organising Large Projects with Sub-Pipelines}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This vignette introduces `rxp_pipeline()`, a function for organising large
projects into logical sub-pipelines. This feature is particularly useful when
working on complex projects with multiple phases (e.g., ETL, Modelling, Reporting)
or when collaborating in teams where different members work on different parts
of the pipeline.

## Large Pipelines Become Unwieldy

As pipelines grow, a single `gen-pipeline.R` file can become difficult to
manage. Consider a data science project with:
- Data extraction and cleaning (ETL)
- Feature engineering
- Model training
- Model evaluation
- Report generation

Putting all derivations in one file makes it hard to:

- Navigate the code
- Understand which derivations belong to which phase
- Collaborate across team members
- Reuse pipeline components in other projects

To solve this issue, you can define your project using sub-pipelines and join
them into a master pipeline using `rxp_pipeline()`.

This allows you to:

1. **Organise** derivations into named groups
2. **Colour-code** groups for visual distinction in DAG visualisations
3. **Modularise** your code across multiple R scripts

### Basic Usage

A project with sub-pipelines would look something like this:


```
my-project/
├── default.nix           # Nix environment (generated by rix)
├── gen-env.R             # Script to generate default.nix
├── gen-pipeline.R        # MASTER SCRIPT: combines all sub-pipelines
└── pipelines/
    ├── 01_data_prep.R    # Data preparation sub-pipeline
    ├── 02_analysis.R     # Analysis sub-pipeline
    └── 03_reporting.R    # Reporting sub-pipeline
```

Each sub-pipeline file returns a list of derivations:

```{r sub-pipeline-1, eval = FALSE}
# Data Preparation Sub-Pipeline
# pipelines/01_data_prep.R
library(rixpress)

list(
  rxp_r(name = raw_mtcars, expr = mtcars),
  rxp_r(name = clean_mtcars, expr = dplyr::filter(raw_mtcars, am == 1)),
  rxp_r(name = selected_mtcars, expr = dplyr::select(clean_mtcars, mpg, cyl, hp, wt))
)
```

The `rxp_pipeline()` function takes:

- **name**: A descriptive name for this group of derivations
- **path**: Either a **file path** to an R script returning a list of derivations (recommended), or a list of derivation objects.
- **color**: Optional CSS color name or hex code for DAG visualisation

The second sub-pipeline:

```{r sub-pipeline-2, eval = FALSE}
# Analysis Sub-Pipeline
# pipelines/02_analysis.R
library(rixpress)

list(
  rxp_r(name = summary_stats, expr = summary(selected_mtcars)),
  rxp_r(name = mpg_model, expr = lm(mpg ~ hp + wt, data = selected_mtcars)),
  rxp_r(name = model_coefs, expr = coef(mpg_model))
)
```

The master script becomes very clean, as `rxp_pipeline` handles sourcing the files:

```{r master-script, eval = FALSE}
# gen-pipeline.R
library(rixpress)

# Create named pipelines with colours by pointing to the files
pipe_data_prep <- rxp_pipeline(
  name = "Data Preparation",
  path = "pipelines/01_data_prep.R",
  color = "#E69F00"
)

pipe_analysis <- rxp_pipeline(
  name = "Statistical Analysis",
  path = "pipelines/02_analysis.R",
  color = "#56B4E9"
)

# Build combined pipeline
rxp_populate(list(pipe_data_prep, pipe_analysis), project_path = ".", build = TRUE)
```

## Visualising Sub-Pipelines

When sub-pipelines are defined, visualisation tools use pipeline colours:

1. **Interactive Network** (`rxp_visnetwork()`) and **Static DAG** (`rxp_ggdag()`) both use a dual-encoding approach:
   - **Node fill (interior)**: Derivation type colour (R = blue, Python = yellow, etc.)
   - **Node border (thick stroke)**: Pipeline group colour
   This allows you to see both what *type* of computation each node is and which *pipeline* it belongs to.

<figure>
<img src="https://raw.githubusercontent.com/ropensci/rixpress/refs/heads/main/vignettes/subpipelines.png" alt="Subpipelines are coloured." />
<figcaption aria-hidden="true">
Subpipelines are coloured.
</figcaption>
</figure>

2. **Trace**: `rxp_trace()` output in the console is coloured by pipeline (using the `cli` package).

<figure>
<img src="https://raw.githubusercontent.com/ropensci/rixpress/refs/heads/main/vignettes/subpipelines_trace.png" alt="If your terminal supports it, derivation names are coloured according to the chosen sub-pipeline colour." />
<figcaption aria-hidden="true">
If your terminal supports it, derivation names are coloured according to the chosen sub-pipeline colour.
</figcaption>
</figure>

### Switching Between Colour Modes

```{r color-modes, eval = FALSE}
# Dual encoding: fill = type, border = pipeline (default when pipelines are defined)
rxp_ggdag(color_by = "pipeline")

# Colour entirely by derivation type (rxp_r, rxp_py, etc.) - original behaviour
rxp_ggdag(color_by = "type")
```

## How It Works Internally

When you call `rxp_populate()` with `rxp_pipeline` objects:

1. **Flattening**: Pipelines are flattened to a single list of derivations
2. **Metadata Preservation**: Each derivation retains `pipeline_group` and `pipeline_color`
3. **DAG Generation**: `dag.json` includes pipeline metadata
4. **Visualisation**: `rxp_visnetwork()` and `rxp_ggdag()` read this metadata

## Best Practices

1. **Use descriptive pipeline names**: "Data Preparation" is better than "ETL"
2. **Choose contrasting colours**: Use [ColorBrewer](https://colorbrewer2.org/) palettes
3. **Keep sub-pipelines focused**: One logical phase per sub-pipeline
4. **Order your files**: Use numeric prefixes (01_, 02_, etc.)

## Conclusion

`rxp_pipeline()` provides a simple yet powerful way to organise complex
pipelines. By grouping derivations into logical units, you can:

- Keep your code organised and maintainable
- Enable team collaboration on different parts of the pipeline
- Visualise the structure of your workflow with meaningful colours
- Reuse sub-pipelines across projects

For a working example, see the `subpipelines` demo in the
[rixpress_demos](https://github.com/b-rodrigues/rixpress_demos) repository.

---
title: "Importing Data Files"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Importing Data Files}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

A crucial first step in any data analysis pipeline is importing data. The
`{rixpress}` package provides a flexible set of functions, `rxp_r_file`,
`rxp_py_file`, and `rxp_jl_file`, to handle various data import scenarios in a
reproducible way. This vignette will guide you through the common use cases.

For more examples, check out the [rixpress_demos repository](https://github.com/b-rodrigues/rixpress_demos/).

## Importing a single local file

The most straightforward case is reading a single data file from your local
project directory. You need to provide a `name` for the resulting R object, the
`path` to the file, and a `read_function` to process it.

```{r, eval = FALSE}
library(rixpress)

list(
  rxp_r_file(
    name = mtcars,
    path = 'data/mtcars.csv',
    read_function = \(x) (read.csv(file = x, sep = "|"))
  ),
...
```

In this example, `rxp_r_file` creates a derivation that:

1. Copies `data/mtcars.csv` into a sandboxed build environment.
2. Executes the provided anonymous function, `\(x) (read.csv(file = x, sep =
   "|"))`, where `x` is the path to the copied file inside the sandbox.
3. Saves the resulting data frame as an object named `mtcars` for subsequent steps in the pipeline.

## Importing a single file from the internet

You can also directly import a file from a URL. Simply provide the URL as the
`path`. `{rixpress}` handles the download and ensures reproducibility by caching
the file using its cryptographic hash.

```{r, eval = FALSE}
library(rixpress)

list(
  rxp_r_file(
    name = mtcars,
    path = 'https://raw.githubusercontent.com/b-rodrigues/rixpress_demos/refs/heads/master/basic_r/data/mtcars.csv',
    read_function = \(x) (read.csv(file = x, sep = "|"))
  ),
...
```

Behind the scenes, `{rixpress}` uses Nix to fetch the file, ensuring that the
exact same version of the file is used every time the pipeline is run. This is
the only time the build sandbox can access a remote file: it's because the file
actually gets downloaded by Nix ahead of time. If you need to access data in
real-time from an API, then you'll need to download the data yourself
outside of `{rixpress}` pipeline, and then import it in the pipeline using
`rxp_r_file()`.

## Importing many files from a directory

Often, you need to import and combine multiple files from a single directory. To
do this, set the `path` argument to the directory's path. Your `read_function`
will then receive the path to this directory inside the build environment and
must contain the logic to handle all the files within.

Here is an example in R that reads all files in the `data` directory:

```{r, eval = FALSE}
library(rixpress)

list(
  rxp_r_file(
    name = mtcars_r,
    path = 'data',
    read_function = \(x) {
      (readr::read_delim(list.files(x, full.names = TRUE), delim = '|'))
    }
  )
) |>
  rxp_populate(project_path = ".")
```

And here's a similar example using Python, which calls a user-defined function
`read_many_csvs` from an external script:

```{r, eval = FALSE}
library(rixpress)

list(
  rxp_py_file(
    name = mtcars_py,
    path = 'data',
    read_function = "read_many_csvs",
    user_functions = "functions.py"
  )
) |>
  rxp_populate(project_path = ".")
```

Here is what the Python function looks like:

```py
import polars
from pathlib import Path

def read_many_csvs(dir_path):
    folder = Path(dir_path)
    csv_files = folder.glob("*.csv")
    return polars.concat([polars.read_csv(f) for f in csv_files])
```

In both cases, the entire `data` directory is copied into the build sandbox, and
the `read_function` is responsible for listing the files and reading them.

## Importing files with dependencies (e.g., Shapefiles)

Some file formats, like the ESRI Shapefile, consist of multiple "sidecar" files
(e.g., `.shp`, `.shx`, `.dbf`) that must be present together for the data to be
read correctly. Even though you might only point the read function to the `.shp`
file, the other component files need to be in the same directory.

`{rixpress}` handles this by allowing you to specify a directory as the `path`.
This ensures all necessary files are copied into the build environment. However,
you must then provide the *full path to the main file inside the build
environment* within your `read_function`.

In a `{rixpress}` pipeline, local files and directories specified in `path` are
copied into a sub-directory called `input_folder`. Therefore, the path to your
data inside the Nix sandbox will be `input_folder/YOUR_PATH`.

The following example shows how to read a shapefile using Python and
`geopandas`:

```{r, eval = FALSE}
library(rixpress)

list(
  rxp_py_file(
    name = gdf,
    # We provide the directory 'data' to ensure all shapefile components are copied.
    path = 'data',
    # The read_function must use the hardcoded path within the build environment.
    read_function = "lambda x: geopandas.read_file('input_folder/data/oceans.shp', driver='ESRI Shapefile')"
  ),

  rxp_py(
    name = sa,
    expr = "gdf.loc[gdf['Oceans'] == 'South Atlantic Ocean']['geometry'].loc[0]"
  )
) |>
  rxp_populate(project_path = ".")
```

Here's what happens:

1. The `path = 'data'` argument tells `{rixpress}` to copy the entire `data`
   directory into the sandbox.
2. Inside the sandbox, the shapefile is located at
   `input_folder/data/oceans.shp`.
3. The `read_function` is a lambda function that explicitly calls
   `geopandas.read_file` with this hardcoded path, allowing it to find the
   `.shp` file and its necessary sidecar files.

A perhaps cleaner alternative is to write a function that takes the path to the
data folder as an input, and then have this function look in that folder for the
shapefile, and pass its path to `geopandas.read_file`. For example

```py
def read_shp(path_folder):
    # Look for files ending with .shp in the given folder
    candidates = glob.glob(os.path.join(path_folder, "*.shp"))
    if not candidates:
        raise FileNotFoundError(f"No .shp file found in {path_folder}")

    shapefile = candidates[0]
    return gpd.read_file(shapefile, driver="ESRI Shapefile")
```

We can then rewrite the derivation like so:

```r
rxp_py_file(
    name = gdf,
    path = 'data',
    read_function = "read_shp",
    user_functions = "functions.py"
  ),
```

(assuming our function is defined in a script called `functions.py`).

Because our Python function also uses `glob` and `os`, we need to import these
functions using `add_import()`. We can add this just after calling
`rxp_populate()`:


```r
  rxp_populate(
    project_path = ".",
    py_imports = c(geopandas = "import geopandas as gpd")
  )

# This is needed for the function defined in functions.py
add_import("import os", "default.nix")
add_import("import glob", "default.nix")
```

## Conclusion

The `rxp_*_file` functions in `{rixpress}` offer a powerful and consistent
interface for ingesting data into your reproducible pipelines, whether your data
lives locally, on the web, as a single file, or as a collection of files. By
understanding how to specify the `path` and tailor the `read_function`, you can
handle a wide variety of data import tasks.

---
title: "Using Chronicler for Error-Tolerant Pipelines"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Using Chronicler for Error-Tolerant Pipelines}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

This vignette explains how to use the `{chronicler}` package with `{rixpress}`
to build error-tolerant pipelines, and how to detect silent failures.

## The Problem: Silent Failures

When building data pipelines, errors typically cause immediate failure. In
`{rixpress}`, if a derivation throws an error, Nix stops the build and reports
what went wrong. This is usually desirable—you want to know when something
breaks.

But sometimes you want pipelines that *continue* even when parts fail. Consider:

- Processing many files where a few might be malformed
- Exploratory analysis where you want to see what works and what doesn't

The `{chronicler}` package provides exactly this: functions that capture errors
and warnings instead of failing, returning a structured result that you can
inspect later.

## What is {chronicler}?
  
`{chronicler}` uses the `{maybe}` package to implement enhanced output that logs
errors and warnings instead of failing. When you "record"
(decorate) a function using `record()`, it returns a `chronicle` object containing:

- **`$value`**: Either `Just(result)` for success, or `Nothing` for failure
- **`$log_df`**: A data frame recording all operations, timings, and messages

Here's a simple example:

```{r, eval = FALSE}
library(chronicler)

# Create a recorded version of sqrt
r_sqrt <- record(sqrt)

# Success case
result <- r_sqrt(4)
result$value
#> Just
#> [1] 2

# Failure case: sqrt(-1) produces a warning "NaNs produced"
# With default strict=2, warnings are treated as failures
result <- r_sqrt(-1)
result$value
#> Nothing
```

## The Challenge with Nix Builds

Here's the catch: when you use chronicler functions in a `{rixpress}` pipeline,
**Nix builds never fail**. Even when a computation produces `Nothing`, it's
still a valid R object that gets serialized successfully.

Consider this pipeline:

```{r, eval = FALSE}
library(rixpress)

list(
  rxp_r(
    name = result,
    expr = r_sqrt(-1),  # This produces a `Nothing` value, not an error!
    user_functions = "functions.R"
  )
) |> rxp_populate(build = FALSE)
```

When you run `rxp_make()`, the build succeeds! But `result` contains `Nothing`,
meaning the computation actually failed. Without checking, you might think
your pipeline worked perfectly.

## The Solution: Automatic Chronicle Checking

When `{chronicler}` is available, `{rixpress}` automatically checks your
pipeline outputs for `Nothing` values after every successful build.
The `rxp_check_chronicles()` function is called automatically by `rxp_make()`:

```{r, eval = FALSE}
# Build the pipeline - chronicle status is checked automatically!
rxp_make()
```

After the build, you'll see the status for each chronicle object:

```
Chronicle status:
✓ filtered_mtcars (chronicle: OK)
✓ mtcars_mpg (chronicle: OK)
✓ mean_mpg (chronicle: OK)
✗ sqrt_of_negative (chronicle: NOTHING)
    Failed: sqrt
    Message: NaNs produced
✗ downstream_of_nothing (chronicle: NOTHING)
    Failed: (anonymous)
    Message: Pipeline failed upstream

Summary: 3 success, 0 with warnings, 2 nothing
Warning: 2 derivation(s) contain Nothing values!
```

## Three-State Status System

Chronicles can be in one of three states:

| Symbol | State | Meaning |
|--------|-------|---------|
| ✓ | **Success** | `Just` value, no warnings or errors |
| ⚠ | **Warning** | `Just` value, but warnings were captured |
| ✗ | **Nothing** | Failed computation, errors captured |

## Complete Example

Here's a complete example demonstrating the pattern. First, create `functions.R`
with your recorded functions:

```{r, eval = FALSE}
# functions.R
library(chronicler)

r_filter <- record(dplyr::filter)
r_pull <- record(dplyr::pull)
r_sqrt <- record(sqrt)
r_mean <- record(mean)
```

Then create your pipeline in `gen-pipeline.R`:

```{r, eval = FALSE}
library(rixpress)

list(
  # Read data (not a chronicle)
  rxp_r_file(
    name = mtcars,
    path = "data/mtcars.csv",
    read_function = \(x) read.csv(file = x, sep = "|")
  ),

  # Filter using chronicler - SUCCESS
  rxp_r(
    name = filtered_mtcars,
    expr = mtcars |> r_filter(am == 1),
    user_functions = "functions.R"
  ),

  # Pull column - SUCCESS
  rxp_r(
    name = mtcars_mpg,
    expr = filtered_mtcars |> bind_record(r_pull, mpg),
    user_functions = "functions.R"
  ),

  # Compute mean - SUCCESS
  rxp_r(
    name = mean_mpg,
    expr = mtcars_mpg |> bind_record(r_mean),
    user_functions = "functions.R"
  ),

  # Intentional failure: sqrt(-1) - NOTHING
  rxp_r(
    name = sqrt_of_negative,
    expr = r_sqrt(-1),
    user_functions = "functions.R"
  ),

  # Downstream of Nothing - also NOTHING
  rxp_r(
    name = downstream_of_nothing,
    expr = sqrt_of_negative |> bind_record(r_mean),
    user_functions = "functions.R"
  )
) |>
  rxp_populate(build = FALSE)
```

Build the pipeline (chronicle status is checked automatically):

```{r, eval = FALSE}
rxp_make()

# You can also manually check chronicles at any time:
# rxp_check_chronicles()
```

## Automatic Warnings

When you read or load a chronicle with `Nothing` using `rxp_read()` or
`rxp_load()`, you'll automatically get a warning:

```{r, eval = FALSE}
rxp_read("sqrt_of_negative")
#> Warning message:
#> Derivation 'sqrt_of_negative' contains a chronicle with Nothing value!
#>   Use chronicler::read_log() on this object for details.
```

This helps catch silent failures even during interactive exploration.

## Best Practices

1. **Chronicle status is checked automatically** after every successful build
   when `{chronicler}` is available. You can also run `rxp_check_chronicles()`
   manually at any time.

2. **Use `chronicler::read_log()` for debugging**. When a chronicle contains
   `Nothing`, the log shows exactly where and why it failed:
   
   ```{r, eval = FALSE}
   result <- rxp_read("sqrt_of_negative")
   read_log(result)
   ```

3. **Consider the `strict` parameter**. By default (`strict = 2`), chronicler
   treats warnings as failures. Use `strict = 1` to allow warnings through
   while still capturing them, or `strict = 0` to ignore warnings entirely.

4. **Chain operations with `bind_record()`**. This properly propagates
   `Nothing` values through the pipeline—if upstream fails, downstream
   automatically becomes `Nothing` too.

## When to Use Chronicler

Use chronicler with rixpress when you want:

- **Resilient pipelines** that continue even when parts fail
- **Detailed logging** of all operations and their outcomes
- **Explicit failure handling** rather than silent errors

Don't use chronicler when:

- You want builds to fail fast on any error (the default Nix behaviour)
- You need maximum performance (chronicler adds overhead)
- Your pipeline is simple and errors are rare

## Further Reading

- [chronicler package documentation](https://b-rodrigues.github.io/chronicler/)
- Example project: `chronicler_example` in the
  [rixpress_demos](https://github.com/b-rodrigues/rixpress_demos) repository

---
title: "Setting Up a Pipeline from Scratch"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Setting Up a Pipeline From Scratch}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

This vignette demonstrates how to build a `{rixpress}` pipeline that only uses R
derivations. For multilingual pipelines, read the next vignette
`vignette("polyglot")`.

If you are more of a visual learner, you can also watch
[this video](https://www.youtube.com/watch?v=IXKd5ySzzSU) which
will walk you through the contents of this vignette.

# Installing the required software

If you're using `{rixpress}`, you're likely already familiar with its sister
package `{rix}`, which helps you set up reproducible development environments
for R or Python using `Nix`. As such, you likely already have `Nix` installed on your
system. If not, read the following [`{rix}` vignette to learn
how to set up `Nix` on Linux or
Windows](https://docs.ropensci.org/rix/articles/installing-r-packages.html),
or [this one if you're on macOS](https://docs.ropensci.org/rix/articles/setting-up-macos.html).

Once `Nix` is installed (as well as `cachix`, and the `rstats-on-nix` cache
configured), you're ready to go.

# Bootstrapping a project

In this vignette, I'm going to assume the following:

- you're already familiar with `Nix` thanks to `{rix}`;
- you only use `Nix` to manage R environments for data science projects (in other
  words, you don't have a system-installed R);
- you have `Nix` installed on your system as well as `cachix` and configured the
  `rstats-on-nix` cache;
- you've read through the previous two vignettes.

If you're already familiar with `{rix}`, you likely know that you can use the
following command in your project's root directory to get access to a temporary
`Nix` shell that provides R, `{rix}`, and `{rixpress}`:

```bash
nix-shell --expr "$(curl -sl https://raw.githubusercontent.com/ropensci/rix/main/inst/extdata/default.nix)"
```

This will start a `Nix` session. Type `R` to start an R session, and then type:

```{r, eval = FALSE}
library(rixpress)

rxp_init()
```

This will write two files to the root of your project, `gen-env.R` and
`gen-pipeline.R`. `gen-env.R` will look like this:

```{r, eval = FALSE}
# This script defines the default environment the pipeline runs in.
# Add the required packages to execute the code necessary for each derivation.
# If you want to create visual representations of the pipeline, consider adding
# `{visNetwork}` and `{ggdag}` to the list of R packages.
library(rix)

# Define execution environment
rix(
  date = NULL,
  r_pkgs = NULL,
  py_conf = NULL,
  git_pkgs = list(
    "package_name" = "rixpress",
    "repo_url" = "https://github.com/ropensci/rixpress",
    "commit" = "HEAD"
  ),
  ide = "none",
  project_path = "."
)
```

To execute a `{rixpress}` pipeline, the environment needs `{rixpress}` to be
available, but you likely require other packages. Let's add `{dplyr}` and
`{igraph}` (which is needed to view the graphical representation of the
pipeline's DAG). Open `gen-env.R` in any text editor and change it to:

```{r, eval = FALSE}
library(rix)

# Define execution environment
rix(
  date = "2025-04-11",
  r_pkgs = c("dplyr", "igraph"),
  git_pkgs = list(
    package_name = "rixpress",
    repo_url = "https://github.com/ropensci/rixpress",
    commit = "HEAD"
  ),
  ide = "rstudio",
  project_path = ".",
  overwrite = TRUE
)
```

I've added a date, which means that the versions of R and R packages will be set
to the versions available on this date from CRAN, and I've added `{dplyr}` and
`{igraph}`. I've also added RStudio as the IDE. I can now go back to my
temporary `Nix` shell and run `Rscript gen-env.R` (or `source("gen-env.R")`
inside an R session). This will create a `default.nix` file in my project's
folder: this is a `Nix` file that defines my environment. I can now leave my
temporary shell by first quitting R (if applicable), then typing CTRL-D or
`exit`. Then I can build my development environment using `nix-build`. Let's now
enter our development shell by typing `nix-shell` and launch RStudio (or
whatever editor you want to use) and open `gen-pipeline.R`:


```{r, eval = FALSE}
library(rixpress)
library(igraph)

list(
  rxp_r_file(
    name = NULL,
    path = NULL,
    read_function = \(x) read.csv(file = x, sep = ",")
  ),
  rxp_r(
    name = NULL,
    expr = NULL
  )
) |> rxp_populate(build = TRUE)
```

This template starts with `rxp_r_file()` to load data, and `rxp_r()` to
define a step. These two derivations are put into a list and then passed to
`rxp_populate()`. Let's start by importing data and filtering it using `{dplyr}`:

```{r, eval = FALSE}
library(rixpress)
library(igraph)

list(
  rxp_r_file(
    name = mtcars,
    path = 'data/mtcars.csv',
    read_function = \(x) (read.csv(file = x, sep = "|"))
  ),

  rxp_r(
    name = filtered_mtcars,
    expr = filter(mtcars, am == 1)
  )
) |> rxp_populate(build = TRUE)
```

Select the pipeline and run it (I highly recommend you use keyboard shortcuts
to quickly run the above code when working interactively). If everything
went well, you will see the following in your R console:

```r
Build process started...

+ > mtcars building
+ > filtered_mtcars building
✓ filtered_mtcars built
✓ mtcars built
✓ pipeline completed [2 completed, 0 errored]
Build successful! Run `rxp_inspect()` for a summary.
Read individual derivations using `rxp_read()` or
load them into the global environment using `rxp_load()`.
```

Let’s inspect the pipeline outputs using `rxp_inspect()`:

```r
       derivation build_success
1 all-derivations          TRUE
2 filtered_mtcars          TRUE
3          mtcars          TRUE
                                                         path       output
1 /nix/store/xc67xyqw8ziggwdk7hhiqsgh7hxfv4k1-all-derivations filtered....
2 /nix/store/njv7zpgiq9p843dixkcg1dpryr5rn6h4-filtered_mtcars filtered....
3          /nix/store/gqf4l5jpbjdv1by0vw6vphhlrfa0600a-mtcars       mtcars
  error_message
1          <NA>
2          <NA>
3          <NA>
```

Three things were built: `mtcars`, which is simply the source data set;
`filtered_mtcars`, which is the filtered data; and an object you didn't
define called `all-derivations`. This last object is mostly for internal
`{rixpress}` use, and you can safely ignore it.

Let's now use `rxp_read()` to take a look at `filtered_mtcars`:

```{r, eval = FALSE}
rxp_read("filtered_mtcars")
```

```r
                mpg cyl  disp  hp drat    wt  qsec vs am gear carb
Mazda RX4      21.0   6 160.0 110 3.90 2.620 16.46  0  1    4    4
Mazda RX4 Wag  21.0   6 160.0 110 3.90 2.875 17.02  0  1    4    4
Datsun 710     22.8   4 108.0  93 3.85 2.320 18.61  1  1    4    1
Fiat 128       32.4   4  78.7  66 4.08 2.200 19.47  1  1    4    1
Honda Civic    30.4   4  75.7  52 4.93 1.615 18.52  1  1    4    2
Toyota Corolla 33.9   4  71.1  65 4.22 1.835 19.90  1  1    4    1
Fiat X1-9      27.3   4  79.0  66 4.08 1.935 18.90  1  1    4    1
Porsche 914-2  26.0   4 120.3  91 4.43 2.140 16.70  0  1    5    2
Lotus Europa   30.4   4  95.1 113 3.77 1.513 16.90  1  1    5    2
Ford Pantera L 15.8   8 351.0 264 4.22 3.170 14.50  0  1    5    4
Ferrari Dino   19.7   6 145.0 175 3.62 2.770 15.50  0  1    5    6
Maserati Bora  15.0   8 301.0 335 3.54 3.570 14.60  0  1    5    8
Volvo 142E     21.4   4 121.0 109 4.11 2.780 18.60  1  1    4    2
```

You can save this object like so:

```{r, eval = FALSE}
filtered_mtcars <- rxp_read("filtered_mtcars")
```

or you can simply use `rxp_load("filtered_mtcars")` which will achieve the same
result.

You can now manipulate this object interactively, as usual. You might still not
quite know what the next step should be (maybe there are missing values you need
to handle somehow, or another such cleaning step), but once you know the next
steps, you can write the next derivation:

```{r, eval = FALSE}
library(rixpress)
library(igraph)

list(
  rxp_r_file(
    name = mtcars,
    path = 'data/mtcars.csv',
    read_function = \(x) (read.csv(file = x, sep = "|"))
  ),

  rxp_r(
    name = filtered_mtcars,
    expr = filter(mtcars, am == 1)
  ),

  rxp_r(
    name = mtcars_mpg,
    expr = select(filtered_mtcars, mpg)
  )
) |>
  rxp_populate(build = TRUE)
```

Rebuild the pipeline. The first two steps are skipped because they're already
available in the `Nix` store. You can now start the *inspect -> try stuff ->
define derivation -> build pipeline* loop again for the next derivation.

You now know the basics of using `{rixpress}` to build a simple pipeline.

The next vignette `vignette("polyglot")` explains how to build a pipeline
with both R and Python derivations, as well as compiling a Quarto document.

If you don't need to work with multiple languages, you can continue with
the vignette `vignette("debugging")`.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxp_read_load.R
\name{rxp_load}
\alias{rxp_load}
\title{Load Output of a Derivation}
\usage{
rxp_load(derivation_name, which_log = NULL, project_path = ".")
}
\arguments{
\item{derivation_name}{Character, the name of the derivation.}

\item{which_log}{Character, defaults to NULL. If NULL the most recent
build log is used. If a string is provided, it's used as a
regular expression to match against available log files.}

\item{project_path}{Character, defaults to ".".
Path to the root directory of the project.}
}
\value{
Nothing, this function has the side effect of loading objects into
the parent frame.
}
\description{
Loads the output of derivations in the parent frame of the
current session, returns a path if reading directly is not possible.
}
\details{
When \code{derivation_name} points to a single R object, it gets loaded
in the current session using \code{assign(..., envir = parent.frame())}, which
corresponds to the global environment in a regular interactive session. If
you're trying to load a Python object and \code{{reticulate}} is available,
\code{reticulate::py_load_object()} is used and then the object gets loaded into
the global environment. In case the derivation is pointing to several
outputs (which can happen when building a Quarto document for example) or
loading fails, the path to the object is returned instead.
}
\examples{
\dontrun{
  # Load an R object
  rxp_load("mtcars")

  # Load a Python object
  rxp_load("my_python_model")

  # Load from a specific build log
  rxp_load("mtcars", which_log = "2025-05-10")
}
}
\seealso{
Other utilities: 
\code{\link{print.rxp_derivation}()},
\code{\link{rxp_check_chronicles}()},
\code{\link{rxp_copy}()},
\code{\link{rxp_gc}()},
\code{\link{rxp_init}()},
\code{\link{rxp_inspect}()},
\code{\link{rxp_list_logs}()},
\code{\link{rxp_read}()},
\code{\link{rxp_trace}()}
}
\concept{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse_pkgs.R
\name{add_import}
\alias{add_import}
\title{Add an Import Statement to Python Files in the _rixpress Folder Matching a
Nix Environment Name}
\usage{
add_import(import_statement, nix_env, project_path = ".")
}
\arguments{
\item{import_statement}{A character string representing the import statement
to be added. For example, \code{"import numpy as np"}.}

\item{nix_env}{A character string naming the Nix environment file (e.g.
\code{"default.nix"} or \code{"py-env.nix"} or similar).}

\item{project_path}{Path to root of project, typically ".".}
}
\value{
No return value; the function performs in-place modifications of the
files.
}
\description{
This function appends a specified import statement to the end of each Python
file within the \verb{_rixpress} folder and its subdirectories, but only for files
whose base name matches the provided Nix environment.
}
\examples{
\dontrun{
add_import("import numpy as np", "default.nix")
add_import("import numpy as np", "default.nix", project_path = "path/to/project")
}
}
\seealso{
Other python import: 
\code{\link{adjust_import}()}
}
\concept{python import}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxp_populate.R
\name{rxp_populate}
\alias{rxp_populate}
\title{Generate Nix Pipeline Code}
\usage{
rxp_populate(derivs, project_path = ".", build = FALSE, py_imports = NULL, ...)
}
\arguments{
\item{derivs}{A list of derivation objects, where each object is a list of
five elements:
\itemize{
\item \emph{name}, name of the derivation,
\item \emph{snippet}, the nix code snippet to build this derivation,
\item \emph{type}, can be R, Python or Quarto,
\item \emph{additional_files}, character vector of paths to files to make available to build sandbox,
\item \emph{nix_env}, path to Nix environment to build this derivation.
A single deriv is the output of \code{rxp_r()}, \code{rxp_qmd()} or \code{rxp_py()}
function.
}}

\item{project_path}{Path to root of project, defaults to ".".}

\item{build}{Logical, defaults to FALSE. Should the pipeline get built right
after being generated? When FALSE, use \code{rxp_make()} to build the pipeline
at a later stage.}

\item{py_imports}{Named character vector of Python import rewrites. Names are
the base modules that rixpress auto-imports as "import name", and values
are the desired import lines. For example: c(numpy = "import numpy as np",
xgboost = "from xgboost import XGBClassifier"). Each entry is applied by
replacing "import name" with the provided string across generated
_rixpress Python library files.}

\item{...}{Further arguments passed down to methods. Use \code{max-jobs} and
\code{cores} to set parallelism during build. See the documentation of
\code{rxp_make()} for more details.}
}
\value{
Nothing, writes a file called \code{pipeline.nix} with the Nix code to
build the pipeline, as well as folder called _rixpress with required
internal files.
}
\description{
Generate Nix Pipeline Code
}
\details{
This function generates a \code{pipeline.nix} file based on a list of derivation
objects. Each derivation defines a build step, and \code{rxp_populate()} chains these
steps and handles the serialization and conversion of Python objects into R
objects (or vice-versa). Derivations are created with \code{rxp_r()}, \code{rxp_py()}
and so on. By default, the pipeline is also immediately built after being
generated, but the build process can be postponed by setting \code{build} to
FALSE. In this case, the pipeline can then be built using \code{rxp_make()} at
a later stage.
The generated \code{pipeline.nix} expression includes:
\itemize{
\item the required imports of environments, typically \code{default.nix} files generated by
the \code{rix} package;
\item correct handling of interdependencies of the different derivations;
\item serialization and deserialization of both R and Python objects, and conversion
between them when objects are passed from one language to another;
\item correct loading of R and Python packages, or extra functions needed to build
specific targets
}

The \verb{_rixpress} folder contains:
\itemize{
\item R, Python or Julia scripts to load the required packages that need to be
available to the pipeline.
\item a JSON file with the DAG of the pipeline, used for visualisation, and to
allow \code{rxp_populate()} to generate the right dependencies between derivations.
\item \code{.rds} files with build logs, required for \code{rxp_inspect()} and \code{rxp_gc()}.
See \code{vignette("debugging")} for more details.
}

Inline Python import adjustments
In some cases, due to the automatic handling of Python packages, users might
want to change import statements. By default if, say, \code{pandas} is needed to
build a derivation, it will be imported with \verb{import pandas}. However, Python
programmers typically use \verb{import pandas as pd}. You can either:
\itemize{
\item use \code{py_imports} to rewrite these automatically during population, or
\item use \code{adjust_import()} and \code{add_import()} for advanced/manual control.
See \code{vignette("polyglot")} for more details.
}
}
\examples{
\dontrun{
# Create derivation objects
d1 <- rxp_r(mtcars_am, filter(mtcars, am == 1))
d2 <- rxp_r(mtcars_head, head(mtcars_am))
list_derivs <- list(d1, d2)

# Generate and build in one go
rxp_populate(derivs = list_derivs, project_path = ".", build = TRUE)

# Or only populate, with inline Python import adjustments
rxp_populate(
  derivs = list_derivs,
  project_path = ".",
  build = FALSE,
  py_imports = c(pandas = "import pandas as pd")
)
# Then later:
rxp_make()
}
}
\seealso{
Other pipeline functions: 
\code{\link{rxp_make}()},
\code{\link{rxp_pipeline}()}
}
\concept{pipeline functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxp_io.R
\name{rxp_r2py}
\alias{rxp_r2py}
\title{Transfer R Object into a Python Session}
\usage{
rxp_r2py(name, expr, nix_env = "default.nix", project_path = ".")
}
\arguments{
\item{name}{Symbol, name of the derivation.}

\item{expr}{Symbol, R object to be saved into a Python pickle.}

\item{nix_env}{Character, path to the Nix environment file, default is
"default.nix".}

\item{project_path}{Character, path to the project root, default is ".".}
}
\value{
An object of class \code{rxp_derivation}.
}
\description{
Transfer R Object into a Python Session
}
\details{
\code{rxp_r2py(my_obj, my_r_object)} saves an R object to a Python pickle
using \code{reticulate::py_save_object()}.
}
\seealso{
Other interop functions: 
\code{\link{rxp_py2r}()}
}
\concept{interop functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxp_read_load.R
\name{rxp_read}
\alias{rxp_read}
\title{Read Output of a Derivation}
\usage{
rxp_read(derivation_name, which_log = NULL, project_path = ".")
}
\arguments{
\item{derivation_name}{Character, the name of the derivation.}

\item{which_log}{Character, defaults to NULL. If NULL the most recent
build log is used. If a string is provided, it's used as a
regular expression to match against available log files.}

\item{project_path}{Character, defaults to ".".
Path to the root directory of the project.}
}
\value{
The derivation's output.
}
\description{
Reads the output of derivations in the current session,
returns a path if reading directly is not possible.
}
\details{
When \code{derivation_name} points to a single R object,
it gets read in the current session using \code{readRDS()}.
If it's a Python object and \code{{reticulate}} is available,
\code{reticulate::py_load_object()} is used. In case
the derivation is pointing to several outputs (which can
happen when building a Quarto document for example) or
neither \code{readRDS()} nor \code{reticulate::py_load_object()}
successfully read the object, the path to the object is
returned instead.
}
\examples{
\dontrun{
  mtcars <- rxp_read("mtcars")

  # Read from a specific build log
  mtcars <- rxp_read("mtcars", which_log = "2025-05-10")
}
}
\seealso{
Other utilities: 
\code{\link{print.rxp_derivation}()},
\code{\link{rxp_check_chronicles}()},
\code{\link{rxp_copy}()},
\code{\link{rxp_gc}()},
\code{\link{rxp_init}()},
\code{\link{rxp_inspect}()},
\code{\link{rxp_list_logs}()},
\code{\link{rxp_load}()},
\code{\link{rxp_trace}()}
}
\concept{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxp_init.R
\name{rxp_init}
\alias{rxp_init}
\title{Initialize Rixpress Project}
\usage{
rxp_init(project_path = ".", skip_prompt = FALSE)
}
\arguments{
\item{project_path}{Character string specifying the project's path.}

\item{skip_prompt}{Logical. If TRUE, skips all confirmation prompts and proceeds
with initialization, useful on continuous integration. Defaults to FALSE.}
}
\value{
Logical. Returns TRUE if initialization was successful, FALSE if the
operation was cancelled by the user.
}
\description{
Generates \code{gen-env.R} and \code{gen-pipeline.R} scripts in the specified project
directory, after asking the user for confirmation. If the user declines, no
changes are made.
}
\details{
Creates (overwriting if they already exist):
\itemize{
\item \code{gen-env.R}: Script to define an execution environment with \code{{rix}}.
\item \code{gen-pipeline.R}: Defines a data pipeline with \code{{rixpress}}.
}
}
\examples{
# Default usage (will prompt before any action)
\dontrun{
  rxp_init()
}

}
\seealso{
Other utilities: 
\code{\link{print.rxp_derivation}()},
\code{\link{rxp_check_chronicles}()},
\code{\link{rxp_copy}()},
\code{\link{rxp_gc}()},
\code{\link{rxp_inspect}()},
\code{\link{rxp_list_logs}()},
\code{\link{rxp_load}()},
\code{\link{rxp_read}()},
\code{\link{rxp_trace}()}
}
\concept{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxp_io.R
\name{rxp_file}
\alias{rxp_file}
\title{Generic Nix Expression Builder for R, Python, and Julia Data Readers}
\usage{
rxp_file(
  lang,
  name,
  path,
  read_function,
  user_functions = "",
  nix_env = "default.nix",
  env_var = NULL,
  encoder = NULL
)
}
\arguments{
\item{lang}{\code{"R"}, \code{"Py"} or \code{"Jl"}.}

\item{name}{Symbol, the name of the derivation.}

\item{path}{Character, the file path to include (e.g., "data/mtcars.shp") or
a folder path (e.g., "data"). See details.}

\item{read_function}{Function, an R function to read the data, taking one
argument (the path). This can be a user-defined function that is made available
using \code{user_functions}. See details.}

\item{user_functions}{Character vector, user-defined functions to include.
This should be a script (or scripts) containing user-defined functions
to include during the build process for this derivation. It is recommended
to use one script per function, and only include the required script(s) in
the derivation.}

\item{nix_env}{Character, path to the Nix environment file, default is
"default.nix".}

\item{env_var}{List, defaults to NULL. A named list of environment variables
to set before running the R script, e.g., c(VAR = "hello"). Each entry will
be added as an export statement in the build phase.}

\item{encoder}{Function/character, defaults to NULL.
A language-specific serializer to write the loaded object to disk.
\itemize{
\item R: function/symbol/character (e.g., \code{qs::qsave}) taking \verb{(object, path)}. Defaults to \code{saveRDS}.
\item Python: character name of a function taking \verb{(object, path)}. Defaults to using \code{pickle.dump}.
\item Julia: character name of a function taking \verb{(object, path)}. Defaults to using \code{Serialization.serialize}.
}}
}
\value{
An object of class \code{rxp_derivation}.
}
\description{
You should not call it directly, but instead use one of \code{rxp_r_file()},
\code{rxp_py_file()} or \code{rxp_jl_file()}.
}
\details{
Creates a Nix derivation that reads a file or folder of data using R,
Python, or Julia. Handles user-defined functions, environment variables, and Nix
environment specification.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxp_make.R
\name{rxp_import_artifacts}
\alias{rxp_import_artifacts}
\title{Import Nix Store Paths from an Archive}
\usage{
rxp_import_artifacts(archive_file = "_rixpress/pipeline_outputs.nar")
}
\arguments{
\item{archive_file}{Character, path to the archive, defaults to
"_rixpress/pipeline-outputs.nar"}
}
\value{
Nothing, imports the archive contents into the local Nix store.
}
\description{
Imports the store paths contained in an archive file into the local Nix store.
Useful for transferring built outputs between machines.
}
\examples{
\dontrun{
  # Import from the default archive location
  rxp_import_artifacts()

  # Import from a custom archive file
  rxp_import_artifacts("path/to/my_archive.nar")
}
}
\seealso{
Other archive caching functions: 
\code{\link{rxp_export_artifacts}()}
}
\concept{archive caching functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxp_inspect.R
\name{rxp_inspect}
\alias{rxp_inspect}
\title{Inspect the Build Result of a Pipeline}
\usage{
rxp_inspect(project_path = ".", which_log = NULL)
}
\arguments{
\item{project_path}{Character, defaults to ".".
Path to the root directory of the project.}

\item{which_log}{Character, defaults to NULL. If NULL the most recent
build log is used. If a string is provided, it's used as a
regular expression to match against available log files.}
}
\value{
A data frame with derivation names, if their build was successful,
their paths in the /nix/store, and their build outputs.
}
\description{
Returns a data frame with four columns:
- derivation: the name of the derivation
- build_success: whether the build was successful or not
- path: the path of this derivation in the Nix store
- output: the output, if this derivation was built successfully.
Empty outputs mean that this derivation was not built
successfully. Several outputs for a single derivation
are possible.
In the \code{derivation} column you will find an object called \code{all-derivations}.
This object is generated automatically for internal purposes, and you can
safely ignore it.
}
\examples{
\dontrun{
  # Inspect the most recent build
  build_results <- rxp_inspect()

  # Inspect a specific build log
  build_results <- rxp_inspect(which_log = "20250510")

  # Check which derivations failed
  failed <- subset(build_results, !build_success)
}
}
\seealso{
Other utilities: 
\code{\link{print.rxp_derivation}()},
\code{\link{rxp_check_chronicles}()},
\code{\link{rxp_copy}()},
\code{\link{rxp_gc}()},
\code{\link{rxp_init}()},
\code{\link{rxp_list_logs}()},
\code{\link{rxp_load}()},
\code{\link{rxp_read}()},
\code{\link{rxp_trace}()}
}
\concept{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxp_io.R
\name{rxp_py2r}
\alias{rxp_py2r}
\title{Transfer Python Object into an R Session}
\usage{
rxp_py2r(name, expr, nix_env = "default.nix", project_path = ".")
}
\arguments{
\item{name}{Symbol, name of the derivation.}

\item{expr}{Symbol, Python object to be loaded into R.}

\item{nix_env}{Character, path to the Nix environment file, default is
"default.nix".}

\item{project_path}{Character, path to the project root, default is ".".}
}
\value{
An object of class \code{rxp_derivation}.
}
\description{
Transfer Python Object into an R Session
}
\details{
\code{rxp_py2r(my_obj, my_python_object)} loads a serialized Python
object and saves it as an RDS file using \code{reticulate::py_load_object()}.
}
\seealso{
Other interop functions: 
\code{\link{rxp_r2py}()}
}
\concept{interop functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_dag.R
\name{rxp_dag_for_ci}
\alias{rxp_dag_for_ci}
\title{Export DAG of Pipeline and Prepare It for Rendering on CI}
\usage{
rxp_dag_for_ci(
  nodes_and_edges = get_nodes_edges(),
  output_file = "_rixpress/dag.dot"
)
}
\arguments{
\item{nodes_and_edges}{List, output of \code{get_nodes_edges()}. Defaults to
calling \code{get_nodes_edges()}.}

\item{output_file}{Character, the path where the DOT file should be saved.
Defaults to \code{"_rixpress/dag.dot"}. The directory will be created if it
doesn't exist.}
}
\value{
Nothing, writes the DOT file to the specified \code{output_file}.
}
\description{
This function generates a DOT file representation of the
pipeline DAG, suitable for visualization, potentially on CI platforms. It
is called by \code{rxp_ga()}.
}
\examples{
\dontrun{
  # Generate the default _rixpress/dag.dot
  rxp_dag_for_ci()

}
}
\seealso{
Other ci utilities: 
\code{\link{rxp_ga}()},
\code{\link{rxp_write_dag}()}
}
\concept{ci utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxp_pipeline.R
\name{flatten_derivations}
\alias{flatten_derivations}
\title{Flatten a list of derivations and pipelines}
\usage{
flatten_derivations(derivs)
}
\arguments{
\item{derivs}{A list that may contain \code{rxp_derivation} objects,
\code{rxp_pipeline} objects, or a mix of both.}
}
\value{
A flat list of \code{rxp_derivation} objects.
}
\description{
Takes a potentially nested list containing both \code{rxp_derivation} objects
and \code{rxp_pipeline} objects and returns a flat list of derivations with
metadata preserved.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxp_pipeline.R
\name{print.rxp_pipeline}
\alias{print.rxp_pipeline}
\title{Print Method for rxp_pipeline Objects}
\usage{
\method{print}{rxp_pipeline}(x, ...)
}
\arguments{
\item{x}{An object of class "rxp_pipeline"}

\item{...}{Additional arguments passed to print methods}
}
\value{
Nothing, prints a summary of the pipeline object to the console.
}
\description{
Print Method for rxp_pipeline Objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxp_ga.R
\name{rxp_ga}
\alias{rxp_ga}
\title{Run a Pipeline on GitHub Actions}
\usage{
rxp_ga()
}
\value{
Nothing, copies file to a directory.
}
\description{
Run a Pipeline on GitHub Actions
}
\details{
This function puts a \code{.yaml} file inside the \verb{.github/workflows/}
folder on the root of your project. This workflow file expects both
scripts generated by \code{rxp_init()}, \code{gen-env.R} and \code{gen-pipeline.R} to be
present. If that's not the case, edit the \code{.yaml} file accordingly. Build
artifacts are archived and restored automatically between runs. Make sure
to give read and write permissions to the GitHub Actions bot.
}
\examples{
\dontrun{
  rxp_ga()
}
}
\seealso{
Other ci utilities: 
\code{\link{rxp_dag_for_ci}()},
\code{\link{rxp_write_dag}()}
}
\concept{ci utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/derivs.R
\name{rxp_py}
\alias{rxp_py}
\title{Create a Nix Expression Running a Python Function}
\usage{
rxp_py(
  name,
  expr,
  additional_files = "",
  user_functions = "",
  nix_env = "default.nix",
  encoder = NULL,
  decoder = NULL,
  env_var = NULL,
  noop_build = FALSE
)
}
\arguments{
\item{name}{Symbol, name of the derivation.}

\item{expr}{Character, Python code to generate the expression. Ideally it
should be a call to a pure function. Multi-line expressions are not
supported.}

\item{additional_files}{Character vector, additional files to include
during the build process. For example, if a function expects a certain
file to be available, this is where you should include it.}

\item{user_functions}{Character vector, user-defined functions to include.
This should be a script (or scripts) containing user-defined functions
to include during the build process for this derivation. It is recommended
to use one script per function, and only include the required script(s) in
the derivation.}

\item{nix_env}{Character, path to the Nix environment file, default is
"default.nix".}

\item{encoder}{Character, defaults to NULL. The name of the Python
function used to serialize the object. It must accept two arguments: the
object to serialize (first), and the target file path (second). If NULL,
the default behaviour uses \code{pickle.dump}. Define this function in
\code{functions.py}. See \code{vignette("encoding-decoding")} for more details.}

\item{decoder}{Character or named vector/list, defaults to NULL. Can be:
\itemize{
\item A single string for the Python function to unserialize all upstream objects
\item A named vector/list where names are upstream dependency names and values
are their specific unserialize functions
If NULL, the default uses \code{pickle.load}. See \code{vignette("encoding-decoding")} for more details.
}}

\item{env_var}{Character vector, defaults to NULL. A named vector of
environment variables
before running the Python script, e.g., c(PYTHONPATH = "/path/to/modules").
Each entry will be added as an export statement in the build phase.}

\item{noop_build}{Logical, defaults to FALSE. If TRUE, the derivation
produces a no-op build (a stub output with no actual build steps). Any
downstream derivations depending on a no-op build will themselves also
become no-op builds.}
}
\value{
An object of class derivation which inherits from lists.
}
\description{
Create a Nix Expression Running a Python Function
}
\details{
At a basic level,
\code{rxp_py(mtcars_am, "mtcars.filter(polars.col('am') == 1).to_pandas()")}
is equivalent to
\verb{mtcars_am = mtcars.filter(polars.col('am') == 1).to_pandas()}. \code{rxp_py()}
generates the required Nix boilerplate to output a so-called "derivation"
in Nix jargon. A Nix derivation is a recipe that defines how to create an
output (in this case \code{mtcars_am}) including its dependencies, build steps,
and output paths.
}
\examples{
\dontrun{
  rxp_py(
    mtcars_pl_am,
    expr = "mtcars_pl.filter(polars.col('am') == 1).to_pandas()"
  )

  # Skip building this derivation
  rxp_py(
    data_prep,
    expr = "preprocess_data(raw_data)",
    noop_build = TRUE
  )

  # Custom serialization
  rxp_py(
    mtcars_pl_am,
    expr = "mtcars_pl.filter(polars.col('am') == 1).to_pandas()",
    user_functions = "functions.py",
    encoder = "serialize_model",
    additional_files = "some_required_file.bin")
}
}
\seealso{
Other derivations: 
\code{\link{rxp_jl}()},
\code{\link{rxp_jl_file}()},
\code{\link{rxp_py_file}()},
\code{\link{rxp_qmd}()},
\code{\link{rxp_r}()},
\code{\link{rxp_r_file}()},
\code{\link{rxp_rmd}()}
}
\concept{derivations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/derivs.R
\name{rxp_r}
\alias{rxp_r}
\title{Create a Nix Expression Running an R Function}
\usage{
rxp_r(
  name,
  expr,
  additional_files = "",
  user_functions = "",
  nix_env = "default.nix",
  encoder = NULL,
  decoder = NULL,
  env_var = NULL,
  noop_build = FALSE
)
}
\arguments{
\item{name}{Symbol, name of the derivation.}

\item{expr}{R code to generate the expression. Ideally it should be a call
to a pure function, or a piped expression. Multi-line expressions are not
supported.}

\item{additional_files}{Character vector, additional files to include
during the build process. For example, if a function expects a certain
file to be available, this is where you should include it.}

\item{user_functions}{Character vector, user-defined functions to include.
This should be a script (or scripts) containing user-defined functions
to include during the build process for this derivation. It is recommended
to use one script per function, and only include the required script(s) in
the derivation.}

\item{nix_env}{Character, path to the Nix environment file, default is
"default.nix".}

\item{encoder}{Function or character defaults to NULL. A function used to
encode (serialize) objects for transfer between derivations. It must accept two
arguments: the object to encode (first), and the target file path
(second). If your function has a different signature, wrap it to match this
interface. By default, \code{saveRDS()} is used, but this may yield unexpected
results, especially for complex objects like machine learning models. For
instance, for \code{{keras}} models, use \code{keras::save_model_hdf5()} to capture
the full model (architecture, weights, training config, optimiser state,
etc.). See \code{vignette("encoding-decoding")} for more details.}

\item{decoder}{Function, character, or named vector/list,
defaults to NULL. Can be:
\itemize{
\item A single function/string to decode (unserialize) all upstream objects (e.g., \code{readRDS})
\item A named vector/list where names are upstream dependency names and values
are their specific decoding functions (e.g.,
\code{c(mtcars_tail = "qs::qread", mtcars_head = "read.csv")})
By default, \code{readRDS()} is used. See \code{vignette("encoding-decoding")} for more details.
}}

\item{env_var}{Character vector, defaults to NULL. A named vector of
environment variables to set before running the R script, e.g.,
\verb{c("CMDSTAN" = "$\{defaultPkgs.cmdstan\}/opt/cmdstan)"}.
Each entry will be added as an export statement in the build phase.}

\item{noop_build}{Logical, defaults to FALSE. If TRUE, the derivation
produces a no-op build (a stub output with no actual build steps). Any
downstream derivations depending on a no-op build will themselves also
become no-op builds.}
}
\value{
An object of class derivation which inherits from lists.
}
\description{
Create a Nix Expression Running an R Function
}
\details{
At a basic level, \code{rxp_r(mtcars_am, filter(mtcars, am == 1))} is
equivalent to \code{mtcars_am <- filter(mtcars, am == 1)}. \code{rxp_r()} generates the
required Nix boilerplate to output a so-called "derivation" in Nix jargon.
A Nix derivation is a recipe that defines how to create an output (in this
case \code{mtcars_am}) including its dependencies, build steps, and output
paths.
}
\examples{
\dontrun{
  # Basic usage
  rxp_r(name = filtered_mtcars, expr = filter(mtcars, am == 1))

  # Skip building this derivation
  rxp_r(
    name = turtles,
    expr = occurrence(species, geometry = atlantic),
    noop_build = TRUE
  )

  # Serialize object using qs
  rxp_r(
   name = filtered_mtcars,
   expr = filter(mtcars, am == 1),
   encoder = qs::qsave
  )
  # Unerialize using qs::qread in the next derivation
  rxp_r(
   name = mtcars_mpg,
   expr = select(filtered_mtcars, mpg),
   decoder = qs::qread
  )
}
}
\seealso{
Other derivations: 
\code{\link{rxp_jl}()},
\code{\link{rxp_jl_file}()},
\code{\link{rxp_py}()},
\code{\link{rxp_py_file}()},
\code{\link{rxp_qmd}()},
\code{\link{rxp_r_file}()},
\code{\link{rxp_rmd}()}
}
\concept{derivations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxp_copy.R
\name{rxp_copy}
\alias{rxp_copy}
\title{Copy Derivations From the Nix Store to Current Working Directory}
\usage{
rxp_copy(derivation_name = NULL, dir_mode = "0755", file_mode = "0644")
}
\arguments{
\item{derivation_name}{The name of the derivation to copy. If empty, then all
the derivations are copied.}

\item{dir_mode}{Character, default "0755". POSIX permission mode to apply to
directories under the copied output (including the top-level output
directory).}

\item{file_mode}{Character, default "0644". POSIX permission mode to apply to
files under the copied output.}
}
\value{
Nothing, the contents of the Nix store are copied to the current
working directory.
}
\description{
When Nix builds a derivation, its output is saved in the
Nix store located under \verb{/nix/store/}. Even though you can import the
derivations into the current R session using \code{rxp_read()} or \code{rxp_load()},
it can be useful to copy the outputs to the current working directory. This
is especially useful for Quarto documents, where there can be more than one
input, as is the case for \code{html} output.
}
\examples{
\dontrun{
  # Copy all derivations to the current working directory
  rxp_copy()

  # Copy a specific derivation
  rxp_copy("mtcars")

  # Copy with custom permissions (e.g., make scripts executable)
  rxp_copy("my_deriv", dir_mode = "0755", file_mode = "0644")

  # Copy a Quarto document output with multiple files
  rxp_copy("my_quarto_doc")
}
}
\seealso{
Other utilities: 
\code{\link{print.rxp_derivation}()},
\code{\link{rxp_check_chronicles}()},
\code{\link{rxp_gc}()},
\code{\link{rxp_init}()},
\code{\link{rxp_inspect}()},
\code{\link{rxp_list_logs}()},
\code{\link{rxp_load}()},
\code{\link{rxp_read}()},
\code{\link{rxp_trace}()}
}
\concept{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxp_write_dag.R
\name{rxp_write_dag}
\alias{rxp_write_dag}
\title{Generate a DAG From a List of Derivations}
\usage{
rxp_write_dag(rxp_list, output_file = "_rixpress/dag.json")
}
\arguments{
\item{rxp_list}{A list of derivations.}

\item{output_file}{Path to the output JSON file.
Defaults to "_rixpress/dag.json".}
}
\value{
Nothing, writes a JSON file representing the DAG.
}
\description{
Creates a JSON representation of a directed acyclic graph (DAG)
based on dependencies between derivations. Is automatically called
by \code{rxp_populate()}.
}
\examples{
\dontrun{
  rxp_write_dag(rxp_list)
}
}
\seealso{
Other ci utilities: 
\code{\link{rxp_dag_for_ci}()},
\code{\link{rxp_ga}()}
}
\concept{ci utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxp_trace.R
\name{rxp_trace}
\alias{rxp_trace}
\title{Trace Lineage of Derivations}
\usage{
rxp_trace(
  name = NULL,
  dag_file = file.path("_rixpress", "dag.json"),
  transitive = TRUE,
  include_self = FALSE
)
}
\arguments{
\item{name}{Character, defaults to NULL. Name of the derivation to inspect.
If NULL, the function prints the whole pipeline (inverted global view).}

\item{dag_file}{Character, defaults to "_rixpress/dag.json". Path to dag.json.}

\item{transitive}{Logical, defaults to TRUE. If TRUE, show transitive closure and
mark transitive-only nodes with "*". If FALSE, show immediate neighbours only.}

\item{include_self}{Logical, defaults to FALSE. If TRUE, include \code{name} itself in the results.}
}
\value{
Invisibly, a named list mapping each inspected derivation name to a
list with elements:
- dependencies
- reverse_dependencies
The function also prints a tree representation to the console.
}
\description{
Trace Lineage of Derivations
}
\seealso{
Other utilities: 
\code{\link{print.rxp_derivation}()},
\code{\link{rxp_check_chronicles}()},
\code{\link{rxp_copy}()},
\code{\link{rxp_gc}()},
\code{\link{rxp_init}()},
\code{\link{rxp_inspect}()},
\code{\link{rxp_list_logs}()},
\code{\link{rxp_load}()},
\code{\link{rxp_read}()}
}
\concept{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxp_inspect.R
\name{rxp_list_logs}
\alias{rxp_list_logs}
\title{List All Available Build Logs}
\usage{
rxp_list_logs(project_path = ".")
}
\arguments{
\item{project_path}{Character, defaults to ".".
Path to the root directory of the project.}
}
\value{
A data frame with log filenames, modification times, and file sizes.
}
\description{
Returns a data frame with information about all build logs
in the project's _rixpress directory.
}
\examples{
\dontrun{
  # List all build logs in the current project
  logs <- rxp_list_logs()

  # List logs from a specific project directory
  logs <- rxp_list_logs("path/to/project")
}
}
\seealso{
Other utilities: 
\code{\link{print.rxp_derivation}()},
\code{\link{rxp_check_chronicles}()},
\code{\link{rxp_copy}()},
\code{\link{rxp_gc}()},
\code{\link{rxp_init}()},
\code{\link{rxp_inspect}()},
\code{\link{rxp_load}()},
\code{\link{rxp_read}()},
\code{\link{rxp_trace}()}
}
\concept{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/derivs.R
\name{print.rxp_derivation}
\alias{print.rxp_derivation}
\title{Print Method for Derivation Objects}
\usage{
\method{print}{rxp_derivation}(x, ...)
}
\arguments{
\item{x}{An object of class "rxp_derivation"}

\item{...}{Additional arguments passed to print methods}
}
\value{
Nothing, prints a summary of the derivation object to the console.
}
\description{
Print Method for Derivation Objects
}
\examples{
\dontrun{
# d0 is a previously defined derivation
  print(d0)
}
}
\seealso{
Other utilities: 
\code{\link{rxp_check_chronicles}()},
\code{\link{rxp_copy}()},
\code{\link{rxp_gc}()},
\code{\link{rxp_init}()},
\code{\link{rxp_inspect}()},
\code{\link{rxp_list_logs}()},
\code{\link{rxp_load}()},
\code{\link{rxp_read}()},
\code{\link{rxp_trace}()}
}
\concept{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_dag.R
\name{rxp_visnetwork}
\alias{rxp_visnetwork}
\title{Create a Directed Acyclic Graph (DAG) Representing the Pipeline
Using \code{{visNetwork}}}
\usage{
rxp_visnetwork(
  nodes_and_edges = get_nodes_edges(),
  color_by = c("pipeline", "type"),
  colour_by = NULL
)
}
\arguments{
\item{nodes_and_edges}{List, output of \code{get_nodes_edges()}.}

\item{color_by}{Character, either "pipeline" (default) or "type".
When "pipeline", nodes show type as fill colour and pipeline as border.
When "type", nodes are colored by their derivation type (rxp_r, rxp_py, etc.).}

\item{colour_by}{Character, alias for \code{color_by}.}
}
\value{
Nothing, this function opens a new tab in your browser with
the DAG generated using \code{{visNetwork}}.
}
\description{
Uses \code{{visNetwork}} to generate the plot. \code{{visNetwork}} is a
soft dependency of \code{{rixpress}} so you need to install it to use this
function. When derivations are organized into pipelines using
\code{rxp_pipeline()}, nodes use a dual-encoding approach: the interior fill
shows the derivation type (R, Python, etc.) while the border shows
the pipeline group colour.
}
\examples{
\dontrun{
  rxp_visnetwork()
  rxp_visnetwork(colour_by = "type")  # Color by derivation type instead
}
}
\seealso{
Other visualisation functions: 
\code{\link{rxp_ggdag}()}
}
\concept{visualisation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/derivs.R
\name{rxp_rmd}
\alias{rxp_rmd}
\title{Render an R Markdown Document as a Nix Derivation}
\usage{
rxp_rmd(
  name,
  rmd_file,
  additional_files = "",
  nix_env = "default.nix",
  params = NULL,
  env_var = NULL,
  noop_build = FALSE
)
}
\arguments{
\item{name}{Symbol, derivation name.}

\item{rmd_file}{Character, path to .Rmd file.}

\item{additional_files}{Character vector, additional files to include, for
example a folder containing the pictures to include in the R Markdown
document.}

\item{nix_env}{Character, path to the Nix environment file, default is
"default.nix".}

\item{params}{List, parameters to pass to the R Markdown document. Default is
NULL.}

\item{env_var}{List, defaults to NULL. A named list of environment variables
to set before running the R Markdown render command, e.g., c(RSTUDIO_PANDOC
= "/path/to/pandoc"). Each entry will be added as an export statement in
the build phase.}

\item{noop_build}{Logical, defaults to FALSE. If TRUE, the derivation
produces a no-op build (a stub output with no actual build steps). Any
downstream derivations depending on a no-op build will themselves also
become no-op builds.}
}
\value{
An object of class derivation which inherits from lists.
}
\description{
Render an R Markdown Document as a Nix Derivation
}
\details{
To include objects built in the pipeline,
\code{rxp_read("derivation_name")} should be put in the .Rmd file.
}
\examples{
\dontrun{
  # Compile a .Rmd file to a pdf
  # `images` is a folder containing images to include in the R Markdown doc
  rxp_rmd(
    name = report,
    rmd_file = "report.Rmd",
    additional_files = "images"
  )

  # Skip building this derivation
  rxp_rmd(
    name = draft_report,
    rmd_file = "draft.Rmd",
    noop_build = TRUE
  )
}
}
\seealso{
Other derivations: 
\code{\link{rxp_jl}()},
\code{\link{rxp_jl_file}()},
\code{\link{rxp_py}()},
\code{\link{rxp_py_file}()},
\code{\link{rxp_qmd}()},
\code{\link{rxp_r}()},
\code{\link{rxp_r_file}()}
}
\concept{derivations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxp_pipeline.R
\name{rxp_pipeline}
\alias{rxp_pipeline}
\title{Create a Named Pipeline of Derivations}
\usage{
rxp_pipeline(name, path, color = NULL, ...)
}
\arguments{
\item{name}{Character, the name of the pipeline (e.g., "ETL", "Model").}

\item{path}{Character path to an R script returning a list of derivations, OR
a list of derivation objects created by \code{rxp_r()}, \code{rxp_py()}, etc.}

\item{color}{Character, optional. A CSS color name (e.g., "darkorange") or
hex code (e.g., "#FF5733") to use for this pipeline's nodes in DAG
visualizations. If NULL, a default color will be assigned.}

\item{...}{Additional arguments (currently unused, reserved for future use).}
}
\value{
An object of class \code{rxp_pipeline} containing the derivations with
pipeline metadata attached.
}
\description{
Groups multiple derivations into a named pipeline for organizational
purposes. This allows you to structure large projects into logical
sub-pipelines (e.g., "ETL", "Model", "Report") that are visually
distinguished in DAG visualizations.
}
\details{
The \code{rxp_pipeline()} function is used to organize derivations into logical
groups. When passed to \code{rxp_populate()}, the derivations are flattened but
retain their group and color metadata, which is then used in DAG
visualizations (\code{rxp_visnetwork()} and \code{rxp_ggdag()}) to distinguish
different parts of your workflow.

This pattern enables a "Master Script" workflow where you can define
sub-pipelines in separate R scripts that each return a list of derivations.
You then pass the paths to these scripts to \code{rxp_pipeline()}:
}
\examples{
\dontrun{
  # Define derivations in separate scripts
  # pipelines/01_etl.R returns: list(rxp_r(...), rxp_r(...))
  # pipelines/02_model.R returns: list(rxp_r(...), rxp_r(...))

  # Master script (run.R):

  # Create named pipelines with colors by pointing to the files
  pipe_etl <- rxp_pipeline("ETL", "pipelines/01_etl.R", color = "darkorange")
  pipe_model <- rxp_pipeline("Model", "pipelines/02_model.R", color = "dodgerblue")

  # Build the combined pipeline
  rxp_populate(list(pipe_etl, pipe_model))
  rxp_make()

  # Visualize - ETL nodes will be orange, Model nodes will be blue
  rxp_visnetwork()
}

}
\seealso{
Other pipeline functions: 
\code{\link{rxp_make}()},
\code{\link{rxp_populate}()}
}
\concept{pipeline functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse_pkgs.R
\name{adjust_import}
\alias{adjust_import}
\title{Adjust Python Import Statements}
\usage{
adjust_import(old_import, new_import, project_path = ".")
}
\arguments{
\item{old_import}{A character string representing the import statement to be
replaced. For example, \code{"import pillow"}.}

\item{new_import}{A character string representing the new import statement to
replace with. For example, \code{"from PIL import Image"}.}

\item{project_path}{Path to root of project, typically ".".}
}
\value{
No return value; the function performs in-place
modifications of the files.
}
\description{
When calling \code{rxp_populate()}, a file containing Python import statements is
automatically generated inside the \verb{_rixpress} folder. For example, if the
\code{numpy} package is needed, the file will include a line like
\code{"import numpy"}. However, Python programmers often write
\code{"import numpy as np"} instead.
}
\details{
In some cases, the correct import statement is entirely different. For
example, for the \code{pillow} package, the generated file will contain
\code{"import pillow"}, which is incorrect—Python code should import from the
\code{PIL} namespace instead, e.g., \code{"from PIL import Image"}.

Because these adjustments cannot be automated reliably, the \code{adjust_import()}
function allows you to search and replace import statements programmatically.
It reads each file in the \verb{_rixpress} folder, performs the replacement, and
writes the modified content back to the file.
}
\examples{
\dontrun{
adjust_import("import pillow", "from PIL import Image")
adjust_import("import pillow", "from PIL import Image", project_path = "path/to/project")
}
}
\seealso{
Other python import: 
\code{\link{add_import}()}
}
\concept{python import}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxp_gc.R
\name{rxp_gc}
\alias{rxp_gc}
\title{Garbage Collect Rixpress Build Artifacts and Logs}
\usage{
rxp_gc(
  keep_since = NULL,
  project_path = ".",
  dry_run = FALSE,
  timeout_sec = 300,
  verbose = FALSE,
  ask = TRUE
)
}
\arguments{
\item{keep_since}{Date or character string (YYYY-MM-DD format). If provided,
only build logs older than this date will be targeted for deletion, along
with their associated Nix store paths. If \code{NULL}, performs a full
Nix garbage collection. Default is \code{NULL}.}

\item{project_path}{Character string specifying the path to the project
directory containing the \code{_rixpress} folder with build logs.
Default is \code{"."} (current directory).}

\item{dry_run}{Logical. If \code{TRUE}, shows what would be deleted without
actually performing any deletions. Useful for previewing the cleanup
operation. Default is \code{FALSE}.}

\item{timeout_sec}{Numeric. Timeout in seconds for individual Nix commands.
Also used for concurrency lock expiration. Default is 300 seconds.}

\item{verbose}{Logical. If \code{TRUE}, provides detailed output including
full paths, command outputs, and diagnostic information about references
preventing deletion. Default is \code{FALSE}.}

\item{ask}{Logical. If \code{TRUE}, ask for user confirmation before
performing deleting artifacts. Default is \code{TRUE}.}
}
\value{
Invisibly returns a list with cleanup summary information:
\itemize{
\item \code{kept}: Vector of build log filenames that were kept
\item \code{deleted}: Vector of build log filenames targeted for deletion
\item \code{protected}: Number of store paths protected via GC roots (date-based mode)
\item \code{deleted_count}: Number of store paths successfully deleted
\item \code{failed_count}: Number of store paths that failed to delete
\item \code{referenced_count}: Number of store paths skipped due to references
\item \code{log_files_deleted}: Number of build log files successfully deleted
\item \code{log_files_failed}: Number of build log files that failed to delete
\item \code{dry_run_details}: List of detailed information when dry_run = TRUE
}
}
\description{
This function performs garbage collection on Nix store paths and build log files
generated by rixpress. It can operate in two modes: full garbage collection
(when \code{keep_since = NULL}) or targeted deletion based on log file age.
}
\details{
The function operates in two modes:

\strong{Full Garbage Collection Mode} (\code{keep_since = NULL}):
\itemize{
\item Runs \code{nix-store --gc} to delete all unreferenced store paths
\item Does not delete any build log files
\item Suitable for complete cleanup of unused Nix store paths
}

\strong{Targeted Deletion Mode} (\code{keep_since} specified):
\itemize{
\item Identifies build logs older than the specified date
\item Extracts store paths from old logs using \code{rxp_inspect()}
\item Protects recent store paths by creating temporary GC roots
\item Attempts to delete old store paths individually using \code{nix-store --delete}
\item Deletes the corresponding build log \code{.json} files from \code{_rixpress/}
\item Handles referenced paths gracefully (paths that cannot be deleted due to dependencies)
}

\strong{Concurrency Safety:}
The function uses a lock file mechanism to prevent multiple instances from
running simultaneously, which could interfere with each other's GC root management.

\strong{Reference Handling:}
Some store paths may not be deletable because they are still referenced by:
\itemize{
\item User or system profile generations
\item Active Nix shell environments
\item Result symlinks in project directories
\item Other store paths that depend on them
}
These paths are reported but not considered errors.
}
\examples{
\dontrun{
# Preview what would be deleted (dry run)
rxp_gc(keep_since = "2025-08-01", dry_run = TRUE, verbose = TRUE)

# Delete artifacts from builds older than August 1st, 2025
rxp_gc(keep_since = "2025-08-01")

# Full garbage collection of all unreferenced store paths
rxp_gc()

# Clean up artifacts older than 30 days ago
rxp_gc(keep_since = Sys.Date() - 30)
}

}
\seealso{
\code{\link{rxp_list_logs}}, \code{\link{rxp_inspect}}

Other utilities: 
\code{\link{print.rxp_derivation}()},
\code{\link{rxp_check_chronicles}()},
\code{\link{rxp_copy}()},
\code{\link{rxp_init}()},
\code{\link{rxp_inspect}()},
\code{\link{rxp_list_logs}()},
\code{\link{rxp_load}()},
\code{\link{rxp_read}()},
\code{\link{rxp_trace}()}
}
\concept{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxp_chronicler.R
\name{rxp_check_chronicles}
\alias{rxp_check_chronicles}
\title{Check Pipeline Outputs for Chronicle Status}
\usage{
rxp_check_chronicles(project_path = ".", which_log = NULL)
}
\arguments{
\item{project_path}{Character, defaults to ".".
Path to the root directory of the project.}

\item{which_log}{Character, defaults to NULL. If NULL the most recent
build log is used. If a string is provided, it's used as a
regular expression to match against available log files.}
}
\value{
A data frame with columns: derivation, chronicle_state,
num_operations, num_failed, failed_functions, messages.
Returns NULL invisibly if chronicler is not installed or no
chronicle objects are found.
}
\description{
Scans all derivation outputs for chronicle objects and
reports their status: success (Just, no warnings), warning (Just with
warnings), or nothing (failed computation). Only active when chronicler
is installed.
}
\details{
This function is useful when using the \code{{chronicler}} package in your
rixpress pipeline. Because chronicler catches errors and warnings,
returning \code{Nothing} values instead of failing, Nix builds will always
succeed. This function helps identify derivations that contain failed
computations.

The function displays one of three symbols for each chronicle:
\itemize{
\item \code{checkmark} Success: Just value, no warnings or errors
\item \code{warning sign} Warning: Just value, but warnings were captured
\item \code{X mark} Nothing: Failed computation, errors captured
}
}
\examples{
\dontrun{
  # After building a pipeline with chronicler functions
  rxp_check_chronicles()

  # Check a specific build log
  rxp_check_chronicles(which_log = "20250131")
}
}
\seealso{
Other utilities: 
\code{\link{print.rxp_derivation}()},
\code{\link{rxp_copy}()},
\code{\link{rxp_gc}()},
\code{\link{rxp_init}()},
\code{\link{rxp_inspect}()},
\code{\link{rxp_list_logs}()},
\code{\link{rxp_load}()},
\code{\link{rxp_read}()},
\code{\link{rxp_trace}()}
}
\concept{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxp_make.R
\name{rxp_make}
\alias{rxp_make}
\title{Build Pipeline Using Nix}
\usage{
rxp_make(verbose = 0L, max_jobs = 1, cores = 1)
}
\arguments{
\item{verbose}{Integer, defaults to 0L. Verbosity level: 0 = show progress
indicators only, 1+ = show nix output with increasing verbosity. 0:
"Progress only", 1: "Informational", 2: "Talkative", 3: "Chatty", 4: "Debug",
5: "Vomit". Values higher than 5 are capped to 5.
Each level adds one --verbose flag to nix-store command.}

\item{max_jobs}{Integer, number of derivations to be built in parallel.}

\item{cores}{Integer, number of cores a derivation can use during build.}
}
\value{
A character vector of paths to the built outputs.
}
\description{
Runs \code{nix-build} with a quiet flag, outputting to \verb{_rixpress/result}.
}
\details{
When the \code{{chronicler}} package is available, \code{rxp_make()} automatically
calls \code{rxp_check_chronicles()} after a successful build to check for
\code{Nothing} values in chronicle objects. This helps detect silent failures
in pipelines that use chronicler's \code{record()} decorated functions.
See \code{vignette("chronicler")} for more details.
}
\examples{
\dontrun{
  # Build the pipeline with progress indicators (default)
  rxp_make()

  # Build with verbose output and parallel execution
  rxp_make(verbose = 2, max_jobs = 4, cores = 2)

  # Maximum verbosity
  rxp_make(verbose = 3)
}
}
\seealso{
Other pipeline functions: 
\code{\link{rxp_pipeline}()},
\code{\link{rxp_populate}()}
}
\concept{pipeline functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxp_make.R
\name{rxp_export_artifacts}
\alias{rxp_export_artifacts}
\title{Export Nix Store Paths to an Archive}
\usage{
rxp_export_artifacts(
  archive_file = "_rixpress/pipeline_outputs.nar",
  which_log = NULL,
  project_path = "."
)
}
\arguments{
\item{archive_file}{Character, path to the archive, defaults to
"_rixpress/pipeline-outputs.nar"}

\item{which_log}{Character or NULL, regex pattern to match a specific log file.
If NULL (default), the most recent log file will be used.}

\item{project_path}{Character, defaults to ".".
Path to the root directory of the project.}
}
\value{
Nothing, creates an archive file at the specified location.
}
\description{
Creates a single archive file containing the specified Nix store paths
and their dependencies.
This archive can be transferred to another machine and imported into
its Nix store.
}
\examples{
\dontrun{
  # Export the most recent build to the default location
  rxp_export_artifacts()

  # Export a specific build to a custom location
  rxp_export_artifacts(
    archive_file = "my_archive.nar",
    which_log = "20250510"
  )
}
}
\seealso{
Other archive caching functions: 
\code{\link{rxp_import_artifacts}()}
}
\concept{archive caching functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxp_io.R
\name{rxp_jl_file}
\alias{rxp_jl_file}
\title{Creates a Nix Expression That Reads In a File (or Folder of Data) Using Julia}
\usage{
rxp_jl_file(...)
}
\arguments{
\item{...}{
  Arguments passed on to \code{\link[=rxp_file]{rxp_file}}
  \describe{
    \item{\code{name}}{Symbol, the name of the derivation.}
    \item{\code{path}}{Character, the file path to include (e.g., "data/mtcars.shp") or
a folder path (e.g., "data"). See details.}
    \item{\code{read_function}}{Function, an R function to read the data, taking one
argument (the path). This can be a user-defined function that is made available
using \code{user_functions}. See details.}
    \item{\code{user_functions}}{Character vector, user-defined functions to include.
This should be a script (or scripts) containing user-defined functions
to include during the build process for this derivation. It is recommended
to use one script per function, and only include the required script(s) in
the derivation.}
    \item{\code{nix_env}}{Character, path to the Nix environment file, default is
"default.nix".}
    \item{\code{env_var}}{List, defaults to NULL. A named list of environment variables
to set before running the R script, e.g., c(VAR = "hello"). Each entry will
be added as an export statement in the build phase.}
    \item{\code{encoder}}{Function/character, defaults to NULL.
A language-specific serializer to write the loaded object to disk.
\itemize{
\item R: function/symbol/character (e.g., \code{qs::qsave}) taking \verb{(object, path)}. Defaults to \code{saveRDS}.
\item Python: character name of a function taking \verb{(object, path)}. Defaults to using \code{pickle.dump}.
\item Julia: character name of a function taking \verb{(object, path)}. Defaults to using \code{Serialization.serialize}.
}}
  }}
}
\value{
An object of class \code{rxp_derivation}.
}
\description{
Creates a Nix Expression That Reads In a File (or Folder of Data) Using Julia
}
\details{
The basic usage is to provide a path to a file, and the function
to read it. For example: \code{rxp_r_file(mtcars, path = "data/mtcars.csv", read_function = read.csv)}.
It is also possible instead to point to a folder that contains many
files that should all be read at once, for example:
\verb{rxp_r_file(many_csvs, path = "data", read_function = \\(x)(readr::read_csv(list.files(x, full.names = TRUE, pattern = ".csv$"))))}.
See the \code{vignette("importing-data")} vignette for more detailed examples.
}
\seealso{
Other derivations: 
\code{\link{rxp_jl}()},
\code{\link{rxp_py}()},
\code{\link{rxp_py_file}()},
\code{\link{rxp_qmd}()},
\code{\link{rxp_r}()},
\code{\link{rxp_r_file}()},
\code{\link{rxp_rmd}()}
}
\concept{derivations}
